/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope;

import org.apache.commons.cli.ParseException;
import org.apache.commons.io.FilenameUtils;

import ru.ispras.fortress.solver.Environment;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Configuration;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.EngineEventHandler;
import ru.ispras.retrascope.basis.EngineRegistry;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.Parameters;
import ru.ispras.retrascope.basis.ToolChain;
import ru.ispras.retrascope.engine.cfg.extractor.iface.CfgInterfaceExtractor;
import ru.ispras.retrascope.engine.cfg.printer.graphml.CfgGraphMlPrinter;
import ru.ispras.retrascope.engine.cfg.transformer.cgaa.CfgCgaaTransformer;
import ru.ispras.retrascope.engine.cgaa.printer.graphml.CgaaGraphMlPrinter;
import ru.ispras.retrascope.engine.cgaa.transformer.efsm.CgaaEfsmTransformer;
import ru.ispras.retrascope.engine.conflict.printer.xml.ConflictXmlPrinter;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.EfsmConflictExtractor;
import ru.ispras.retrascope.engine.efsm.generator.test.EfsmTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.fate.EfsmFatePlusTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.fate.EfsmFateTestGenerator;
import ru.ispras.retrascope.engine.efsm.printer.graphml.EfsmGraphMlPrinter;
import ru.ispras.retrascope.engine.test.printer.xml.TestXmlPrinter;
import ru.ispras.retrascope.parser.test.XmlTestParser;
import ru.ispras.retrascope.parser.verilog.VerilogParser;
import ru.ispras.retrascope.parser.vhdl.VhdlParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * The main class of the project.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class Retrascope {

  /**
   * The tool run information.
   */
  public static class Run {
    // The tool chain.
    private ToolChain toolChain;
    // The source entities.
    private Map<EntityType, Entity> sources;

    /**
     * Constructs a tool run.
     * 
     * @param toolChain the tool chain to be used.
     * @param sources the source entities to be analyzed.
     */
    public Run(final ToolChain toolChain, final Map<EntityType, Entity> sources) {
      this.toolChain = toolChain;
      this.sources = sources;
    }

    /**
     * Construct a tool run with unspecified tool chain.
     */
    public Run() {
      this(null, null);
    }

    /**
     * Returns the tool chain to be used.
     * 
     * @return the tool chain.
     */
    public ToolChain getEngine() {
      return toolChain;
    }

    /**
     * Returns the sources entities to be analyzed.
     * 
     * @return the source entities.
     */
    public Map<EntityType, Entity> getSources() {
      return sources;
    }

    /**
     * Starts the tool chain with the given source entities.
     * 
     * @param eventHandler the engine event handler.
     */
    public void start(final EngineEventHandler eventHandler) {
      if (toolChain != null) {
        toolChain.start(sources, eventHandler);
      }
    }

  }

  /** The command line format. */
  public static final String FORMAT = "[options] files";

  /** The name of the solver path property. */
  public static final String SOLVER_PATH_PROPERTY = "smt.solver.path";
  /** The name of the solver path environment variable. */
  public static final String SOLVER_PATH_VARIABLE = "SMT_SOLVER_PATH";

  /** The command line parameters. */
  public static final Parameter TARGET_PARAM =
      new Parameter("target", true, "Set a target entity");
  public static final Parameter ENGINE_PARAM =
      new Parameter("engine", true, "Set a subset of engines");
  public static final Parameter LOG_PARAM =
      new Parameter("log", true, "Set a log file");
  public static final Parameter LOG_LEVEL =
      new Parameter("log-level", true, "Set a log level");
  public static final Parameter HELP_PARAM =
      new Parameter("help", false, "Show this message");
  public static final Parameter SOLVER_DEBUG_PARAM =
      new Parameter("solver-debug", false, "Set debug mode for SMT solver");

  public static final Parameters PARAMETERS = new Parameters();
  static {
    PARAMETERS.addParameter(TARGET_PARAM);
    PARAMETERS.addParameter(ENGINE_PARAM);
    PARAMETERS.addParameter(LOG_PARAM);
    PARAMETERS.addParameter(LOG_LEVEL);
    PARAMETERS.addParameter(HELP_PARAM);
    PARAMETERS.addParameter(SOLVER_DEBUG_PARAM);
  }

  /** The engine registry. */
  private static EngineRegistry engineRegistry = new EngineRegistry();

  static {
    // HDL Parsers.
    engineRegistry.add(new VerilogParser());
    engineRegistry.add(new VhdlParser());
    // CFG Engines.
    engineRegistry.add(new CfgGraphMlPrinter());
    engineRegistry.add(new CfgCgaaTransformer());
    engineRegistry.add(new CfgInterfaceExtractor());
    // CGAA Engines.
    engineRegistry.add(new CgaaGraphMlPrinter());
    engineRegistry.add(new CgaaEfsmTransformer());
    // EFSM Engines.
    engineRegistry.add(new EfsmConflictExtractor());
    engineRegistry.add(new EfsmGraphMlPrinter());
    engineRegistry.add(new EfsmFateTestGenerator());
    engineRegistry.add(new EfsmFatePlusTestGenerator());
    engineRegistry.add(new EfsmTestGenerator());
    // XML Engines.
    engineRegistry.add(new ConflictXmlPrinter());
    engineRegistry.add(new TestXmlPrinter());
    engineRegistry.add(new XmlTestParser());
  }

  /**
   * Returns the engine registry.
   * 
   * @return the engine registry.
   */
  public static EngineRegistry getEngineRegistry() {
    return engineRegistry;
  }

  /** The tool configuration */
  private static Configuration config = new Configuration();

  /**
   * Returns the tool configuration.
   *
   * @return the tool configuration
   */
  public static Configuration getConfiguration() { return config; }

  /**
   * Analyzes the command line and returns the run information.
   * 
   * @param args the command line.
   * @return the run information.
   */
  public static Run getRun(final String[] args) {
    final Arguments arguments = new Arguments(PARAMETERS);

    try {
      // Parse the command line.
      arguments.setCommandLine(args);
    } catch (ParseException e) {
      return new Run();
    }

    // Fill the configuration.
    config.configure(args);

    // Get the target entity types.
    final Set<EntityType> targets = new HashSet<>();

    if (arguments.contains(TARGET_PARAM)) {
      final StringTokenizer tokenizer =
          new StringTokenizer(arguments.getValue(TARGET_PARAM), File.pathSeparator);

      while (tokenizer.hasMoreTokens()) {
        targets.add(EntityType.get(tokenizer.nextToken()));
      }
    }

    // Get the engines to be used.
    final Set<Engine> engines = new HashSet<>();

    if (arguments.contains(ENGINE_PARAM)) {
      final StringTokenizer tokenizer =
          new StringTokenizer(arguments.getValue(ENGINE_PARAM), File.pathSeparator);

      while (tokenizer.hasMoreTokens()) {
        final String id = tokenizer.nextToken();
        final Engine engine = engineRegistry.getEngine(id);

        if (engine == null) {
          Log.getLogger().log(LogLevel.ERROR, String.format("The engine '%s' is unknown", id));
        } else {
          engines.add(engine);
        }
      }
    }

    // Get the list of file names.
    final String[] files = arguments.getArguments();

    // Construct the map of source entities.
    final Map<EntityType, Entity> sources = new HashMap<>();
    {
      // Add the arguments as one of the entities.
      sources.put(arguments.getEntityType(), arguments);

      // Divide the file names into groups according to the extension.
      for (final String file : files) {
        final String extension = FilenameUtils.getExtension(file);
        final EntityType source = EntityType.get(extension);

        FileNames fileNames = (FileNames) sources.get(source);

        if (fileNames == null) {
          fileNames = new FileNames(extension);
        }

        fileNames.add(file);
        sources.put(source, fileNames);
      }
    }

    // Get the tool chain for the specified target.
    final ToolChain toolChain = engineRegistry.getEngine(targets, engines, sources.keySet());

    return new Run(toolChain, sources);
  }

  /**
   * Configures the Fortress environment.
   */
  private static void configureSolverPath() {
    String solverPath = System.getProperties().getProperty(SOLVER_PATH_PROPERTY);

    if (solverPath == null) {
      solverPath = System.getenv(SOLVER_PATH_VARIABLE);
    }

    if (solverPath == null) {
      throw new RuntimeException(String.format("Neither property '%s' nor variable '%s' is set",
          SOLVER_PATH_PROPERTY, SOLVER_PATH_VARIABLE));
    }

    Environment.setSolverPath(solverPath);
  }
  
  /**
   * Configures and runs the tool chain.
   * 
   * @param args the command line.
   * @param eventHandler the engine event handler.
   */
  public static void main(final String[] args, final EngineEventHandler eventHandler) {
    final Arguments arguments = new Arguments(PARAMETERS);

    try {
      // Configure the solver path.
      configureSolverPath();
      // Parse the command line.
      arguments.setCommandLine(args);
    } catch (ParseException e) {
      PARAMETERS.help(FORMAT);
    }

    if (arguments.contains(LOG_LEVEL)) {
      Log.setLevel(LogLevel.parse(arguments.getValue(LOG_LEVEL)));
    }

    if (arguments.contains(LOG_PARAM)) {
      Log.addFileOutput(arguments.getValue(LOG_PARAM));
    }

    if (arguments.contains(SOLVER_DEBUG_PARAM)) {
      Environment.setDebugMode(true);
    }

    // Handle the help option.
    if (arguments.contains(HELP_PARAM)) {
      PARAMETERS.help(FORMAT);
    }

    // Handle the target option.
    try {
      Log.getLogger().log(LogLevel.INFO, "Retrascope is starting");

      Run run = getRun(args);

      if (run.getEngine() != null) {
        run.start(eventHandler);
      } else {
        Log.getLogger().log(LogLevel.ERROR,
            String.format("The targets '%s' with engines '%s' are unreachable",
                arguments.contains(TARGET_PARAM) ? arguments.getValue(TARGET_PARAM) : "unknown",
                arguments.contains(ENGINE_PARAM) ? arguments.getValue(ENGINE_PARAM) : "unknown"));
      }
    } catch (Exception e) {
      Log.getLogger().log(LogLevel.ERROR, "The exception has been encountered", e);
      throw e;
    } finally {
      Log.getLogger().log(LogLevel.INFO, "Retrascope is shutting down");
      Log.closeFileHandlers();
    }
  }

  /**
   * Configures and runs the tool chain.
   * 
   * @param args the command line.
   */
  public static void main(final String[] args) {
    main(args, null);
  }
}
