/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import java.util.ListIterator;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.ParseException;

/**
 * This class represents command line parameters and implements command line parsing. The use case
 * is as follows:
 * 
 * 1) Specify available parameters by creating constants of the <code>Parameter</code> type:
 * 
 * <pre>
 * public static final PATH_PARAM = new Parameter("I", "path", true,  "Adds a directory");
 * public static final HELP_PARAM = new Parameter("h", "help", false, "Shows this message");
 * </pre>
 *
 * 2) Create a <code>Parameters</code> instance and add the specified parameters:
 * 
 * <pre>
 * Parameters params = new Parameters();
 * params.addParameter(PATH_PARAM);
 * params.addParameter(HELP_PARAM);
 * </pre>
 * 
 * 3) Create a <code>Arguments</code> instance and parse a command line:
 * 
 * <pre>
 * Arguments args = new Arguments(params);
 * args.setCommandLine(strings);
 * </pre>
 * 
 * 4) Get arguments using the <code>contains</code> and <code>getValue</code> methods.
 * 
 * <pre>
 *    if(args.contains(HELP_PARAM) {
 *      params.help();
 *    }
 *    if(args.contains(PATH_PARAM)) {
 *      addDirectory(args.getValue(PATH_PARAM));
 *    }
 * </pre>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class Arguments extends Entity {

  /**
   * This class implements a command line parser on the base of Apache Commons CLI.
   */
  public static class ExtendedGnuParser extends GnuParser {
    // The flag to ignore the unrecognized option error.
    private boolean ignoreUnrecognizedOption;

    /**
     * Creates a command line parser.
     *
     * @param ignoreUnrecognizedOption the flag to ignore the unrecognized option error.
     */
    public ExtendedGnuParser(final boolean ignoreUnrecognizedOption) {
      this.ignoreUnrecognizedOption = ignoreUnrecognizedOption;
    }

    /**
     * Creates a command line parser.
     */
    public ExtendedGnuParser() {
      this(true);
    }

    @Override
    @SuppressWarnings("rawtypes") 
    protected void processOption(final String arg, final ListIterator iterator)
        throws ParseException {
      boolean hasOption = getOptions().hasOption(arg);

      if (hasOption || !ignoreUnrecognizedOption) {
          super.processOption(arg, iterator);
      }
    }
  }  

  // The entity type
  public static final EntityType TYPE = EntityType.get("args");

  // The command line parameters specification.
  private Parameters parameters;
  // The command line parameters (raw).
  private String[] commandLine;
  // The command line parameters (parsed).
  private CommandLine parserResults;

  /**
   * Constructs an arguments entity with the given specification of parameters.
   * 
   * @param parameters the parameters specification.
   */
  public Arguments(final Parameters parameters) {
    super(Arguments.TYPE);
    
    this.parameters = parameters;
  }

  /**
   * Constructs an arguments entity with the given specification of parameters.
   * 
   * @param parameters the parameters specification.
   * @param args the command line to be parsed.
   * @throws ParseException the parse exception.
   */
  public Arguments(final Parameters parameters, final String[] args) throws ParseException {
    this(parameters);

    setCommandLine(args);
  }

  /**
   * Parses the given command line.
   * 
   * @param args the command line to be parsed.
   * @throws ParseException the parse exception.
   */
  public void setCommandLine(final String[] args) throws ParseException {
    final CommandLineParser parser = new ExtendedGnuParser();
    parserResults = parser.parse(parameters.getOptions(), commandLine = args);
  }

  /**
   * Returns the command line having been set before.
   * 
   * @return the unparsed array of strings.
   */
  public String[] getCommandLine() {
    return commandLine;
  }

  /**
   * Checks whether the given parameter is specified.
   * 
   * @param parameter the parameter to be checked.
   * @return <code>true</code> if the parameter is specified; <code>false</code> otherwise.
   */
  public boolean contains(final Parameter parameter) {
    return parserResults.hasOption(parameter.getLongName());
  }

  /**
   * Returns the parameter value.
   * 
   * @param parameter the parameter which value to be returned.
   * @return the option value if the parameter is specified; <code>null</code> otherwise.
   */
  public String getValue(final Parameter parameter) {
    return parserResults.getOptionValue(parameter.getLongName());
  }

  /**
   * Returns the parameter value.
   * 
   * @param parameter the parameter which value to be returned.
   * @param defaultValue the value to be returned if the parameter is absent
   * @return the parameter value.
   */
  public String getValue(final Parameter parameter, final String defaultValue) {
    return parserResults.getOptionValue(parameter.getLongName(), defaultValue);
  }

  /**
   * Returns the unparsed part of the command line (typically, list of files).
   * 
   * @return the unparsed part of the command line.
   */
  public String[] getArguments() {
    return parserResults.getArgs();
  }

  @Override
  public String toString() {
    if(commandLine == null) {
      return "";
    }

    StringBuffer buffer = new StringBuffer();

    for (int i = 0; i < commandLine.length; i++) {
      buffer.append(commandLine[i]);

      if(i < commandLine.length - 1) {
        buffer.append(" ");
      }
    }

    return buffer.toString();
  }
}
