/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.commons.cli.ParseException;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.util.Log;

/**
 * This is a base class of an engine (tool) that transforms one {@link Entity} entity to another.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class Engine {
  /// The engine identifier.
  private String id;

  /// The input types of the engine.
  private Set<EntityType> inputTypes = new HashSet<>();
  /// The output type of the engine.
  private EntityType outputType;

  // The parameters of the engine.
  private Parameters parameters = new Parameters();

  // The engine event handler.
  private EngineEventHandler eventHandler;

  // The engine backends
  private List<Backend> backends;

  /**
   * Constructs an engine.
   * 
   * @param id the engine identifier.
   */
  public Engine(final String id) {
    this.id = id;
  }

  /**
   * Constructs an engine.
   * 
   * @param id the engine identifier.
   * @param inputType the input type of the engine.
   * @param outputType the output type of the engine.
   */
  public Engine(final String id, final EntityType inputType, final EntityType outputType) {
    this(id);

    this.inputTypes.add(inputType);
    this.outputType = outputType;
  }

  /**
   * Returns the unique engine identifier.
   * 
   * @return the unique engine identifier.
   */
  public final String getId() {
    return id;
  }

  /**
   * Returns the input types of the engine (i.e., the types of input entities).
   * 
   * @return the input types of the engine.
   */
  public final Set<EntityType> getInputTypes() {
    return inputTypes;
  }

  /**
   * Adds the given input type of the engine.
   * 
   * @param inputType the input type to be added.
   */
  public final void addInputType(final EntityType inputType) {
    this.inputTypes.add(inputType);
  }

  /**
   * Checks whether the engine uses an entity of the given type.
   * 
   * @param inputType the input type to be checked.
   * @return <code>true</code> if it uses; <code>false</code> otherwise.
   */
  public final boolean uses(final EntityType inputType) {
    return this.inputTypes.contains(inputType);
  }

  /**
   * Returns the output type of the engine (i.e., the type of output entities).
   * 
   * @return the output type of the engine.
   */
  public final EntityType getOutputType() {
    return outputType;
  }

  /**
   * Sets the given output type of the engine.
   * 
   * @param outputType the output type to be set.
   */
  public final void setOutputType(final EntityType outputType) {
    this.outputType = outputType;
  }

  /**
   * Checks whether the engine defines an entity of the given type.
   * 
   * @param outputType the output type to be checked.
   * @return <code>true</code> if it defines; <code>false</code> otherwise.
   */
  public final boolean defines(final EntityType outputType) {
    return this.outputType.equals(outputType);
  }

  /**
   * Returns the parameters of the engine.
   * 
   * @return the engine parameters.
   */
  public final Parameters getParameters() {
    return parameters;
  }

  /**
   * Adds the given parameter to the engine.
   * 
   * @param parameter the parameter to be added.
   */
  public final void addParameter(final Parameter parameter) {
    parameters.addParameter(parameter);
  }

  /**
   * Parses the command line.
   * 
   * @param args the command line.
   * @return the arguments
   * @throws ParseException the parse exception.
   */
  public final Arguments parseCommandLine(final String[] args) throws ParseException {
    return parameters.parseCommandLine(args);
  }

  /**
   * Returns the logger of the engine.
   * 
   * @return the logger of the engine.
   */
  protected Logger getLogger() {
    return Log.getLogger(this.getClass());
  }

  /**
   * Processes entities of the input types and produces an entity of the output type.
   * 
   * @param inputs the entities to be processed.
   * @return the result of the processing.
   */
  public abstract Entity start(final Map<EntityType, Entity> inputs);

  /**
   * Processes entities of the input types and produces an entity of the output type.
   * While calling this method the engine backends (if exist) will be also executed.
   * 
   * @param inputs the entities to be processed.
   * @param eventHandler the engine event handler.
   * @return the result of the processing.
   */
  public final Entity start(
      final Map<EntityType, Entity> inputs, final EngineEventHandler eventHandler) {
    // Set up the event handler.
    this.eventHandler = eventHandler;

    // Run the main method & backends if they are exist.
    Entity result = start(inputs);
    if (hasBackends()) {
      runBackends(result);
    }

    return result;
  }

  //------------------------------------------------------------------------------------------------
  // Event Handlers
  //------------------------------------------------------------------------------------------------

  /**
   * Sets the engine progress.
   * 
   * @param progress the engine progress (from 0 to 100).
   */
  public final void setProgress(int progress) {
    if (eventHandler != null) {
      eventHandler.onSetProgress(this, progress);
    }
  }

  //------------------------------------------------------------------------------------------------
  // Backends Methods
  //------------------------------------------------------------------------------------------------

  /**
   * Returns <code>true</code> if engine has backends, <code>false</code> otherwise.
   * @return rather engine has backends or not
   */
  public boolean hasBackends() { return this.backends != null && !this.backends.isEmpty(); }

  /**
   * Adds backend.
   *
   * @param backend parser backend
   */
  public void addBackend(Backend backend) {
    InvariantChecks.checkNotNull(backend);
    if (this.backends == null) {
      this.backends = new ArrayList<>();
    }
    this.backends.add(backend);
  }

  /**
   * Runs all the backends that were added to this engine.
   *
   * @param entity Entity to be processed by backends
   */
  public void runBackends(Entity entity) {
    InvariantChecks.checkNotNull(entity);
    if (this.backends == null || this.backends.isEmpty())
      throw new IllegalStateException("No backends are available");

    for (Backend backend : this.backends) {
      backend.start(entity);
    }
  }

  //------------------------------------------------------------------------------------------------
  // Auxiliary Methods
  //------------------------------------------------------------------------------------------------

  @Override
  public String toString() {
    return id;
  }

  @Override
  public boolean equals(Object obj) {
    if (!(obj instanceof Engine)) {
      return false;
    }

    Engine rhs = (Engine) obj;

    return id.equals(rhs.id);
  }

  @Override
  public int hashCode() {
    return id.hashCode();
  }
}
