/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * This class implements an {@link Engine} engine registry. The main function is to compose a
 * {@link ToolChain} tool chain for given targets (entities to be generated).
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class EngineRegistry {

  /** Enables/disables cycle dependencies between engines. */
  public static boolean ENABLE_CYCLE_DEPENDENCIES = true;

  /** All registered engines. */
  private Map<String, Engine> existingEngines = new HashMap<>();

  /** Maps entity types to sets of engines (engines that generate entities of the given type). */
  private Map<EntityType, Set<Engine>> outputToEngines = new HashMap<>();

  /**
   * Returns an engine with the given identifier.
   * 
   * @param id the engine identifier.
   * @return an engine with the given identifier or <code>null</code> if such engine does not exist. 
   */
  public Engine getEngine(final String id) {
    return existingEngines.get(id);
  }

  /**
   * Adds an engine into the configuration.
   * 
   * @param engine the engine to be added.
   */
  public void add(final Engine engine) {
    // Check whether the engine addition produces a cycle dependency.
    if (!ENABLE_CYCLE_DEPENDENCIES && producesCycle(engine)) {
      throw new IllegalArgumentException(
        String.format("Addition of engine '%s' causes a cycle dependency ", engine.getId()));
    }

    // Get the set of engines that generate entities of the given type.
    Set<Engine> engines = outputToEngines.get(engine.getOutputType()); {
      // This is the first engine of the given type.
      if (engines == null) {
        engines = new HashSet<>();
      }

      // Add the given engine to the set of engines.
      engines.add(engine);
    }

    existingEngines.put(engine.getId(), engine);
    outputToEngines.put(engine.getOutputType(), engines);
  }

  /**
   * Returns the collection of all available engines.
   * 
   * @return the collection of engines.
   */
  public Collection<Engine> getEngines() {
    return existingEngines.values();
  }

  /**
   * Returns the collection of all available targets (output types of engines).
   * 
   * @return the collection of targets.
   */
  public Collection<EntityType> getTargets() {
    return outputToEngines.keySet();
  }

  /**
   * Returns an engine (a tool chain) capable to generate targets on the base of the sources.
   * 
   * @param targets the entities to be generated.
   * @param engines the engines to be used.
   * @param sources the given entities.
   * @return the engine to generate targets.
   */
  public ToolChain getEngine(final Set<EntityType> targets, final Set<Engine> engines,
    final Set<EntityType> sources) {

    // Initialize the source propagation map.
    final Map<EntityType, Set<EntityType>> flow = new HashMap<>(); {
      for (final EntityType source : sources) {
        final Set<EntityType> itself = new HashSet<>(); {
          itself.add(source);
        }
        flow.put(source, itself);
      }
    }

    // Modify the set of targets (add output entities of the specified engines).
    final Set<EntityType> newTargets = new HashSet<>(); {
      newTargets.addAll(targets);
      for (final Engine engine : engines) {
        newTargets.add(engine.getOutputType());
      }
    }

    return getEngine(newTargets, engines, sources, flow);
  }

  /**
   * Checks whether the engine addition produces a cycle dependency. The implementation is not
   * effective, but performance is not required.
   * 
   * @param engine the engine to be added.
   */
  private boolean producesCycle(final Engine engine) {
    Set<EntityType> inputs = new HashSet<>(); {
      inputs.addAll(engine.getInputTypes());
    }

    // Check whether the output depends on itself.
    return producesCycle(engine.getOutputType(), inputs);
  }

  /**
   * Checks whether the engine addition produces a cycle dependency. The auxiliary method.
   * 
   * @param output the output of the engine to be added.
   * @param inputs the set of reachable inputs.
   */
  private boolean producesCycle(final EntityType output, final Set<EntityType> inputs) {
    // The output depends on itself and, thus, produces a cycle dependency.
    if (inputs.contains(output)) {
      return true;
    }

    final Set<EntityType> closure = new HashSet<>();

    // For each input determine how it can be generated.
    for (final EntityType input : inputs) {
      final Set<Engine> engines = outputToEngines.get(input);

      if (engines == null) {
        continue;
      }

      for (final Engine engine : engines) {
        closure.addAll(engine.getInputTypes());
      }
    }

    // Return false when the transitive closure is fully constructed.
    return inputs.addAll(closure) && producesCycle(output, inputs);
  }

  /**
   * Propagates the source from the given inputs to the given output. 
   * @param output the output.
   * @param inputs the inputs.
   * @param flow the source propagation map.
   * @return <code>false</code> if the there is a conflict; <code>true</code> otherwise.
   */
  private boolean propagateSources(final EntityType output, final Set<EntityType> inputs,
      final Map<EntityType, Set<EntityType>> flow) {

    // The propagation result (set of sources used to produce the output).
    Set<EntityType> result = flow.get(output);

    if (result == null) {
      result = new HashSet<>();
    }

    // Propagate the sources from the inputs to the output.
    for (final EntityType input : inputs) {
      final Set<EntityType> sources = flow.get(input);

      if (sources == null) {
        continue;
      }

      for (final EntityType source : sources) {
        if (result.contains(source)) {
          // A conflict has been detected (the same source is used to produce the same result).
          return false;
        }

        result.add(source);
      }
    }

    // Modify the source propagation map.
    flow.put(output, result);
    return true;
  }

  /**
   * Returns an engine (a tool chain) capable to generate targets on the base of the sources.
   * 
   * @param targets the entities to be generated.
   * @param engines the engines to be used.
   * @param sources the given entities.
   * @param flow the source propagation map.
   * @return the engine to generate targets.
   */
  private ToolChain getEngine(final Set<EntityType> targets, final Set<Engine> engines,
      final Set<EntityType> sources, final Map<EntityType, Set<EntityType>> flow) {

    // If the set of targets is empty, return the empty tool chain or null.
    if (targets.isEmpty()) {
      // Null is returned if some engines are specified but not used.
      return engines.isEmpty() ? new ToolChain() : null;
    }

    // The loop is used to get one of the targets (at most one iteration is executed).
    for (final EntityType target : targets) {
      // Reduce the set of targets.
      final Set<EntityType> reduced = new HashSet<>(); {
        reduced.addAll(targets);
        reduced.remove(target);
      }

      // If the target belongs to the sources, it is ignored.
      if (sources.contains(target)) {
        // Construct a tool chain for the reduced set of targets.
        return getEngine(reduced, engines, sources, flow);
      } else {
        // Get the engines capable to generate the target.
        final Set<Engine> variants = outputToEngines.get(target);

        // The target is unreachable.
        if (variants == null) {
          return null;
        }

        // Modify the set of sources (add the target to the set of sources).
        final Set<EntityType> newSources = new HashSet<>(); {
          newSources.addAll(sources);
          newSources.add(target);
        }

        // Try different engines to generate the target.
        ToolChain toolChain = null;

        for (final Engine engine : variants) {
          // Get the engine inputs.
          final Set<EntityType> inputs = engine.getInputTypes();

          // Modify the set of targets (replace the target with the engine's inputs).
          final Set<EntityType> newTargets = new HashSet<>(); {
            newTargets.addAll(reduced);
            newTargets.addAll(inputs);
          }

          // Modify the set of engines (remove the engine from the given set).
          final Set<Engine> newEngines = new HashSet<>(); {
            newEngines.addAll(engines);
            newEngines.remove(engine);
          }

          // Create a local copy of the source propagation map.
          final Map<EntityType, Set<EntityType>> newFlow = new HashMap<>(); {
            newFlow.putAll(flow);
          }

          // Try to construct a tool chain for the modified targets and engines.
          ToolChain subChain = getEngine(newTargets, newEngines, newSources, newFlow);

          // If there is a tool chain for the modified set:
          if (subChain != null) {
            // Propagate the sources and checks whether there is a conflict.
            if(propagateSources(target, inputs, newFlow)) {
              // The constructed tool chain in couple with the engine reaches the target.
              subChain.addEngine(engine);
              flow.putAll(newFlow);

              // Before returning the tool chain, try to apply the rest of the engines.
              if (toolChain == null) {
                // This is the first solution.
                toolChain = subChain;
              } else {
                // If the engine does not conflict with the existing solution, it is added.
                toolChain.addEngine(engine);
              }
            }
          }
        } // for engine.

        // Return the constructed tool chain or null if the target is unreachable.
        return toolChain;
      }
    } // for target.

    // The targets are unreachable.
    return null;
  }
}
