/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.basis;

import java.math.BigDecimal;

/**
 * The representation of an abstract test coverage.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public abstract class TestCoverage {
  /**
   * The number representing full coverage percentage, that is, {@code BigDecimal.valueOf(100)}.
   */
  public static final BigDecimal FULL_COVERAGE_PERCENTAGE = BigDecimal.valueOf(100);

  /**
   * Returns a percentage representation of this test coverage with a default precision.
   * Method of computation of this value and metrics is used for it depend on a specific type of
   * this test coverage. It is guaranteed that the value is precise enough to reflect the smallest
   * possible change in a coverage information.
   *
   * @return the percentage of this coverage
   */
  public abstract BigDecimal getPercentage();

  /**
   * Returns a percentage representation of this test coverage with the specified precision.
   * Method of computation of this value and metrics is used for it depend on a specific type of
   * this test coverage.
   *
   * @param precision a number of decimal places in the result
   *
   * @return the percentage of this coverage
   */
  public abstract BigDecimal getPercentage(int precision);

  /**
   * Tests if this coverage is full. Method of testing and metrics is used for it depend on a
   * specific type of this test coverage.
   *
   * @return {@code true} if this coverage is full. Otherwise returns {@code false}
   */
  public abstract boolean isFull();

  /**
   * Resets this coverage to the initial state
   */
  public abstract void reset();

  /**
   * Returns a string representing this coverage.
   *
   * @return the string representation of this coverage
   */
  @Override
  public abstract String toString();
}
