/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorAlgorithm;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.*;
import ru.ispras.retrascope.model.cfg.Process;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * The class containing utility methods for models transformation.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class TransformUtils {

  /**
   * Performs backward substitution of the specified list of blocks
   * to the specified node expression.
   *
   * @param blocks blocks collection
   * @param node expression to be transformed
   * @return the specified expression where all the variables, that are defined in
   * at least one block from the specified collection, will be substituted by the
   * expressions that are assigned to them.
   */
  public static Node substituteBackward(List<BasicBlock> blocks, Node node) {

    Node result = node.deepCopy();

    /* Iterate blocks & assignments in backward direction. */
    for (int i = blocks.size() - 1; i >= 0; i--) {
      List<Assignment> assignments = blocks.get(i).getAssignments();

      for (int j = assignments.size() - 1; j >= 0; j--) {

        /* Get assignment fields. */
        Assignment assignment = assignments.get(j);
        RangedVariable rVariable = assignment.getRangedVariable();
        NodeVariable variable = rVariable.getVariable();
        Node expression = assignment.getExpression();

        /* Get the variables that are used in target expression. */
        Collection<NodeVariable> variables = ExprUtils.getVariables(result);

        if (variables.contains(variable)) {

          /* Ranged variable case. */
          if (rVariable.isRanged()) {

            Node young = rVariable.getRange().getYoung();
            Node old = rVariable.getRange().getOld();

            /* Elaborate this case depending on the data type. */
            if (variable.isType(DataTypeId.BIT_VECTOR)) {

              /* Select operation identifiers & constants depending on operand types. */
              Enum<?> equalOp, greaterOp, lessOp, addOp, subOp;
              NodeValue one, zero, max;

              if (young.isType(DataType.INTEGER) && old.isType(DataType.INTEGER)) {
                equalOp = StandardOperation.EQ;
                greaterOp = StandardOperation.GREATER;
                lessOp = StandardOperation.LESS;
                addOp = StandardOperation.ADD;
                subOp = StandardOperation.SUB;

                zero = NodeValue.newInteger(0);
                one = NodeValue.newInteger(1);
                max = NodeValue.newInteger(variable.getDataType().getSize() - 1);

              } else if (young.isType(DataTypeId.BIT_VECTOR) && old.isType(DataTypeId.BIT_VECTOR)) {
                equalOp = StandardOperation.EQ;
                greaterOp = StandardOperation.BVUGT;
                lessOp = StandardOperation.BVULT;
                addOp = StandardOperation.BVADD;
                subOp = StandardOperation.BVSUB;

                zero = NodeValue.newBitVector(BitVector.valueOf(0, young.getDataType().getSize()));
                one = NodeValue.newBitVector(BitVector.valueOf(1, young.getDataType().getSize()));
                BitVector maxVector = BitVector.newEmpty(young.getDataType().getSize());
                BitVectorAlgorithm.fill(maxVector, (byte)1);
                max = NodeValue.newBitVector(maxVector);
              } else throw new UnsupportedOperationException(
                  String.format(
                      "Data types: young=%s, old=%s.", young.getDataType(), old.getDataType()));

              /* The bit vector case: create an every cases-covering expression. */

              /* young == 0 && old == max */
              NodeOperation condition1 =
                  new NodeOperation(
                      StandardOperation.AND,
                      new NodeOperation(equalOp, young, zero),
                      new NodeOperation(equalOp, old, max));

              /* young == 0 && old < max */
              NodeOperation condition2 =
                  new NodeOperation(
                      StandardOperation.AND,
                      new NodeOperation(equalOp, young, zero),
                      new NodeOperation(lessOp, old, max));

              /* young > 0 && old == max */
              NodeOperation condition3 =
                  new NodeOperation(
                      StandardOperation.AND,
                      new NodeOperation(greaterOp, young, zero),
                      new NodeOperation(equalOp, old, max));

              /* Check rather above conditions are satisfiable (using SAT solver). */
              boolean checkCondition1 = ExprUtils.isSAT(condition1);
              boolean checkCondition2 = ExprUtils.isSAT(condition2);
              boolean checkCondition3 = ExprUtils.isSAT(condition3);

              /* The formulas to which the bit vector ranged variable will be transformed
              * when the corresponding condition (and only this condition) is SAT. */
              Node condition2ThenCase = new NodeOperation(
                  StandardOperation.BVCONCAT,
                  new NodeOperation(
                      StandardOperation.BVEXTRACT,
                      max,
                      new NodeOperation(addOp, old, one),
                      variable),
                      expression);
              Node condition3ThenCase = new NodeOperation(
                  StandardOperation.BVCONCAT,
                      expression,
                  new NodeOperation(
                      StandardOperation.BVEXTRACT,
                      new NodeOperation(subOp, young, one),
                      zero,
                      variable));

              /* The case when no condition above is true. */
              Node elseCase = new NodeOperation(
                  StandardOperation.BVCONCAT,
                  new NodeOperation(
                      StandardOperation.BVEXTRACT,
                      max,
                      new NodeOperation(addOp, old, one),
                      variable),
                  expression,
                  new NodeOperation(
                      StandardOperation.BVEXTRACT,
                      new NodeOperation(subOp, young, one),
                      zero,
                      variable));

              Node transform;

              if (checkCondition1 && !checkCondition2 && !checkCondition3) {
                transform = expression;
              } else if (!checkCondition1 && checkCondition2 && !checkCondition3) {
                transform = condition2ThenCase;
              } else if (!checkCondition1 && !checkCondition2 && checkCondition3) {
                transform = condition3ThenCase;
              } else if (!checkCondition1 && !checkCondition2) {
                transform = elseCase;
              } else {

                /* The most general case (for variable ranges)
                 * when two or more conditions become true.*/
                transform =
                    new NodeOperation(
                        StandardOperation.ITE,
                        condition1,
                        variable,
                        new NodeOperation(
                            StandardOperation.ITE,
                            condition2,
                            condition2ThenCase,
                            new NodeOperation(
                                StandardOperation.ITE,
                                condition3,
                                condition3ThenCase,
                                elseCase)));
              }
              result = Transformer.substitute(result, variable.getName(), transform);

            } else if (variable.isType(DataTypeId.MAP)) {

              /* The array case: substitute x by (STORE x i expr). */
              result =
                  Transformer.substitute(
                      result,
                      variable.getName(),
                      new NodeOperation(StandardOperation.STORE, variable, young, expression));

            } else
                throw new IllegalArgumentException(
                    String.format("Unsupported data type id: %s.", variable.getDataTypeId()));
          } else {
            /* Non-ranged variable case. */
            result = Transformer.substitute(result, variable.getName(), expression);
          }
        }
      }
    }

    /* Call the reducer to simplify the substitution result. */
    return Transformer.reduce(ReduceOptions.NEW_INSTANCE, result);
  }

  /**
   * Constructs a sequence of connected nodes
   * between the specified begin and end nodes.
   *
   * Constructs a nodes sequence, to which control flow goes through
   * the specified start node and from which it goes through the specified end node.
   * The path does not contain copies of the specified begin & end nodes.
   *
   * @param begin the before-first node of the path
   * @param end the after-last node of the path
   *
   * @return sub-graph
   */
  public static List<CfgNode> getPath(final CfgNode begin, final CfgNode end) {

    class CfgNodeCollector implements CfgVisitor {

      private Status status;

      private List<CfgNode> nodes;

      public CfgNodeCollector() {
        this.nodes = new ArrayList<>();
        this.status = Status.OK;
      }

      public void setStatus(Status status) { this.status = status; }

      @Override
      public Status getStatus() { return this.status; }

      @Override
      public void onRootBegin() { /* Empty. */ }

      @Override
      public void onRootEnd() { /* Empty. */ }

      @Override
      public void onCfgModelBegin() { /* Empty. */ }

      @Override
      public void onCfgModelEnd() { /* Empty. */ }

      @Override
      public void onModuleBegin(Module module) { /* Empty. */ }

      @Override
      public void onModuleEnd(Module module) { /* Empty. */ }

      @Override
      public void onProcessBegin(Process process) { /* Empty. */ }

      @Override
      public void onProcessEnd(Process process) { /* Empty. */ }

      @Override
      public void onCfgBegin(Cfg cfg) { /* Empty. */ }

      @Override
      public void onCfgEnd(Cfg cfg) { /* Empty. */ }

      @Override
      public void onBasicBlockBegin(BasicBlock basicBlock) { tryAdd(basicBlock); }

      @Override
      public void onBasicBlockEnd(BasicBlock basicBlock) { resetStatus(); }

      @Override
      public void onCaseBegin(Case aCase) { tryAdd(aCase); }

      @Override
      public void onCaseEnd(Case aCase) { resetStatus(); }

      @Override
      public void onSwitchBegin(Switch aSwitch) { tryAdd(aSwitch); }

      @Override
      public void onSwitchEnd(Switch aSwitch) { resetStatus(); }

      @Override
      public void onMergeBegin(Merge merge) { tryAdd(merge); }

      @Override
      public void onMergeEnd(Merge merge) { resetStatus(); }

      @Override
      public void onSrcBegin(Source source) { /* Empty. */ }

      @Override
      public void onSrcEnd(Source source) { /* Empty. */ }

      @Override
      public void onSink(Sink sink) { /* Empty. */ }

      @Override
      public void onWaitBegin(Wait wait) { /* Empty. */ }

      @Override
      public void onWaitEnd(Wait wait) { /* Empty. */ }

      private void resetStatus() {
        if (getStatus().equals(Status.SKIP)) {
          setStatus(Status.OK);
        }
      }

      private void tryAdd(CfgNode node) {
        if (node.equals(end)) {
          setStatus(Status.ABORT);
        }

        if (!this.nodes.contains(node) && !node.equals(begin) && !node.equals(end)) {
          this.nodes.add(node);
        }
      }
    }

    final CfgNodeCollector collector = new CfgNodeCollector();
    final CfgWalker walker = new CfgWalker(collector, CfgWalkerMode.DFS);

    walker.visitNode(begin);

    /* Copy nodes. */
    final Map<CfgNode, CfgNode> oldNewMap = new LinkedHashMap<>();

    final List<CfgNode> newNodes = new ArrayList<>();

    for (CfgNode node : collector.nodes) {
      CfgNode copy = (CfgNode) node.deepcopy();
      newNodes.add(copy);
      oldNewMap.put(node, copy);
    }

    /* Copy links between them. */

    for (int i = 0; i < collector.nodes.size(); i++) {

      CfgNode iNode = collector.nodes.get(i);

      for (int j = 0; j < collector.nodes.size(); j++) {
        CfgNode jNode = collector.nodes.get(j);

        if (iNode.getChildren().contains(jNode)) {
          oldNewMap.get(iNode).addChild(oldNewMap.get(jNode));
        }
      }
    }

    return newNodes;
  }

  /**
   * Substitutes usages of variable with the specified {@link java.lang.String} name
   * by the specified {@link ru.ispras.fortress.expression.Node} expression
   * in all the elements of the specified collection.
   *
   * @param list collection which nodes will be copied & transformed
   * @param name variable name
   * @param node expression
   *
   * @return a collection of node copies of the specified collection; in which node
   * the every usage of the variable with the specified name will be substituted
   * by the specified expression
   */
  public static List<CfgNode> substitute(final List<CfgNode> list, final String name,
    final Node node) {

    final int listSize = list.size();
    final List<CfgNode> result = new ArrayList<>(listSize);
    final Map<CfgNode, CfgNode> map = new LinkedHashMap<>(listSize);

    /* Substitute variable with the specified name by the specified node. */
    for (CfgNode elem : list) {
      CfgNode copy;

      if (elem instanceof Switch) {
        copy = new Switch(Transformer.substitute(((Switch) elem).getCondition(), name, node));

      } else if (elem instanceof BasicBlock) {

        final List<Assignment> elemAssignments = ((BasicBlock) elem).getAssignments();
        final List<Assignment> assignments = new ArrayList<>(elemAssignments.size());

        for (Assignment elemAssign : elemAssignments) {
          final Node value = elemAssign.getExpression();
          final RangedVariable target = elemAssign.getRangedVariable();
          Assignment copyAssignment;
          final Node copyValue = Transformer.substitute(value, name, node);
          if (target.isRanged()) {
            final Node old = target.getRange().getOld();
            final Node young = target.getRange().getYoung();

            final Range copyRange =
                new Range(
                    Transformer.reduce(ReduceOptions.NEW_INSTANCE, Transformer.substitute(young, name, node)),
                    Transformer.reduce(ReduceOptions.NEW_INSTANCE, Transformer.substitute(old, name, node)));
            copyAssignment =
                new Assignment(new RangedVariable(target.getVariable(), copyRange), copyValue);
          } else {
            copyAssignment = new Assignment(target.getVariable(), copyValue);
          }
          assignments.add(copyAssignment);
        }

        copy = new BasicBlock(assignments, ((BasicBlock) elem).isConcurrent());
      } else {
        copy = (CfgNode) elem.deepcopy();
      }
      map.put(elem, copy);
      result.add(copy);
    }

    /* Establish parent-child links between nodes. */
    for (int i = 0; i < listSize; i++) {
      CfgNode iNode = list.get(i);
      for (CfgNode jNode : list) {
        if (iNode.hasChildren() && iNode.getChildren().contains(jNode)) {
          map.get(iNode).addChild(map.get(jNode));
        }
      }
    }

    return result;
  }
}
