/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.CfgWalker;
import ru.ispras.retrascope.model.cfg.CfgWalkerMode;
import ru.ispras.retrascope.model.cfg.ModelProduceCfgVisitor;

import java.util.Map;

/**
 * Abstract engine for control flow graph model elaboration.
 *
 * Runs walker subsequently with different visitors and in different modes
 * on the control flow graph model.
 *
 * Note: while executed, the visitors & modes will be used
 * in the same order they appeared in constructor.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgEngine extends Engine {

  /* The model visitor configurations. */
  private final CfgVisitorConfig[] configs;

  /**
   * Constructs object with the specified identifier, input/output types
   * and visitors.
   *
   * All the defined visitors will be run on the walker with default
   * walking mode {@link ru.ispras.retrascope.model.cfg.CfgWalkerMode}.
   *
   * @param id engine identifier
   * @param inputType engine input type
   * @param outputType engine output type
   * @param visitors control flow graph visitors
   *
   * @throws java.lang.NullPointerException when the specified visitors array is <code>null</code>.
   */
  public CfgEngine(
      String id,
      EntityType inputType,
      EntityType outputType,
      CfgVisitor ... visitors) {
    super(id);

    InvariantChecks.checkNotNull(visitors);

    this.addInputType(inputType);
    this.setOutputType(outputType);
    this.configs = new CfgVisitorConfig[visitors.length];
    for (int i = 0; i < visitors.length; i++) {
      this.configs[i] = new CfgVisitorConfig(visitors[i], CfgWalkerMode.getDefaultMode());
    }
 }

  /**
   * Constructs object with the specified identifier, input/output types,
   * visitors configurations.
   *
   * @param id engine identifier
   * @param inputType engine input type
   * @param outputType engine output type
   * @param configs visitors configurations
   *
   */
  public CfgEngine(
      String id,
      EntityType inputType,
      EntityType outputType,
      CfgVisitorConfig ... configs) {
    super(id);

    InvariantChecks.checkNotNull(configs);

    this.addInputType(inputType);
    this.setOutputType(outputType);

    this.configs = configs;
  }

  /**
   * Initializes engine with the specified inputs.
   *
   * @param inputs Engine inputs
   */
  public abstract void initialize(Map<EntityType, Entity> inputs);

  @Override
  public Entity start(Map<EntityType, Entity> inputs) {

    initialize(inputs);

    CfgModel cfgModel = (CfgModel) inputs.get(CfgModel.TYPE);

    for (CfgVisitorConfig config : this.configs) {
      CfgVisitor visitor = config.getVisitor();
      CfgWalker walker = new CfgWalker(visitor, config.getWalkerMode());
      walker.start(cfgModel);

      //TODO: refactor it
      if (visitor instanceof ModelProduceCfgVisitor) {
        cfgModel = ((ModelProduceCfgVisitor) visitor).produceModel();
      }
    }
    return getOutput();
  }

  /**
   * Returns engine output entity.
   *
   * @return Output entity
   */
  public abstract Entity getOutput();

  /**
   * Returns visitors configurations.
   *
   * @return visitors configurations
   */
  public final CfgVisitorConfig[] getConfigs() { return this.configs; }

  /**
   * Returns initial visitor.
   *
   * @return visitor
   */
  public final CfgVisitor getInitVisitor() { return this.configs[0].getVisitor(); }
}
