/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.printer.graphml;

import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;
import ru.ispras.retrascope.result.graphml.GraphMlUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Collection;

/**
 * Control flow graph visitor for GraphML printer.
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class CfgGraphMlVisitor implements CfgVisitor {

  // print-to-GraphMl-file writer
  private PrintWriter printWriter;

  // expression printer
  private ExprTreePrinter exprPrinter;

  /**
   * Constructs a CFG visitor.
   */
  public CfgGraphMlVisitor() {}

  /**
   * Initializes object with the specified file name.
   * 
   * @param fileName GraphML file name
   */
  public void initialize(String fileName) {

    try {
      // Create a GraphML file.
      File graphmlFile = FileCreator.newFile(fileName);

      FileWriter fileWriter = new FileWriter(graphmlFile.getAbsoluteFile());
      printWriter = new PrintWriter(new BufferedWriter(fileWriter));
    } catch (Exception e) {
      e.printStackTrace();
    }
    // expression printer initialization
    this.exprPrinter = Retrascope.getConfiguration().getStylePrinter();
  }

  @Override
  public Status getStatus() { return Status.OK; }

  /**
   * Prints the GraphML title tag to the GraphML file.
   */
  public void onRootBegin() { GraphMlUtils.onGraphMlBegin(printWriter); }

  /**
   * Prints the GraphML close tag to the GraphML file.
   */
  public void onRootEnd() {
    GraphMlUtils.onGraphMlEnd(printWriter);
    GraphMlUtils.closeGraphMlFile(printWriter);
  }

  @Override
  public void onCfgModelBegin() { GraphMlUtils.onGraphBegin(printWriter); }

  @Override
  public void onCfgModelEnd() { GraphMlUtils.onGraphEnd(printWriter); }

  /**
   * Prints the node, gets the list of parents for this node and prints the edges.
   * 
   * @param child the CFG model node.
   */
  private void printNodeBegin(CfgModelNode child) {
    printNodeBegin(child, GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_PURPLE);
  }

  /**
   * Prints the node, gets the list of parents for this node and prints the edges.
   * 
   * @param child the CFG model node.
   * @param form the node form.
   * @param color the node color.
   */
  private void printNodeBegin(CfgModelNode child, String form, String color) {

    // Add the node to the GraphML file.
    GraphMlUtils.printNode(printWriter, child.getId(), child.getDescription(this.exprPrinter),
        form, color);

    // Add the edges to the GraphML file.
    if (child.hasChildren()) {
      Collection<CfgModelNode> children = child.getChildren();
      for (CfgModelNode nodeChild : children) {

        // Case nodes will not be added to GraphML
        if (nodeChild.getType() == CfgNodeType.CASE) {
          Collection<CfgModelNode> caseChildCollection = nodeChild.getChildren();
          if (caseChildCollection.size() == 1) {
            CfgModelNode caseChild = (CfgModelNode) caseChildCollection.toArray()[0];
            GraphMlUtils.printEdge(printWriter, child.getId(),
                nodeChild.getDescription(this.exprPrinter), caseChild.getId());
          } else {
            Log.getLogger().log(LogLevel.ERROR,
                String.format("Case node must have only one child. Node id: %s.", child.getId()));
          }
        } else {
          GraphMlUtils.printEdge(printWriter, child.getId(), nodeChild.getId());
        }
      }
    }
  }

  @Override
  public void onModuleBegin(Module module) {
    printNodeBegin(module, GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_YELLOW);
  }

  @Override
  public void onModuleEnd(Module module) { /* Empty. */ }

  @Override
  public void onProcessBegin(Process process) {
    // Add the node to the GraphML file.
    GraphMlUtils.printNode(printWriter, process.getId(), process.getDescription(this.exprPrinter),
        GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_BLUE);

    CfgModelNode cfgNode = (CfgModelNode) process.getChildren().toArray()[0];
    if (process.getChildren().size() != 1)
      Log.getLogger()
          .log(
              LogLevel.ERROR,
              String.format("Process node must " + "have only one child. Node id: %s.",
                  process.getId()));
    CfgModelNode source = (CfgModelNode) cfgNode.getChildren().toArray()[0];
    if (cfgNode.getChildren().size() != 1)
      Log.getLogger().log(LogLevel.ERROR,
          String.format("Cfg node must have " + "only one child. Node id: %s.", process.getId()));
    CfgModelNode node = (CfgModelNode) source.getChildren().toArray()[0];
    if (source.getChildren().size() != 1)
      Log.getLogger()
          .log(
              LogLevel.ERROR,
              String.format("Source node must have " + "only one child. Node id: %s.",
                  process.getId()));

    GraphMlUtils.printEdge(printWriter, process.getId(), node.getId());
  }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { /* Empty. */ }

  @Override
  public void onCfgEnd(Cfg cfg) { /* Empty. */ }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { printNodeBegin(basicBlock); }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) {
    printNodeBegin(aSwitch, GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_RED);
  }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) {
    printNodeBegin(merge, GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_WHITE);
  }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) {
    printNodeBegin(sink, GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_YELLOW);
  }

  @Override
  public void onWaitBegin(Wait wait) {
    printNodeBegin(wait, GraphMlUtils.FORM_OCTAGON, GraphMlUtils.COLOR_ORANGE);
  }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
