/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.engine.basis.TransformUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

/**
 * Clocked guarded atomic actions decision diagram builder
 * for {@link ru.ispras.retrascope.engine.cfg.transformer.cgaa.CfgCgaaTransformer} transformer.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgCgaaTransformerBuilder {

  /* The mapping of module to transformations. */
  private final Map<Module, Map<Process, List<CfgTransformation>>> moduleMap;

  /**
   * Constructs object with the specified mapping.
   *
   * @param moduleMap case -> basic blocks collection
   *
   * @throws java.lang.NullPointerException when argument is <code>null</code>.
   */
  public CfgCgaaTransformerBuilder(Map<Module, Map<Process, List<CfgTransformation>>> moduleMap) {
    InvariantChecks.checkNotNull(moduleMap);

    this.moduleMap = moduleMap;
  }

  /**
   * Builds guarded actions decision diagram
   * from control flow graph model {@link ru.ispras.retrascope.model.cfg.CfgModel} representation.
   *
   * @return guarded actions decision diagram.
   */
  public CfgModel build() {
    CfgModel model = new CfgModel();

    for (Module module : this.moduleMap.keySet()) {
      buildModule(model, module);
    }
    return model;
  }

  private void buildModule(CfgModel model, Module module) {

    /* The mapping for "old" case nodes and linked sequences of block nodes. */
    Map<Case, List<BasicBlock>> terminations = new LinkedHashMap<>();

      /* The mapping for old cases and new cases. */
    Map<Case, List<Case>> caseMap = new LinkedHashMap<>();

      /* The module container. */
    Module newModule = new Module(module.getName());
    newModule.setVariablesMapping(module.getVariablesMapping());
    model.addModule(newModule);

      /* The transformations of the current module's process. */
    Map<Process, List<CfgTransformation>> ptList = this.moduleMap.get(module);
    applyTransformations(terminations, caseMap, newModule, ptList);
  }

  private void applyTransformations(
      Map<Case, List<BasicBlock>> terminations,
      Map<Case, List<Case>> caseMap,
      Module module,
      Map<Process, List<CfgTransformation>> ptMap) {

    for (Process ptProcess : ptMap.keySet()) {

      /* The process container. */
      Process process = new Process();
      process.setSensitivityList(ptProcess.getSensitivityList());
      process.setVariablesMapping(ptProcess.getVariablesMapping());
      module.addChild(process);

      /* The process control flow graph that contains a decision diagram. */
      Cfg cfg = new Cfg();
      process.addChild(cfg);

      /* Build the control flow graph from the list of transformations. */
      buildCfgControlPart(cfg, terminations, caseMap, ptMap.get(ptProcess));

      /* Continue terminal case nodes with sequences of basic blocks. */
      buildCfgDataPart(cfg, terminations);
    }
  }

  private void buildCfgDataPart(Cfg cfg, Map<Case, List<BasicBlock>> terminations) {

    for (Case newCase : terminations.keySet()) {
      if (!newCase.hasChildren()) {
        List<BasicBlock> termBlocks = terminations.get(newCase);
        addBlocksAfterCase(cfg, newCase, termBlocks);
      }
    }
  }

  private void addBlocksAfterCase(Cfg cfg, Case newCase, List<BasicBlock> termBlocks) {

    CfgNode sink = cfg.getSink();

    CfgNode current = null;
    if (termBlocks.isEmpty()) {
      if (!sink.hasParents()) {
        Merge merge = new Merge();
        cfg.addNode(merge);
        merge.addChild(sink);
        newCase.addChild(merge);
      } else {
        CfgModelNode sinkParent = sink.getOnlyParent();

        assert(sinkParent instanceof Merge);

        newCase.addChild(sinkParent);
      }
    } else {
      for (int i = 0; i < termBlocks.size(); i++) {
        CfgNode newBlock;
        BasicBlock termBlock = termBlocks.get(i);

        /* Create a deep copy of terminal block. */
        List<Assignment> newAssignments = new ArrayList<>();
        List<Assignment> termAssignments = termBlock.getAssignments();
        for (Assignment assignment : termAssignments) {
          newAssignments.add(assignment.deepcopy());
        }
        newBlock = new BasicBlock(newAssignments, termBlock.isConcurrent());

        cfg.addNode(newBlock);

        newBlock.addParent((i == 0) ? newCase : current);

        current = newBlock;

        if (i == termBlocks.size() - 1) {

          if (!sink.hasParents()) {
            Merge merge = new Merge();
            cfg.addNode(merge);
            sink.addParent(merge);
            newBlock.addChild(merge);
          } else {
            CfgModelNode sinkParent = sink.getOnlyParent();
            if (sinkParent instanceof Merge) {
              newBlock.addChild(sinkParent);
            } else
              throw new IllegalStateException(
                String.format("Incorrect type: %s.", sinkParent.getType()));
          }
        }
      }
    }
  }

  private void buildCfgControlPart(Cfg cfg,
    Map<Case, List<BasicBlock>> terminations,
    Map<Case, List<Case>> caseMap,
    List<CfgTransformation> transformations) {

    /* Get source node. */
    CfgNode src = cfg.getSource();

    for (CfgTransformation transform : transformations) {

      Switch oldSwitch = transform.getSwitchNode();
      Stack<Case> tCases = transform.getCases();
      List<BasicBlock> tBlocks = transform.getBlocks();

      if (!transform.hasSwitch()) {

        /* The basic block-only control flow graph - no transformation needed. */

        /* Remove all the connections with source's children. */
        Collection<CfgModelNode> srcChildren = src.getChildren();
        for (CfgModelNode child : srcChildren) {
          src.removeChild(child);
        }

        CfgNode current = src;

        assert (!tBlocks.isEmpty());

        for (BasicBlock tBlock : tBlocks) {
          CfgNode newBlock = new BasicBlock(tBlock.getAssignments(), tBlock.isConcurrent());
          cfg.addNode(newBlock);
          newBlock.addParent(current);
          current = newBlock;
        }
        current.addChild(cfg.getSink());
      }
      else {

        /* Create new switch with the modified expression. */

        Switch newSwitch;
        if (oldSwitch.getEvent() == null) {
          newSwitch = new Switch(
            TransformUtils.substituteBackward(tBlocks, oldSwitch.getCondition().deepCopy()));
        } else {
          newSwitch = new Switch(oldSwitch.getEvent());
        }
        cfg.addNode(newSwitch);

        /* Add all the cases of old switch to terminals and create new case nodes. */
        Collection<CfgModelNode> oldSwitchChildren = oldSwitch.getChildren();
        for (CfgModelNode oldCase : oldSwitchChildren) {

          /* Create a new child case node & add it to the control flow graph. */
          Case newCase = new Case(makeValue(oldSwitch, (Case)oldCase));
          cfg.addNode(newCase);
          newSwitch.addChild(newCase);

          /* Add new case nodes to old-new mapping. */
          if (!caseMap.keySet().contains(oldCase)) {
            caseMap.put((Case) oldCase, new ArrayList<>(Collections.singleton(newCase)));
          } else {
            caseMap.get(oldCase).add(newCase);
          }
        }

        /* Determine the parent node for the node switch. */
        if (tCases.isEmpty()) {

          /* Remove source's children. */
          Collection<CfgModelNode> srcChildren = src.getChildren();
          for (CfgModelNode srcChild : srcChildren) {
            srcChild.removeParent(src);
          }

          /* Add new switch as child for source. */
          newSwitch.addParent(src);
        } else {

          /* The new incarnation of old case is already created. */
          Case tCasesPeek = tCases.peek();

          /* Check rather mapping contains old case. */
          assert(caseMap.containsKey(tCasesPeek));

          List<Case> terminalCases = caseMap.get(tCasesPeek);

          /* Get cases which does not have children. */
          List<Case> noChildrenCases = new ArrayList<>();
          for (Case termCase : terminalCases) {
            if (!termCase.hasChildren()) {
              noChildrenCases.add(termCase);
            }
          }

          /* No-child-having cases should be the parents of the new switch node. */
          if (noChildrenCases.size() == 1) {
            Case terminalCase = noChildrenCases.get(0);
            newSwitch.addParent(terminalCase);
          } else if (noChildrenCases.size() > 1) {
            Merge merge = new Merge();
            cfg.addNode(merge);
            newSwitch.addParent(merge);
            for (Case noChildren : noChildrenCases) {
              noChildren.addChild(merge);
            }
          }
        }
      }
      if (transform.hasMap()) {
        Map<Case, List<BasicBlock>> transformMap = transform.getMap();

        for (Case terminalCase : transformMap.keySet()) {
          List<Case> newCases = caseMap.get(terminalCase);

          for (Case newCase : newCases) {
            if (!newCase.hasChildren() && !terminations.containsKey(newCase)) {
              terminations.put(newCase, transformMap.get(terminalCase));
            }
          }
        }
      }
    }
  }

  private Set<NodeValue> makeValue(Switch aSwitch, Case aCase) {
    Node value = null;

    if (aCase.isDefault()) {
      Collection<CfgModelNode> children = aSwitch.getChildren();
      for (CfgModelNode child : children) {
        if (!((Case)child).isDefault()) {
          if (value == null) {
            Set<NodeValue> childValues = ((Case) child).getValues();
            for (NodeValue childValue : childValues) {
              if (value == null) {
                value = new NodeOperation(StandardOperation.NOT, childValue);
              } else {
                value = new NodeOperation(
                    StandardOperation.AND,
                    value,
                    new NodeOperation(StandardOperation.NOT, childValue));
              }
            }
          } else {
            Set<NodeValue> childValues = ((Case) child).getValues();
            for (NodeValue childValue : childValues) {
              value = new NodeOperation(
                  StandardOperation.AND,
                  value,
                  new NodeOperation(StandardOperation.NOT, childValue));
            }
          }
        }
      }
    } else {
      return aCase.getValues();
    }
    if (value != null) {
      Node reduced = Transformer.reduce(ReduceOptions.NEW_INSTANCE, value);

      assert(reduced instanceof NodeValue);

      Set<NodeValue> set = new LinkedHashSet<>();
      set.add((NodeValue) reduced);
      return set;
    } else
      throw new NullPointerException();
  }
}

