/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

/**
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgCgaaTransformerVisitor implements CfgVisitor {

  /* The current module. */
  private final Stack<Module> modules = new Stack<>();

  /* The current basic blocks collection. */
  private final List<BasicBlock> blocks;

  /* The current cases stack. */
  private final Stack<Case> cases;

  /* The current process transformation. */
  private List<CfgTransformation> transformations;

  /* The mapping of module data to transformations. */
  private final Map<Module, Map<Process, List<CfgTransformation>>> moduleMap = new LinkedHashMap<>();

  /**
   * Constructs the visitor for the {@link ru.ispras.retrascope.engine.cfg.transformer.cgaa.CfgCgaaTransformer} engine.
   */
  public CfgCgaaTransformerVisitor() {
    this.blocks = new ArrayList<>();
    this.cases = new Stack<>();
  }

  /**
   * Returns mapping that keeps the transformed control flow graph model.
   *
   * @return transformed control flow graph model
   */
  public Map<Module, Map<Process, List<CfgTransformation>>> getModuleMap() { return this.moduleMap; }

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) {
    this.moduleMap.put(module, new LinkedHashMap<Process, List<CfgTransformation>>());
    this.modules.push(module);
  }

  @Override
  public void onModuleEnd(Module module) {
    this.modules.pop();
  }

  @Override
  public void onProcessBegin(Process process) { this.transformations = new ArrayList<>(); }

  @Override
  public void onProcessEnd(Process process) {
    this.moduleMap.get(this.modules.peek()).put(process, this.transformations);
  }

  @Override
  public void onCfgBegin(Cfg cfg) { /* Empty. */ }

  @Override
  public void onCfgEnd(Cfg cfg) { /* Empty. */ }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) {

    if (this.blocks.contains(basicBlock)) {
      Log.getLogger().log(LogLevel.WARNING, "Loop is found");
    }
    this.blocks.add(basicBlock);
  }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) {

    /* Remove the last block. */
    int lastBlockNum = this.blocks.size() - 1;
    this.blocks.remove(lastBlockNum);
  }

  @Override
  public void onCaseBegin(Case aCase) {
    if (this.cases.contains(aCase)) {
      Log.getLogger().log(LogLevel.WARNING, "Loop is found");
    }
    this.cases.push(aCase);
  }

  @Override
  public void onCaseEnd(Case aCase) { this.cases.pop(); }

  @Override
  public void onSwitchBegin(Switch aSwitch) { addTransformation(this.blocks, this.cases, aSwitch); }

  private void addTransformation(List<BasicBlock> blocks, Stack<Case> cases, Switch aSwitch) {
    this.transformations.add(
      new CfgTransformation(cases, blocks, aSwitch, new LinkedHashMap<Case, List<BasicBlock>>()));
  }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) { /* Empty. */ }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) {

    if (!this.cases.isEmpty()) {

      /* Get last control flow graph transformation. */
      int lastProcessTrSize = this.transformations.size();
      CfgTransformation tr = this.transformations.get(lastProcessTrSize - 1);

      /* Create a copy of current blocks list. */
      List<BasicBlock> blocks = new ArrayList<>();
      blocks.addAll(this.blocks);

      tr.addTermination(this.cases.peek(), blocks);
    } else {
      /* There is no switch\case nodes. */
      this.transformations.add(new CfgTransformation(this.cases, this.blocks, null, null));
    }
  }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
