/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

/**
 * Control flow graph transformation.
 *
 * This keep information about transformation
 * of control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} model transformation.
 * The guarded actions decision diagram builder {@link CfgCgaaTransformerBuilder}
 * forms this as intermediate step of it's work.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgTransformation {
  private Stack<Case> cases; // case nodes
  private List<BasicBlock> blocks; // block nodes
  private Switch switchNode; // current switch node
  private Map<Case, List<BasicBlock>> map; // map for case node and following block nodes

  /**
   * Constructs object with the specified case {@link ru.ispras.retrascope.model.cfg.Case}
   * and block  nodes, switch {@link ru.ispras.retrascope.model.cfg.Switch} node and mapping.
   *
   * @param cases case nodes
   * @param blocks block nodes
   * @param aSwitch switch node
   * @param map mapping
   */
  public CfgTransformation(Stack<Case> cases, List<BasicBlock> blocks,
      Switch aSwitch, Map<Case, List<BasicBlock>> map) {
    this.cases = new Stack<>();
    this.cases.addAll(cases);
    this.blocks = new ArrayList<>();
    this.blocks.addAll(blocks);
    this.switchNode = aSwitch;
    this.map = map;
  }

  /**
   * Returns case {@link ru.ispras.retrascope.model.cfg.Case} nodes stack.
   *
   * @return cases stack
   */
  public Stack<Case> getCases() { return this.cases; }

  /**
   * Returns block nodes.
   *
   * Block nodes can be of different type.
   * Their main property is: they have only one parent node and only one child node.
   *
   * @return basic blocks stack
   */
  public List<BasicBlock> getBlocks() { return this.blocks; }

  /**
   * Returns switch {@link ru.ispras.retrascope.model.cfg.Switch} node.
   *
   * @return switch node
   */
  public Switch getSwitchNode() { return this.switchNode; }

  /**
   * Returns mapping of case {@link ru.ispras.retrascope.model.cfg.Case}
   * and {@link ru.ispras.retrascope.model.cfg.BasicBlock} basic block nodes.
   *
   * @return cases/basic block nodes mapping
   */
  public Map<Case, List<BasicBlock>> getMap() { return this.map; }

  /**
   * Adds terminal case {@link ru.ispras.retrascope.model.cfg.Case}
   * and {@link ru.ispras.retrascope.model.cfg.BasicBlock} basic block nodes.
   *
   * @param aCase case node
   * @param blocks basic blocks
   */
  public void addTermination(Case aCase, List<BasicBlock> blocks) { this.map.put(aCase, blocks); }

  /**
   * Checks rather object contains a switch {@link ru.ispras.retrascope.model.cfg.Switch} node.
   *
   * @return <code>true</code> if object contains non-null switch node,
   * <code>false</code> otherwise
   */
  public boolean hasSwitch() { return this.switchNode != null; }

  /**
   * Checks rather object contains mapping.
   *
   * @return <code>true</code> if object contains non-null mapping,
   * <code>false</code> otherwise
   */
  public boolean hasMap() { return this.map != null && !this.map.isEmpty(); }
}
