/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.engine.cfg.CfgVisitorBackend;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgWalkerMode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

/**
 * HDL parser backend that converts clocked guarded atomic actions model
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
 * into Static Signal Assignment (SSA) form.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaSSABackend extends CfgVisitorBackend {

  /* Module variables names & their versions. */
  private final Map<String, Stack<NodeVariable>> versions = new LinkedHashMap<>();

  /* The current module. */
  private Module module;

  /* The blocks of current action. */
  private final List<BasicBlock> blocks = new ArrayList<>();

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) {

    this.module = module;

    /* Initialize versions mapping with variables. */
    List<NodeVariable> vars = module.getVariables();

    for (NodeVariable var : vars) {
      Stack<NodeVariable> stack = new Stack<>();
      stack.push(var);
      this.versions.put(var.getName(), stack);
    }
  }

  @Override
  public void onModuleEnd(Module module) { this.versions.clear(); }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { /* Empty. */ }

  @Override
  public void onCfgEnd(Cfg cfg) { /* Empty. */ }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { this.blocks.add(basicBlock); }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { this.blocks.remove(basicBlock); }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) { /* Empty. */ }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) {

    /* Check rather the current diagram branch contains versions. */
    boolean versionsExist = false;

    /* Get block/assignment numbers where variables are defined last time. */
    Map<RangedVariable, Map.Entry<Integer, Integer>> defNumbers =
      new LinkedHashMap<>();

    for (int i = 0; i < this.blocks.size(); i++) {

      List<Assignment> assignments = this.blocks.get(i).getAssignments();

      for (int j = 0; j < assignments.size(); j++) {
        RangedVariable define = assignments.get(j).getRangedVariable();
        if (defNumbers.containsKey(define)) {
          defNumbers.remove(define);
          versionsExist = true;
        }
        defNumbers.put(define, new AbstractMap.SimpleEntry<>(i, j));
      }
    }

    /* If the current diagram branch contains variable versions. */
    if (versionsExist) {

      /* Substitute defined variables by their versions except last defines. */
      for (int i = 0; i < this.blocks.size(); i++) {

        BasicBlock block = this.blocks.get(i);
        List<Assignment> assignments = block.getAssignments();
        for (int j = 0; j < assignments.size(); j++) {
          Assignment assignment = assignments.get(j);

          /* Assignment define processing. */
          RangedVariable define = assignment.getRangedVariable();
          NodeVariable defineVariable = define.getVariable();

          Map.Entry<Integer, Integer> lastDefNumbers = defNumbers.get(define);

          /* If the last definition of this variable goes later. */
          if (!lastDefNumbers.getKey().equals(i) || !lastDefNumbers.getValue().equals(j)) {

            /* create and add new version of the defined variable. */
            NodeVariable newVariable = this.module.addVariableVersion(define);

            if (!this.module.containsVariable(newVariable)) {
              this.module.addVariable(newVariable);
            }

            /* Put version to the stack. */
            Stack<NodeVariable> stack = this.versions.get(defineVariable.getName());
            stack.push(newVariable);
            this.versions.remove(defineVariable.getName());
            this.versions.put(defineVariable.getName(), stack);

            assignment.setVariable(new RangedVariable(newVariable));
          }

          /* Assignment uses processing. */
          Set<NodeVariable> uses = assignment.getUses();

          if (!uses.isEmpty()) {

            List<NodeBinding.BoundVariable> bindings = new ArrayList<>();

            for (NodeVariable use : uses) {
              NodeVariable lastVersion = this.versions.get(use.getName()).peek();
              if (!lastVersion.equals(use)) {
                bindings.add(NodeBinding.bindVariable(use, lastVersion));
              }
            }
            if (!bindings.isEmpty()) {
              assignment.setExpression(
                Transformer.substituteAllBindings(
                  new NodeBinding(assignment.getExpression(), bindings)));
            }
          }
        }
      }
    }
  }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }

  @Override
  public CfgWalkerMode getWalkerMode() { return CfgWalkerMode.DFS; }
}
