/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Visitor of clocked guarded actions model {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
 * for clock-like signals extracting.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaClockVarExtractVisitor extends MapCfgVisitor {

  /* The clock-like variables for current module. */
  private List<NodeVariable> currentClockList;

  /**
   * Returns rather the specified variable is a one-bit long signal.
   *
   * @param variable variable
   *
   * @throws java.lang.NullPointerException when any of arguments is <code>null</code>.
   *
   * @return <code>true</code> if variable is one bit long signal,
   * <code>false</code> otherwise
   */
  public boolean isOneBitSignal(NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    Object variableData = variable.getUserData();
    if (variableData instanceof VariableData) {
      boolean isSignal = VariableType.isSignal(((VariableData) variableData).getVariableType());
      DataType dataType = variable.getDataType();

      return isSignal
            && (dataType.equals(DataType.BOOLEAN) || dataType.equals(DataType.BIT_VECTOR(1)));
    } else
      throw new IllegalArgumentException(
        String.format("Incorrect type: %s.",variableData.getClass()));
  }

  @Override
  public void onRootBegin() { this.currentClockList = new ArrayList<>(); }

  @Override
  public void onModuleBegin(Module module) { this.currentClockList = new ArrayList<>(); }

  @Override
  public void onModuleEnd(Module module) {
    Set<NodeVariable> clocks = new LinkedHashSet<>();
    clocks.addAll(this.currentClockList);
    this.getClockMap().put(module, clocks);
    this.currentClockList.clear();
    this.currentClockList = null;
  }

  @Override
  public void onProcessBegin(Process process) {

    /* Analyse sensitivity list. */
    EventList eventList = process.getSensitivityList();
    if (eventList != null && !eventList.isEmpty()) {
      Collection<Event> events = eventList.getEvents();
      for (Event event : events) {
        NodeVariable variable = event.getRangedVariable().getVariable();
        if (isOneBitSignal(variable) && !this.currentClockList.contains(variable)) {
          this.currentClockList.add(variable);
        }
      }
    }
  }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) {
    List<Assignment> statements = basicBlock.getAssignments();
    for (Assignment statement : statements) {

      /* Analyse the signals are used in basic blocks assignments. */
      Collection<NodeVariable> stmtUses = statement.getUses();
      if (stmtUses != null && !stmtUses.isEmpty()) {
        for (NodeVariable stmtUse : stmtUses) {
          if (this.currentClockList.contains(stmtUse)) {
            this.currentClockList.remove(stmtUse);
          }
        }
      }

      /* Analyse the signals are defined in basic blocks assignments. */
      Collection<NodeVariable> stmtDefines = statement.getDefines();
      if (stmtDefines != null && !stmtDefines.isEmpty()) {
        for (NodeVariable stmtDefine : stmtDefines) {
          if (this.currentClockList.contains(stmtDefine)) {
            this.currentClockList.remove(stmtDefine);
          }
        }
      }
    }
  }
}
