/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualHashBidiMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.logic.Clause;
import ru.ispras.fortress.logic.NormalForm;
import ru.ispras.fortress.logic.Orthogonalizer;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.engine.basis.TransformUtils;
import ru.ispras.retrascope.engine.cfg.CfgEngine;
import ru.ispras.retrascope.engine.cfg.CfgVisitorConfig;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgVariableContainerNode;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.CfgWalkerMode;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.VariableDeclaration;
import ru.ispras.retrascope.model.cgaa.CgaaModel;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The transformer of clocked guarded actions container
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel} into
 * extended finite state machine {@link ru.ispras.retrascope.model.efsm.EfsmModel}.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CgaaEfsmTransformer extends CfgEngine {

  /* The engine identifier. */
  public static final String ENGINE_ID = "cgaa-efsm-transformer";

  /* The clock-like variables extractor. */
  private final CgaaClockVarExtractVisitor clockVisitor;

  /* The state-like variables extractor. */
  private final CgaaStateVarExtractVisitor stateVisitor;

  /* The state-like variables statistics keeping visitor. */
  private final CgaaStateVarStatVisitor stateVarStatVisitor;

  /* The depending-on-states switches splitter. */
  private final CgaaSwitchSplitVisitor splitVisitor;

  /* The splitting rules' applicator. */
  private final CgaaRuleApplyVisitor ruleVisitor;

  /* The states expressions' extractor. */
  private final CgaaStateExprVisitor stExprVisitor;

  /**
   * Default constructor.
   */
  public CgaaEfsmTransformer() {
    super(
      CgaaEfsmTransformer.ENGINE_ID,
      CgaaModel.TYPE,
      EfsmModel.TYPE,
      new CfgVisitorConfig(new CgaaClockVarExtractVisitor(), CfgWalkerMode.DFS_WITHOUT_REPEAT),
      new CfgVisitorConfig(new CgaaStateVarExtractVisitor(), CfgWalkerMode.DFS),
      new CfgVisitorConfig(new CgaaStateVarStatVisitor(), CfgWalkerMode.DFS),
      new CfgVisitorConfig(new CgaaSwitchSplitVisitor(), CfgWalkerMode.DFS),
      new CfgVisitorConfig(new CgaaRuleApplyVisitor(), CfgWalkerMode.DFS),
      new CfgVisitorConfig(new CgaaStateExprVisitor(), CfgWalkerMode.DFS));

    CfgVisitorConfig[] configs = this.getConfigs();

    /* Initialize visitor components. */
    CgaaClockVarExtractVisitor clkVisitor = null;
    CgaaStateVarExtractVisitor sVisitor = null;
    CgaaStateVarStatVisitor statVisitor = null;
    CgaaSwitchSplitVisitor cndVisitor = null;
    CgaaRuleApplyVisitor rVisitor = null;
    CgaaStateExprVisitor eVisitor = null;

    for (CfgVisitorConfig config : configs) {

      CfgVisitor visitor = config.getVisitor();

      if (visitor instanceof CgaaClockVarExtractVisitor) {
        clkVisitor = (CgaaClockVarExtractVisitor) visitor;
      }
      else if (visitor instanceof CgaaStateVarExtractVisitor) {
        sVisitor = (CgaaStateVarExtractVisitor) visitor;
      }
      else if (visitor instanceof CgaaStateVarStatVisitor) {
        statVisitor = (CgaaStateVarStatVisitor) visitor;
      }
      else if (visitor instanceof CgaaSwitchSplitVisitor) {
        cndVisitor = (CgaaSwitchSplitVisitor) visitor;
      }
      else if (visitor instanceof CgaaRuleApplyVisitor) {
        rVisitor = (CgaaRuleApplyVisitor) visitor;
      }
      else if (visitor instanceof CgaaStateExprVisitor) {
        eVisitor = (CgaaStateExprVisitor) visitor;
      }
    }
    this.clockVisitor  = clkVisitor;
    this.stateVisitor  = sVisitor;
    this.stateVarStatVisitor = statVisitor;
    this.splitVisitor  = cndVisitor;
    this.ruleVisitor   = rVisitor;
    this.stExprVisitor = eVisitor;
  }

  @Override
  public void initialize(Map<EntityType, Entity> inputs) {

    // TODO: refactoring is needed - substitute by front-ends

    /* Extract control flow graph from wrapper. */

    final CgaaModel cgaaModel = (CgaaModel) inputs.get(CgaaModel.TYPE);
    inputs.remove(CgaaModel.TYPE);
    inputs.put(CfgModel.TYPE, cgaaModel.getModel());

    /* Initialize the internal visitors. */

    Map<Module, Set<NodeVariable>> clockMap = new LinkedHashMap<>();
    this.clockVisitor.setClockMap(clockMap);
    this.stExprVisitor.setClockMap(clockMap);

    Map<Module, Set<NodeVariable>> stateMap = new LinkedHashMap<>();
    this.stateVisitor.setStateMap(stateMap);
    this.stateVarStatVisitor.setStateMap(stateMap);
    this.splitVisitor.setStateMap(stateMap);
    this.stExprVisitor.setStateMap(stateMap);

    Map<CfgModelNode, CfgModelNode> map = new LinkedHashMap<>();
    this.splitVisitor.setOldNewMap(map);
    this.ruleVisitor.setOldNewMap(map);
  }

  @Override
  public Entity getOutput() {

    EfsmModel model = new EfsmModel();

    Map<Process, List<EfsmDataContainer>> processDataMap =
        this.stExprVisitor.getProcessDataMap();

    Log.getLogger().log(LogLevel.INFO, "======================================");
    int efsmCount = 0;

    int procNum = 0;
    for (Process process : processDataMap.keySet()) {

      Module module = (Module) process.getOnlyParent();

      /* Print clock-like variables. */

      List<String> clockNames = new ArrayList<>();
      Map<Module, Set<NodeVariable>> clockMap = this.clockVisitor.getClockMap();
      Set<NodeVariable> clockVars = clockMap.get(module);

      for (NodeVariable var : clockVars) {
        clockNames.add(var.getName());
      }

      Log.getLogger().log(LogLevel.INFO, "Clock-like variables: " + clockNames.toString());

      Log.getLogger().log(
        LogLevel.INFO,
        "Transforming the process of module: " +
            module.getName());

      List<EfsmDataContainer> list = processDataMap.get(process);

      Efsm efsm = new Efsm(module.getName() + "_" + procNum);

      /* Copy invariants from process & parent modules. */
      addInvariants(efsm, process);

      /* Construct EFSM states. */
      addStates(efsm, process, list);

      if (efsm.containsStates()) {
        addTransitions(efsm, list);
        model.addEfsm(efsm);
        efsmCount++;
      }
      procNum++;
    }
    Log.getLogger().log(LogLevel.INFO, efsmCount + " EFSM(s) are extracted.");
    return model;
  }

  private void addInvariants(Efsm efsm, Process process) {
    CfgVariableContainerNode current = process;
    while (current != null) {
      Map<String, VariableDeclaration> decMap = current.getVariablesMapping();
      for (String name : decMap.keySet()) {
        Node invariant = decMap.get(name).getInvariant();
        if (!efsm.containsInvariant(name) && invariant != null) {
          efsm.addInvariant(name, invariant);
        }
      }
      current = (CfgVariableContainerNode) current.getOnlyParent();
    }
  }

  private void addStates(Efsm efsm, Process process, List<EfsmDataContainer> list) {

    /* Variables for EFSM extraction statistics keeping. */
    int stateCount = 0;
    List<String> stateVarNames = new ArrayList<>();

    /* Create the convenience collection of pre-states. */
    Set<Set<Node>> preStates = new LinkedHashSet<>();

    for (EfsmDataContainer data : list) {
      Set<Node> dataState = data.getState();
      if (!dataState.isEmpty()) {
        preStates.add(dataState);
      } else
        throw new IllegalStateException("Empty state");
    }

    List<Node> states = new ArrayList<>();

    /* Create a bi-directional mapping for literals and expressions. */
    BidiMap<Integer, Node> literalMap = new DualHashBidiMap<>();

    if (!preStates.isEmpty() && preStates.size() > 1) {

      /* Transform data containers' pre-states into clauses. */
      List<Clause> clauses = createClauses(preStates, literalMap);

      /* Make clauses orthogonal. */
      NormalForm normalForm = new NormalForm(NormalForm.Type.DNF, clauses);
      NormalForm orthogonalForm = Orthogonalizer.orthogonalize(normalForm);
      List<Clause> ortClauses = orthogonalForm.getClauses();

      /* Transform orthogonal clauses into EFSM states. */
      createStates(process, states, literalMap, ortClauses);
    }

    /* If states are not extracted, a trivial EFSM should be created. */
    if (states.isEmpty()) {
      states.add(NodeValue.newBoolean(true));
    }

    /* Add states to EFSM model. */
    for (Node state : states) {

      EfsmState efsmState;

      if (state instanceof NodeOperation) {
        efsmState = new EfsmState(state);

        /* Store state-like variables names (for statistics only). */
        Collection<NodeVariable> stateVariables = ExprUtils.getVariables(state);

        for (NodeVariable stateVar : stateVariables) {
          String name = stateVar.getName();
          if (!stateVarNames.contains(name)) {
            stateVarNames.add(name);
          }
        }
      } else if (state instanceof NodeValue) {
        efsmState = new EfsmState(state);
      } else throw new IllegalArgumentException(
        String.format("Incorrect state: %s.", state.toString()));

      efsm.addState(efsmState);
      stateCount++;
    }

    /* Debug printing. */
    Log.getLogger().log(LogLevel.INFO, stateCount + " states are extracted.");
    Log.getLogger().log(LogLevel.INFO, "The state-like variables are: " + stateVarNames.toString());
  }

  private void createStates(
      Process process,
      List<Node> states,
      BidiMap<Integer, Node> literalMap,
      List<Clause> ortClauses) {

    for (Clause clause : ortClauses) {

      Node state = null;
      Set<Integer> clauseLiterals = clause.getVars();
      for (Integer literal : clauseLiterals) {
        Node node = literalMap.get(literal);

        if (clause.getSign(literal)) {
          node = new NodeOperation(StandardOperation.NOT, node);
        }
        state = state == null ? node : new NodeOperation(StandardOperation.AND, state, node);
      }

      /* Add variable invariants to state constraint. */
      Collection<NodeVariable> stateVars = ExprUtils.getVariables(state);
      Collection<Node> stateVarInvariants =
          CgaaEfsmTransformerUtils.getInvariants(process, stateVars);
      Node invConjunction = CgaaEfsmTransformerUtils.getConjunction(stateVarInvariants);

      /* If state expression is satisfiable, then it is a correct state. */
      if (ExprUtils.isSAT(state) &&
          (invConjunction == null || ExprUtils.areCompatible(state, invConjunction))) {
        states.add(state);
      }
    }
  }

  private List<Clause> createClauses(Set<Set<Node>> preStates, BidiMap<Integer, Node> literalMap) {
    List<Clause> clauses = new ArrayList<>();

    for (Set<Node> preState : preStates) {

      Clause clause = new Clause();

      for (Node node : preState) {

        boolean negated = false;

        if (node instanceof NodeOperation) {

          Enum<?> opId = ((NodeOperation) node).getOperationId();

          if (opId == StandardOperation.NOT) {
            node = ((NodeOperation) node).getOperand(0);
            negated = true;
          }

          int literal;

          if (!literalMap.values().contains(node)) {
            literal = literalMap.keySet().size();
            literalMap.put(literal, node);
          } else {
            literal = literalMap.getKey(node);
          }
          clause.add(literal, negated);
        }
      }
      clauses.add(clause);
    }
    return clauses;
  }

  private void addTransitions(Efsm efsm, List<EfsmDataContainer> list) {

    /* Collection for candidates on initial model state. */
    Map<EfsmState, Set<NodeVariable>> initialResetMap = new LinkedHashMap<>();

    /* Variables for EFSM extraction statistics keeping. */
    int transitionCount = 0;

    /* Create EFSM transitions. */
    for (EfsmDataContainer data : list) {

      GuardedAction guardedAction = data.getGuardedAction();
      Action action = guardedAction.getAction();
      Collection<Event> clocks = guardedAction.getClocks();

      Guard guard = guardedAction.getGuard();
      Node guardNode = guard.toNode();

      List<BasicBlock> blocks = new ArrayList<>();

      BasicBlock block = new BasicBlock(action.getStatements(), false);
      blocks.add(block);

      EfsmState srcState;

      NodeOperation expression = new NodeOperation(
          StandardOperation.AND,
          data.getState().toArray(new Node[data.getState().size()]));

      Set<EfsmState> efsmStates = efsm.getStates();
      for (EfsmState efsmState : efsmStates) {

        Node stateExpr = efsmState.getExpression();

        if (ExprUtils.areCompatible(expression, stateExpr)) {

          /* Detect source state. */
          srcState = efsmState;

          /* Detect possible destination states. */
          List<EfsmState> possible = new ArrayList<>();
          for (EfsmState possibleDst : efsmStates) {

            Node posExpr = possibleDst.getExpression();
            Node substitution = posExpr.deepCopy();

            /* Create a weakest precondition and check the compatibility with source state. */
            if (ExprUtils.areCompatible(
                srcState.getExpression(),
                TransformUtils.substituteBackward(blocks, substitution))) {
              possible.add(possibleDst);
            }
          }

          /* Add possible destination states. */
          if (possible.size() == 0) {
            Log.getLogger().log(
                LogLevel.WARNING,
                String.format(
                  "Lockup found for: src-state=%s, guard=%s, action=%s.",
                  srcState.toString(), guard.toString(), action.toString()));

          } else if (possible.size() == 1) {

            /* Get the only possible destination state. */
            EfsmState dst = possible.get(0);

            /* Create transition's guarded action. */
            GuardedAction transitionGA = new GuardedAction(new Guard(guardNode), action);
            transitionGA.setClocks(clocks);

            efsm.createTransition(srcState, dst, transitionGA);

            /* Check rather transition leads to initial state. */
            if (leadsToInitialState(transitionGA)) {
              initialResetMap.put(dst, getReset(transitionGA));
            }

            /* Increment transition counter. */
            transitionCount++;
          } else {

            /* More than one possible destination state. */
            for (EfsmState dst : possible) {

              /* Refine the transition guard. */
              Node dstExprCopy = dst.getExpression().deepCopy();
              Node refinedGuard = Transformer.standardize(ExprUtils.getConjunction(
                guardNode,
                Transformer.reduce(
                    ReduceOptions.NEW_INSTANCE,
                    TransformUtils.substituteBackward(
                        Arrays.asList((BasicBlock)action.toBasicBlock()),
                        dstExprCopy))));

              GuardedAction transitionGA = new GuardedAction(new Guard(refinedGuard), action);
              transitionGA.setClocks(clocks);

              efsm.createTransition(srcState, dst, transitionGA);

              /* Check rather transition leads to initial state. */
              if (leadsToInitialState(transitionGA)) {
                initialResetMap.put(dst, getReset(transitionGA));
              }

              /* Increment transition counter. */
              transitionCount++;
            }
          }
        }
      }
    }
    Log.getLogger().log(LogLevel.INFO, transitionCount + " transitions are extracted.");

    if (initialResetMap.keySet().isEmpty()) {

      /* No initial state has been found. */

      Log.getLogger().log(LogLevel.WARNING, "No initial state found");

      EfsmState first = efsm.getStates().iterator().next();

      efsm.setInitialState(first);

      setResetGuardedAction(efsm, first);

      Log.getLogger().log(LogLevel.WARNING, "The first is chosen as initial: " + first.toString());
    } else if (initialResetMap.keySet().size() == 1) {

      /* Only one candidate is extracted - set is as initial. */

      EfsmState initialState = initialResetMap.keySet().iterator().next();

      Log.getLogger().log(LogLevel.INFO, "The initial state is: " + initialState.toString() +
          ", reset signal is: " + initialResetMap.get(initialState));

      efsm.setInitialState(initialState);

      setResetGuardedAction(efsm, initialState);
    } else {

      /* More than one candidate on initial state has been found. */

      Log.getLogger().log(
          LogLevel.WARNING,
          "The following candidates on initial state are detected:");

      for (EfsmState state : initialResetMap.keySet()) {
        Log.getLogger().log(LogLevel.INFO, state.toString());
      }

      EfsmState first = initialResetMap.keySet().iterator().next();

      Log.getLogger().log(LogLevel.WARNING, "The first one is chosen as initial: " + first);

      efsm.setInitialState(first);

      setResetGuardedAction(efsm, first);
    }
  }

  /* Set as resetting guarded action the transition that leads to the specified state. */
  private void setResetGuardedAction(Efsm efsm, EfsmState first) {

    Set<EfsmTransition> incoming = efsm.getIncomingTransitions(first);

    /* Set the first one. */
    efsm.setResetGuardedAction(incoming.iterator().next().getGuardedAction());
  }

  /* Returns reset signals that probably exist in the specified guarded action. */
  private Set<NodeVariable> getReset(GuardedAction guardedAction) {
    Set<NodeVariable> resets = new LinkedHashSet<>();

    EventList eventList = guardedAction.getEventList();
    Collection<Event> clocks = eventList.getEvents();
    Set<NodeVariable> guardUses = guardedAction.getGuard().getUses();

    if (guardUses.isEmpty()) {

      /* Add all clock-like variables for this guarded action. */
      for (Event clock : clocks) {
        resets.add(clock.getRangedVariable().getVariable());
      }

    } else if (guardUses.size() == 1) {
      NodeVariable guardUse = guardUses.iterator().next();
      if (eventList.contains(guardUse)) {
        resets.add(guardUse);
      }
    }
    return resets;
  }

  private boolean leadsToInitialState(GuardedAction guardedAction) {

    /* Check rather guard depends only from one clock-like variable */

    Collection<Event> clocks = guardedAction.getClocks();
    Guard guard = guardedAction.getGuard();
    boolean checkGuard = false;
    Set<NodeVariable> guardUses = guard.getUses();
    if (guardUses.isEmpty()) {
      checkGuard = true;
    }
    else if (guardUses.size() == 1) {

      NodeVariable use = guardUses.iterator().next();

      for (Event clock : clocks) {
        if (clock.getRangedVariable().getVariable().equals(use)) {
          checkGuard = true;
          break;
        }
      }
    }

    /* Check rather action contains only assignments like var := constant. */

    Action action = guardedAction.getAction();
    boolean checkAction = false;
    for (Assignment assignment : action.getStatements()) {
      if (assignment.getUses().isEmpty()) {
        checkAction = true;
      } else {
        checkAction = false;
        break;
      }
    }
    return (checkGuard && checkAction);
  }
}
