/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgVariableContainerNode;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * The class containing auxiliary methods for
 * {@link ru.ispras.retrascope.engine.cgaa.transformer.efsm.CgaaEfsmTransformer}
 * and it's sub-components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaEfsmTransformerUtils {

  /**
   * Returns a conjunction of the specified collection of nodes.
   *
   * @param nodes nodes collection
   *
   * @throws java.lang.NullPointerException when argument is <code>null</code>.
   *
   * @return nodes conjunction
   */
  public static Node getConjunction(Collection<Node> nodes) {
    InvariantChecks.checkNotNull(nodes);

    Node result = null;

    for (Node node : nodes) {
      result = (result == null) ? node : new NodeOperation(StandardOperation.AND, node, result);
    }

    return result;
  }

  /**
   * Constructs an expression from the expression of the specified switch
   * {@link ru.ispras.retrascope.model.cfg.Switch} node and it's possible values
   * in the specified case {@link ru.ispras.retrascope.model.cfg.Case} node.
   *
   * @param aCase case node
   * @param aSwitch switch node
   *
   * @return expression
   */
  public static Node getExpression(Case aCase, Switch aSwitch) {

    Node expr;

    if (aCase.isDefault()) {

      /* The 'default' case processing. */
      Collection<CfgModelNode> children = aSwitch.getChildren();
      List<NodeOperation> list = new ArrayList<>();
      for (CfgModelNode child : children) {
        if (!child.equals(aCase)) {
          list.add(constructExpr((Case)child, aSwitch));
        }
      }
      expr = ExprUtils.getComplement(list.toArray(new Node[list.size()]));
    } else {

      /* The 'normal' case processing. */
      expr = constructExpr(aCase, aSwitch);
    }

    /* Transform expression into standard form. */
    return Transformer.standardize(expr);
  }

  private static NodeOperation constructExpr(Case aCase, Switch parent) {
    NodeOperation expr = null;

    for (NodeValue value: aCase.getValues()) {
      if (expr == null) {
        expr = new NodeOperation(StandardOperation.EQ, parent.getCondition(), value);
      } else {
        expr = new NodeOperation(
          StandardOperation.OR,
          expr,
          new NodeOperation(StandardOperation.EQ, parent.getCondition(), value));
      }
    }
    return expr;
  }

  /**
   * Returns invariants of the specified variables that are defined/used in the specified process.
   *
   * @param process process
   * @param variables variables
   *
   * @throws java.lang.NullPointerException when any of the arguments is <code>null</code>.
   *
   * @return collection of variable invariants
   */
  public static Collection<Node> getInvariants(
      Process process,
      Collection<NodeVariable> variables) {

    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(variables);

    /* The returned collection of invariants. */
    List<Node> invariants = new ArrayList<>();

    for (NodeVariable var : variables) {
      Node invariant = getInvariant(process, var);
      if (invariant != null) {
        invariants.add(invariant);
      }
    }

    return invariants;
  }

  private static Node getInvariant(Process process, NodeVariable var) {
    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(var);

    /* The current variable declarations' container. */
    CfgVariableContainerNode current = process;

    do {
      if (current.containsVariable(var)) {
        return current.getDeclaration(var).getInvariant();
      }

      current = (CfgVariableContainerNode) current.getOnlyParent();
    } while (current != null);
    return null;
  }
}
