/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.Collection;

/**
 * Visitor for applying transformation rules to control flow graph model.
 * Rules are constructed by {@link CgaaSwitchSplitVisitor} visitor
 * as a part of EFSM extraction algorithm.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaRuleApplyVisitor extends MapCfgVisitor {

  /* The model to which rules should be applied. */
  private CfgModel cfgModel;

  /* The current graph. */
  private Cfg currentCfg;

  /**
   * Returns model with rules applied.
   *
   * @return model
   */
  public Entity getModel() { return this.cfgModel; }

  @Override
  public void onRootBegin() {
    this.cfgModel = null;
    this.currentCfg = null;
  }

  @Override
  public void onCfgModelBegin() { this.cfgModel = new CfgModel(); }

  @Override
  public void onModuleBegin(Module module) {
    Module moduleNode = (Module)getOldNewMap().get(module);
    this.cfgModel.addModule(moduleNode);
    if (module.hasParents()) {
      Collection<CfgModelNode> moduleParents = module.getParents();
      for (CfgModelNode parent : moduleParents) {
        moduleNode.addParent(getOldNewMap().get(parent));
      }
    }
  }

  @Override
  public void onProcessBegin(Process process) {
    addParentsToNode(process, getOldNewMap().get(process));
  }

  @Override
  public void onCfgBegin(Cfg cfg) {
    Cfg cfgNode = (Cfg) getOldNewMap().get(cfg);
    addParentsToNode(cfg, cfgNode);
    this.currentCfg = cfgNode;
  }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) {
    BasicBlock blockNode = (BasicBlock) getOldNewMap().get(basicBlock);
    addParentsToNode(basicBlock, blockNode);
  }

  @Override
  public void onCaseBegin(Case aCase) {
    CfgModelNode newCase = getOldNewMap().get(aCase);
    addParentsToNode(aCase, newCase);
  }

  @Override
  public void onSwitchBegin(Switch aSwitch) {
    CfgModelNode switchCfg = getOldNewMap().get(aSwitch);

    if (switchCfg instanceof Switch) {
      addParentsToNode(aSwitch, switchCfg);
    }

    /* This switch was previously transformed into sub-graph. */
    else if (switchCfg instanceof Cfg) {
      CfgNode cfgSource = ((Cfg) switchCfg).getSource();

      /* Get "root" switch in the sub-graph. */
      CfgModelNode mainSwitch = cfgSource.getOnlyChild();

      /* Re-link "root" switch to main graph parent node. */
      mainSwitch.removeParent(cfgSource);
      mainSwitch.addParent(getOldNewMap().get(aSwitch.getOnlyParent()));

      /* Get sub-graph nodes. */
      Collection<CfgNode> nodes = ((Cfg) switchCfg).getNodes();

      /* Add non-source & non-sink nodes. */
      for (CfgNode node : nodes) {
        CfgNodeType type = node.getType();
        if (!type.equals(CfgNodeType.SOURCE) && !type.equals(CfgNodeType.SINK)) {
          this.currentCfg.addNode(node);
        }
      }
    }
  }

  @Override
  public void onMergeBegin(Merge merge) {
    CfgModelNode newMerge = getOldNewMap().get(merge);
    addParentsToNode(merge, newMerge);
  }

  @Override
  public void onSink(Sink sink) {
    CfgModelNode newSink = getOldNewMap().get(sink);
    addParentsToNode(sink, newSink);
  }

  private void addParentsToNode(CfgModelNode oldNode, CfgModelNode newNode) {
    if (oldNode.hasParents() && !newNode.hasParents()) {
      Collection<CfgModelNode> parents = oldNode.getParents();
      for (CfgModelNode parent : parents) {
        InvariantChecks.checkNotNull(parent);
        InvariantChecks.checkNotNull(getOldNewMap().get(parent));
        getOldNewMap().get(parent).addChild(newNode);
      }
    }
    if (oldNode instanceof CfgNode) {
      this.currentCfg.addNode(newNode);
    }
  }
}
