/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
* Visitor that extracts data for constructing EFSM states, guards and transitions
* from clocked guarded actions decision diagram.
*
* @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
*/
final class CgaaStateExprVisitor extends MapCfgVisitor {

  /* Processes and corresponding guarded actions with associated conditions on state-like variables */
  private Map<Process, List<EfsmDataContainer>> processDataMap;

  /* Used-in-the-walking-process fields */

  /* The current module. */
  private Module currentModule;

  /* The current process. */
  private Process currentProcess;

  /* The data containers for current process. */
  private List<EfsmDataContainer> dataList;

  /* The current state expression. */
  private List<Node> currentState;

  /* The current list of events. */
  private List<Event> currentEventList;

  /* The current guard. */
  private List<Node> currentGuard;

  /* The current action. */
  private List<BasicBlock> currentAction;

  /* Be set to true when guard processing has been finished. */
  private boolean isGuardFinished;

  /* Be set to true when guard & state are satisfiable & compatible. */
  private boolean isActionReachable;

  /* Cases that used into state-like variables' depending expressions. */
  private List<Case> stateCases;

  /* The current clock. */
  private Event clock;

  /* The current case. */
  private Case clockCase;

  /* Diagram paths number (for statistics keeping). */
  private int pathNum;

  public CgaaStateExprVisitor() {
    this.processDataMap = new LinkedHashMap<>();
    this.currentState = new ArrayList<>();
    this.currentEventList = new ArrayList<>();
    this.currentGuard = new ArrayList<>();
    this.currentAction = new ArrayList<>();
    this.isGuardFinished = false;
    this.isActionReachable = true;
    this.stateCases = new ArrayList<>();

    this.pathNum = 0;
  }

  /**
   * Returns mapping of processes into data containers.
   *
   * @return mapping of processes into data containers
   */
  public Map<Process, List<EfsmDataContainer>> getProcessDataMap() {
    Log.getLogger().log(LogLevel.INFO, "Number of GADD paths: " + this.pathNum);
    return this.processDataMap;
  }

  @Override
  public void onRootBegin() {
    this.processDataMap = new LinkedHashMap<>();
    this.currentState = new ArrayList<>();
    this.currentEventList = new ArrayList<>();
    this.currentGuard = new ArrayList<>();
    this.currentAction = new ArrayList<>();
    this.isGuardFinished = false;
    this.isActionReachable = true;
    this.stateCases = new ArrayList<>();
    this.pathNum = 0;
  }

  @Override
  public void onModuleBegin(Module module) { this.currentModule = module; }

  @Override
  public void onProcessBegin(Process process) {
    this.currentProcess = process;
    this.dataList = new ArrayList<>();
    this.currentEventList.addAll(process.getSensitivityList().getEvents());
  }

  @Override
  public void onProcessEnd(Process process) {
    this.processDataMap.put(this.currentProcess, this.dataList);
    this.currentProcess = null;
    this.currentEventList.clear();
  }

  @Override
  public void onCaseBegin(Case aCase) {
    Node expr;
    Switch caseParent = (Switch)aCase.getOnlyParent();

    /* Construct a condition expression for the current case node. */
    expr = CgaaEfsmTransformerUtils.getExpression(aCase, caseParent);

    if (isStateVarUsingExpr(expr)) {
      this.currentState.add(expr);
      this.stateCases.add(aCase);
    } else {
      this.currentGuard.add(expr);
    }

    /* If switch has event, save it as clock. */
    if (caseParent.getEvent() != null) {
      if (this.clock == null) {
        this.clock = caseParent.getEvent();
        this.clockCase = aCase;
      } else {
        Log.getLogger().log(
            LogLevel.WARNING,
            String.format(
                "Unreachable path for: event1=%s, event2=%s.",
                this.clock.toString(),
                caseParent.getEvent().toString()));
      }
    }
  }

  private boolean isStateVarUsingExpr(Node expr) {
    return CollectionUtils.complementSets(
        new LinkedHashSet<>(ExprUtils.getVariables(expr)),
        this.getStateMap().get(this.currentModule)).isEmpty();
  }

  @Override
  public void onCaseEnd(Case aCase) {
    if (this.stateCases.contains(aCase)) {
      this.currentState.remove(this.currentState.size() - 1);
      this.stateCases.remove(aCase);
    } else {
      this.currentGuard.remove(this.currentGuard.size() - 1);
    }

    /* Visitor is ready again to process guard part elaboration. */
    if (this.isGuardFinished) {
      this.isGuardFinished = false;
    }

    if (this.clockCase != null && this.clockCase.equals(aCase)) {
      this.clockCase = null;
      this.clock = null;
    }
  }

  @Override
  public void onBasicBlockBegin(BasicBlock block) {

    checkConditionsIfNot();

    if (this.isActionReachable) {
      this.currentAction.add(block);
    }
  }

  private void checkConditionsIfNot() {
    if (!this.isGuardFinished) {
      this.isActionReachable =
          (this.currentState.isEmpty() ||
              checkInvCompliance(this.currentProcess, this.currentState))
              && (this.currentGuard.isEmpty() ||
              checkInvCompliance(this.currentProcess, this.currentGuard));
      this.isGuardFinished = true;
    }
  }

  /* Check the compliance of nodes from the specified collection to the invariants
  * on used variables at the specified process.*/
  private static boolean checkInvCompliance(Process process, List<Node> list) {
    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(list);

    boolean isSat;

    Node conjunction = ExprUtils.getConjunction(list.toArray(new Node[list.size()]));

    if (!ExprUtils.areCompatible(conjunction)) {
      isSat = false;
    } else {
      Collection<Node> invNodes =
        CgaaEfsmTransformerUtils.getInvariants(process, ExprUtils.getVariables(conjunction));

      for (Node invNode : invNodes) {
        conjunction = new NodeOperation(StandardOperation.AND, invNode, conjunction);
      }
      isSat = ExprUtils.areCompatible(conjunction);
    }
    return isSat;
  }

  @Override
  public void onSink(Sink sink) {

    this.pathNum++;
    if (this.isActionReachable) {

      Action action;

      if (!this.currentAction.isEmpty()) {
        List<Assignment> assignments = new ArrayList<>();
        for (BasicBlock block : this.currentAction) {
          assignments.addAll(block.getAssignments());
        }
        action = new Action(assignments);
      } else {
        action = new Action(new ArrayList<Assignment>());
      }

      Set<Node> states = getNodes(this.currentState);
      Set<Node> guard = getNodes(this.currentGuard);

      if (states.isEmpty()) {
        states.add(NodeValue.newBoolean(true));
      }

      /* Check rather guard is satisfiable. */
      Guard guardExpr = new Guard(guard);
      if (ExprUtils.isSAT(guardExpr.toNode())) {
        GuardedAction guardedAction = new GuardedAction(guardExpr, action);
        if (this.clock == null) {
          guardedAction.setClocks(this.currentEventList);
        } else {
          guardedAction.setClock(this.clock.deepcopy());
        }
        EfsmDataContainer container =
            new EfsmDataContainer(this.currentProcess, states, guardedAction);

        this.dataList.add(container);
      }
    }
    this.currentAction.clear();
  }

  private Set<Node> getNodes(List<Node> list) {
    Set<Node> nodes = new LinkedHashSet<>();
    if (list.isEmpty()) {
      nodes.add(NodeValue.newBoolean(true));
    } else {
      nodes.addAll(list);
    }
    return nodes;
  }
}
