/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Visitor for state-like variables extracting.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaStateVarExtractVisitor extends MapCfgVisitor {

  /* The control parts of CFG paths. */
  private List<Switch> controlPath;

  /* The data parts of CFG paths. */
  private List<BasicBlock> dataPath;

  /* The state-like variables. */
  private List<NodeVariable> stateVariables;

  /**
   * The default constructor.
   */
  public CgaaStateVarExtractVisitor() {
    this.controlPath = new ArrayList<>();
    this.dataPath = new ArrayList<>();
  }

  @Override
  public void onRootBegin() {
    this.controlPath = new ArrayList<>();
    this.dataPath = new ArrayList<>();
    this.stateVariables = new ArrayList<>();
  }

  @Override
  public void onModuleEnd(Module module) {
    Set<NodeVariable> states = new LinkedHashSet<>();
    states.addAll(this.stateVariables);
    this.getStateMap().put(module, states);
    this.stateVariables.clear();
  }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) {
    if (this.dataPath.contains(basicBlock)) {
      Log.getLogger().log(LogLevel.WARNING, "Loop is found");
    }
    this.dataPath.add(basicBlock);
  }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) {
    if (this.dataPath.contains(basicBlock)) {
      this.dataPath.remove(basicBlock);
    } else
        throw new IllegalStateException(
            String.format("No such node: %s.", basicBlock.toString()));
  }

  @Override
  public void onSwitchBegin(Switch aSwitch) {
    if (this.controlPath.contains(aSwitch)) {
      Log.getLogger().log(LogLevel.WARNING, "Loop is found");
    }
    this.controlPath.add(aSwitch);
  }

  @Override
  public void onSwitchEnd(Switch aSwitch) {
    if (this.controlPath.contains(aSwitch)) {
      this.controlPath.remove(aSwitch);
    } else
        throw new IllegalStateException(
            String.format("No such node: %s.", aSwitch.toString()));
  }

  @Override
  public void onSink(Sink sink) {

    /* Get variables that are defined in data parts of graph paths. */
    Set<NodeVariable>definedDataVars = getDefinedVariables(this.dataPath);

    /* Get variables that are used in control parts of graph paths. */
    List<NodeVariable> usedControlVars = getUsedVariables(this.controlPath);

     /* The weak heuristic is applied. */
    for (NodeVariable var : definedDataVars) {

      if (var.getUserData() != null && var.getUserData() instanceof VariableData) {

        final VariableType variableType = ((VariableData) var.getUserData()).getVariableType();
        if (!VariableType.isSignal(variableType) &&
            usedControlVars.contains(var) &&
            !this.stateVariables.contains(var)) {
          this.stateVariables.add(var);
        }
      }
    }
  }

  private Set<NodeVariable> getDefinedVariables(List<BasicBlock> dataPath) {

    /* Get those defined variables that do not depend on input signals. */
    Set<NodeVariable> defined = new LinkedHashSet<>();

    Set<NodeVariable> forbidden = new LinkedHashSet<>();

    for (BasicBlock block : dataPath) {
      List<Assignment> assignments = block.getAssignments();

      for (Assignment assignment : assignments) {
        Collection<NodeVariable> stmtUses = assignment.getUses();

        boolean dependOnSignal = false;
        for (NodeVariable use : stmtUses) {
          if (VariableType.isInputSignal(((VariableData) use.getUserData()).getVariableType())) {
            dependOnSignal = true;
            break;
          }
        }
        Set<NodeVariable> defines = assignment.getDefines();
        if (dependOnSignal) {

          /* Variables that are defined here, cannot be returned. */
          forbidden.addAll(defines);
        } else {
          defined.addAll(CollectionUtils.complementSets(defines, forbidden));
        }
      }

      Collection<NodeVariable> blockDefines = block.getDefines();
      for (NodeVariable blockDefined : blockDefines) {
        if (!forbidden.contains(blockDefined)) {
          defined.add(blockDefined);
        }
      }
    }
    return defined;
  }

  private List<NodeVariable> getUsedVariables(List<Switch> controlPath) {
    List<NodeVariable> used = new ArrayList<>();

    for (Switch node : controlPath) {
      Collection<NodeVariable> nodeUses = node.getUses();
      for (NodeVariable nodeUse : nodeUses) {
        if (!used.contains(nodeUse)) {
          used.add(nodeUse);
        }
      }
    }
    return used;
  }
}
