/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The visitor for {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
 * that collects statistical information about state-like variables.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaStateVarStatVisitor extends MapCfgVisitor {

  /* The current module. */
  private Module currentModule;

  /* The control part of current CFG path. */
  private Map<Case, Node> controlPath;

  /* The data part of current CFG path. */
  private List<BasicBlock> dataPath;

  /* Class for keeping the statistical information. */
  class VarStatInfo {

    /* The number of all process paths. */
    int allPathNum;

    /* The number of paths in which the variable is used. */
    int usePathNum;

    /* The number of paths' control parts in which the variable is used. */
    int controlUsePathNum;

    /* The number of paths' data parts in which the variable is used. */
    int dataUsePathNum;

    /* The number of paths in which the variable is defined. */
    int defPathNum;

    /* The number of paths in which the variable is both used and defined. */
    int useDefPathNum;

    VarStatInfo() {

      this.allPathNum = 0;
      this.usePathNum = 0;
      this.controlUsePathNum = 0;
      this.dataUsePathNum = 0;
      this.defPathNum = 0;
      this.useDefPathNum = 0;
    }
    /* Getter-methods. */

    int getAllPathNum() { return this.allPathNum; }
    int getUsePathNum() { return this.usePathNum; }
    int getControlUsePathNum() { return this.controlUsePathNum; }
    int getDataUsePathNum() { return this.dataUsePathNum; }
    int getDefPathNum() { return this.defPathNum; }
    int getUseDefPathNum() { return this.useDefPathNum; }

    /* Incrementation methods. */

    void incTotalPathNum() { this.allPathNum ++; }
    void incUsePathNum() { this.usePathNum ++; }
    void incControlUsePathNum() { this.controlUsePathNum++; }
    void incDataUsePathNum() { this.dataUsePathNum++; }
    void incDefPathNum() { this.defPathNum ++; }
    void incUseDefPathNum() { this.useDefPathNum ++; }

    @Override
    public String toString() {
      return "{all-path-num=" + getAllPathNum() +
          ", use-path-num=" + getUsePathNum() +
          ", control-use-path-num=" + getControlUsePathNum() +
          ", data-use-path-num=" + getDataUsePathNum() +
          ", def-path-num=" + getDefPathNum() +
          ", def-use-path-num=" + getUseDefPathNum() + "}";
    }
  }

  /* The mapping for keeping the statistical info about state-like variables. */
  private Map<NodeVariable, VarStatInfo> varStatInfoMap;

  /* The mapping for statistical info about state-like variables of processes. */
  private Map<Process, Map<NodeVariable, VarStatInfo>> infoMap;

  @Override
  public void onRootBegin() {
    this.controlPath = new LinkedHashMap<>();
    this.dataPath = new ArrayList<>();
  }

  @Override
  public void onCfgModelBegin() { this.infoMap = new LinkedHashMap<>(); }

  @Override
  public void onCfgModelEnd() {

    /* Print the collected statistical information to the tool output log. */
    //Log.getLogger().log(LogLevel.INFO, this.infoMap.toString());
    this.infoMap = null;
  }

  @Override
  public void onModuleBegin(Module module) { this.currentModule = module; }

  @Override
  public void onModuleEnd(Module module) { this.currentModule = null; }

  @Override
  public void onProcessBegin(Process process) { this.varStatInfoMap = new LinkedHashMap<>(); }

  @Override
  public void onProcessEnd(Process process) {

    /* Create a copy of internal mapping. */
    LinkedHashMap<NodeVariable, VarStatInfo> infoMapCopy = new LinkedHashMap<>();
    infoMapCopy.putAll(this.varStatInfoMap);

    this.infoMap.put(process, infoMapCopy);
    this.varStatInfoMap = null;
  }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { this.dataPath.add(basicBlock); }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { this.dataPath.remove(basicBlock); }

  @Override
  public void onCaseBegin(Case aCase) {
    Switch parent = (Switch) aCase.getOnlyParent();
    Node expr = CgaaEfsmTransformerUtils.getExpression(aCase, parent);
    this.controlPath.put(aCase, expr);
  }

  @Override
  public void onCaseEnd(Case aCase) { this.controlPath.remove(aCase); }

  @Override
  public void onSink(Sink sink) {

    /* Get state-like variables of the current module. */
    Set<NodeVariable> stateVariables = this.getStateMap().get(this.currentModule);

    /* The collection of variables that are used in the control part of current path. */
    Set<NodeVariable> controlUsed = new LinkedHashSet<>();

    Collection<Node> controlNodes = this.controlPath.values();

    for (Node expr : controlNodes) {
      Collection<NodeVariable> usedVars = ExprUtils.getVariables(expr);
      controlUsed.addAll(usedVars);
    }

    Set<NodeVariable> dataUsed = new LinkedHashSet<>();
    Set<NodeVariable> dataDefined = new LinkedHashSet<>();

    for (BasicBlock block : this.dataPath) {
      dataUsed.addAll(block.getUses());
      dataDefined.addAll(block.getDefines());
    }

    for (NodeVariable state : stateVariables) {

      if (!this.varStatInfoMap.containsKey(state)) {
        this.varStatInfoMap.put(state, new VarStatInfo());
      }

      VarStatInfo statInfo = this.varStatInfoMap.get(state);

      boolean usedInControl = controlUsed.contains(state);
      boolean usedInData = dataUsed.contains(state);

      if (usedInControl || usedInData) {
        if (usedInControl) {
          statInfo.incControlUsePathNum();
        }

        if (dataUsed.contains(state)) {
          statInfo.incDataUsePathNum();
        }
        statInfo.incUsePathNum();
      }

      if (dataDefined.contains(state)) {
        statInfo.incDefPathNum();
        if (usedInControl || usedInData) {
          statInfo.incUseDefPathNum();
        }
      }

      statInfo.incTotalPathNum();
    }
  }
}
