/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.ExprTreeVisitor;
import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Visitor for splitting switch nodes, that are depend on state-like variables
 * and can be split into many-nodes-containing sub-graph. Such splitting
 * can be done when top-level operation identifier in switch node's expression
 * is of Boolean type (like NOT, OR, AND, IML, XOR). If the switch node's expression
 * also contains internal Boolean identifiers, the splitting procedure will be applied
 * to them also only if one of their sub-expressions depends on state-like variable.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaSwitchSplitVisitor extends MapCfgVisitor {

  /* The module to be processed. */
  private Module current;

  @Override
  public void onRootBegin() { this.current = null; }

  @Override
  public void onModuleBegin(Module module) {
    this.current = module;

    /* Make a copy of module. */
    Module newModule = new Module(module.getName());
    newModule.setVariablesMapping(module.getVariablesMapping());

    getOldNewMap().put(module, newModule);
  }

  @Override
  public void onProcessBegin(Process process) {

    /* Make a copy of process. */
    Process newProcess = new Process(process.isInitial());
    newProcess.setSensitivityList(process.getSensitivityList());

    getOldNewMap().put(process, newProcess);
  }

  @Override
  public void onCfgBegin(Cfg cfg) {
    Cfg newCfg = new Cfg();
    getOldNewMap().put(cfg, newCfg);
    CfgNode newSrc = newCfg.getSource();
    CfgNode newSink = newCfg.getSink();
    getOldNewMap().put(cfg.getSource(), newSrc);
    getOldNewMap().put(cfg.getSink(), newSink);

    /* Remove the default link. */
    newSrc.removeChild(newSink);
  }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) {
    BasicBlock newBlock = new BasicBlock(basicBlock.getAssignments(), basicBlock.isConcurrent());
    getOldNewMap().put(basicBlock, newBlock);
  }

  @Override
  public void onSwitchBegin(Switch aSwitch) {
    CfgModelNode newSwitch = splitIntoSubGraph(aSwitch);
    if (newSwitch instanceof Switch && !newSwitch.hasChildren())
      throw new NullPointerException();
    getOldNewMap().put(aSwitch, newSwitch);
  }

  private CfgModelNode splitIntoSubGraph(Switch aSwitch) {
    CfgModelNode subGraph;
    Node expr = aSwitch.getCondition();
    Collection<CfgModelNode> sChildren = aSwitch.getChildren();

    if (isAbleToBeSplit(expr)) {

      subGraph = new Cfg();

      /* Every control flow graph contains linked source and sink - delete that link. */
      CfgNode cfgSource = ((Cfg)subGraph).getSource();
      CfgNode cfgSink = ((Cfg)subGraph).getSink();
      cfgSource.removeChild(cfgSink);

       /* Switch expression is boolean
       * so this node has only two case children
       * with 'true'/'false' values.
       * These children are transformed into merge nodes.*/
      Merge trueMerge = null;
      Merge falseMerge = null;
      for (CfgModelNode child : sChildren) {
        Set<NodeValue> childValues = ((Case) child).getValues();
        if (childValues.size() == 1) {
          NodeValue cValue = childValues.iterator().next();
          if (cValue.getData().equals(Data.newBoolean(true))) {
            trueMerge = new Merge();
            getOldNewMap().put(child, trueMerge);
          } else if (cValue.getData().equals(Data.newBoolean(false))) {
            falseMerge = new Merge();
            getOldNewMap().put(child, falseMerge);
          }
        } else
          throw new IllegalStateException(
              "Able-to-bet-split switch cannot be processed: multiple values at child case: " +
              childValues.toString());
      }

      /* Check that merge nodes a linked with cases. */
      if (trueMerge == null || falseMerge == null)
        throw new IllegalStateException("Able-to-be-split switch cannot be processed: " +
            expr.toString());

      ((Cfg)subGraph).addNode(falseMerge);
      ((Cfg)subGraph).addNode(trueMerge);

      makeSubGraph((NodeOperation) expr, (Cfg)subGraph, cfgSource, trueMerge, falseMerge);
    } else {
      if (aSwitch.getEvent() == null) {
        subGraph = new Switch(expr);
      } else {
        subGraph = new Switch(aSwitch.getEvent());
      }

      /* All child nodes are not to be transformed. */
      for (CfgModelNode child : sChildren) {
        Case newCase = new Case(((Case) child).getValues());
        subGraph.addChild(newCase);
        getOldNewMap().put(child, newCase);
      }
    }
    return subGraph;
  }

  private void makeSubGraph(NodeOperation expr, Cfg cfg,
    CfgNode root, Merge trueMerge, Merge falseMerge) {

    int opCount = expr.getOperandCount();

    if (opCount == 1) {

     /* Unary operation. */
      makeUnaryOpSubGraph(expr, cfg, root, trueMerge, falseMerge);
    } else if (opCount == 2) {

      /* Binary operation. */
      makeBinOpSubGraph(expr, cfg, root, trueMerge, falseMerge);
    } else {

      /* No additional elaboration is needed here. */
      makeAtomicSubGraph(expr, cfg, root, trueMerge, falseMerge);
    }
  }

  private void makeUnaryOpSubGraph(
    NodeOperation expr,
    Cfg cfg,
    CfgNode root,
    Merge trueMerge,
    Merge falseMerge) {

    Enum<?> opId = expr.getOperationId();
    Node operand = expr.getOperand(0);

    if (isAbleToBeSplit(operand)) {
      if (opId.equals(StandardOperation.NOT)) {
        makeSubGraph((NodeOperation) operand, cfg, root, falseMerge, trueMerge);
      }
    } else {
      makeAtomicSubGraph(operand, cfg, root, trueMerge, falseMerge);
    }
  }

  private void makeBinOpSubGraph(
    NodeOperation expr,
    Cfg cfg,
    CfgNode root,
    Merge trueMerge,
    Merge falseMerge) {
    /**
     * The constructed sub-graph has the following structure:
     *
     *                       firstSwitch
     *                      |           |
     *                  tCase         fCase
     *                   |               |
     *         trueSecondSwitch      falseSecondSwitch
     *         |               |     |                |
     *      ttCase           tfCase ftCase          ffCase
     *      |                  |     |                |
     *      <--- link layer that is operation specific>
     *        trueMerge                   falseMerge
     **/
    Enum<?> opId = expr.getOperationId();
    Node first = expr.getOperand(0);
    Node second = expr.getOperand(1);

    CfgNode tSecondRoot, fSecondRoot;

    if(isAbleToBeSplit(first)) {

      /* First operand of binary operation to be split. */
      Merge firstTrueMerge = new Merge();
      tSecondRoot = firstTrueMerge;
      cfg.addNode(firstTrueMerge);

      Merge firstFalseMerge = new Merge();
      fSecondRoot = firstFalseMerge;
      cfg.addNode(firstFalseMerge);

      makeSubGraph((NodeOperation) first, cfg, root, firstTrueMerge, firstFalseMerge);
    } else {

      /* First operand of binary operation cannot be split. */
      Switch newSwitch = new Switch(first);
      cfg.addNode(newSwitch);
      root.addChild(newSwitch);

      Case tCase = new Case(NodeValue.newBoolean(true));
      tSecondRoot = tCase;
      newSwitch.addChild(tCase);
      cfg.addNode(tCase);

      Case fCase = new Case(NodeValue.newBoolean(false));
      fSecondRoot = fCase;
      newSwitch.addChild(fCase);
      cfg.addNode(fCase);
    }

    CfgNode ttNode, tfNode, ftNode, ffNode;
    if (isAbleToBeSplit(second)) {

      /* Second operand of binary operation to be split. */
      Merge ttMerge = new Merge();
      ttNode = ttMerge;
      cfg.addNode(ttMerge);

      Merge tfMerge = new Merge();
      tfNode = tfMerge;
      cfg.addNode(tfMerge);

      Merge ftMerge = new Merge();
      ftNode = ftMerge;
      cfg.addNode(ftMerge);

      Merge ffMerge = new Merge();
      ffNode = ffMerge;
      cfg.addNode(ffMerge);

      makeSubGraph((NodeOperation) second, cfg, tSecondRoot, ttMerge, tfMerge);
      makeSubGraph((NodeOperation) second, cfg, fSecondRoot, ftMerge, ffMerge);
    } else {

      /* Second operand of binary operation cannot be split. */
      Switch trueSecondSwitch = new Switch(second);
      cfg.addNode(trueSecondSwitch);
      trueSecondSwitch.addParent(tSecondRoot);

      Switch falseSecondSwitch = new Switch(second);
      cfg.addNode(falseSecondSwitch);
      falseSecondSwitch.addParent(fSecondRoot);

      Case ttCase = new Case(NodeValue.newBoolean(true));
      ttNode = ttCase;
      cfg.addNode(ttCase);
      trueSecondSwitch.addChild(ttCase);

      Case tfCase = new Case(NodeValue.newBoolean(false));
      tfNode = tfCase;
      cfg.addNode(tfCase);
      trueSecondSwitch.addChild(tfCase);

      Case ftCase = new Case(NodeValue.newBoolean(true));
      ftNode = ftCase;
      cfg.addNode(ftCase);
      falseSecondSwitch.addChild(ftCase);

      Case ffCase = new Case(NodeValue.newBoolean(false));
      ffNode = ffCase;
      cfg.addNode(ffCase);
      falseSecondSwitch.addChild(ffCase);
    }
    /* The operation-id dependent link layer. */
    if (opId.equals(StandardOperation.AND)) {
      ttNode.addChild(trueMerge);
      tfNode.addChild(falseMerge);
      ftNode.addChild(falseMerge);
      ffNode.addChild(falseMerge);
    } else if (opId.equals(StandardOperation.OR)) {
      ttNode.addChild(trueMerge);
      tfNode.addChild(trueMerge);
      ftNode.addChild(trueMerge);
      ffNode.addChild(falseMerge);
    } else if (opId.equals(StandardOperation.XOR)) {
      ttNode.addChild(falseMerge);
      tfNode.addChild(trueMerge);
      ftNode.addChild(trueMerge);
      ffNode.addChild(falseMerge);
    } else if (opId.equals(StandardOperation.IMPL)) {
      ttNode.addChild(trueMerge);
      tfNode.addChild(falseMerge);
      ftNode.addChild(trueMerge);
      ffNode.addChild(trueMerge);
    } else
        throw new IllegalArgumentException(String.format("Unexpected operation: %s.", opId));
  }

  private void makeAtomicSubGraph(
      Node expr,
      Cfg cfg,
      CfgNode root,
      Merge trueMerge,
      Merge falseMerge) {

    Switch newSwitch = new Switch(expr);
    root.addChild(newSwitch);
    cfg.addNode(newSwitch);

    Case trueCase = new Case(NodeValue.newBoolean(true));
    newSwitch.addChild(trueCase);
    trueCase.addChild(trueMerge);
    cfg.addNode(trueCase);

    Case falseCase = new Case(NodeValue.newBoolean(false));
    newSwitch.addChild(falseCase);
    falseCase.addChild(falseMerge);
    cfg.addNode(falseCase);
  }

  private boolean isAbleToBeSplit(Node expr) {

    if (expr instanceof NodeOperation) {
      final Set<NodeVariable> stateVars = this.getStateMap().get(this.current);

      final ExprTreeVisitor visitor = new ExprTreeVisitorDefault() {

        @Override
        public void onOperationBegin(NodeOperation op) {
          Collection<NodeVariable> variables = ExprUtils.getVariables(op);

          Set<NodeVariable> varSet = new LinkedHashSet<>(variables);

          if (variables.size() == 1 && CollectionUtils.areIntersectedSets(varSet, stateVars))
            setStatus(Status.ABORT);
        }
      };

      ExprTreeWalker walker = new ExprTreeWalker(visitor);
      walker.visit(expr);
      return ((visitor.getStatus().equals(ExprTreeVisitor.Status.ABORT))
          && isAbleToBeSplit(((NodeOperation) expr).getOperationId()));
    } else {
      return false;
    }
  }

  private boolean isAbleToBeSplit(Enum<?> opId) {
    return (opId.equals(StandardOperation.NOT)
        || opId.equals(StandardOperation.AND)
        || opId.equals(StandardOperation.OR)
        || opId.equals(StandardOperation.XOR)
        || opId.equals(StandardOperation.IMPL));
  }

  @Override
  public void onMergeBegin(Merge merge) {
    Merge newMerge = new Merge();
    getOldNewMap().put(merge, newMerge);
  }

  @Override
  public void onSink(Sink sink) {
    Sink newSink = new Sink();
    getOldNewMap().put(sink, newSink);
  }
}
