/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.Map;
import java.util.Set;

/**
 * Basic class for {@link ru.ispras.retrascope.model.cfg.CfgVisitor} visitors
 * that are used in {@link ru.ispras.retrascope.engine.cgaa.transformer.efsm.CgaaEfsmTransformer}.
 * All the successors of this class can use data containers (mappings) of different semantics
 * for transmitting data to each other.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
class MapCfgVisitor implements CfgVisitor {

  /* The mapping of old nodes to performed new nodes. */
  private Map<CfgModelNode, CfgModelNode> oldNewMap;

  /* The mapping of modules to their clock-like variables. */
  private Map<Module, Set<NodeVariable>> clockMap;

  /* The mapping of modules to their state-like variables. */
  private Map<Module, Set<NodeVariable>> stateMap;

  /**
   * Sets mapping of old model nodes into new model nodes.
   *
   * @param map mapping
   *
   * @throws java.lang.NullPointerException when argument is <code>null</code>.
   */
  public void setOldNewMap(Map<CfgModelNode, CfgModelNode> map) {
    InvariantChecks.checkNotNull(map);

    this.oldNewMap = map;
  }

  /**
   * Returns mapping of old (non-performed) model nodes to new (performed) model nodes.
   *
   * @return mapping
   */
  public Map<CfgModelNode, CfgModelNode> getOldNewMap() { return this.oldNewMap; }

  /**
   * Sets mapping of modules and their clock-like variables.
   *
   * @param map mapping of modules and their clock-like variables
   *
   * @throws java.lang.NullPointerException when argument is <code>null</code>.
   */
  public void setClockMap( Map<Module, Set<NodeVariable>> map) {
    InvariantChecks.checkNotNull(map);

    this.clockMap = map;
  }

  /**
   * Returns the mapping of modules and their clock-like variables.
   *
   * @return mapping of modules and their clock-like variables
   */
  public Map<Module, Set<NodeVariable>> getClockMap() { return this.clockMap; }

  /**
   * Sets mapping of modules and their state-like variables.
   *
   * @param map mapping of modules and their state-like variables
   *
   * @throws java.lang.NullPointerException when argument is <code>null</code>.
   */
  public void setStateMap(Map<Module, Set<NodeVariable>> map) {
    InvariantChecks.checkNotNull(map);

    this.stateMap = map;
  }

  /**
   * Returns the mapping of modules and their state-like variables.
   *
   * @return mapping of modules and their state-like variables
   */
  public Map<Module, Set<NodeVariable>> getStateMap() { return this.stateMap; }

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) { /* Empty. */ }

  @Override
  public void onModuleEnd(Module module) { /* Empty. */ }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { /* Empty. */ }

  @Override
  public void onCfgEnd(Cfg cfg) { /* Empty. */ }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) { /* Empty. */ }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { /* Empty. */ }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
