/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.extractor.conflict;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmVisitor;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This class is used to examine EFSMs for conflict detection.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class EfsmConflictVisitor implements EfsmVisitor {

  // The list of variables.
  private ArrayList<NodeVariable> variables;
  
  // The list of defined variables.
  private ArrayList<NodeVariable> defsList;
  
  // The list of used variables.
  private ArrayList<NodeVariable> usesList;
  
  // The map of duplicating variables.
  private HashMap<NodeVariable, List<EfsmTransition>> duplicatingVariables;

  public EfsmConflictVisitor() {
    clearVariableLists();
  }

  /**
   * Returns the list of defined variables.
   * 
   * @return The {@link List} of defined variables.
   */
  public List<NodeVariable> getDefsList() {
    return defsList;
  }

  /**
   * Returns the list of used variables.
   * 
   * @return The {@link List} of used variables.
   */
  public List<NodeVariable> getUsesList() {
    return usesList;
  }

  /**
   * Returns the list of variables.
   * 
   * @return The {@link List} of variables.
   */
  public List<NodeVariable> getVariables() {
    return variables;
  }

  /**
   * Returns the map of duplicating variables.
   * 
   * @return The {@link Map} of duplicating variables.
   */
  public Map<NodeVariable, List<EfsmTransition>> getDuplicatingVariables() {
    return duplicatingVariables;
  }
  
  /**
   * Resets the lists of variables.
   */
  public void clearVariableLists() {
    variables = new ArrayList<>();
    defsList = new ArrayList<>();
    usesList = new ArrayList<>();
    duplicatingVariables = new HashMap<>();
  }


  @Override
  public void onEfsmBegin(Efsm efsm) {

  }

  @Override
  public void onEfsmEnd(Efsm efsm) {

  }

  @Override
  public void onStateBegin(EfsmState state) {

  }

  @Override
  public void onStateEnd(EfsmState state) {

  }

  @Override
  public void onTransitionBegin(EfsmTransition transition) {

    List<NodeVariable> defs = new ArrayList<>();
    if (transition.getGuardedAction().getAction() != null) {
      Action action = transition.getGuardedAction().getAction();
      List<Assignment> statements = action.getStatements();
      for (int i = 0; i < statements.size(); i++) {
        for (int j = i + 1; j < statements.size(); j++) {
          if (CollectionUtils.areIntersectedSets(statements.get(i).getDefines(), statements.get(j)
              .getDefines())) {
            defs.addAll(CollectionUtils.intersectSets(statements.get(i).getDefines(),
              statements.get(j).getDefines()));
          }
        }
      }

    }
    for (NodeVariable var : defs) {
      if (duplicatingVariables.containsKey(var)) {
        List<EfsmTransition> tmp = duplicatingVariables.get(var);
        tmp.add(transition);
        duplicatingVariables.put(var, tmp);
      } else {
        List<EfsmTransition> tmp = new ArrayList<>();
        tmp.add(transition);
        duplicatingVariables.put(var, tmp);
      }
    }

  }

  @Override
  public void onTransitionEnd(EfsmTransition transition) {

  }

  @Override
  public void onGuardedActionBegin(GuardedAction guardedAction) {

  }

  @Override
  public void onGuardedActionEnd(GuardedAction guardedAction) {

  }

  @Override
  public void onEvent(Event event) {

  }

  @Override
  public void onGuardBegin(Guard guard) {
    Set<NodeVariable> tmpUsesList = guard.getUses();
    for (NodeVariable item : tmpUsesList) {
      if (!usesList.contains(item)) {
        usesList.add(item);
      }
    }
  }

  @Override
  public void onGuardEnd(Guard guard) {

  }

  @Override
  public void onAssignmentActionBegin(Action action) {

  }

  @Override
  public void onAssignmentActionEnd(Action action) {

  }

  @Override
  public void onAssignmentBegin(Assignment statement) {
    Set<NodeVariable> tmpDefsList = statement.getDefines();
    for (NodeVariable item : tmpDefsList) {
      if (!defsList.contains(item)) {
        defsList.add(item);
      }
    }
    Set<NodeVariable> tmpUsesList = statement.getUses();
    for (NodeVariable item : tmpUsesList) {
      if (!usesList.contains(item)) {
        usesList.add(item);
      }
    }

    RangedVariable item = statement.getRangedVariable();
    if (!variables.contains(item.getVariable())) {
      variables.add(item.getVariable());
    }

  }

  @Override
  public void onExpression(Node expression) {
    if (expression != null) {
      Collection<NodeVariable> tmpVarList = ExprUtils.getVariables(expression);

      for (NodeVariable item : tmpVarList) {
        if (!variables.contains(item)) {
          variables.add(item);
        }
      }
    }
  }

  @Override
  public void onAssignmentEnd(Assignment statement) {


  }

}
