/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import ru.ispras.fortress.data.Random;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.NearestUncoveredEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.pathfinders.dfs.DepthFirstSearchPathfinder;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.testbench.Sequence;
import ru.ispras.retrascope.result.testbench.Transaction;
import ru.ispras.retrascope.result.testbench.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * Directed test generator for single EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see RandomGenerator
 * @see EfsmTestGenerator
 */
final class DirectedGenerator extends AtomicTestGenerator {
  private final Logger logger;
  private final String logEntryHeader;

  private final EfsmSimulator simulator;
  private int loopLimit;

  private final EfsmTransitionCoverage coverage;

  /**
   * Constructs a new EFSM atomic test generator for the specified EFSM.
   *
   * @param efsm the EFSM
   * @param initialCoverage a coverage reached on previous steps of test generation
   * @param loopLimit max number of single EFSM loop traversing in one path
   */
  DirectedGenerator(Efsm efsm, EfsmTransitionCoverage initialCoverage, int loopLimit) {
    this.logEntryHeader = "EFSM.TestGenerator.DirectedGenerator." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.loopLimit = loopLimit;

    this.simulator = new EfsmSimulator(efsm);

    this.coverage = new EfsmTransitionCoverage(initialCoverage);
  }

  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    NearestUncoveredEfsmTransitionChooser chooser =
        new NearestUncoveredEfsmTransitionChooser(this.simulator.getStaticModel(), this.coverage);

    for (EfsmTransition transition = chooser.chooseNext(); transition != null;
        transition = chooser.chooseNext()) {
      Sequence sequence = this.generateSequenceDirectionally(transition);

      if (sequence != null) {
        this.coverage.include(this.simulator.getCoverage());

        return sequence.getVectors().iterator();
      }
    }

    return null;
  }

  @Override
  public EfsmTransitionCoverage getCoverage() {
    return this.coverage;
  }

  /**
   * Returns EFSM which this generated is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getStaticModel();
  }

  private Sequence generateSequenceDirectionally(EfsmTransition transition) {
    Efsm efsm = this.simulator.getStaticModel();

    this.logger.log(LogLevel.INFO, this.logEntryHeader
        + ": starting to cover the transition: " + transition);

    try (DepthFirstSearchPathfinder pathfinder =
        new DepthFirstSearchPathfinder(efsm, transition, this.loopLimit)) {

      for (List<EfsmTransition> path = pathfinder.getNext(); path != null;
          path = pathfinder.getNext()) {
        this.logger.log(LogLevel.INFO, this.logEntryHeader
            + ": testing the path for traversability: " + path);

        List<NodeOperation> pathConstraints = pathfinder.getConstraints();
        int constraintsAmount = pathConstraints.size();
        NodeOperation pathConstraint;

        if (constraintsAmount > 1) {
          NodeOperation[] constraintsArray = new NodeOperation[constraintsAmount];
          pathConstraint =
              new NodeOperation(StandardOperation.AND, pathConstraints.toArray(constraintsArray));
        } else {
          pathConstraint = pathConstraints.get(0);
        }

        this.simulator.reset();

        for (NodeVariable variable: ExprUtils.getVariables(pathConstraint)) {
          String variableName = variable.getName();

          if (efsm.containsInnerVariable(variableName)) {
            NodeValue defaultValue = this.simulator.getInnerValue(variableName);
            pathConstraint =
                (NodeOperation) Transformer.substitute(pathConstraint, variableName, defaultValue);
          }
        }

        this.logger.log(LogLevel.INFO, this.logEntryHeader
            + ": the corresponding constraint: " + pathConstraint);

        SolverResult result = ConstraintUtils.solve(ConstraintUtils.newConstraint(pathConstraint));

        if (result.getStatus().equals(Status.SAT)) {
          this.logger.log(LogLevel.INFO, this.logEntryHeader
              + ": the path is traversable. Traversing it");

          int pathLength = path.size();
          Map<Integer, Map<String, Node>> inputValues = new HashMap<>();

          for (Variable variable: result.getVariables()) {
            String annotatedName = variable.getName();
            String signalName =
                DepthFirstSearchPathfinder.deannotateName(annotatedName);
            Integer cycleNumber = Integer.valueOf(pathLength -
                DepthFirstSearchPathfinder.getCycleNumberForAnnotatedInput(annotatedName) - 1);

            if (!inputValues.containsKey(Integer.valueOf(cycleNumber))) {
              inputValues.put(cycleNumber, new HashMap<String, Node>());
            }

            inputValues.get(cycleNumber).put(signalName, new NodeValue(variable.getData()));
          }

          Sequence sequence = new Sequence();

          for (Integer i = 0; i < pathLength; i++) {
            Vector vector = new Vector();

            if (inputValues.containsKey(i)) {
              Transaction transaction = new Transaction(inputValues.get(i));
              vector.addTransaction(transaction);
            }

            EfsmTransition nextTransition = path.get(i);
            vector.addEvents(nextTransition.getGuardedAction().getClocks());
            vector = this.fillUnusedInputsRandomly(efsm, vector);

            this.simulator.setInputValues(vector.getTransactions());
            this.simulator.processEvents(vector.getEventList());

            EfsmState expectedState = nextTransition.getDestinationState();
            EfsmState currentState = this.simulator.getCurrentState();

            if (!currentState.equals(expectedState)) {
                throw new RuntimeException("Unexpected simulation result. Expected state " +
                expectedState + ". Current state: " + currentState);
            }

            sequence.addVector(vector);
          }

          if (this.simulator.getCoverage().isCovered(transition)) {
            this.logger.log(LogLevel.INFO,
                this.logEntryHeader + ": the transition has been successfully covered");

            return sequence;
          } else {
            throw new RuntimeException("Unexpected simulation result. The target transition is not covered");
          }
        } else {
          this.logger.log(LogLevel.INFO, this.logEntryHeader
              + ": the path is untraversable. Skipping it");
        }
      }
    } catch (Exception e) {
      throw new RuntimeException(e);
    }

    this.logger.log(LogLevel.INFO, this.logEntryHeader
        + ": the transition " + transition + " hasn't been covered");

    return null;
  }

  private Vector fillUnusedInputsRandomly(Efsm efsm, Vector inputVector) {
    Transaction randomlyFilledTransaction = new Transaction();

    this.logger.log(LogLevel.INFO, this.logEntryHeader
        + ": generating random values for unused inputs");

    Collection<Event> events = inputVector.getEventList().getEvents();
    Set<String> eventNames = new HashSet<>(events.size());

    for (Event event: events) {
      eventNames.add(event.getRangedVariable().getName());
    }

    for (String inputName: efsm.getInputDataTypes().keySet()) {
      if (!inputVector.containsVariable(inputName) && !eventNames.contains(inputName)) {
        randomlyFilledTransaction.addVariable(inputName,
            new NodeValue(Random.newValue(efsm.getInputDataType(inputName))));
      }
    }

    if (!randomlyFilledTransaction.isEmpty()) {
      inputVector.addTransaction(randomlyFilledTransaction);

      this.logger.log(LogLevel.INFO, this.logEntryHeader
          + ": randomly generated transaction: " + randomlyFilledTransaction);
    } else {
      this.logger.log(LogLevel.INFO, this.logEntryHeader
          + ": all inputs are already set");
    }

    return inputVector;
  }
}
