/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.SemiRandomEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.testbench.Transaction;
import ru.ispras.retrascope.result.testbench.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * Random-based test generator for single EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see DirectedGenerator
 * @see EfsmTestGenerator
 */
final class RandomGenerator extends AtomicTestGenerator {
  private final Logger logger;
  private final String logEntryHeader;

  private final EfsmSimulator simulator;

  private final EfsmTransitionCoverage coverage;

  private Vector inputConstraint;
  private SemiRandomEfsmTransitionChooser chooser;

  private long seed;

  private final Set<EfsmTransition> priorityTransitions;

  /**
   * Constructs a new EFSM atomic test generator for the specified EFSM.
   *
   * @param efsm the EFSM
   * @param loopLimit max number of single EFSM loop traversing in one path
   */
  RandomGenerator(Efsm efsm) {
    this.logEntryHeader = "EFSM.TestGenerator.RandomGenerator." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.simulator = new EfsmSimulator(efsm);

    this.inputConstraint = null;

    this.coverage = new EfsmTransitionCoverage(efsm);
    this.priorityTransitions = new HashSet<>(this.coverage.getUncoveredTransitions());

    this.chooser = null;
    this.seed = 0L;
  }

  /**
   * Sets the specified constraint on the inputs of EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  void setAdditionalConstraintOnIputs(Vector constraint) {
    this.inputConstraint = constraint;
  }

  private boolean isCurrentSequenceFinished() {
    return this.coverage.isFull() || ((this.chooser != null) && !this.chooser.hasNext());
  }

  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    return new RandomSequenceIterator();
  }

  @Override
  public EfsmTransitionCoverage getCoverage() {
    return this.coverage;
  }

  /**
   * Returns EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getStaticModel();
  }

  /**
   * Feeds the specified vector to the inner EFSM simulator.
   *
   * @return the EFSM
   */
  void submitVector(Vector vector) {
    this.simulator.setInputValues(vector.getTransactions());
    this.simulator.processEvents(vector.getEventList());

    this.chooser = new SemiRandomEfsmTransitionChooser(this.simulator.getStaticModel(),
        this.simulator.getCurrentState(), this.coverage, this.priorityTransitions, this.seed);
  }

  private Vector generateInputVectorRandomly() {
    this.logger.log(LogLevel.INFO, this.logEntryHeader + ": generating a new input vector");

    Vector vector = null;

    if (this.chooser == null) {
      vector = this.reset();
    } else {
      for (EfsmTransition transition = this.chooser.chooseNext(); transition != null;
          transition = this.chooser.chooseNext()) {
        vector = this.traverseTransition(transition);

        if (vector != null) {
          break;
        }
      }
    }

    return vector;
  }

  private Vector reset() {
    this.logger.log(LogLevel.INFO, this.logEntryHeader + ": resetting the EFSM");

    Vector vector =
        this.enableGuardedAction(this.simulator.getStaticModel().getResetGuardedAction());

    if (vector != null) {
      this.logger.log(LogLevel.INFO,
          this.logEntryHeader + ": the EFSM has been reset successfully");
    } else {
      this.logger.log(LogLevel.WARNING,
          this.logEntryHeader + ": failed to reset the EFSM");
    }

    return vector;
  }

  private Vector traverseTransition(EfsmTransition transition) {
    this.logger.log(LogLevel.INFO,
        this.logEntryHeader + ": traversing the transition " + transition);

    Vector inputVector = this.enableGuardedAction(transition.getGuardedAction());

    if (inputVector != null) {
      this.logger.log(LogLevel.INFO,
          this.logEntryHeader + ": the transition has been traversed successfully");

      this.coverage.markTransitionAsCovered(transition);
      this.priorityTransitions.remove(transition);
      this.chooser =
          new SemiRandomEfsmTransitionChooser(this.simulator.getStaticModel(),
              this.simulator.getCurrentState(), this.coverage, this.priorityTransitions, this.seed);
    } else {
      this.logger.log(LogLevel.INFO, this.logEntryHeader + ": failed to traverse the transition");
    }

    return inputVector;
  }

  private Vector enableGuardedAction(GuardedAction guardedAction) {
    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": enabling the guarded action: " + guardedAction);

    if (this.inputConstraint != null) {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": fixed inputs: " + this.inputConstraint);

      EventList gaEventList = new EventList();
      gaEventList.addEvents(guardedAction.getClocks());

      if (!gaEventList.isCompatible(this.inputConstraint.getEventList())) {
        this.logger.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the transition isn't compatible with the fixed events");

        return null;
      }
    }

    Node guardExpression = guardedAction.getGuard().toNode();
    Collection<NodeVariable> usedVariables = ExprUtils.getVariables(guardExpression);
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = usedVariable.getName();
      NodeValue variableValue = null;
      Map<String, NodeValue> innerVariables = this.simulator.getInnerValues();

      if (innerVariables.containsKey(variableName)) {
        variableValue = innerVariables.get(variableName);
      } else {
        if ((this.inputConstraint != null) && this.inputConstraint.containsVariable(variableName)) {
          variableValue = (NodeValue) this.inputConstraint.getValue(variableName);
        }
      }

      if (variableValue != null) {
        bindingList.add(NodeBinding.bindVariable(usedVariable, variableValue));
      }
    }

    guardExpression =
        Transformer.substituteAllBindings(new NodeBinding(guardExpression, bindingList));

    Constraint targetConstraint = ConstraintUtils.newConstraint(guardExpression);
    SolverResult solverResult = ConstraintUtils.solve(targetConstraint);

    if (solverResult.getStatus().equals(Status.SAT)) {
      Efsm efsm = this.simulator.getStaticModel();
      Vector inputVector = new Vector();
      Transaction transaction = new Transaction();

      for (Variable variable : solverResult.getVariables()) {
        String variableName = variable.getName();

        if (efsm.getInputDataTypes().containsKey(variableName)) {
          transaction.addVariable(variableName, new NodeValue(variable.getData()));
        }
      }

      inputVector.addEvents(guardedAction.getClocks());

      if (!transaction.isEmpty()) {
        inputVector.addTransaction(transaction);
      }

      if (this.inputConstraint != null) {
        inputVector.addEvents(this.inputConstraint.getEventList().getEvents());

        for (Transaction fixedTransaction: this.inputConstraint.getTransactions()) {
          inputVector.addTransaction(fixedTransaction);
        }
      }

      this.logger.log(LogLevel.INFO, this.logEntryHeader
          + ": the guarded action has been enabled successfully. Input vector: " + inputVector);

      return inputVector;
    } else {
      this.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": failed to enable the guarded action");

      return null;
    }
  }

  private class RandomSequenceIterator implements Iterator<Vector> {
    RandomSequenceIterator() {
      RandomGenerator.this.simulator.reset();
      RandomGenerator.this.coverage.reset();
      RandomGenerator.this.chooser = null;
      RandomGenerator.this.seed++;
    }

    @Override
    public boolean hasNext() {
      return !RandomGenerator.this.isCurrentSequenceFinished();
    }

    @Override
    public Vector next() {
      if (this.hasNext()) {
        return RandomGenerator.this.generateInputVectorRandomly();
      } else {
        return null;
      }
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException("remove");
    }
  }
}
