/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

import edu.uci.ics.jung.graph.DirectedSparseGraph;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Random;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintCombiner;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.engine.efsm.generator.test.Substitutor;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.NearestUncoveredEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulationSnapshot;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis;
import ru.ispras.retrascope.result.testbench.Sequence;
import ru.ispras.retrascope.result.testbench.Transaction;
import ru.ispras.retrascope.result.testbench.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * The FATE test generator which deals with single EFSM and operates at the vector-level.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmFateTestGenerator
 */
public class DirectedFatePlusGenerator extends AtomicTestGenerator {
  private final Logger logger;
  private final String logEntryHeader;

  private final Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityInformation;
  private final Map<EfsmTransition, Set<ReachabilityInformation>> testedReachabilityInformation;

  private EfsmTransitionCoverage coverage;

  private final EfsmSimulator simulator;

  private final DirectedSparseGraph<EfsmTransition, VariableDependency> controlDependencyGraph;
  private final Map<NodeVariable, Set<List<EfsmTransition>>> counters;

  private NearestUncoveredEfsmTransitionChooser directedChooser;

  private final int loopLimit;

  /**
   * Constructs a test generator for the specified EFSM with the specified parameters.
   *
   * @param efsm the EFSM
   * @param loopLimit a depth limit for loop unfolding
   */
  public DirectedFatePlusGenerator(Efsm efsm, int loopLimit,
      Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityInformation) {
    this.logEntryHeader = "EFSM.FATE+.DirectedGenerator." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.simulator = new EfsmSimulator(efsm);

    EfsmAnalyser efsmAnalyser = new EfsmAnalyser(efsm);
    this.controlDependencyGraph = efsmAnalyser.getControlDependencyGraph();
    this.counters = efsmAnalyser.getCounters();

    this.coverage = new EfsmTransitionCoverage(efsm);

    for (EfsmTransition coveredTransition: reachabilityInformation.keySet()) {
      this.coverage.markTransitionAsCovered(coveredTransition);
    }

    this.reachabilityInformation = new HashMap<>(reachabilityInformation);
    this.testedReachabilityInformation = new HashMap<>();

    this.directedChooser = null;

    this.loopLimit = loopLimit;
  }

  /**
   * Returns an EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getStaticModel();
  }

  /**
   * Returns a coverage reached during generation.
   *
   * @return the coverage
   */
  @Override
  public EfsmTransitionCoverage getCoverage() {
    return this.coverage;
  }

  /**
   * Returns an iterator enabling to retrieve next sequence of vectors produced by this generator.
   *
   * @return the iterator. If this generator is already finished generation, {@code null} is
   * returned
   */
  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    if (!this.coverage.isFull()) {
      if (this.directedChooser == null) {
        this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
            this.simulator.getStaticModel(), this.coverage);
      }

      for (EfsmTransition transition = this.directedChooser.chooseNext(); transition != null;
          transition = this.directedChooser.chooseNext()) {
        Sequence sequence = this.generateSequence(transition);

        if (sequence != null) {
          this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
              this.simulator.getStaticModel(), this.coverage);

          return sequence.getVectors().iterator();
        }
      }
    }

    return null;
  }

  private Vector reset() {
    this.logger.log(LogLevel.INFO, this.logEntryHeader + ": resetting the EFSM");

    this.simulator.reset();

    Vector vector =
        this.enableGuardedAction(this.simulator.getStaticModel().getResetGuardedAction(), null);

    if (vector != null) {
      this.logger.log(LogLevel.INFO,
          this.logEntryHeader + ": the EFSM has been reset successfully");
    } else {
      this.logger.log(LogLevel.WARNING,
          this.logEntryHeader + ": failed to reset the EFSM");
    }

    return vector;
  }

  private Sequence generateSequence(EfsmTransition transition) {
    Efsm efsm = this.simulator.getStaticModel();
 
    this.logger.log(LogLevel.INFO, this.logEntryHeader
        + ": generating a sequence to cover the transition: " + transition);

    if (transition.getSourceState().equals(this.simulator.getStaticModel().getInitialState())) {
      Vector vector = this.reset();

      if (vector != null) {
        this.simulator.setInputValues(vector.getTransactions());
        this.simulator.processEvents(vector.getEventList());
        Sequence sequence = new Sequence();
        sequence.addVector(vector);
        List<EfsmTransition> pseudoPath = new ArrayList<>(1);
        pseudoPath.add(transition);
        List<Vector> vectors = this.traversePath(pseudoPath);

        if (vectors != null) {
          vector = vectors.get(0);
          
          EfsmSimulationSnapshot sourceStateSnapshot = this.simulator.makeSimulationSnapshot();

          this.simulator.setInputValues(vector.getTransactions());
          this.simulator.processEvents(vector.getEventList());

          if (!this.simulator.getCurrentState().equals(transition.getDestinationState())) {
            throw new RuntimeException("Unexpected simulator state: " +
                this.simulator.getCurrentState() + " instead of " +
                transition.getDestinationState());
          }

          Set<ReachabilityInformation> transitionReachabilityInformation =
              this.reachabilityInformation.get(transition);

          if (transitionReachabilityInformation == null) {
            transitionReachabilityInformation = new HashSet<>();
            this.reachabilityInformation.put(transition, transitionReachabilityInformation);
          }

          transitionReachabilityInformation.add(new ReachabilityInformation(sequence,
              sequence.getLength(), sourceStateSnapshot, this.simulator.makeSimulationSnapshot()));
          this.coverage.markTransitionAsCovered(transition);
          sequence.addVector(vector);

          return sequence;
        }
      }
    } else {
      Set<EfsmTransition> coveredTransitions = this.coverage.getCoveredTransitions();

      for (EfsmTransition previousTransition:
          efsm.getIncomingTransitions(transition.getSourceState())) {
        if (coveredTransitions.contains(previousTransition)) {
          for (ReachabilityInformation reachabilityInformation:
              this.reachabilityInformation.get(previousTransition)) {
            if (this.testedReachabilityInformation.containsKey(transition) &&
                this.testedReachabilityInformation.get(transition).contains(reachabilityInformation)) {
              continue;
            }

            this.logger.log(LogLevel.INFO, this.logEntryHeader
                + ": simulating the model to the end state of the transition: " + previousTransition);

            this.reach(reachabilityInformation, true);

            List<EfsmTransition> pseudoPath = new ArrayList<>(1);
            pseudoPath.add(transition);
            List<Vector> vectors = this.traversePath(pseudoPath);

            if (vectors != null) {
              Vector vector = vectors.get(0);
              Sequence sequence = new Sequence();

              for (int i = 0; i <= reachabilityInformation.vectorNumber; i++) {
                sequence.addVector(reachabilityInformation.sequence.getVector(i));
              }

              EfsmSimulationSnapshot sourceStateSnapshot = this.simulator.makeSimulationSnapshot();

              this.simulator.setInputValues(vector.getTransactions());
              this.simulator.processEvents(vector.getEventList());

              if (!this.simulator.getCurrentState().equals(transition.getDestinationState())) {
                throw new RuntimeException("Unexpected simulator state: " +
                    this.simulator.getCurrentState() + " instead of " +
                    transition.getDestinationState());
              }

              Set<ReachabilityInformation> transitionReachabilityInformation =
                  this.reachabilityInformation.get(transition);

              if (transitionReachabilityInformation == null) {
                transitionReachabilityInformation = new HashSet<>();
                this.reachabilityInformation.put(transition, transitionReachabilityInformation);
              }

              transitionReachabilityInformation.add(new ReachabilityInformation(sequence,
                  sequence.getLength(), sourceStateSnapshot, this.simulator.makeSimulationSnapshot()));
              this.coverage.markTransitionAsCovered(transition);
              sequence.addVector(vector);

              if (this.testedReachabilityInformation.containsKey(transition)) {
                this.testedReachabilityInformation.remove(transition);
              }

              return sequence;
            }
          }
        }
      }
    }

    return this.coverByBackjumping(transition);
  }

  private Sequence coverByBackjumping(EfsmTransition transition) {
    this.logger.log(LogLevel.INFO, this.logEntryHeader
        + ": using backjumping to cover the transition: " + transition);
 
    Efsm efsm = this.simulator.getStaticModel();

    for (VariableDependency dependency: this.controlDependencyGraph.getInEdges(transition)) {
      EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);
      NodeVariable dependencyVariable = dependency.getVariable();
 
      if (this.coverage.getCoveredTransitions().contains(definingTransition)) {
        for (ReachabilityInformation reachabilityInformation:
            this.reachabilityInformation.get(definingTransition)) {
          if (this.testedReachabilityInformation.containsKey(transition) &&
              this.testedReachabilityInformation.get(transition).contains(reachabilityInformation)) {
            continue;
          }

          this.logger.log(LogLevel.INFO, this.logEntryHeader
              + ": the transition " + definingTransition + " defining the variable " +
              dependencyVariable.getName() + " is covered. Trying it");

          this.reach(reachabilityInformation, false);

          boolean isCounter = false;

          if (this.counters.containsKey(dependencyVariable)) {
            for (List<EfsmTransition> loop: this.counters.get(dependencyVariable)) {
              if (loop.get(0).equals(definingTransition)) {
                isCounter = true;

                break;
              }
            }
          }

          LinkedList<EfsmTransition> path =
              new LinkedList<>(efsm.getShortestPath(definingTransition.getDestinationState(),
              transition.getSourceState(), TransitionWeightBasis.GUARD_REGISTER_AMOUNT));

          this.logger.log(LogLevel.INFO, this.logEntryHeader + ": the shortest path from the "
              + "destination state of the defining transition to the source state of the target "
              + "transition: " + path);

          path.addLast(transition);

          List<Vector> vectors;

          if (isCounter) {
            this.logger.log(LogLevel.INFO, this.logEntryHeader + ": the variable " +
                dependencyVariable.getName() + " is a counter");

            vectors = this.traverseLoop(dependencyVariable, definingTransition, path);
          } else {
            path.addFirst(definingTransition);
            vectors = this.traversePath(path);
          }

          if (vectors != null) {
            Sequence sequence = new Sequence();

            for (int i = 0; i < reachabilityInformation.vectorNumber; i++) {
              sequence.addVector(reachabilityInformation.sequence.getVector(i));
            }

            this.logger.log(LogLevel.INFO,
                this.logEntryHeader + ": simulating the generated vectors");

            for (ListIterator<EfsmTransition> i = path.listIterator(); i.hasNext();) {
              int index = i.nextIndex();
              EfsmTransition pathTransition = i.next();

              Vector vector = vectors.get(index);
              EfsmSimulationSnapshot sourceStateSnapshot = this.simulator.makeSimulationSnapshot();

              this.simulator.setInputValues(vector.getTransactions());
              this.simulator.processEvents(vector.getEventList());

              if (!this.simulator.getCurrentState().equals(pathTransition.getDestinationState())) {
                throw new RuntimeException("Unexpected simulator state: " +
                    this.simulator.getCurrentState() + " instead of " +
                    pathTransition.getDestinationState());
              }

              Set<ReachabilityInformation> transitionReachabilityInformation =
                  this.reachabilityInformation.get(pathTransition);

              if (transitionReachabilityInformation == null) {
                transitionReachabilityInformation = new HashSet<>();
                this.reachabilityInformation.put(pathTransition, transitionReachabilityInformation);
              }

              transitionReachabilityInformation.add(new ReachabilityInformation(sequence,
                  sequence.getLength(), sourceStateSnapshot, this.simulator.makeSimulationSnapshot()));
              this.coverage.markTransitionAsCovered(pathTransition);
              sequence.addVector(vector);
            }

            return sequence;
          } else {
            Set<ReachabilityInformation> testedRi = this.testedReachabilityInformation.get(transition);

            if (testedRi == null) {
              testedRi = new HashSet<>();
              this.testedReachabilityInformation.put(transition, testedRi);
            }

            testedRi.add(reachabilityInformation);
          }
        }
      } else {
        this.logger.log(LogLevel.DEBUG, this.logEntryHeader
            + ": the transition " + definingTransition + " defining the variable " +
            dependency.getVariable().getName() + " isn't covered. Skipping it");
      }
    }

    return null;
  }

  private void reach(ReachabilityInformation reachabilityInformation, boolean withLastTransition) {
    if (withLastTransition) {
      this.simulator.loadSnapshot(reachabilityInformation.destinationStateSnapshot);
    } else {
      this.simulator.loadSnapshot(reachabilityInformation.sourceStateSnapshot);
    }
  }

  private List<Vector> traversePath(List<EfsmTransition> path) {
    this.logger.log(LogLevel.INFO,
        this.logEntryHeader + ": the path to be traversed: " + path);

    Efsm efsm = this.simulator.getStaticModel();
    Substitutor substitutor = new Substitutor(efsm);
    Node pathConstraint = substitutor.backtrack(path);

    this.logger.log(LogLevel.INFO,
        this.logEntryHeader + ": the traverse constraint: " + pathConstraint);

    pathConstraint = this.fillExpression(pathConstraint);

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the constraint to resolve: " + pathConstraint);

    SolverResult solverResult =
        ConstraintUtils.solve(ConstraintUtils.newConstraint(pathConstraint));

    if (solverResult.getStatus().equals(Status.SAT)) {
      this.logger.log(LogLevel.INFO,
          this.logEntryHeader + ": the path is traversable. Generating input vectors");

      Map<Integer, Map<String, Data>> inputs =
          Substitutor.deannotateInputs(solverResult.getVariables());

      List<Vector> vectors = new ArrayList<>(path.size());

      for (ListIterator<EfsmTransition> i = path.listIterator(); i.hasNext();) {
        int index = i.nextIndex();
        EfsmTransition transition = i.next();

        Map<String, Data> vectorInputs = inputs.get(Integer.valueOf(index));
        Transaction transaction = new Transaction();

        if (vectorInputs != null) {
          for (Entry<String, Data> input: vectorInputs.entrySet()) {
            transaction.addVariable(input.getKey(), new NodeValue(input.getValue()));
          }
        }

        Vector vector = new Vector();
        vector.addTransaction(transaction);
        vector.addEvents(transition.getGuardedAction().getClocks());
        vector = this.fillUnusedInputsRandomly(vector);
        vectors.add(vector);

        this.logger.log(LogLevel.INFO,
            this.logEntryHeader + ": generated vector: " + vector);
      }

      return vectors;
    } else {
      this.logger.log(LogLevel.INFO,
          this.logEntryHeader + ": the path is untraversable");

      return null;
    }
  }

  private Node fillExpression(Node expression) {
    Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());
    Set<Node> inputConstraints = new HashSet<>();
    Efsm efsm = this.simulator.getStaticModel();

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = Substitutor.deannotateName(usedVariable.getName());

      if (efsm.containsInput(variableName)) {
        Node variableConstraint = efsm.getInvariantCondition(variableName);

        if (variableConstraint != null) {
          inputConstraints.add(Transformer.substitute(variableConstraint, variableName, usedVariable));
        }
      } else {
        Map<String, NodeValue> innerVariables = this.simulator.getInnerValues();

        if (innerVariables.containsKey(variableName)) {
          bindingList.add(NodeBinding.bindVariable(usedVariable, innerVariables.get(variableName)));
        }
      }
    }

    expression =
        Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));

    if (!inputConstraints.isEmpty()) {
      Node inputConstraint =
          ExprUtils.getConjunction(inputConstraints.toArray(new Node[inputConstraints.size()]));
      expression = Node.AND(expression, inputConstraint);
    }

    return expression;
  }

  private List<Vector> traverseLoop(NodeVariable counter, EfsmTransition definingTransition,
      List<EfsmTransition> tailPath) {
    Set<List<EfsmTransition>> loops = this.counters.get(counter);
    Set<List<EfsmTransition>> completeLoops = new HashSet<>(loops.size());

    for (List<EfsmTransition> loop: loops) {
      if (loop.get(0).equals(definingTransition)) {
        LinkedList<EfsmTransition> completeLoop = new LinkedList<>(loop);

        for (ListIterator<EfsmTransition> i = completeLoop.listIterator(); i.hasNext();) {
          EfsmTransition currentTransition = i.next();

          EfsmState sourceState = currentTransition.getDestinationState();
          EfsmState destinationState;

          if (i.hasNext()) {
            EfsmTransition nextTransition = i.next();
            destinationState = nextTransition.getSourceState();
            i.previous();
          } else {
            destinationState = completeLoop.getFirst().getSourceState();
          }

          for (EfsmTransition transition:
              this.simulator.getStaticModel().getShortestPath(sourceState, destinationState,
                  TransitionWeightBasis.GUARD_REGISTER_AMOUNT)) {
            i.add(transition);
          }
        }

        if (completeLoop.size() > 1) {
          completeLoop.addLast(completeLoop.removeFirst());
        }

        completeLoops.add(completeLoop);
      }
    }

    if (completeLoops.size() > 1) {
      this.logger.log(LogLevel.INFO, this.logEntryHeader + ": traversing the loop");

      List<EfsmTransition> loopPath = new LinkedList<EfsmTransition>();
      loopPath.add(definingTransition);

      return this.traverseMixedLoop(completeLoops, new LinkedList<List<EfsmTransition>>(),
          tailPath);
    }

    return null;
  }

  private List<Vector> traverseMixedLoop(Set<List<EfsmTransition>> loops,
      LinkedList<List<EfsmTransition>> currentPath, List<EfsmTransition> tailPath) {
    if (currentPath.size() == this.loopLimit) {
      return null;
    }

    for (List<EfsmTransition> loop: loops) {
      currentPath.addLast(loop);
      List<EfsmTransition> loopPath = new LinkedList<EfsmTransition>();
      loopPath.add(loop.get(loop.size() - 1));

      for (List<EfsmTransition> previousLoop: currentPath) {
        loopPath.addAll(previousLoop);
      }

      this.logger.log(LogLevel.INFO,
          this.logEntryHeader + ": checking the loop for consistency: " + loopPath);

      if (this.traversePath(loopPath) != null) {
        this.logger.log(LogLevel.INFO,
            this.logEntryHeader + ": the loop is cosistent. Trying to reach the target transition");

        loopPath.addAll(tailPath);

        List<Vector> vectors = this.traversePath(loopPath);

        if (vectors != null) {
          tailPath.clear();
          tailPath.addAll(loopPath);

          return vectors;
        } else {
          if (currentPath.size() < this.loopLimit) {
            vectors = this.traverseMixedLoop(loops, currentPath, tailPath);

            if (vectors != null) {
              return vectors;
            }
          }
        }
      } else {
        this.logger.log(LogLevel.INFO, this.logEntryHeader + ": the loop is incosistent");
      }

      currentPath.removeLast();
    }

    return null;
  }

  private Vector fillUnusedInputsRandomly(Vector inputVector) {
    Efsm efsm = this.simulator.getStaticModel();
    Transaction randomlyFilledTransaction = new Transaction();

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": generating random values for unused inputs");

    Collection<Event> events = inputVector.getEventList().getEvents();
    Set<String> eventNames = new HashSet<>(events.size());

    for (Event event: events) {
      eventNames.add(event.getRangedVariable().getName());
    }

    for (String inputName: efsm.getInputDataTypes().keySet()) {
      if (!inputVector.containsVariable(inputName)) {
        NodeValue value;
        
        Node constraint = efsm.getInvariantCondition(inputName);

        if (constraint == null) {
          value = new NodeValue(Random.newValue(efsm.getInputDataType(inputName)));
        } else {
          if ((efsm.getInputDataType(inputName).getTypeId() == DataTypeId.LOGIC_INTEGER) &&
              (constraint instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperandCount() == 2) &&
              (((NodeOperation) constraint).getOperand(0) instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperand(1) instanceof NodeOperation)) {
            NodeOperation firstCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(0);
            NodeOperation secondCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(1);

            if (firstCondition.getOperationId().equals(StandardOperation.GREATEREQ) &&
                secondCondition.getOperationId().equals(StandardOperation.LESSEQ)) {
              int min = ((NodeValue)firstCondition.getOperand(1)).getInteger().intValue();
              int max = ((NodeValue)secondCondition.getOperand(1)).getInteger().intValue();

              NodeValue module = new NodeValue(Data.newInteger(max - min));
              NodeValue minValueNode = new NodeValue(Data.newInteger(min));
              value = new NodeValue(Random.newValue(DataType.INTEGER));
              NodeVariable targetVariable = new NodeVariable(inputName, DataType.INTEGER);

              NodeOperation moduledValueComputation =
                  new NodeOperation(StandardOperation.MOD, value, module);
              NodeOperation valueComputation =
                  new NodeOperation(StandardOperation.ADD, moduledValueComputation, minValueNode);
              NodeOperation assignment =
                  new NodeOperation(StandardOperation.EQ, targetVariable, valueComputation);
              SolverResult result =
                  ConstraintUtils.solve(ConstraintUtils.newConstraint(assignment));

              if (result.getStatus() == Status.SAT) {
                value = new NodeValue(result.getVariables().iterator().next().getData());
              } else {
                throw new RuntimeException("Failed to resolve invariant constraint on input " +
                    inputName + " due to the errors: " + result.getErrors());
              }
            } else {
              throw new IllegalArgumentException("Such invariant is currently unsupported: " +
                  constraint);
            }
          } else {
            SolverResult result = ConstraintUtils.solve(ConstraintUtils.newConstraint(constraint));

            if (result.getStatus() == Status.SAT) {
              value = new NodeValue(result.getVariables().iterator().next().getData());
            } else {
              throw new RuntimeException("Failed to resolve invariant constraint on input " +
                  inputName + " due to the errors: " + result.getErrors());
            }
          }
        }

        randomlyFilledTransaction.addVariable(inputName, value);
      }
    }

    if (!randomlyFilledTransaction.isEmpty()) {
      inputVector.addTransaction(randomlyFilledTransaction);

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": randomly generated transaction: " + randomlyFilledTransaction);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": all inputs are already set");
    }

    return inputVector;
  }

  private Vector enableGuardedAction(GuardedAction guardedAction, Node additionalExpression) {
    Node guardExpression = guardedAction.getGuard().toNode();
    Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(guardExpression));
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());
    Set<Node> inputConstraints = new HashSet<>();
    Efsm efsm = this.simulator.getStaticModel();

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = usedVariable.getName();

      if (efsm.containsInput(variableName)) {
        Node variableConstraint = efsm.getInvariantCondition(variableName);

        if (variableConstraint != null) {
          inputConstraints.add(variableConstraint);
        }
      } else {
        Map<String, NodeValue> innerVariables = this.simulator.getInnerValues();

        if (innerVariables.containsKey(variableName)) {
          bindingList.add(NodeBinding.bindVariable(usedVariable, innerVariables.get(variableName)));
        }
      }
    }

    guardExpression =
        Transformer.substituteAllBindings(new NodeBinding(guardExpression, bindingList));

    Constraint targetConstraint = ConstraintUtils.newConstraint(guardExpression);

    StringBuilder logMessageBuilder = new StringBuilder(this.logEntryHeader);
    logMessageBuilder.append(": trying to generate an input vector for the guard expression ");
    logMessageBuilder.append(guardExpression);

    if (additionalExpression != null)
    {
      logMessageBuilder.append("\n").append("Additional constraint: ").append(additionalExpression);

      Constraint additionalConstraint = ConstraintUtils.newConstraint(additionalExpression);
      targetConstraint = ConstraintCombiner.makeConjunction(targetConstraint, additionalConstraint);

      for (NodeVariable usedVariable: ExprUtils.getVariables(additionalExpression)) {
        String variableName = usedVariable.getName();

        if (efsm.containsInput(variableName)) {
          Node variableConstraint = efsm.getInvariantCondition(variableName);

          if (variableConstraint != null) {
            inputConstraints.add(variableConstraint);
          }
        }
      }
    }

    if (!inputConstraints.isEmpty()) {
      Node inputConstraint =
          ExprUtils.getConjunction(inputConstraints.toArray(new Node[inputConstraints.size()]));

      logMessageBuilder.append("\n").append("Constraint on inputs: ").append(inputConstraint);

      targetConstraint = ConstraintCombiner.makeConjunction(targetConstraint,
          ConstraintUtils.newConstraint(inputConstraint));
    }

    this.logger.log(LogLevel.DEBUG, logMessageBuilder.toString());

    SolverResult solverResult = ConstraintUtils.solve(targetConstraint);

    if (solverResult.getStatus().equals(Status.SAT)) {
        Vector inputVector = new Vector();
        Transaction transaction = new Transaction();

        for (Variable variable : solverResult.getVariables()) {
          String variableName = variable.getName();

          if (efsm.getInputDataTypes().containsKey(variableName)) {
            transaction.addVariable(variableName, new NodeValue(variable.getData()));
          }
        }

        inputVector.addEvents(guardedAction.getClocks());

        if (!transaction.isEmpty()) {
          inputVector.addTransaction(transaction);
        }

        this.logger.log(LogLevel.DEBUG, this.logEntryHeader
            + ": the input vector has been generated successfully: " + inputVector);

        return inputVector;
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": failed to generate an input vector");

      return null;
    }
  }
}
