/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.Set;

import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;

/**
 * The class representing a random-based heuristic for choosing a transitions to traverse.
 * <br>
 * Each instance of {@code SemiRandomEfsmTransitionChooser} is bounded to the specified
 * EFSM, its state, and test coverage.
 * <br>
 * This class is very similar to {@link RandomEfsmTransitionChooser} but during choosing it uses
 * coverage information in order to firstly choose uncovered transitions.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see RandomEfsmTransitionChooser
 */
public final class SemiRandomEfsmTransitionChooser extends EfsmTransitionChooser {
  private final List<EfsmTransition> priorityTransitions;
  private final List<EfsmTransition> uncoveredTransitions;
  private final List<EfsmTransition> coveredTransitions;

  private final Random randomNumberGenerator;

  /**
   * Creates a new SemiRandomEfsmTransitionChooser for the specified EFSM, its state, and coverage.
   * <br>
   * Note that list of choosable transitions is constructed only during a
   * SemiRandomEfsmTransitionChooser instance initialisation, so any future changes in the
   * specified EFSM, state, and coverage instances will not affect a result of choosing by that
   * {@code SemiRandomEfsmTransitionChooser} instance. 
   * 
   * @param efsm the EFSM
   * @param state the state of the EFSM
   * @param coverage the coverage of the EFSM
   * @param priorityTransitions a set of transitions to be chosen firstly. Those transitions from
   * this set which aren't outgoing from the specified state are ignored
   * @param seed a seed for an inner random number generator
   */
  public SemiRandomEfsmTransitionChooser(Efsm efsm, EfsmState state,
      EfsmTransitionCoverage coverage, Set<EfsmTransition> priorityTransitions, long seed) {
    this.randomNumberGenerator = new Random(seed);
    this.priorityTransitions = new ArrayList<>();
    this.uncoveredTransitions = new ArrayList<>();
    this.coveredTransitions = new ArrayList<>();

    for (EfsmTransition transition: efsm.getOutgoingTransitions(state)) {
      if (priorityTransitions.contains(transition)) {
        this.priorityTransitions.add(transition);
      } else {
        if (coverage.isCovered(transition)) {
          this.coveredTransitions.add(transition);
        } else {
          this.uncoveredTransitions.add(transition);
        }
      }
    }
  }

  /**
   * Chooses and returns a transition for traversing.
   * <br>
   * The transition is chosen at random among previously unchosen transitions outgoing from the
   * EFSM state which this SemiRandomEfsmTransitionChooser is bounded to.
   * <br>
   * Firstly SemiRandomEfsmTransitionChooser chooses priority transitions which were specified
   * during construction of this chooser. If there are no such previously unchosen transitions, it
   * switches subsequently to uncovered transitions and finally to covered transitions.
   *
   * @return the transition proposed for traversing. {@code null} is returned if all choosable
   * transitions were chosen or there are no choosable transitions at all
   */
  @Override
  public EfsmTransition chooseNext() {
    if (!this.priorityTransitions.isEmpty()) {
      int chosenIndex = this.randomNumberGenerator.nextInt(this.priorityTransitions.size());

      return this.priorityTransitions.remove(chosenIndex);
    } else {
      if (!this.uncoveredTransitions.isEmpty()) {
        int chosenIndex = this.randomNumberGenerator.nextInt(this.uncoveredTransitions.size());

        return this.uncoveredTransitions.remove(chosenIndex);
      } else {
        if (!this.coveredTransitions.isEmpty()) {
          int chosenIndex = this.randomNumberGenerator.nextInt(this.coveredTransitions.size());

          return this.coveredTransitions.remove(chosenIndex);
        } else {
          return null;
        }
      }
    }
  }

  @Override
  public boolean hasNext() {
    return !this.priorityTransitions.isEmpty() || !this.uncoveredTransitions.isEmpty() ||
        !this.coveredTransitions.isEmpty();
  }
}
