/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test.heuristic.pathfinders.dfs;

import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

final class DepthFirstSearchTraverser implements Runnable {
  private final static String RENAMING_SUFFIX_START = "<";
  private final static String RENAMING_SUFFIX_END = ">";

  private final static Pattern NAME_PATTERN = Pattern.compile(".*(?=" +
       DepthFirstSearchTraverser.RENAMING_SUFFIX_START + ".*" +
       DepthFirstSearchTraverser.RENAMING_SUFFIX_END + ")");
  private final static Pattern SUFFIX_PATTERN = Pattern.compile("(?<=" +
      DepthFirstSearchTraverser.RENAMING_SUFFIX_START + ").*(?=" +
      DepthFirstSearchTraverser.RENAMING_SUFFIX_END + ")");

  private final Efsm efsm;
  private final EfsmTransition transition;

  private int currentLoopLimit;
  private final int maxLoopLimit;

  private final LinkedList<EfsmTransition> path;
  private final Map<EfsmTransition, Integer> visitCounters;
  private final LinkedList<List<NodeOperation>> constraints;

  private volatile boolean finish;
  private volatile boolean getNext;
  private volatile boolean done;
  private boolean isStarted;

  private final Map<String, NodeVariable> annotatedVariables;

  public DepthFirstSearchTraverser(Efsm efsm, EfsmTransition transition, int loopLimit) {
    this.efsm = efsm;
    this.transition = transition;

    this.currentLoopLimit = 1;
    this.maxLoopLimit = loopLimit;

    this.visitCounters = new HashMap<>();
    this.path = new LinkedList<>();
    this.constraints = new LinkedList<>();
    
    this.finish = false;
    this.getNext = false;
    this.done = false;
    this.isStarted = false;

    this.annotatedVariables = new HashMap<>();
  }

  @Override
  public void run() {
    this.isStarted = true;
    this.traverse(this.transition);
  }

  boolean isStarted() {
    return this.isStarted;
  }

  void finish() {
    this.finish = true;
  }

  boolean isFinished() {
    return this.finish;
  }

  void getNext() {
    this.done = false;
    this.getNext = true;
  }

  boolean isNextReady() {
    return this.done;
  }

  List<EfsmTransition> getPath() {
    return this.path;
  }

  List<NodeOperation> getConstraints() {
    if (!this.constraints.isEmpty()) {
      return this.constraints.getFirst();
    } else {
      return null;
    }
  }

  static int getCycleNumberForAnnotatedInput(String name) {
    Matcher suffixMatcher = DepthFirstSearchTraverser.SUFFIX_PATTERN.matcher(name);

    if (suffixMatcher.find()) {
      return Integer.valueOf(suffixMatcher.group());
    } else {
      return -1;
    }
  }

  static String deannotateName(String name) {
    Matcher nameMatcher = DepthFirstSearchTraverser.NAME_PATTERN.matcher(name);

    if (nameMatcher.find()) {
      return nameMatcher.group();
    } else {
      return name;
    }
  }

  void traverse(EfsmTransition transition) {
    if (!this.finish) {
      if (!this.visitCounters.containsKey(transition)) {
        this.visitCounters.put(transition, 0);
      }

      int visitCounter = this.visitCounters.get(transition);

      if (visitCounter < this.currentLoopLimit) {
        this.path.addFirst(transition);
        this.visitCounters.put(transition, visitCounter + 1);
        int cycleNumber = this.path.size() - 1;
        this.simulateAction(transition, cycleNumber);
        List<NodeOperation> newTargetPredicates;

        if (this.constraints.isEmpty()) {
          newTargetPredicates = new ArrayList<>();
        } else {
          newTargetPredicates = new ArrayList<>(this.constraints.getFirst()); 
        }

        this.constraints.addFirst(newTargetPredicates);

        for (Node node: transition.getGuardedAction().getGuard().getGuardExpr()) {
          if (node.getKind() == Node.Kind.OPERATION) {
            NodeOperation annotatedPredicate =
                this.annotateInputsWithCycleNumbers((NodeOperation) node, cycleNumber);
            newTargetPredicates.add(annotatedPredicate);
          }
        }

        EfsmState startState = transition.getSourceState();

        if (startState.equals(this.efsm.getInitialState())) {
          this.done = true;

          while (!this.finish) {
            try {
              if (this.getNext) {
                this.done = false;
                this.getNext = false;
                this.backtrack();

                break;
              }

              Thread.sleep(10);
            } catch (InterruptedException e) {
              this.finish = true;

              break;
            }
          }
        } else {
          boolean innerDependent = false;

          for (NodeOperation operation: this.constraints.getFirst()) {
            for (NodeVariable variable: ExprUtils.getVariables(operation)) {
              if (this.efsm.containsInnerVariable(variable.getName())) {
                innerDependent = true;

                break;
              }
            }
          }

          if (!innerDependent) {
            List<EfsmTransition> shortestPath =
                this.efsm.getShortestPath(this.efsm.getInitialState(), startState,
                    TransitionWeightBasis.CONSTANT);
            this.traverse(shortestPath.get(shortestPath.size() - 1));
          }

          for (EfsmTransition nextTransition: this.efsm.getIncomingTransitions(startState)) {
            this.traverse(nextTransition);
          }

          this.backtrack();

          if (this.path.isEmpty()) {
            this.traverse(transition);
          }
        }
      }
    }
  }

  void simulateAction(EfsmTransition transition, int cycleNumber) {
    if (!this.constraints.isEmpty() && !this.constraints.getFirst().isEmpty()) {
      Action action = transition.getGuardedAction().getAction();

      if (action != null) {
        Map<NodeVariable, Node> substitutions = action.getResultingSubstitutions();

        for (NodeVariable variable: substitutions.keySet()) {
          Node expression = substitutions.get(variable);

          if (expression.getKind() == Node.Kind.OPERATION) {
            expression = this.annotateInputsWithCycleNumbers((NodeOperation) expression, cycleNumber);
          } else {
            if (expression.getKind() == Node.Kind.VARIABLE) {
              expression = this.annotateInputWithCycleNumber((NodeVariable) expression, cycleNumber);
            }
          }

          substitutions.put(variable, expression);
        }

        for (NodeVariable variable: substitutions.keySet()) {
          Node substitution = substitutions.get(variable);

          if (substitution.getKind() == Node.Kind.OPERATION) {
            substitution = this.annotateInnerSignals((NodeOperation) substitution);
          } else if (substitution.getKind() == Node.Kind.VARIABLE) {
            substitution = this.annotateInnerSignal((NodeVariable) substitution);
          }

          substitutions.put(variable, substitution);
        }

        LinkedList<NodeOperation> newTargetPredicates = new LinkedList<>();

        for (NodeOperation predicate: this.constraints.removeFirst()) {
          NodeOperation newPredicate = predicate;

          for (NodeVariable variable: substitutions.keySet()) {
            newPredicate = (NodeOperation) Transformer.substitute(newPredicate, variable.getName(),
                substitutions.get(variable));
          }

          newTargetPredicates.add(this.deannotateInnerSignals(newPredicate));
        }
        
        this.constraints.addFirst(newTargetPredicates);
      } else {
        throw new NullPointerException();
      }
    }
  }

  void backtrack() {
    this.constraints.removeFirst();
    EfsmTransition transition = this.path.removeFirst();
    int visitCounter = this.visitCounters.get(transition);
    this.visitCounters.put(transition, visitCounter - 1);

    if (this.path.isEmpty()) {
      if (this.currentLoopLimit < this.maxLoopLimit) {
        this.currentLoopLimit++;
      } else {
        this.finish();
      }
    }
  }

  NodeOperation annotateInputsWithCycleNumbers(NodeOperation expression, int number) {
    NodeOperation annotatedExpression = expression;

    for (NodeVariable variable: ExprUtils.getVariables(expression)) {
      String originalName = variable.getName();

      if (this.efsm.getInputDataTypes().containsKey(originalName)) {
        NodeVariable annotatedVariable = this.annotateInputWithCycleNumber(variable, number);
        annotatedExpression =
            (NodeOperation) Transformer.substitute(annotatedExpression, originalName, annotatedVariable);
      }
    }

    return annotatedExpression;
  }

  NodeVariable annotateInputWithCycleNumber(NodeVariable variable, int number) {
    String originalName = variable.getName();

    if (this.efsm.getInputDataTypes().containsKey(originalName)) {
      String newName = originalName + DepthFirstSearchTraverser.RENAMING_SUFFIX_START +
          number + DepthFirstSearchTraverser.RENAMING_SUFFIX_END;

      if (this.annotatedVariables.containsKey(newName)) {

        return this.annotatedVariables.get(newName);
      } else {
        NodeVariable annotatedVariable = new NodeVariable(new Variable(newName, variable.getData()));
        this.annotatedVariables.put(newName, annotatedVariable);

        return annotatedVariable;
      }
    } else {
      return variable;
    }
  }

  NodeOperation annotateInnerSignals(NodeOperation expression) {
    NodeOperation annotatedExpression = expression;

    for (NodeVariable variable: ExprUtils.getVariables(expression)) {
      String originalName = variable.getName();

      if (this.efsm.getInnerDataTypes().containsKey(originalName)) {
        NodeVariable annotatedVariable = this.annotateInnerSignal(variable);
        annotatedExpression =
            (NodeOperation) Transformer.substitute(annotatedExpression, originalName, annotatedVariable);
      }
    }

    return annotatedExpression;
  }

  NodeOperation deannotateInnerSignals(NodeOperation expression) {
    NodeOperation deannotatedExpression = expression;

    for (NodeVariable variable: ExprUtils.getVariables(expression)) {
      String originalName = variable.getName();

      NodeVariable deannotatedVariable = this.deannotateInnerSignal(variable);
      deannotatedExpression =
          (NodeOperation) Transformer.substitute(deannotatedExpression, originalName, deannotatedVariable);
    }

    return deannotatedExpression;
  }

  NodeVariable annotateInnerSignal(NodeVariable variable) {
    String originalName = variable.getName();

    if (this.efsm.getInnerDataTypes().containsKey(originalName)) {
      String newName = originalName + DepthFirstSearchTraverser.RENAMING_SUFFIX_START +
          DepthFirstSearchTraverser.RENAMING_SUFFIX_END;

      return new NodeVariable(new Variable(newName, variable.getData()));
    } else {
      return variable;
    }
  }

  NodeVariable deannotateInnerSignal(NodeVariable variable) {
    String originalName = DepthFirstSearchTraverser.deannotateName(variable.getName());

    if (this.efsm.getInnerDataTypes().containsKey(originalName)) {
      return new NodeVariable(new Variable(originalName, variable.getData()));
    } else {
      return variable;
    }
  }
}
