/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.printer.graphml;

import ru.ispras.fortress.expression.Node;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModelVisitor;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.graphml.GraphMlUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;

/**
 * EFSM visitor for GraphML printer.
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class EfsmGraphMlVisitor implements EfsmModelVisitor {

  private PrintWriter printWriter; // GraphML output file

  /** Map of identifiers for transitions. */
  private Map<EfsmTransition, Integer> transitionMap;

  /** Map of identifiers for states. */
  private Map<EfsmState, Integer> stateMap;

  /** Model identifier. */
  private int modelId;

  @Override
  public void onEfsmBegin(Efsm efsm) { 
    GraphMlUtils.onGraphBegin(printWriter, Integer.toString(modelId));
    stateMap = new HashMap<>();
    transitionMap = new HashMap<>();
    modelId++;  
  }

  @Override
  public void onEfsmEnd(Efsm efsm) { GraphMlUtils.onGraphEnd(printWriter); }

  @Override
  public void onStateBegin(EfsmState state) {
    final int idState = getStateId(state);

    GraphMlUtils.printNode(printWriter, 
        Integer.toHexString(idState) + Integer.toString(modelId), state.toString(),
        GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_YELLOW);
  }

  private int getStateId(EfsmState state) {
    final Integer idState = stateMap.get(state);

    if (idState == null) {
      final int sizeMap = stateMap.size();
      stateMap.put(state, sizeMap);
      return sizeMap;
    }

    return idState;
  }

  @Override
  public void onStateEnd(EfsmState state) { /* Empty. */ }

  @Override
  public void onTransitionBegin(EfsmTransition transition) {
    Integer idTransition = transitionMap.get(transition);

    if (idTransition == null) {
      final int sizeMap = transitionMap.size();
      transitionMap.put(transition, sizeMap);
      idTransition = sizeMap;
    }

    // Get child & parent nodes
    final int sourceState = getStateId(transition.getSourceState());
    final GuardedAction guardedAction = transition.getGuardedAction();

    final String guargId = "Guard" + Integer.toHexString(idTransition);
    GraphMlUtils.printNode(printWriter, guargId, guardedAction.getGuard().toString(),
        GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_BLUE, GraphMlUtils.CROPPING_TEXT);
    
    GraphMlUtils.printEdge(
        printWriter,
        Integer.toHexString(sourceState)  + Integer.toString(modelId), guargId);

    final String actionId = "Action" + Integer.toHexString(idTransition);
    GraphMlUtils.printNode(printWriter, actionId, guardedAction.getAction().toString(),
        GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_GREEN, GraphMlUtils.CROPPING_TEXT);
    GraphMlUtils.printEdge(printWriter, guargId, actionId);

    final int destinationState = getStateId(transition.getDestinationState());
    
    GraphMlUtils.printEdge(
        printWriter, actionId,
        Integer.toHexString(destinationState)  + Integer.toString(modelId));
  }

  @Override
  public void onTransitionEnd(EfsmTransition transition) { /* Empty. */ }

  @Override
  public void onGuardedActionBegin(GuardedAction guardedAction) { /* Empty. */ }

  @Override
  public void onGuardedActionEnd(GuardedAction guardedAction) { /* Empty. */ }

  @Override
  public void onEvent(Event event) { /* Empty. */ }

  @Override
  public void onGuardBegin(Guard guard) { /* Empty. */ }

  @Override
  public void onGuardEnd(Guard guard) { /* Empty. */ }

  @Override
  public void onAssignmentActionBegin(Action action) { /* Empty. */ }

  @Override
  public void onAssignmentActionEnd(Action action) { /* Empty. */ }

  @Override
  public void onAssignmentBegin(Assignment statement) { /* Empty. */ }

  @Override
  public void onAssignmentEnd(Assignment statement) { /* Empty. */ }

  @Override
  public void onExpression(Node expression) { /* Empty. */ }

  public void initialize(String string) {
    try {
      // Create GML file
      File graphmlFile = FileCreator.newFile(string);

      modelId= 0;
      FileWriter fileWriter = new FileWriter(graphmlFile.getAbsoluteFile());
      printWriter = new PrintWriter(new BufferedWriter(fileWriter));
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  @Override
  public void onEfsmModelBegin(EfsmModel efsmModel) { GraphMlUtils.onGraphMlBegin(printWriter); }

  @Override
  public void onEfsmModelEnd(EfsmModel efsmModel) {
    GraphMlUtils.onGraphMlEnd(printWriter);

    GraphMlUtils.closeGraphMlFile(printWriter);
  }
}
