/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.engine.basis.Simulator;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.memory.HashMapMemory;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.testbench.Sequence;
import ru.ispras.retrascope.result.testbench.Transaction;
import ru.ispras.retrascope.result.testbench.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.logging.Logger;

/**
 * The EFSM-based representation of a hardware model which encapsulates a runtime state along with a
 * static structure of an original device.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class EfsmSimulator extends Simulator {
  /**
   * The engine ID of all {@code EfsmSimulator} instances.
   */
  public static final String ENGINE_ID = "efsm-simulator".intern();

  private final Logger logger;
  private final String logEntryHeader;

  private final String name;
  final Efsm efsm;

  EfsmState currentState;
  HashMapMemory inputsMemory;
  HashMapMemory outputsMemory;
  HashMapMemory innerVariablesMemory;

  EfsmTransitionCoverage coverage;

  /**
   * Constructs a simulator above the specified EFSM.
   * <br>
   * Immediately after construction the simulator is in the uninitialised state. To drive it to the
   * initial state of the EFSM it is required to feed such inputs which enable so-called reset
   * guarded action of the EFSM.
   *
   * @param efsm an EFSM which the simulator is based on
   *
   * @see Efsm#getInitialState()
   * @see Efsm#getResetGuardedAction()
   */
  public EfsmSimulator(Efsm efsm) {
    super(EfsmSimulator.ENGINE_ID, EntityType.get(Sequence.ENTITY_TYPE_ID),
        EntityType.get(Sequence.ENTITY_TYPE_ID));

    this.logger = Log.getLogger(this.getClass());
    this.logEntryHeader = "EFSM.Simulator." + efsm.getName();

    this.name = efsm.getName();
    this.efsm = efsm;

    this.inputsMemory = new HashMapMemory(efsm.getInputDataTypes().keySet());
    this.innerVariablesMemory = new HashMapMemory(efsm.getInnerDataTypes().keySet());
    this.outputsMemory = new HashMapMemory(efsm.getOutputDataTypes().keySet());

    this.coverage = new EfsmTransitionCoverage(efsm);

    this.currentState = null;
  }

  /**
   * Returns a snapshot representing a current state of the simulation.
   *
   * @return the snapshot
   */
  public EfsmSimulationSnapshot makeSimulationSnapshot() {
    return new EfsmSimulationSnapshot(this);
  }

  /**
   * Restores a simulation state from the specified snapshot.
   *
   * @param snapshot the snapshot
   */
  public void loadSnapshot(EfsmSimulationSnapshot snapshot) {
    this.logger.log(LogLevel.INFO, this.logEntryHeader
        + ": loading a simulation snapshot");

    this.currentState = snapshot.currentState;

    this.inputsMemory = new HashMapMemory(snapshot.inputsMemory);
    this.innerVariablesMemory = new HashMapMemory(snapshot.innerVariablesMemory);
    this.outputsMemory = new HashMapMemory(snapshot.outputsMemory);

    this.coverage = snapshot.coverage;

    this.logInputDump();
    this.logInnerDump();
  }

  /**
   * Returns an EFSM which this simulator is based on.
   *
   * @return the EFSM representing a static description of this simulator
   */
  public Efsm getStaticModel() {
    return this.efsm;
  }

  /**
   * Returns a name of this simulator.
   *
   * @return the name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns a state of an underlying EFSM which this simulator is currently in.
   *
   * @return the current state of this simulator. Returns {@code null} if called immediately after
   * construction/reset of this simulator
   *
   * @see #reset()
   * @see Efsm#getResetGuardedAction()
   */
  @Override
  public EfsmState getCurrentState() {
    return this.currentState;
  }

  /**
   * Returns a test coverage reached during simulation since last reset.
   *
   * @return the test coverage
   */
  @Override
  public EfsmTransitionCoverage getCoverage() {
    return new EfsmTransitionCoverage(this.coverage);
  }

  /**
   * Resets this simulator, that is, returns it to an uninitialised state.
   */
  public void reset() {
    this.logger.log(LogLevel.INFO, this.logEntryHeader + ": resetting EFSM");

    this.currentState = null;

    this.resetInputs();
    this.resetInnerVariables();
    this.resetOutputs();

    this.coverage.reset();

    this.checkInnerConsistency();
  }

  /**
   * Simulates the specified input sequence producing an output sequence.
   *
   * @param inputs  a map containing mappings between entity types and corresponding entities
   *
   * @return an output sequence
   *
   * @throws IllegalArgumentException if there is no {@link Sequence} object in the {@code inputs}
   * parameter
   */
  @Override
  public Sequence start(Map<EntityType, Entity> inputs) {
    for (Entity entity: inputs.values()) {
      if (entity instanceof Sequence) {
          Sequence inputSequence = (Sequence) entity;
        Sequence outputSequence = new Sequence(inputSequence.getLength());
  
        for (Vector inputVector: inputSequence.getVectors()) {
          this.setInputValues(inputVector.getTransactions());
          this.processEvents(inputVector.getEventList());
          Vector vector = new Vector();
          Transaction transaction = new Transaction(this.getOutputValues());
          vector.addTransaction(transaction);
          outputSequence.addVector(vector);
        }
  
        return outputSequence;
      }
    }

    throw new IllegalArgumentException("No Sequence objects have been found among arguments");
  }

  /**
   * Handles the specified events. During this process current state of this simulator as well as
   * values of input, inner, and output variables are updated.
   *
   * @param events the events to be processed
   */
  @Override
  public void processEvents(EventList events) {
    this.logger.log(LogLevel.INFO, this.logEntryHeader
        + ": processing the events: " + events);
    this.logInputDump();

    this.checkInputConsistency();

    if (this.currentState == null) {
      this.initialise(events);
    } else {
      List<EfsmTransition> possibleTransitions = this.getTraversableTransitions(events);

      if (possibleTransitions.isEmpty()) {
        this.logger.log(LogLevel.INFO, this.logEntryHeader
            + ": there are no traversable transitions now. Finishing event handling");
      } else {
        EfsmTransition traversedTransition = this.chooseTransitionToTraverse(possibleTransitions);

        this.logger.log(LogLevel.INFO,
            this.logEntryHeader + ": traversing the transition " + traversedTransition);

        this.executeAction(traversedTransition.getGuardedAction().getAction());
        this.currentState = traversedTransition.getDestinationState();
        this.coverage.markTransitionAsCovered(traversedTransition);

        this.logger.log(LogLevel.INFO,
            this.logEntryHeader + ": traversing has been finished successfully");
      }
    }

    this.logInnerDump();
    this.checkInnerConsistency();

    this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
        ": model has finished to process the events");
  }

  private void initialise(EventList events) {
    GuardedAction resetGuardedAction = this.efsm.getResetGuardedAction();
    this.logger.log(LogLevel.INFO, this.logEntryHeader +
        ": initialising the simulator. The reset guarded action: " + resetGuardedAction);

    if (resetGuardedAction.isSensible(events)) {
      Constraint resetConstraint = this.guardToConstraint(resetGuardedAction.getGuard());
      SolverResult result = ConstraintUtils.solve(resetConstraint);

      if (result.getStatus().equals(Status.SAT)) {
        this.logger.log(LogLevel.INFO, this.logEntryHeader + ": the reset guard is enabled");

        this.executeAction(resetGuardedAction.getAction());
        this.currentState = this.efsm.getInitialState();
      } else {
        this.logger.log(LogLevel.INFO, this.logEntryHeader + ": the reset guard isn't enabled");
      }
    } else {
      this.logger.log(LogLevel.INFO, this.logEntryHeader
          + ": the guarded action isn't sensible to the occurred events");
    }
  }

  private List<EfsmTransition> getTraversableTransitions(EventList events) {
    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": performing transitions traversability test");

    List<EfsmTransition> possibleTransitions =
        new ArrayList<>(this.efsm.getOutgoingTransitions(this.currentState));

    for (Iterator<EfsmTransition> possibleTransitionsIterator = possibleTransitions.iterator();
        possibleTransitionsIterator.hasNext();) {
      EfsmTransition testedTransition = possibleTransitionsIterator.next();

      this.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": testing the transition " + testedTransition);

      GuardedAction testedGuardedAction = testedTransition.getGuardedAction();

      if (!testedGuardedAction.isSensible(events)) {
        this.logger.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the transition isn't sensible to the occurred events");

        possibleTransitionsIterator.remove();
      } else {
        Constraint testedConstraint = this.guardToConstraint(testedGuardedAction.getGuard());
        SolverResult result = ConstraintUtils.solve(testedConstraint);

        if (result.getStatus().equals(Status.SAT)) {
          this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the transition is traversable");
        } else {
          this.logger.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the transition isn't traversable now");

          possibleTransitionsIterator.remove();
        }
      }
    }

    return possibleTransitions;
  }

  private EfsmTransition chooseTransitionToTraverse(List<EfsmTransition> traversableTransitions) {
    EfsmTransition traversedTransition;

    if (traversableTransitions.size() == 1) {
      traversedTransition = traversableTransitions.get(0);
    } else {
      this.logger.log(LogLevel.WARNING, this.logEntryHeader
          + ": there are multiple traversable transitions now. Choosing one of them at "
          + "random");
      for (EfsmTransition transition: traversableTransitions) {
        this.logger.log(LogLevel.WARNING,
            this.logEntryHeader + ": possible transition: " + transition);
      }

      Random randomGenerator = new Random(traversableTransitions.size());

      traversedTransition =
          traversableTransitions.get(randomGenerator.nextInt(traversableTransitions.size()));
    }

    return traversedTransition;
  }

  private Constraint guardToConstraint(Guard guard) {
    Node guardExpression =
        this.substituteVariables(guard.toNode(), Collections.<String, NodeValue>emptyMap());

    return ConstraintUtils.newConstraint(guardExpression);
  }

  @Override
  public void setInputValue(String name, NodeValue value) {
    if (this.efsm.containsInput(name)) {
      this.inputsMemory.setValue(name, value);
    }
  }

  @Override
  public void setInputValues(Collection<Transaction> transactions) {
    for (Transaction transaction: transactions) {
      if (!transaction.isEmpty()) {
        Set<String> variableNames = transaction.getValues().keySet();
  
        if (this.inputsMemory.containsVariables(variableNames)) {
          for (String variableName: variableNames) {
            Node variableValue = transaction.getValues().get(variableName);
  
            if (variableValue.getKind() == Node.Kind.VALUE) {
              this.inputsMemory.setValue(variableName, ((NodeValue) variableValue));
            } else {
              throw new IllegalArgumentException(
                  "All variable values must be represented by NodeValue instances");
            }
          }
        }
      }
    }
  }

  @Override
  public NodeValue getInnerValue(String name) throws IllegalArgumentException {
    if (this.efsm.containsInnerVariable(name)) {
      return this.innerVariablesMemory.getValue(name);
    } else {
      throw new IllegalArgumentException(
          "The specified variable isn't found among inner variables of this simulator");
    }
  }

  @Override
  public Map<String, NodeValue> getInnerValues(Collection<String> names) {
    if (this.innerVariablesMemory.containsVariables(names)) {
      Map<String, NodeValue> variables = new HashMap<>(names.size());

      for (String variableName: names) {
        variables.put(variableName, this.innerVariablesMemory.getValue(variableName));
      }

      return variables;
    } else {
      throw new IllegalArgumentException(
          "Some of the specified variables aren't found among inner variables of this simulator");
    }
  }

  @Override
  public Map<String, NodeValue> getInnerValues() {
    return new HashMap<>(this.innerVariablesMemory.getValues());
  }

  @Override
  public NodeValue getOutputValue(String name) throws IllegalArgumentException {
    if (this.efsm.containsOutput(name)) {
      return this.outputsMemory.getValue(name);
    } else {
      throw new IllegalArgumentException(
          "The specified variable isn't found among outputs of this simulator");
    }
  }

  @Override
  public Map<String, NodeValue> getOutputValues(Collection<String> names) {
    if (this.outputsMemory.containsVariables(names)) {
      Map<String, NodeValue> outputs = new HashMap<>(names.size());

      for (String variableName : names) {
        outputs.put(variableName, this.outputsMemory.getValue(variableName));
      }

      return outputs;
    } else {
      throw new IllegalArgumentException(
          "Some of the specified variables aren't found among outputs of this simulator");
    }
  }

  @Override
  public Map<String, NodeValue> getOutputValues() {
    return new HashMap<>(this.outputsMemory.getValues());
  }

  private void resetInputs() {
    for (String variableName : this.efsm.getInputDataTypes().keySet()) {
      this.inputsMemory.setValue(variableName, null);
    }
  }

  private void resetInnerVariables() {
    for (String variableName : this.efsm.getInnerDataTypes().keySet()) {
      this.innerVariablesMemory.setValue(variableName, null);
    }
  }

  private void resetOutputs() {
    for (String variableName : this.efsm.getOutputDataTypes().keySet()) {
      this.outputsMemory.setValue(variableName, null);
    }
  }

  private void checkInputConsistency() {
    Set<Node> invariants = new HashSet<>();

    for (String variableName: this.efsm.getInputDataTypes().keySet()) {
      if (this.inputsMemory.getValue(variableName) != null) {
        Node invariant = efsm.getInvariantCondition(variableName);

        if (invariant != null) {
          invariants.add(invariant);
        }
      }
    }

    if (!invariants.isEmpty()) {
      Node consistencyConstraint =
          ExprUtils.getConjunction(invariants.toArray(new Node[invariants.size()]));

      this.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": checking input consistency: " + consistencyConstraint);

      consistencyConstraint =
          this.substituteVariables(consistencyConstraint, Collections.<String, NodeValue>emptyMap());
      SolverResult result =
          ConstraintUtils.solve(ConstraintUtils.newConstraint(consistencyConstraint));

      if (result.getStatus() != Status.SAT) {
        throw new RuntimeException("Input consistency constraint violation: " +
            consistencyConstraint);
      }
    }
  }

  private void checkInnerConsistency() {
    if (this.currentState != null) {
      Set<Node> invariants = new HashSet<>();

      for (String variableName: this.efsm.getInnerDataTypes().keySet()) {
        if (this.innerVariablesMemory.getValue(variableName) != null) {
          Node invariant = this.efsm.getInvariantCondition(variableName);

          if (invariant != null) {
            invariants.add(invariant);
          }
        }
      }

      for (String variableName: this.efsm.getOutputDataTypes().keySet()) {
        if (this.outputsMemory.getValue(variableName) != null) {
          Node invariant = efsm.getInvariantCondition(variableName);

          if (invariant != null) {
            invariants.add(invariant);
          }
        }
      }

      invariants.add(this.currentState.getExpression());

      if (!invariants.isEmpty()) {
        Node consistencyConstraint =
            ExprUtils.getConjunction(invariants.toArray(new Node[invariants.size()]));

        this.logger.log(LogLevel.DEBUG,
            this.logEntryHeader + ": checking inner consistency: " + consistencyConstraint);

        consistencyConstraint = this.substituteVariables(consistencyConstraint,
            Collections.<String, NodeValue>emptyMap());
        SolverResult result =
            ConstraintUtils.solve(ConstraintUtils.newConstraint(consistencyConstraint));

        if (result.getStatus() != Status.SAT) {
          throw new RuntimeException("Inner consistency constraint violation: " +
              consistencyConstraint);
        }
      }
    }
  }

  private void executeAction(Action action) {
    InvariantChecks.checkNotNull(action);

    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": executing the action: " + action);

    List<Assignment> assignments = action.getStatements();
    Map<String, NodeValue> bufferedVariables = new HashMap<>(assignments.size());

    for (Assignment assignment: assignments) {
      bufferedVariables = this.executeAssignment(assignment, bufferedVariables);
    }

    for (Entry<String, NodeValue> bufferedVariable: bufferedVariables.entrySet()) {
      String variableName = bufferedVariable.getKey();
      NodeValue variableValue = bufferedVariable.getValue();

      if (this.efsm.getInnerDataTypes().containsKey(variableName)) {
        this.innerVariablesMemory.setValue(variableName, variableValue);
      } else {
        if (this.efsm.getOutputDataTypes().containsKey(variableName)) {
          this.outputsMemory.setValue(variableName, variableValue);
        }
      }
    }

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the action has been executed successfully: " + action);
  }

  private Map<String, NodeValue> executeAssignment(Assignment assignment,
      Map<String, NodeValue> bufferedVariables) {
    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": executing the assignment: " + assignment);

    RangedVariable rangedVariable = assignment.getRangedVariable();
    Node expression = assignment.getExpression();

    if ((rangedVariable == null) && (expression == null)) {
      return Collections.emptyMap();
    }

    InvariantChecks.checkNotNull(rangedVariable);
    InvariantChecks.checkNotNull(expression);

    NodeVariable nodeVariable = rangedVariable.getVariable();
    DataType dataType = nodeVariable.getDataType();

    if (rangedVariable.isRanged()) {
      if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR) &&
          !dataType.getTypeId().equals(DataTypeId.MAP)) {
        throw new IllegalArgumentException("Unsupported data type of ranged variable " +
            rangedVariable.getName() +  ": " + dataType);
      }

      Range range = rangedVariable.getRange();

      Node leftBoundExpression = this.substituteVariables(range.getYoung(), bufferedVariables);
      Node rightBoundExpression = this.substituteVariables(range.getOld(), bufferedVariables);

      NodeValue leftBound = this.computeExpression(leftBoundExpression, DataType.INTEGER);
      NodeValue rightBound = this.computeExpression(rightBoundExpression, DataType.INTEGER);

      int leftBoundInteger = leftBound.getInteger().intValue();
      int rightBoundInteger = rightBound.getInteger().intValue();

      int variableSize = dataType.getSize();

      if (leftBoundInteger < rightBoundInteger) {
        throw new IllegalArgumentException("This type of subvectors is currently unsupported");
      }

      NodeOperation mostSignificantPart = null;
      NodeOperation lessSignificantPart = null;
      
      if (leftBoundInteger < variableSize - 1) {
        NodeValue leftIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), 10));
        NodeValue rightIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBoundInteger + 1), 10));

        mostSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
            nodeVariable);
      }

      if (rightBoundInteger > 0) {
        NodeValue leftIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBoundInteger - 1), 10));
        NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", 10));

        lessSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
            nodeVariable);
      }

      if ((mostSignificantPart != null) && (lessSignificantPart != null)) {
        expression = new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression,
            lessSignificantPart);
      } else {
        if (mostSignificantPart != null) {
          expression =
              new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression);
        } else {
          if (lessSignificantPart != null) {
            expression =
                new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignificantPart);
          }
        }
      }
    }

    expression = this.substituteVariables(expression, bufferedVariables);
    NodeValue assignedValue = this.computeExpression(expression, dataType);
    bufferedVariables.put(rangedVariable.getName(), assignedValue);

    this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
        ": the assigned value: " + assignedValue);

    return bufferedVariables;
  }

  private Node substituteVariables(Node expression, Map<String, NodeValue> bufferedVariables) {
    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": substituting variables in the expression: " + expression);

    Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = usedVariable.getName();
      NodeValue variableValue = null;

      if (bufferedVariables.containsKey(variableName)) {
        variableValue = bufferedVariables.get(variableName);
      } else {
        if (this.inputsMemory.containsVariable(variableName)) {
          variableValue = this.inputsMemory.getValue(variableName);
        } else {
          if (this.innerVariablesMemory.containsVariable(variableName)) {
            variableValue = this.innerVariablesMemory.getValue(variableName);
          } else {
            if (this.outputsMemory.containsVariable(variableName)) {
              variableValue = this.outputsMemory.getValue(variableName);
            } else {
              throw new RuntimeException("The variable " + variableName + " isn't found among"
                  + "inputs, inner variables, outputs and buffer variables of this action");
            }
          }
        }
      }

      bindingList.add(NodeBinding.bindVariable(usedVariable, variableValue));
    }

    Node result = Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the expression after substitution: " + result);

    return result;
  }

  private NodeValue computeExpression(Node expression, DataType resultDataType) {
    if (expression instanceof NodeValue) {
      NodeValue value = (NodeValue) expression;

      if (!value.getDataType().equals(resultDataType)) {
        throw new IllegalArgumentException("Uncompatible data types: " + value.getDataType() +
            " vs " + resultDataType);
      }

      return value;
    } else {
      if (!ExprUtils.getVariables(expression).isEmpty()) {
        throw new IllegalArgumentException("The expression to be computed (" + expression +
            ") contains unevaluated variables: " + ExprUtils.getVariables(expression));
      }

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": computing the expression " + expression);

      expression = new NodeOperation(StandardOperation.EQ,
          new NodeVariable("assignmentResult", resultDataType), expression);

      SolverResult result = ConstraintUtils.solve(ConstraintUtils.newConstraint(expression));

      if (result.getStatus().equals(Status.SAT)) {
        NodeValue resultValue = new NodeValue(result.getVariables().iterator().next().getData());
        
        this.logger.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the computation result: " + resultValue);

        return resultValue;
      } else {
        throw new RuntimeException("Failed to resolve the assignment constraint due to errors: " +
            result.getErrors());
      }
    }
  }

  private void logInputDump() {
    this.logger.log(LogLevel.INFO, this.logEntryHeader + ": input dump: " + this.inputsMemory);
  }

  private void logInnerDump() {
    this.logger.log(LogLevel.INFO, this.logEntryHeader + ": current state: " + this.currentState);
    this.logger.log(LogLevel.INFO,
        this.logEntryHeader + ": inner variables' dump: " + this.innerVariablesMemory);
    this.logger.log(LogLevel.INFO, this.logEntryHeader + ": output dump: " + this.outputsMemory);
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": current coverage: " + this.coverage);
  }
}
