/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.testbench;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.Set;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.result.testbench.Sequence;
import ru.ispras.retrascope.result.testbench.Test;
import ru.ispras.retrascope.result.testbench.Vector;

/**
 * The class for printing tests to files in the DECIDER-like manner.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class DeciderLikeTestPatternPrinter {
  private static final String PATTERN_HEADER = "#Pattern: ";
  private static final String SEQUENCE_HEADER = "#New sequence";
  private static final String SIGNAL_DELIMETER = " ";
  private static final String VALUE_FOR_UNSPECIFIED_SIGNALS = "X";

  private DeciderLikeTestPatternPrinter() {}

  /**
   * Prints the specified test to the specified file.
   *
   * @param test the test to be printed
   * @param file the file where the test will be printed to
   * @param inputNames a set of all input signals of a tested module
   *
   * @throws IOException if the specified file already exists or if an I/O error occurred
   */
  public static void print(Test test, Path file, Set<String> inputNames) throws IOException {
    if (Files.exists(file)) {
      throw new FileAlreadyExistsException("The specified file already exists");
    }

    try (BufferedWriter testWriter = Files.newBufferedWriter(file, Charset.defaultCharset())) {
      testWriter.append(DeciderLikeTestPatternPrinter.PATTERN_HEADER);

      String clockSignalName = test.getSequences().get(0).getVector(0).getEventList().getEvents().
          iterator().next().getRangedVariable().getName();

      Set<String> nonClockInputNames = new HashSet<>(inputNames.size());

      for (String inputName: inputNames) {
        if (!inputName.equals(clockSignalName)) {
          nonClockInputNames.add(inputName);
          testWriter.append(inputName);
          testWriter.append(DeciderLikeTestPatternPrinter.SIGNAL_DELIMETER);
        }
      }

      testWriter.newLine();

      for (Sequence sequence: test.getSequences()) {
        testWriter.append(DeciderLikeTestPatternPrinter.SEQUENCE_HEADER);
        testWriter.newLine();

        for (Vector vector: sequence.getVectors()) {
          for (String inputName: nonClockInputNames) {
            if (vector.containsVariable(inputName)) {
              Node value = vector.getValue(inputName);

              if (value instanceof NodeValue) {
                testWriter.append(String.valueOf(((NodeValue) value).getData().getValue()));
              } else {
                throw new IllegalArgumentException("Unsupported value type: " +
                    value.getClass().getSimpleName());
              }
            } else {
              testWriter.append(DeciderLikeTestPatternPrinter.VALUE_FOR_UNSPECIFIED_SIGNALS);
            }

            testWriter.append(DeciderLikeTestPatternPrinter.SIGNAL_DELIMETER);
          }

          testWriter.newLine();
        }
      }

      testWriter.flush();
    }
  }
}
