/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Assignment statement.
 *
 * Can be used both for usual assignment and for multi-assignment
 * statement of HDL. All the statement's assignments should be interpreted
 * as concurrently executable.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Assignment extends MetaInfo implements UseDef {

  /* The target (defined) variable. */
  private RangedVariable rangedVariable;

  /* The value (used) expression. */
  private Node expression;

  /**
   * The default constructor.
   */
  public Assignment() { /* Empty. */ }

  /**
   * Constructs object with the specified ranged variable and expression.
   *
   * @param variable variable
   * @param expr syntax node
   *
   * @throws java.lang.NullPointerException when any of arguments is {@code null}.
   */
  public Assignment(RangedVariable variable, Node expr) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(expr);

    this.rangedVariable = variable;
    this.expression = expr;
  }

  /**
   * Constructs object with the specified non-ranged variable and expression.
   *
   * @param var variable which expression is assigned
   * @param expr expression to be assigned
   *
   * @throws java.lang.NullPointerException when any of arguments is {@code null}.
   */
  public Assignment(NodeVariable var, Node expr) {
    InvariantChecks.checkNotNull(var);
    InvariantChecks.checkNotNull(expr);

    this.rangedVariable = new RangedVariable(var);
    this.expression = expr;
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return full copy of the current object
   */
  public Assignment deepcopy() {
    Assignment deepcopy =
        new Assignment(this.rangedVariable.deepcopy(), this.expression.deepCopy());
    deepcopy.addMetaInfo(this.getMetaInfo());
    return deepcopy;
  }

  /**
   * Returns variables that are to be assigned to.
   *
   * @return variables
   */
  public RangedVariable getRangedVariable() { return this.rangedVariable; }

  /**
   * Sets the specified variable as objects' target.
   *
   * @param variable variable to be added
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setVariable(RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.rangedVariable = variable;
  }

  /**
   * Returns value expression.
   *
   * @return value expression syntax node
   */
  public Node getExpression() { return this.expression; }

  /**
   * Sets new value expression.
   *
   * @param expression new value expression syntax node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setExpression(Node expression) {
    InvariantChecks.checkNotNull(expression);

    this.expression = expression;
  }

  /**
   * Returns names of variables used in this assignment.
   *
   * @return set of the variables' names.
   */
  public Set<String> getUsedVariableNames() {
    Collection<NodeVariable> usedVariables = this.getUses();

    if (!usedVariables.isEmpty()) {
      Set<String> usedVariableNames = new LinkedHashSet<>(usedVariables.size());

      for (NodeVariable variable: usedVariables) {
        usedVariableNames.add(variable.getName());
      }

      return usedVariableNames;
    } else {
      return new LinkedHashSet<>();
    }
  }

  /**
   * Returns a {@link java.lang.String} description of object
   * stored in format of the specified printer.
   *
   * @param printer expression printer
   * @param separator assignment separator between defined and used variables
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   *
   * @return string description
   */
  public String getDescription(ExprTreePrinter printer, String separator) {
    InvariantChecks.checkNotNull(printer);
    InvariantChecks.checkNotNull(separator);

    return this.rangedVariable.toString() +
        separator +
        printer.toString(this.expression);
  }

  @Override
  public Set<NodeVariable> getUses() {
    if (this.expression != null) {
      return new LinkedHashSet<>(ExprUtils.getVariables(this.expression));
    } else {
      return new LinkedHashSet<>();
    }
  }

  @Override
  public Set<NodeVariable> getDefines() {
    Set<NodeVariable> defines = new LinkedHashSet<>();
  
    defines.add(getRangedVariable().getVariable());
    return defines;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((this.expression == null) ? 0 : this.expression.hashCode());
    result = prime * result + this.rangedVariable.hashCode();

    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof Assignment)) {
      return false;
    }

    Assignment other = (Assignment) obj;

    if (this.expression == null) {
      if (other.expression != null) {
        return false;
      }
    } else if (!this.expression.equals(other.expression)) {
      return false;
    }

    return this.rangedVariable.equals(other.rangedVariable);
  }

  @Override
  public String toString() {

    /* HACK: relevant separator should be used here. */
    String separator = ":=";

    return this.rangedVariable + " " + separator + " " + this.expression;
  }
}
