/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.jaxb.JaxbEventAdapter;

import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/**
 * Fact that occurs with variable (usually with RTL signal).
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
@XmlJavaTypeAdapter(JaxbEventAdapter.class)
public class Event extends MetaInfo {

  /* The event type. */
  private final EventType type;

  /* The event variable. */
  private final RangedVariable variable;

  /**
   * Constructs an object with the specified variable and event type.
   *
   * @param variable variable
   * @param type event type
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public Event(RangedVariable variable, EventType type) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(type);

    this.type = type;
    this.variable = variable;
  }

  /**
   * Constructs an object with the specified variable and of any edge event type.
   *
   * @param variable variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Event(RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.variable = variable;
    this.type = EventType.ANY_EDGE;
  }

  /**
   * Returns event type.
   *
   * @return event type
   */
  public EventType getType() { return this.type; }

  /**
   * Returns event ranged variable.
   *
   * @return ranged variable
   */
  public RangedVariable getRangedVariable() { return this.variable; }

  /**
   * Checks if this event is always triggered along with the specified one (for example, if this
   * event is any edge of some variable while other is positive edge of the same variable).
   *
   * @param event the event to be tested
   *
   * @return {@code true} if this event is always triggered along with the specified one, otherwise
   * returns {@code false}
   */
  public boolean isTriggeredBy(Event event) {
    return this.variable.equals(event.variable) &&
        ((this.type == event.type) || (this.type == EventType.ANY_EDGE));
  }

  /**
   * Checks if this event is compatible with the specified one, that is, they aren't mutually
   * exclusive.
   *
   * @param event the event to be tested for compatibility with this one
   *
   * @return {@code true} if the events are compatible, otherwise returns {@code false}
   */
  public boolean isCompatible(Event event) {
    return !this.variable.equals(event.variable) ||
        (this.type == event.type) ||
        (this.type == EventType.ANY_EDGE) ||
        (event.type == EventType.ANY_EDGE);
  }

  /**
   * Returns a deep copy of this object.
   *
   * @return object's deep copy
   */
  public Event deepcopy() { return new Event(this.variable.deepcopy(), this.type); }

  @Override
  public String toString() {
    if (this.getType() == EventType.ANY_EDGE) {
      return this.getRangedVariable().toString();
    } else {
      return this.getType().toCompactString() + " of " + this.getRangedVariable().toString();
    }
  }

  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o == null || getClass() != o.getClass())
      return false;

    Event event = (Event) o;

    return this.type == event.type && this.variable.equals(event.variable);
  }

  @Override
  public int hashCode() {
    int result = this.type.hashCode();
    result = 31 * result + this.variable.hashCode();
    return result;
  }
}
