/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.jaxb.JaxbEventListAdapter;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import javax.xml.bind.annotation.XmlType;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/**
 * Event collection.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
@XmlType(name = "eventList")
@XmlJavaTypeAdapter(JaxbEventListAdapter.class)
public class EventList {

  /* The events collection. */
  private final Set<Event> events;

  /**
   * The default constructor.
   */
  public EventList() { this.events = new LinkedHashSet<>(); }

  /**
   * Constructs object with the specified events collection.
   *
   * @param events events collection
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public EventList(Set<Event> events) {
    InvariantChecks.checkNotNull(events);

    this.events = events;
  }

  /**
   * Checks rather event is in this class.
   *
   * @param event event object
   *
   * @return {@code true} if event is here, {@code false} otherwise
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public boolean contains(Event event) {
    InvariantChecks.checkNotNull(event);

    return this.events.contains(event);
  }

  /**
   * Checks rather object contains an event with the specified variable.
   *
   * @param variable variable of an event of the object to be checked
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @return {@code true} if object contains an event with the specified variable,
   * {@code false} otherwise
   */
  public boolean contains(NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);
    boolean contains = false;

    for (Event event : this.events) {
      if (event.getRangedVariable().getVariable().equals(variable)) {
        contains = true;
        break;
      }
    }

    return contains;
  }

  /**
   * Adds event.
   *
   * If this object already contains an event with the same
   * {@link ru.ispras.retrascope.model.basis.RangedVariable} as the specified
   * {@link ru.ispras.retrascope.model.basis.Event} has, but their
   * {@link ru.ispras.retrascope.model.basis.EventType} types are different,
   * the {@link ru.ispras.retrascope.model.basis.Event} object with the same
   * {@link ru.ispras.retrascope.model.basis.RangedVariable} and
   * {@code EventType.ANY_EDGE} event type.
   *
   * @param event event object
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void add(Event event) {
    InvariantChecks.checkNotNull(event);

    RangedVariable eVariable = event.getRangedVariable();
    EventType eType = event.getType();
    Event sameVarEvent = null;
    for (Event thisEvent : this.events) {
      if (thisEvent.getRangedVariable().equals(eVariable) && !thisEvent.getType().equals(eType)) {
        sameVarEvent = thisEvent;
        break;
      }
    }

    if (sameVarEvent != null) {
      this.events.remove(sameVarEvent);
      final Event combination = new Event(event.getRangedVariable(), EventType.ANY_EDGE);
      this.events.add(combination);
    } else {
      this.events.add(event);
    }
  }

  /**
   * Removes an event with the specified variable.
   *
   * @param variable event's variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void remove(RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    Event variableEvent = null;
    for (Event event : this.events) {
      if (event.getRangedVariable().equals(variable)) {
        variableEvent = event;
        break;
      }
    }

    if (variableEvent != null) {
      this.events.remove(variableEvent);
    }
  }

  /**
   * Adds a collection of events.
   *
   * @param events event collection
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addEvents(Collection<? extends Event> events) {
    InvariantChecks.checkNotNull(events);

    for (Event event : events) {
      add(event);
    }
  }

  /**
   * Checks if this event list is empty.
   * 
   * @return {@code true} if this event list doesn't contain any events, {@code false} otherwise.
   */
  public boolean isEmpty() { return this.events.isEmpty(); }

  /**
   * Returns number of events.
   *
   * @return number of events
   */
  public int eventNum() { return this.events.size(); }

  /**
   * Returns a collection of events.
   *
   * @return event collection
   */
  public Collection<Event> getEvents() { return this.events; }

  /**
   * Applies expressions to bounded event variables.
   *
   * @param bindings "event's variable name - bounded expression" map
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void applyToEventList(Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    EventList eventList = new EventList();
    boolean isBounded = false;
    for (Event event : this.events) {

      for (String name : bindings.keySet()) {
        if (event.getRangedVariable().getVariable().getName().equals(name)) {
          isBounded = true;

          Collection<NodeVariable> vars = ExprUtils.getVariables(bindings.get(name));
          for (NodeVariable variable : vars) {
            if (eventList.isEmpty() || !eventList.contains(variable)) {

              /* Add any-edge event. */
              eventList.add(new Event(new RangedVariable(variable)));
            }
          }
        }
      }
      if (!isBounded) {
        eventList.add(new Event(event.getRangedVariable(), event.getType()));
      } else {
        isBounded = false;
      }
    }
  }

  /**
   * Checks if this event list is compatible with the specified one, that is, they aren't mutually
   * exclusive.
   *
   * @param eventList the event list to be tested for compatibility with this one
   *
   * @return {@code true} if the lists are compatible, otherwise returns {@code false}
   *
   * @see Event#isCompatible(Event)
   */
  public boolean isCompatible(EventList eventList) {
    for (Event event: this.getEvents()) {
      for (Event testedEvent: eventList.getEvents()) {
        if (!event.isCompatible(testedEvent)) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * Checks if this event list is always triggered along with the specified one (for example, if
   * this event is any edge of some variable while other is positive edge of the same variable).
   * To trigger this event list it is necessary and sufficient to trigger at least one of its
   * events. 
   *
   * @param event the event to be tested
   *
   * @return {@code true} if this event is always triggered along with the specified one, otherwise
   * returns {@code false}
   */
  public boolean isTriggeredBy(Event event) {
    if (this.getEvents().isEmpty()) {
      return true;
    }

    for (Event expectedEvent: this.getEvents()) {
      if (expectedEvent.isTriggeredBy(event)) {
        return true;
      }
    }

    return false;
  }

  /**
   * Returns a deep copy of this object.
   *
   * @return object's deep copy
   */
  public EventList deepcopy() {
    EventList copy = new EventList();

    for (Event event : this.events) {
      Event eventCopy = event.deepcopy();

      copy.add(eventCopy);
    }

    return copy;
  }

  /**
   * Removes all the internal {@link ru.ispras.retrascope.model.basis.Event} events.
   */
  public void clear() { this.events.clear(); }

  @Override
  public String toString() {
    StringBuilder text = new StringBuilder("{");

    Iterator<Event> iterator = this.events.iterator();
    while (iterator.hasNext()) {
      text.append(iterator.next().toString());
      if (iterator.hasNext()) {
        text.append(",");
      }
    }

    text.append("}");
    return text.toString();
  }

  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o == null || getClass() != o.getClass())
      return false;

    EventList that = (EventList) o;

    return this.events.equals(that.events);
  }

  @Override
  public int hashCode() { return this.events.hashCode(); }
}
