/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Class that stores met-information of different types into different values.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class MetaInfo {

  /* The mapping of meta-information types into values. */
  private final Map<String, String> metaInfoMap;

  /**
   * Constructs meta-information container.
   */
  public MetaInfo () { this.metaInfoMap = new LinkedHashMap<>(); }

  /**
   * Contains rather object contains meta-information for the specified type.
   *
   * @return type key
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public boolean hasMetaInfo(Enum<?> type) {
    InvariantChecks.checkNotNull(type);

    return this.metaInfoMap.containsKey(type.name());
  }

  /**
   * Contains rather object contains meta-information for the specified type.
   *
   * @return type key
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public boolean hasMetaInfo(String type) {
    InvariantChecks.checkNotNull(type);

    return this.metaInfoMap.containsKey(type);
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * It is supposed that this method is used for the predefined types and values.
   *
   * @param type information type
   * @param value information value
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(Enum<?> type, Enum<?> value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value);

    this.metaInfoMap.put(type.name(), value.name());
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * It is supposed that this method is used for the predefined types.
   *
   * @param type information type
   * @param value information value
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(Enum<?> type, String value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value);

    this.metaInfoMap.put(type.name(), value);
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * @param type information type
   * @param value information value
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(String type, String value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value);

    this.metaInfoMap.put(type, value);
  }

  /**
   * Adds meta-information as mapping.
   *
   * @param map meta-information mapping
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addMetaInfo(Map<String, String> map) {
    InvariantChecks.checkNotNull(map);

    this.metaInfoMap.putAll(map);
  }

  /**
   * Removes information of the specified type.
   *
   * @param type information type
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void removeMetaInfo(String type) {
    InvariantChecks.checkNotNull(type);

    this.metaInfoMap.remove(type);
  }

  /**
   * Removes information of the specified type.
   *
   * @param type information type
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void removeMetaInfo(Enum<?> type) {
    InvariantChecks.checkNotNull(type);

    this.metaInfoMap.remove(type.name());
  }

  /**
   * Returns meta-information of the specified type.
   *
   * @param key information type
   * @return information if an object contains it, {@code null} otherwise.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public String getMetaInfo(Enum<?> key) {
    InvariantChecks.checkNotNull(key);

    return this.metaInfoMap.get(key.name());
  }

  /**
   * Returns meta-information of the specified type.
   *
   * @param key information type
   * @return information if an object contains it, {@code null} otherwise.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public String getMetaInfo(String key) {
    InvariantChecks.checkNotNull(key);

    return this.metaInfoMap.get(key);
  }

  /**
   * Returns all the meta-information.
   *
   * @return meta-information map
   */
  public Map<String, String> getMetaInfo() { return this.metaInfoMap; }
}
