/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import javax.xml.bind.annotation.XmlAttribute;

/**
 * The object that keep a pair of values. It is proposed that one value
 * is less or equal to another value.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Range {

  /* The delimiter for object's string representation. */
  private static final String RANGE_DELIMITER = ":";

  /* Left (less or equal) index. */
  @XmlAttribute private final Node young;

  /* Right (greater or equal) index. */
  @XmlAttribute private final Node old;

  /**
   * The default constructor.
   *
   * @param young young index
   * @param old old index
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   *
   * @throws java.lang.IllegalArgumentException when the second argument is always less than
   * first argument.
   */
  public Range(Node young, Node old) {
    InvariantChecks.checkNotNull(young);
    InvariantChecks.checkNotNull(old);

    Enum<?> opId;
    if (young.isType(DataTypeId.BIT_VECTOR) && old.isType(DataTypeId.BIT_VECTOR)) {
      opId = StandardOperation.BVUGE;
    } else {
      opId = StandardOperation.GREATEREQ;
    }

    if (!ExprUtils.isSAT(new NodeOperation(opId, old, young))) {
      throw new IllegalArgumentException(
          String.format("The %s is less than %s.", old.toString(), young.toString()));
    }

    this.young = young;
    this.old = old;
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return a full copy of the current object
   */
  public Range deepcopy() { return new Range(this.young.deepCopy(), this.old.deepCopy()); }

  /**
   * Returns young index.
   *
   * @return young index
   */
  public Node getYoung() { return young; }

  /**
   * Returns old index.
   *
   * @return old index
   */
  public Node getOld() { return old; }

  @Override
  public String toString() { return this.young + RANGE_DELIMITER + this.old; }

  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o == null || getClass() != o.getClass())
      return false;

    Range range = (Range) o;

    return this.young == range.young && this.old == range.old;
  }

  @Override
  public int hashCode() {
    int result = this.young.hashCode();
    result = 31 * result + this.old.hashCode();
    return result;
  }
}
