/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * The RangedVariable class represents both variable and it's "range". The range will be
 * not-null if not all variable is in use but only it's sub-array.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class RangedVariable {

  /* The variable. */
  private NodeVariable variable;

  /* The range, */
  private Range range;

  /**
   * The default constructor.
   */
  public RangedVariable() {
    this.variable = null;
    this.range = null;
  }

  /**
   * Constructs object from the specified variable.
   *
   * @param variable variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public RangedVariable(NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.variable = variable;
    this.range = null;
  }

  /**
   * Constructs object from the specified variable and it's range.
   *
   * NOTE: The ranged variable should be either of bit vector data type
   * or of map data type.
   *
   * @param variable variable
   * @param range range
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   *
   * @throws java.lang.IllegalArgumentException when variable argument is neither of
   * bit vector data type nor of map (array) data type.
   *
   * @throws java.lang.IllegalArgumentException when variable id of bit vector type
   * but range old/young bit numbers are out of bit vector size.
   *
   * @throws java.lang.IllegalArgumentException when variable is of map type
   * but range old/young bit numbers are not equal.
   */
  public RangedVariable(NodeVariable variable, Range range) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(range);

    if (!variable.isType(DataTypeId.BIT_VECTOR) && !variable.isType(DataTypeId.MAP))
      throw new IllegalArgumentException(
          String.format("Illegal ranged variable type id: %s.", variable.getDataTypeId()));

    Node young = range.getYoung();
    Node old = range.getOld();

    if (variable.isType(DataTypeId.BIT_VECTOR)) {

      /* Check rather range does not spread outside the 0..size-1 bit numbers. */

      NodeValue zero = NodeValue.newInteger(0);
      NodeValue max = NodeValue.newInteger(variable.getDataType().getSize());

      if (!ExprUtils.isSAT(
          new NodeOperation(
              StandardOperation.AND,
              new NodeOperation(StandardOperation.GREATEREQ, young, zero),
              new NodeOperation(StandardOperation.LESSEQ, old, max))))
        throw new IllegalArgumentException(
            String.format("Incorrect range: %s < 0 or %s > %s.", young, old, max));
    }

    if (variable.isType(DataTypeId.MAP) &&
        !ExprUtils.isSAT(new NodeOperation(StandardOperation.EQ, young, old)))
      throw new IllegalArgumentException(
          String.format("Incorrect range: %s != %s for map", old, young));

    this.variable = variable;
    this.range = range;
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return a full copy of the current object
   */
  public RangedVariable deepcopy() {
    RangedVariable deepcopy = new RangedVariable();
    deepcopy.setVariable((NodeVariable) this.variable.deepCopy());
    if (this.isRanged()) {
      deepcopy.setRange(this.range.deepcopy());
    }
    return deepcopy;
  }

  /**
   * Returns variable.
   *
   * @return variable
   */
  public NodeVariable getVariable() { return this.variable; }

  /**
   * Returns variable {@link java.lang.String} string name.
   *
   * @return variable name
   */
  public String getName() { return this.getVariable().getName(); }

  /**
   * Sets new variable.
   *
   * NOTE: corresponding range is set to {@code null}.
   *
   * @param nodeVariable variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setVariable(NodeVariable nodeVariable) {
    InvariantChecks.checkNotNull(nodeVariable);

    this.variable = nodeVariable;
    this.range = null;
  }

  /**
   * Returns range.
   *
   * @return range
   */
  public Range getRange() { return this.range; }

  /**
   * Sets new range.
   *
   * @param range range
   */
  public void setRange(Range range) { this.range = range; }

  /**
   * Checks rather this is really ranged.
   *
   * @return {@code true} if this contains nontrivial range,{@code false} otherwise
   */
  public boolean isRanged() {
    return this.range != null && this.range.getYoung() != null && this.range.getOld() != null;
  }

  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder();
    builder.append(this.variable.toString());
    if (this.range != null) {
      builder.append("[");
      builder.append(this.range.toString());
      builder.append("]");
    }
    return builder.toString();
  }

  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o == null || getClass() != o.getClass())
      return false;

    boolean result;
    RangedVariable that = (RangedVariable) o;

    if (this.range != null && that.getRange() != null) {
      result = this.variable.equals(that.variable) && this.range.equals(that.range);
    } else {
      result = this.variable.equals(that.variable);
    }
    return result;
  }

  @Override
  public int hashCode() {
    int result = this.variable.hashCode();
    if (this.range != null) {
      result = 31 * result + this.range.hashCode();
    }
    return result;
  }
}
