/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprPrinter;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Basic block node of process' control flow graph.
 *
 * Basic blocks can be either of concurrent (non-blocking) type
 * or of sequential (blocking) type. The concurrent type means
 * that all the assignments should be executed simultaneously
 * in such a concurrent mode. The sequential type means that
 * all the assignments should be executed sequentially in the same order
 * they have in it's collection.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class BasicBlock extends CfgNode implements UseDef {

  /* The assignments. */
  private final List<Assignment> assignments;

  /* The concurrent type flag. */
  private boolean isConcurrent;

  /* The assignment separators depending on HDL and concurrent type. */

  private static final String VHDL_CONCURRENT = "<=";
  private static final String VHDL_NON_CONCURRENT = ":=";
  private static final String VERILOG_CONCURRENT = "<=";
  private static final String VERILOG_NON_CONCURRENT = "=";

  /**
   * Constructs object of the specified type flag.
   *
   * @param isConcurrent block type flag
   */
  public BasicBlock(boolean isConcurrent) {
    this.assignments = new ArrayList<>();
    this.isConcurrent = isConcurrent;
  }

  /**
   * Constructs object with the specified assignments and the type flag.
   *
   * @param assignments assignments
   * @param isConcurrent block type flag
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public BasicBlock(List<Assignment> assignments, boolean isConcurrent) {
    InvariantChecks.checkNotNull(assignments);

    this.assignments = assignments;
    this.isConcurrent = isConcurrent;
  }

  /**
   * Constructs object with the specified assignment and the flag type.
   *
   * @param assignment assignment
   * @param isConcurrent block type flag
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public BasicBlock(Assignment assignment, boolean isConcurrent) {
    InvariantChecks.checkNotNull(assignment);

    this.assignments = new ArrayList<>();
    this.assignments.add(assignment);
    this.isConcurrent = isConcurrent;
  }

  /**
   * Constructs object with the specified assignment and of non-concurrent type.
   *
   * @param assignment assignment
   */
  public BasicBlock(Assignment assignment) { this(assignment, false); }

  /**
   * Adds assignment statement.
   *
   * Assignment will be added with on 1 greater number
   * than other assignments have.
   *
   * @param stmt assignment to be added
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addAssignment(Assignment stmt) {
    InvariantChecks.checkNotNull(stmt);

    this.assignments.add(stmt);
  }

  /**
   * Appends assignments list to the last assignment of basic block.
   *
   * @param statements list of concurrent assignments
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void appendAssignments(List<Assignment> statements) {
    InvariantChecks.checkNotNull(statements);

    this.assignments.addAll(statements);
  }

  /**
   * Returns all block assignments.
   *
   * @return assignments list
   */
  public List<Assignment> getAssignments() { return this.assignments; }

  /**
   * Removes all block statements.
   */
  public void clearStatements() { this.assignments.clear(); }

  /**
   * Checks rather object contains only concurrent assignments.
   *
   * @return {@code true} if contains only concurrent assignments,
   * {@code false} otherwise.
   */
  public boolean isConcurrent() { return this.isConcurrent; }

  /**
   * Sets the concurrent flag to the specified value.
   *
   * @param concurrent concurrent flag
   */
  public void setConcurrent(boolean concurrent) { this.isConcurrent = concurrent; }

  @Override
  public CfgModelNode deepcopy() { return new BasicBlock(this.assignments, this.isConcurrent); }

  @Override
  public String getDescription(ExprTreePrinter printer) {
    StringBuilder description = new StringBuilder("");

    Iterator<Assignment> iterator = this.assignments.iterator();
    while (iterator.hasNext()) {
      Assignment stmt = iterator.next();

      /* Determine assignment kind. */
      String kind = null;
      if (printer instanceof ExprPrinter) {
        if (((ExprPrinter) printer).name().equals(ExprPrinter.VERILOG.name())) {

          if (isConcurrent()) {
            kind = VERILOG_CONCURRENT;
          } else {
            kind = VERILOG_NON_CONCURRENT;
          }
        } else if (((ExprPrinter) printer).name().equals(ExprPrinter.VHDL.name())) {
          if (isConcurrent()) {
            kind = VHDL_CONCURRENT;
          } else {
            kind = VHDL_NON_CONCURRENT;
          }
        }
      }

      if (kind != null) {
        description.append(stmt.getDescription(printer, kind));
      } else
          throw new IllegalArgumentException("Unknown language");

      if (iterator.hasNext()) {
        description.append("\n");
      }
    }
    return description.toString();
  }

  @Override
  public CfgNodeType getType() { return CfgNodeType.BASIC_BLOCK; }

  @Override
  public CfgModelNode getOnlyParent() { return this.parentNode; }

  @Override
  public CfgModelNode getOnlyChild() { return this.childNode; }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> uses = new LinkedHashSet<>();

    for (Assignment assignment : this.assignments) {
      Collection<NodeVariable> assignmentUses = assignment.getUses();

      for (NodeVariable assignmentUse : assignmentUses) {
        if (!uses.contains(assignmentUse)) {
          uses.add(assignmentUse);
        }
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    Set<NodeVariable> defines = new LinkedHashSet<>();

    for (Assignment assignment : this.assignments) {
      Collection<NodeVariable> assignmentDefines = assignment.getDefines();

      for (NodeVariable assignmentDefine : assignmentDefines) {
        if (!defines.contains(assignmentDefine)) {
          defines.add(assignmentDefine);
        }
      }
    }
    return defines;
  }

//  @Override
//  public boolean equals(Object o) {
//    if (this == o)
//      return true;
//    if (o == null || getClass() != o.getClass())
//      return false;
//
//    BasicBlock that = (BasicBlock) o;
//
//    return isConcurrent == that.isConcurrent && assignments.equals(that.assignments);
//  }
//
//  @Override
//  public int hashCode() {
//    int result = assignments.hashCode();
//    result = 31 * result + (isConcurrent ? 1 : 0);
//    return result;
//  }
}
