/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * Control flow graph node that keeps condition\case branch.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Case extends CfgNode {

  /* The parent node. */
  private Switch parent;

  /* The values that parent node's expression can be equal to. */
  private Set<NodeValue> values;

  /* The default node flag: such object does not have any values. */
  private boolean isDefault;

  /* The default node string description. */
  private static final String DEFAULT = "default";

  /**
   * The default constructor.
   */
  public Case() {
    this.values = new LinkedHashSet<>();
    this.isDefault = false;
  }

  /**
   * Constructs node with the specified possible value.
   *
   * @param value value
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Case(NodeValue value) {
    InvariantChecks.checkNotNull(value);

    this.values = new LinkedHashSet<>();
    this.values.add(value);
    this.isDefault = false;
  }

  /**
   * Constructs node with the specified possible values.
   *
   * @param values values
   */
  public Case(Set<NodeValue> values) {
    this.values = values;
    this.isDefault = false;
  }

  /**
   * Constructs node with the specified values and default flag.
   *
   * Note: if values collection is non-empty,
   * than the default flag must be set to {@code false}.
   * Otherwise the default flag must be set to {@code true}.
   *
   * @param values values
   * @param isDefault default flag
   *
   * @throws java.lang.NullPointerException when values argument is {@code null}.
   */
  public Case(Set<NodeValue> values, boolean isDefault) {
    InvariantChecks.checkNotNull(values);

    if (values.isEmpty() != isDefault)
      throw new IllegalArgumentException(String.format("Can't create %s node.",this.getType()));
    this.values = values;
    this.isDefault = isDefault;
  }

  /**
   * Checks rather case node is default.
   *
   * @return {@code true} if node is default,
   * {@code false} otherwise
   */
  public boolean isDefault() { return this.isDefault; }

  /**
   * Make node default.
   *
   * Default node must not have any possible values.
   *
   * @throws java.lang.IllegalStateException method is called for object
   * with non-empty internal collection of values.
   */
  public void setDefault() {
    if (this.values.size() > 0)
      throw new IllegalStateException(
          String.format("Cannot set node %s as default with non-empty values collection: %s.",
        this.getType(), this.values.toString()));

    this.isDefault = true;
  }

  /**
   * Adds one more possible value.
   *
   * @param value possible value for this
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @throws java.lang.IllegalStateException when the object is default.
   */
  public void addValue(NodeValue value) {
    InvariantChecks.checkNotNull(value);

    if (this.isDefault)
      throw new IllegalStateException(
        String.format("Cannot add value to default node: %s.",
            this.getType()));

    this.values.add(value);
  }

  /**
   * Returns node values.
   *
   * @return values
   */
  public Set<NodeValue> getValues() { return this.values; }

  @Override
  public CfgModelNode deepcopy() { return new Case(this.values, this.isDefault); }

  @Override
  public String getDescription(ExprTreePrinter printer) {
    if (!this.values.isEmpty()) {
      StringBuilder builder = new StringBuilder();
      Iterator<NodeValue> iterator = this.values.iterator();

      while (iterator.hasNext()) {
        NodeValue value = iterator.next();
        builder.append(value.toString());
        if (iterator.hasNext()) {
          builder.append(",");
        }
      }
      return builder.toString();
    } else {
      return Case.DEFAULT;
    }
  }

  @Override
  public CfgNodeType getType() { return CfgNodeType.CASE; }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @throws java.lang.IllegalStateException when this object already
   * has non-null parent node.
   *
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.Switch} type.
   */
  @Override
  public void addParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent != null)
      throw new IllegalStateException(
        String.format("Parent node for %s re-assignment with: %s.",
          this.getType(), node.getType()));

    if (node.getType() == CfgNodeType.SWITCH) {
      this.parent = (Switch) node;
    } else {
      throw new IllegalArgumentException(
        String.format("Unsupported type for parent: %s.", node.getType()));
    }
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalArgumentException when there is no
   * such parent node in this object.
   */
  @Override
  public void removeParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent.equals(node)) {
      this.parent = null;
      node.removeChild(this);
    } else {
      throw new IllegalArgumentException(
        String.format("No such parent node: %s.", node.getId()));
    }
  }

  @Override
  public boolean hasParents() { return this.parent != null; }

  @Override
  public CfgModelNode getOnlyParent() { return this.parent; }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton((CfgModelNode) this.parent);
  }

  @Override
  public CfgModelNode getOnlyChild() { return this.childNode; }

  @Override
  public Set<NodeVariable> getUses() { return new LinkedHashSet<>(); }

  @Override
  public Set<NodeVariable> getDefines() { return new LinkedHashSet<>(); }

//  @Override
//  public boolean equals(Object o) {
//    if (this == o)
//      return true;
//    if (o == null || getClass() != o.getClass())
//      return false;
//
//    Case aCase = (Case) o;
//
//    return isDefault == aCase.isDefault && values.equals(aCase.values);
//  }
//
//  @Override
//  public int hashCode() {
//    int result = values.hashCode();
//    result = 31 * result + (isDefault ? 1 : 0);
//    return result;
//  }
}
