/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

/**
 * Control flow graph of some process.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Cfg extends CfgModelNode implements UseDef {

  /* The graph nodes. */
  private final List<CfgNode> nodes;

  /* The parent process. */
  private Process parent;

  /**
   * The default constructor.
   *
   * By default the control flow graph is not empty.
   * It always contains source and sink nodes,
   * connected as parent and child.
   */
  public Cfg() {
    this.nodes = new ArrayList<>();

    Source src = new Source(this);
    Sink sink = new Sink();
    src.addChild(sink);
    this.nodes.add(src);
    this.nodes.add(sink);
  }

  /**
   * Adds control flow graph node.
   *
   * NOTE: the argument node will only be added to control flow graph
   * without establishing any "child-parent" links with already added nodes.
   *
   * @param node node to be added
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @throws java.lang.IllegalArgumentException when the specified node
   * neither of {@link ru.ispras.retrascope.model.cfg.CfgNode} type
   * nor of {@link ru.ispras.retrascope.model.cfg.Cfg} type.
   */
  public void addNode(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      this.nodes.add((CfgNode) node);
    } else if (node instanceof Cfg) {
      this.nodes.addAll(((Cfg) node).getNodes());
    } else
      throw new IllegalArgumentException(
          String.format("Incorrect node type: %s.", node.getType()));
  }

  /**
   * Adds control flow graph nodes collection.
   *
   * NOTE: the argument nodes will only be added to control flow graph
   * without establishing any "child-parent" links with already added nodes.
   *
   * @param nodes nodes to be added
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addNodes(Collection<CfgNode> nodes) {
    InvariantChecks.checkNotNull(nodes);

    this.nodes.addAll(nodes);
  }

  /**
   * Appends another control flow graph.
   *
   * NOTE: an argument control flow graph nodes
   * will not only be added to this nodes
   * but argument source node will be linked with this sink node
   * in some correct way.
   *
   * @param cfg control flow graph to be appended
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void append(Cfg cfg) {
    InvariantChecks.checkNotNull(cfg);

    // get this graph sink
    CfgNode sink = getSink();

    // get sink node's parent
    CfgModelNode sinkParent = sink.getOnlyParent();

    // remove sink & link with it's parent
    sinkParent.removeChild(sink);
    this.nodes.remove(sink);

    // get argument's source node
    CfgNode cfgSource = cfg.getSource();

    // get source's node child
    CfgModelNode srcChild = cfgSource.getOnlyChild();

    // remove source & link with it's child
    srcChild.removeParent(cfgSource);
    cfg.removeNode(cfgSource);

    // make a link between parent & child
    sinkParent.addChild(srcChild);

    // add argument's nodes
    this.nodes.addAll(cfg.getNodes());
  }

  /**
   * Substitutes internal default {@link ru.ispras.retrascope.model.cfg.Source} node by other node.
   *
   *  Previous SOURCE node will be removed from control flow graph.
   *
   * @param node node to be added as the new "source"
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addAsSource(CfgNode node) {
    InvariantChecks.checkNotNull(node);

    Source src = getSource();
    CfgModelNode srcChild = src.getOnlyChild();
    src.removeChild(srcChild);
    this.removeNode(src);
    node.addChild(srcChild);
  }

  /**
   * Substitutes internal default {@link ru.ispras.retrascope.model.cfg.Sink} node by other node.
   *
   * Previous SINK node will be removed from control flow graph.
   *
   * @param node node to be added as the new "sink"
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addAsSink(CfgNode node) {
    InvariantChecks.checkNotNull(node);

    Sink sink = getSink();
    Collection<CfgModelNode> sinkParents = sink.getParents();
    final Merge merge = new Merge();
    this.addNode(merge);
    for (CfgModelNode parent : sinkParents) {
      parent.removeChild(sink);
      merge.addParent(parent);
    }
    this.removeNode(sink);
    merge.addChild(node);
  }

  /**
   * Removes internal node from control flow graph.
   *
   * NOTE: the argument node will only be removed from control flow graph
   * without removing any "child-parent" links with existing nodes.
   *
   * @param node node to be removed
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void removeNode(CfgNode node) {
    InvariantChecks.checkNotNull(node);

    this.nodes.remove(node);
  }

  /**
   * Returns all internal control flow graph nodes.
   *
   * @return nodes collection
   */
  public Collection<CfgNode> getNodes() { return this.nodes; }

  /**
   * Returns {@link ru.ispras.retrascope.model.cfg.Source} node of this graph.
   *
   * @return source node
   */
  public Source getSource() { return (Source) getNode(CfgNodeType.SOURCE); }

  /**
   * Returns {@link ru.ispras.retrascope.model.cfg.Sink} node of this graph.
   *
   * @return sink node
   */
  public Sink getSink() { return (Sink) getNode(CfgNodeType.SINK); }

  /**
   * Returns first internal node of the specified type.
   *
   * Note: this method does not process sub-nodes
   * of {@link Cfg} control flow graph type nodes.
   *
   * @param type target node type
   *
   * @return control flow graph node
   *
   * @throws java.util.NoSuchElementException if there is no such node
   */
  private CfgNode getNode(CfgNodeType type) {
    for (CfgNode node : this.nodes) {
      if (node.getType() == type) {
        return node;
      }
    }
    throw new NoSuchElementException(
      String.format("No such node with type: %s.", type));
  }

  /**
   * Returns first internal node with the specified identifier.
   *
   * Note: this method does not process sub-nodes
   * of {@link Cfg} control flow graph type nodes.
   *
   * @param id target node identifier
   *
   * @return control flow graph node
   *
   * @throws java.util.NoSuchElementException if there is no such node
   */
  public CfgNode getNode(String id) {
    for (CfgNode node : this.nodes) {
      if (node.getId().equals(id)) {
        return node;
      }
    }
    throw new NoSuchElementException(
      String.format("No such node with id: %s.", id));
  }

  /**
   * Returns last internal node with the specified type.
   *
   * @param type type of target node
   *
   * @return control flow graph node or {@code null} if there is no such node
   */
  public CfgNode getLastNode(CfgNodeType type) {

    int lastNodeIndex = -1;

    for (CfgNode node : this.nodes) {
      int nodeIndex = this.nodes.indexOf(node);
      if (node.getType() == type && nodeIndex > lastNodeIndex) {
        lastNodeIndex = nodeIndex;
      }
    }

    if (lastNodeIndex >= 0) {
      return this.nodes.get(lastNodeIndex);
    } else {
      return null;
    }
  }

  /**
   * Applies bindings to control flow graph internals.
   *
   * @param bindings "variable name - expression" binding
   *
   * @throws NullPointerException when argument is {@code null}.
   * @throws IllegalArgumentException when some argument element
   * has a key that is not of {@link ru.ispras.fortress.expression.NodeVariable} type.
   */
  public void applyToCfg(Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    /* Elaborate only nodes that contain code with variables. */
    for (CfgNode node : this.nodes) {

      /* Elaborate basic block. */
      if (node.getType() == CfgNodeType.BASIC_BLOCK) {
        for (Assignment stmt : ((BasicBlock) node).getAssignments()) {
          RangedVariable rangedVariable = stmt.getRangedVariable();
          String name = rangedVariable.getName();

          /* If defined variable is used-in-bindings. */
          if (bindings.containsKey(name)) {
            Node binding = bindings.get(name);
            if (bindings.get(name) instanceof NodeVariable) {

              rangedVariable.setVariable(((NodeVariable) binding));
            } else {
              throw new IllegalArgumentException(
                String.format("Illegal node kind: %s.", binding.getKind()));
            }
          }

          Set<String> usedVarNames = stmt.getUsedVariableNames();
          for (String usedVarName : usedVarNames) {

            /* If used-in-expression variable is used-in-bindings. */
            if (bindings.containsKey(usedVarName)) {
              stmt.setExpression(
                Transformer.substitute(
                  stmt.getExpression(),
                  usedVarName,
                  bindings.get(usedVarName)));
            }
          }
        }
      } else if (node.getType() == CfgNodeType.SWITCH) {
        Set<NodeVariable> uses = node.getUses();

        for (NodeVariable use : uses) {

          String useName = use.getName();

          /* If used-in-expression variable is used-in-bindings. */
          if (bindings.containsKey(useName)) {
            ((Switch)node).setExpression(
              Transformer.substitute(
                ((Switch) node).getCondition(),
                useName,
                bindings.get(useName)));
          }
        }
      }
    }
  }

  @Override
  public CfgModelNode deepcopy() {
    Cfg copy = new Cfg();

    /*Make a copies of internal nodes. */

    Map<CfgNode, CfgNode> oldNewMap = new LinkedHashMap<>();
    List<CfgNode> copies = new ArrayList<>();

    for (CfgNode node : this.nodes) {
      CfgModelNode nodeCopy = node.deepcopy();
      oldNewMap.put(node, (CfgNode) nodeCopy);
      copies.add((CfgNode) nodeCopy);
    }

    /* Set links between node copies. */
    for (int i = 0; i < this.nodes.size(); i++) {
      for (CfgNode node : this.nodes) {
        CfgNode iNode = this.nodes.get(i);

        if (iNode.getChildren().contains(node)) {
          oldNewMap.get(iNode).addChild(oldNewMap.get(node));
        }
      }
    }

    copy.addNodes(copies);
    return copy;
  }

  /**
   * Throws an exception.
   *
   * @param node child node
   *
   * @throws java.lang.UnsupportedOperationException every time it is called
   */
  @Override
  public void addChild(CfgModelNode node) {
    throw new UnsupportedOperationException(
        String.format("No child for: %s.", this.getType()));
  }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @throws java.lang.IllegalStateException when this object
   * does not have such parent node.
   *
   *  @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.Process} type.
   */
  @Override
  public void addParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.PROCESS) {
      throw new IllegalArgumentException(
        String.format("Cannot add parent: %s.", this.getType()));
    }
    if (this.parent != null && !this.parent.equals(node)) {
      throw new IllegalStateException(
          String.format("Cannot add parent: %s.", this.getType()));
    }
    this.parent = (Process) node;

    if (!node.hasChildren() || !node.getChildren().contains(this)) {
      node.addChild(this);
    }
  }

  /**
   * Throws an exception.
   *
   * @param node child node
   *
   * @throws java.lang.UnsupportedOperationException every time it is called.
   */
  @Override
  public void removeChild(CfgModelNode node) {
    throw new UnsupportedOperationException(
        String.format("No child node for: %s.",
        this.getType()));
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.Process} type
   * @throws java.lang.IllegalStateException when this object
   * does not contain the specified node as parent.
   */
  @Override
  public void removeParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.PROCESS)
      throw new IllegalArgumentException(
        String.format("Incorrect node type: %s.", node.getType()));
    if (!this.parent.equals(node)) {
      throw new IllegalArgumentException(
          String.format("Cannot remove parent node of: %s.",
          this.getType()));
    } else {
      this.parent = null;
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton((CfgModelNode) this.parent);
  }

  @Override
  public CfgModelNode getOnlyChild() { return getSource(); }

  @Override
  public Collection<CfgModelNode> getChildren() {
    return Collections.singleton((CfgModelNode)getSource());
  }

  @Override
  public String getDescription(ExprTreePrinter printer) {
    return this.getType().name().toLowerCase();
  }

  @Override
  public CfgNodeType getType() { return CfgNodeType.CFG; }

  @Override
  public boolean hasParents() { return this.parent != null; }

  @Override
  public boolean hasChildren() { return true; }

  @Override
  public CfgModelNode getOnlyParent() { return this.parent; }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> uses = new LinkedHashSet<>();

    for (CfgModelNode node : this.nodes) {
      Collection<NodeVariable> nodeUses = node.getUses();

      for (NodeVariable use : nodeUses) {
        if (!uses.contains(use)) {
          uses.add(use);
        }
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    Set<NodeVariable> defines = new LinkedHashSet<>();

    for (CfgModelNode node : this.nodes) {

      Iterable<NodeVariable> nodeDefines = node.getDefines();
      for (NodeVariable define : nodeDefines) {
        if (!defines.contains(define)) {
          defines.add(define);
        }
      }
    }
    return defines;
  }

//  @Override
//  public boolean equals(Object o) {
//    if (this == o)
//      return true;
//    if (o == null || getClass() != o.getClass())
//      return false;
//
//    Cfg cfg = (Cfg) o;
//
//    if (nodes.size() != cfg.nodes.size())
//      return false;
//
//    for (int i = 0; i < nodes.size(); i++) {
//      if (!nodes.get(i).equals(cfg.nodes.get(i))) {
//        return false;
//      }
//    }
//    return true;
//  }
//
//  @Override
//  public int hashCode() { return nodes.hashCode(); }
}
