/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;

/**
 * Control flow graph model.
 *
 * A collection of control flow graph modules.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgModel extends Entity {

  /* The entity type. */
  public static final EntityType TYPE = EntityType.get("cfg");

  /* The child modules. */
  private Collection<Module> modules;

  /**
   * The default constructor.
   */
  public CfgModel() {
    super(CfgModel.TYPE);

    this.modules = new LinkedHashSet<>();
  }

  /**
   * Adds the specified entity.
   *
   * @param entity the entity to be added.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @throws java.lang.IllegalArgumentException when the specified entity
   * is not of {@link ru.ispras.retrascope.model.cfg.CfgModel} type.
   */
  @Override
  public void add(Entity entity) {
    InvariantChecks.checkNotNull(entity);

    if (entity instanceof CfgModel) {
      Collection<Module> entityModules = ((CfgModel) entity).getModules();
      this.addModules(entityModules);
    } else {
      throw new IllegalArgumentException(
          String.format("Incorrect type: %s.", entity.getEntityType().getId()));
    }
  }

  /**
   * Returns internal modules.
   *
   * @return internal modules collection
   */
  public Collection<Module> getModules() { return modules; }

  /**
   * Adds modules collection.
   *
   * @param modules modules collection
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addModules(Collection<Module> modules) {
    InvariantChecks.checkNotNull(modules);

    this.modules.addAll(modules);
  }

  /**
   * Adds the specified module to the model.
   *
   * @param module module
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void addModule(Module module) {
    InvariantChecks.checkNotNull(module);

    this.modules.add(module);
  }

  /**
   * Returns sub-module with the specified name.
   *
   * @param moduleName module name
   * @return sub-module with the specified name,
   * or {@code null} if it does not exist
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Module getModule(String moduleName) {
    InvariantChecks.checkNotNull(moduleName);

    for (Module module : this.modules) {
      if (module.getName().equals(moduleName)) {
        return module;
      }
      Module subModule = getSubModule(module, moduleName);
      if (subModule != null) {
        return subModule;
      }
    }
    return null;
  }

  private Module getSubModule(Module module, String moduleName) {
    InvariantChecks.checkNotNull(module);

    List<Module> subModules = module.getModules();

    if (!subModules.isEmpty()) {
      for (Module subModule : subModules) {
        if (subModule.getName().equals(moduleName)) {
          return subModule;
        }
        Module subSubModule = getSubModule(subModule, moduleName);
        if (subSubModule != null) {
          return subSubModule;
        }
      }
    }
    return null;
  }
}
