/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.Collection;

/**
 * Abstract node of control flow graph model.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgModelNode extends MetaInfo implements UseDef {

  /**
   * Returns a copy of this node.
   *
   * The copy does not keep the same links to parent or child nodes
   * as this does.
   *
   * @return deep copy of this node
   */
  public abstract CfgModelNode deepcopy();

  /**
   * Returns unique node identifier in string form.
   *
   * @return node identifier
   */
  public String getId() { return this.toString(); }

  /**
   * Returns node string description in the format of the specified printer.
   *
   * A description contains some significant part of source code
   * that characterizes this node.
   *
   * @param printer printer
   * @return description
   */
  public abstract String getDescription(ExprTreePrinter printer);

  /**
   * Returns node type.
   *
   * By default, node has {@code CfgNodeType.NODE} type.
   *
   * @return node type
   */
  public CfgNodeType getType() { return CfgNodeType.NODE; }

  /**
   * Checks rather node has parent nodes.
   *
   * @return {@code true} if node has at least one parent, {@code false} otherwise.
   */
  public abstract boolean hasParents();

  /**
   * Checks rather node has child nodes.
   *
   * @return {@code true} if node has at least one child,
   * {@code false} otherwise.
   */
  public abstract boolean hasChildren();

  /**
   * Returns the only parent node.
   *
   * @return only parent node
   */
  public abstract CfgModelNode getOnlyParent();

  /**
   * Returns parent nodes.
   *
   * @return parent nodes collection
   */
  public abstract Collection<CfgModelNode> getParents();

  /**
   * Returns the only child node.
   *
   * @return only child node
   */
  public abstract CfgModelNode getOnlyChild();

  /**
   * Return child nodes.
   *
   * @return child nodes collection
   */
  public abstract Collection<CfgModelNode> getChildren();

  /**
   * Adds child node.
   *
   * @param node child node
   */
  public abstract void addChild(CfgModelNode node);

  /**
   * Adds parent node.
   *
   * @param node parent node
   */
  public abstract void addParent(CfgModelNode node);

  /**
   * Removes child node.
   *
   * @param node child node
   */
  public abstract void removeChild(CfgModelNode node);

  /**
   * Removes parent node.
   *
   * @param node parent node
   */
  public abstract void removeParent(CfgModelNode node);
}
