/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.Set;

/**
 * Abstract control flow graph node.
 *
 * Has only one parent node and only one child node.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgNode extends CfgModelNode {

  /* The parent node. */
  protected CfgNode parentNode;

  /* The child node. */
  protected CfgNode childNode;

  @Override
  public String getDescription( ExprTreePrinter printer) {
    return this.getType().name().toLowerCase();
  }

  @Override
  public boolean hasParents() { return this.parentNode != null; }

  @Override
  public boolean hasChildren() { return this.childNode != null; }

  /**
   * Adds the specified node as child.
   *
   * @param node child node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalStateException when this object already
   * has non-null child node.
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   */
  @Override
  public void addChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if(this.childNode == null) {

        this.childNode = (CfgNode) node;

        if (!node.hasParents() || !node.getParents().contains(this)) {
          node.addParent(this);
        }
      } else
        throw new IllegalStateException(
          String.format("Child node for %s re-assignment with: %s.",
            this.getType(), node.getType()));
    } else {
      throw new IllegalArgumentException(
        String.format("Incorrect node type: %s.", node.getType()));
    }
  }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalStateException when this object
   * already has non-null parent node.
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   */
  @Override
  public void addParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if ((node instanceof CfgNode)) {

      if (this.parentNode == null) {
        this.parentNode = (CfgNode) node;

        if (!node.hasChildren() || !node.getChildren().contains(this)) {
          node.addChild(this);
        }
      } else
        throw new IllegalStateException(
          String.format("Parent node for %s re-assignment with: %s.",
            this.getType(), node.getType()));
    } else {
      throw new IllegalArgumentException(
        String.format("Incorrect node type: %s.", node.getType()));
    }
  }

  /**
   * Removes child node.
   *
   * @param node child node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalStateException when this object
   * does not have such child node.
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   */
  @Override
  public void removeChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if (this.childNode.equals(node)) {

        this.childNode = null;

        if (node.hasParents() && node.getParents().contains(this)) {
          node.removeParent(this);
        }
      } else {
        throw new IllegalStateException(
          String.format("No such child node: %s.", node.getId()));
      }
    } else {
      throw new IllegalArgumentException(
        String.format("Incorrect node type: %s.", node.getType()));
    }
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalStateException when this object
   * does not have such parent node.
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   */
  @Override
  public void removeParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if (this.parentNode.equals(node)) {

        this.parentNode = null;

        if (node.hasChildren() && node.getChildren().contains(this)) {
          node.removeChild(this);
        }
      } else {
        throw new IllegalStateException(
          String.format("No such child node: %s.", node.getId()));
      }
    } else {
      throw new IllegalArgumentException(
        String.format("Incorrect node type: %s.", node.getType()));
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton((CfgModelNode) this.parentNode);
  }

  @Override
  public Collection<CfgModelNode> getChildren() {
    return Collections.singleton((CfgModelNode) this.childNode);
  }

  /**
   * Throws an exception.
   *
   * @return {@code null}
   *
   * @throws java.lang.UnsupportedOperationException every time it is called.
   */
  @Override
  public Set<NodeVariable> getUses() {
    throw new UnsupportedOperationException(
        String.format("Not implemented for: %s.", this.getType()));
  }

  /**
   * Throws an exception.
   *
   * @return {@code null}
   *
   * @throws java.lang.UnsupportedOperationException every time it is called.
   */
  @Override
  public Set<NodeVariable> getDefines() {
    throw new UnsupportedOperationException(
        String.format("No child nodes for: %s.", this.getType()));
  }
}
