/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.RangedVariable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Basic abstract class for control flow graph model {@link ru.ispras.retrascope.model.cfg.CfgModel}
 * that contains a collection of variable declarations
 * {@link ru.ispras.retrascope.model.cfg.VariableDeclaration} and a collection of bindings
 * between internal variables and external expressions.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgVariableContainerNode extends CfgModelNode {

  /* Constants for new variables creation. */
  private static final String UNDERSCORE = "_";
  public static final String NEW_VAR_PREFIX = "NEW_VAR" + UNDERSCORE;

  /* Internal variables declarations. */
  private Map<String, VariableDeclaration> variables;

  /* The bindings between internal variables and external expressions. */
  private final Map<String, Node> bindings;

  /**
   * The default constructor.
   */
  public CfgVariableContainerNode() {
    this.variables = new LinkedHashMap<>();
    this.bindings = new LinkedHashMap<>();
  }

  /**
   * Returns a collection of variables declarations.
   *
   * @return declarations of node variables
   */
  public Collection<VariableDeclaration> getDeclarations() { return this.variables.values(); }

  /**
   * Mapping of names of internal variables to variable declarations.
   *
   * @return mapping of variable names to declarations
   */
  public Map<String, VariableDeclaration> getVariablesMapping() { return this.variables; }

  /**
   * Returns module's bindings.
   *
   * @return bindings
   */
  public Map<String, Node> getBindings() { return this.bindings; }

  /**
   * Sets module's bindings.
   *
   * @param bindings new bindings
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setBindings(Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    this.bindings.clear();
    this.bindings.putAll(bindings);
  }

  /**
   * Checks does this module contain the specified variable.
   *
   * @param variable variable to be checked
   *
   * @return {@code true} if module already has such variable declared, {@code false} otherwise.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public boolean containsVariable(NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    return this.variables.containsKey(variable.getName());
  }

  /**
   * Returns the declaration of the specified variable.
   *
   * @param variable variable of which the declaration should be returned
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @return declaration of the specified variable.
   */
  public VariableDeclaration getDeclaration(NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    return this.variables.get(variable.getName());
  }

  /**
   * Adds variable declaration.
   *
   * NOTE: module MUST NOT contain the specified variable.
   * Check this fact by corresponding check-method first.
   *
   * @param variable variable to be added
   *
   * @throws java.lang.NullPointerException when variable argument is {@code null}.
   * @throws java.lang.IllegalStateException when this module object
   * already contains such variable.
   */
  public void addVariable(NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    if (this.variables.containsKey(variable.getName()))
      throw new IllegalStateException(
        String.format("Already contain such node: %s.", variable.getName()));

    this.variables.put(variable.getName(), new VariableDeclaration(variable));
  }

  /**
   * Adds variable declaration.
   *
   * NOTE: module MUST NOT contain the specified variable.
   * Check this fact by corresponding method first.
   *
   * @param variable variable to be added
   * @param invariant variable invariant (can be {@code null})
   *
   * @throws java.lang.NullPointerException when variable argument is {@code null}.
   * @throws java.lang.IllegalStateException when this module object
   * already contains such variable.
   */
  public void addVariable(NodeVariable variable, Node invariant) {
    InvariantChecks.checkNotNull(variable);

    if (invariant == null) {
      addVariable(variable);
    } else {
      if (!this.variables.keySet().contains(variable.getName())) {
        this.variables.put(variable.getName(), new VariableDeclaration(variable, invariant));
      } else {
        throw new IllegalStateException(
          String.format("Already contain such node: %s.", variable.getName()));
      }
    }
  }

  /**
   * Creates new variable of the specified data type and adds it to the declarations of this module.
   *
   * @param dataType new variable data type
   *
   * @return new variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public NodeVariable addNewVariable(DataType dataType) {
    InvariantChecks.checkNotNull(dataType);

    /* Unique variable name creation. */
    String varName = NEW_VAR_PREFIX + (new Integer(this.variables.size())).toString();

    NodeVariable nodeVariable = new NodeVariable(varName, dataType);
//    nodeVariable.setUserData(new VariableData(VariableType.REG));

    /* There is no invariant for new variable. */
    this.addVariable(nodeVariable);
    return nodeVariable;
  }

  /**
   * Creates a version of already declared variable and adds it to the declarations of this module.
   *
   * @param variable an already declared module variable
   *
   * @return a version of variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalArgumentException when this object does not contain
   * the specified variable.
   */
  public NodeVariable addVariableVersion(RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    NodeVariable nodeVariable = variable.getVariable();
    String varName = variable.getName();
    if (!this.variables.containsKey(varName))
      throw new IllegalArgumentException(String.format("No such variable: %s.", varName));

    /* Calculate data type of version of the specified variable. */
    DataType dataType;
    if (!variable.isRanged()) {
      dataType = nodeVariable.getDataType();
    } else {
      DataType varDataType = nodeVariable.getDataType();
      Node right = variable.getRange().getOld();
      Node left = variable.getRange().getOld();

      /* Ranged bit vector. */
      if (varDataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {

        if (left.equals(right)) {
          /* 1-bit size vector. */
          dataType = DataType.BIT_VECTOR(1);
        } else {
          /* Range is constant. */
            throw new IllegalArgumentException(
                String.format("Cannot parse: %s.", variable.toString()));
        }
      } else if (varDataType.getTypeId().equals(DataTypeId.MAP)) {
        if (left.equals(right)) {
          /* Element of array. */
          dataType =
              DataType.BIT_VECTOR(
                  ((DataType)nodeVariable.getData().getType().getParameters()[1]).getSize());
        }
        else
          throw new IllegalArgumentException(
            String.format("Cannot parse: %s.", varDataType.toString()));
      } else
          throw new IllegalArgumentException(
              String.format("Cannot parse: %s.", varDataType.toString()));
    }

    /* Create a unique name of version. */
    String versionName = varName +
      UNDERSCORE +
      (variable.isRanged() ? variable.getRange().toString() : "") +
      UNDERSCORE +
      NEW_VAR_PREFIX +
      (new Integer(this.variables.size())).toString();

    /* Create version and add in to the node. */
    NodeVariable version = new NodeVariable(versionName, dataType);

    version.setUserData(nodeVariable.getUserData());

    /* Version has the same invariant as the specified variable. */
    if (getDeclaration(nodeVariable).hasInvariant()) {
      Node invariant = getDeclaration(nodeVariable).getInvariant();
      addVariable(version, Transformer.substitute(invariant, nodeVariable.getName(), version));
    } else {
      this.addVariable(version);
    }
    return version;
  }

  /**
   * Returns declared variable with specified name.
   *
   * NOTE: the variable must have been already declared.
   *
   * @param name name of variable to be returned
   *
   * @return declared variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public NodeVariable getVariable(String name) {
    InvariantChecks.checkNotNull(name);

    return (NodeVariable) this.variables.get(name).getVarExpression();
  }

  /**
   * Sets declared variables mapping.
   *
   * @param map mapping of variable name to variable declaration
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setVariablesMapping(Map<String, VariableDeclaration> map) {
    InvariantChecks.checkNotNull(map);

    this.variables.putAll(map);
  }

  /**
   * Returns declared variables.
   *
   * @return variables collection
   */
  public List<NodeVariable> getVariables() {
    List<NodeVariable> vars = new ArrayList<>();

    for (String varName : this.variables.keySet()) {
      Node node = this.variables.get(varName).getVarExpression();
      switch (node.getKind()) {
        case VARIABLE:
          vars.add((NodeVariable) node);
          break;
        default:
          break;
      }
    }
    return vars;
  }

  /**
   * Removes declaration of variable with the specified name.
   *
   * @param name variable name
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void removeDeclaration(String name) {
    InvariantChecks.checkNotNull(name);

    this.variables.remove(name);
  }

  /**
   * Adds the variable with the specified name and the specified declaration.
   *
   * @param name variable name
   * @param declaration variable declaration
   *
   *@throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public void addDeclaration(String name, VariableDeclaration declaration) {
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(declaration);

    this.variables.put(name, declaration);
  }
}
