/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Control flow graph model walker.
 *
 * Provides methods to traverse nodes of control flow graph model.
 *
 * <p>
 * NOTE: Actions taken in the traversal process depend on the current status of the visitor (see
 * {@link ru.ispras.retrascope.model.cfg.CfgVisitor.Status}). There are three possible statuses:
 * <ol>
 * <li>OK - continue traversal
 * <li>SKIP - skip child nodes
 * <li>ABORT - stop traversal
 * </ol>
 *
 * <p>
 * The status is checked after calling any visitor method. Once ABORT is set, all traversal methods
 * return. If after a call to a method having the Begin suffix (e.g. onModuleBegin), the SKIP status
 * is set (not ABORT and not OR), nested elements of the visited node (child nodes or subtrees) are
 * not traversed and a corresponding terminating method (that has the End suffix) is called.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgWalker {

  /* The visitor. */
  private final CfgVisitor visitor;

  /* The already-visited node identifiers collection. */
  private final Set<String> visitedIds;

  /* The walking mode. */
  private final CfgWalkerMode mode;

  /**
   * The default constructor.
   *
   * Constructs object with the specified visitor. By default,
   * object's mode is "depth-first search without repetitions".
   *
   * @param visitor visitor to be applied to control flow graph model
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public CfgWalker(CfgVisitor visitor) {
    InvariantChecks.checkNotNull(visitor);

    this.visitor = visitor;

    /* The default walking mode. */
    this.mode = CfgWalkerMode.getDefaultMode();
    this.visitedIds = null;
  }

  /**
   * Constructs object with the specified visitor and mode.
   *
   * @param visitor visitor to be applied to control flow graph model
   * @param mode visiting & walking mode
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public CfgWalker(CfgVisitor visitor, CfgWalkerMode mode) {
    InvariantChecks.checkNotNull(visitor);
    InvariantChecks.checkNotNull(mode);

    this.visitor = visitor;
    this.mode = mode;

    if (mode.equals(CfgWalkerMode.DFS_WITHOUT_REPEAT)) {
      this.visitedIds = new LinkedHashSet<>();
    } else {
      this.visitedIds = null;
    }
  }

  /**
   * Checks whether the current status of the visitor equals the specified status.
   *
   * @param status Status to be checked for equality with the current status of the visitor.
   * @return <code>true</code> if the visitor status equals the specified status, or
   *         <code>false</code> otherwise.
   */

  private boolean isStatus(CfgVisitor.Status status) {
    return this.visitor.getStatus() == status;
  }

  /**
   * Starts control flow graph model walking.
   *
   * @param model control flow graph model
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void start(CfgModel model) {
    InvariantChecks.checkNotNull(model);

    visitor.onRootBegin();

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visitCfgModel(model);
      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    visitor.onRootEnd();
  }

  /**
   * Visits a sequence of control flow graph model nodes.
   *
   * @param nodes a sequence of nodes to be visited
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void visit(Iterable<CfgModelNode> nodes) {
    InvariantChecks.checkNotNull(nodes);

    for (CfgModelNode tree : nodes) {
      switch (this.mode) {
        case DFS_WITHOUT_REPEAT:
          if (this.visitedIds.isEmpty() || !this.visitedIds.contains(tree.getId())) {
            this.visitedIds.add(tree.getId());
            visitNode(tree);

            if (isStatus(CfgVisitor.Status.ABORT)) {
              return;
            }
          }
          break;
        default:
          visitNode(tree);

          if (isStatus(CfgVisitor.Status.ABORT)) {
            return;
          }

          break;
      }
    }
  }

  /**
   * Visits the specified node.
   *
   * @param node node to be visited
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void visitNode(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    switch (node.getType()) {
      case MODULE:
        visitModule((Module) node);
        break;
      case PROCESS:
        visitProcess((Process) node);
        break;
      case BASIC_BLOCK:
        visitBasicBlock((BasicBlock) node);
        break;
      case CFG:
        visitCfg((Cfg) node);
        break;
      case CASE:
        visitCase((Case) node);
        break;
      case MERGE:
        visitMerge((Merge) node);
        break;
      case SWITCH:
        visitSwitch((Switch) node);
        break;
      case SINK:
        visitSink((Sink) node);
        break;
      case SOURCE:
        visitSource((Source) node);
        break;
      case WAIT:
        visitWait((Wait) node);
        break;
      default:
        throw new IllegalArgumentException(
          String.format("Unknown node type: %s.", node.getType()));
    }
  }

  private void visitCfgModel(CfgModel model) {

    this.visitor.onCfgModelBegin();

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      Collection<Module> modules = model.getModules();
      for (Module module : modules) {
        visitModule(module);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }
    this.visitor.onCfgModelEnd();
  }

  private void visitModule(Module node) {

    this.visitor.onModuleBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      List<Process> processes = node.getProcesses();
      for (Process process : processes) {
        visitProcess(process);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }

      List<Module> modules = node.getModules();
      for (Module module : modules) {
        visitModule(module);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }

    this.visitor.onModuleEnd(node);
  }

  private void visitProcess(Process node) {

    this.visitor.onProcessBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visitCfg(node.getCfg());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onProcessEnd(node);
  }

  private void visitCfg(Cfg node) {

    this.visitor.onCfgBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onCfgEnd(node);
  }

  private void visitSource(Source node) {

    this.visitor.onSrcBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onSrcEnd(node);
  }

  private void visitSink(Sink node) { this.visitor.onSink(node); }

  private void visitSwitch(Switch node) {

    this.visitor.onSwitchBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onSwitchEnd(node);
  }

  private void visitMerge(Merge node) {

    this.visitor.onMergeBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onMergeEnd(node);
  }

  private void visitCase(Case node) {

    this.visitor.onCaseBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onCaseEnd(node);
  }

  private void visitBasicBlock(BasicBlock node) {

    this.visitor.onBasicBlockBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onBasicBlockEnd(node);
  }

  private void visitWait(Wait node) {

    this.visitor.onWaitBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onWaitEnd(node);
  }
}
