/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Merge node of control flow graph model.
 *
 * It is a reversed version of switch node.
 * This node has only one child node and more than one parent node.
 * Every time no more than one parent node is active.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Merge extends CfgNode {

  /* The parent nodes. */
  private final Collection<CfgNode> parents;

  /**
   * The default constructor.
   */
  public Merge() { this.parents = new LinkedHashSet<>(); }

  @Override
  public CfgModelNode deepcopy() { return new  Merge(); }

  @Override
  public CfgNodeType getType() { return CfgNodeType.MERGE; }

  @Override
  public boolean hasParents() { return !this.parents.isEmpty(); }

  @Override
  public CfgModelNode getOnlyParent() { return getParents().iterator().next(); }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   */
  @Override
  public void addParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      this.parents.add((CfgNode) node);
      if (!node.hasChildren() || !node.getChildren().contains(this)) {
        node.addChild(this);
      }
    } else {
      throw new IllegalArgumentException(
          String.format("Incorrect node type: %s.", node.getType()));
    }
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   *
   * @throws java.lang.IllegalStateException when the object does not contain
   * the specified parent node.
   *
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   */
  @Override
  public void removeParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if (this.hasParents() && this.getParents().contains(node)) {
        this.parents.remove(node);
        if (node.hasChildren() && node.getChildren().contains(this)) {
          node.removeChild(this);
        }
      } else {
        throw new IllegalStateException(String.format("No such parent: %s.", node.getId()));
      }
    } else {
      throw new IllegalArgumentException(
          String.format("Incorrect node type: %s.", node.getType()));
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    Set<CfgModelNode> parents = new LinkedHashSet<>();
    parents.addAll(this.parents);

    return parents;
  }

  @Override
  public CfgModelNode getOnlyChild() { return this.childNode; }

  @Override
  public Collection<CfgModelNode> getChildren() {
    return Collections.singleton((CfgModelNode) this.childNode);
  }

  @Override
  public Set<NodeVariable> getUses() { return new LinkedHashSet<>(); }

  @Override
  public Set<NodeVariable> getDefines() { return new LinkedHashSet<>(); }
}
