/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * HDL module.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Module extends CfgVariableContainerNode {

  /* Parent node. */
  private CfgModelNode parent;

  /* The module name. */
  private final String name;

  /* The module's instance name. */
  private String instanceName;

  /* The child processes. */
  private final List<Process> processes;

  /* The child sub-modules. */
  private final List<Module> modules;

  /**
   * The default constructor.
   *
   * Constructs module object with the specified name.
   *
   * @param name module name
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Module(String name) {
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.instanceName = null;
    this.processes = new ArrayList<>();
    this.modules = new ArrayList<>();
  }

  /**
   * Sets instance name.
   *
   * @param instanceName name of current instance
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setInstanceName(String instanceName) {
    InvariantChecks.checkNotNull(instanceName);

    this.instanceName = instanceName;
  }

  /**
   * Returns module's processes.
   *
   * @return collection of module's processes
   */
  public List<Process> getProcesses() { return this.processes; }

  /**
   * Returns sub-modules.
   *
   * @return collection of sub-modules
   */
  public List<Module> getModules() { return this.modules; }

  /**
   * Constructs an instance of this module
   *
   * @param instanceName instance name
   * @param bindings instance bindings
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public void instantiate(String instanceName, Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    setInstanceName(instanceName);
    setBindings(bindings);
    applyBindings(bindings);
  }

  /**
   * Applies bindings to module's internals.
   *
   * @param bindings "variable name - expression" binding
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void applyBindings(Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    for (Module module : this.modules) {
      module.applyBindings(bindings);
    }

    for (Process process : this.processes) {
      process.applyBindings(bindings);
    }

    // change declarations
    for (String varName : bindings.keySet()) {
      if (this.getVariablesMapping().containsKey(varName)) {

        /* Remove old declaration for the variable that is substituted by bindings. */
        this.removeDeclaration(varName);
      }
    }
  }

  /**
   * Returns module name.
   *
   * @return module name
   */
  public String getName() {
    String name = this.name;

    if (this.instanceName != null) {
      name += "." + this.instanceName;
    }
    return name;
  }

  /**
   * Returns all the module variables,
   * that are of input signal type,
   * or empty collection, if there are no such.
   *
   * @return input signals
   */
  public List<NodeVariable> getInputs() {
    return getVariables(VariableType.IN, VariableType.INOUT);
  }

  /**
   * Returns all the module variables,
   * that are of output signal type,
   * or empty collection, if there are no such.
   *
   * @return output signals
   */
  public List<NodeVariable> getOutputs() {
    return getVariables(VariableType.INOUT, VariableType.OUT);
  }

  private List<NodeVariable> getVariables(VariableType ... types) {
    List<NodeVariable> variables = new ArrayList<>();

    for (VariableDeclaration declaration : this.getDeclarations()) {
      Node var = declaration.getVarExpression();
      if (var instanceof NodeVariable) {

        // extract and analyse variable data
        Object varData = var.getUserData();
        if (varData instanceof VariableData) {
          VariableType varType = ((VariableData) varData).getVariableType();
          for (VariableType type : types) {
            if (varType.equals(type)) {
              variables.add((NodeVariable) var);
              break;
            }
          }
        }
      } else
        throw new IllegalStateException("Incorrect kind: " + var.getKind());
    }
    return variables;
  }

  /**
   * Remove module's processes.
   */
  public void removeProcesses() {
    for (int i = 0; i < this.processes.size(); i++) {
      this.processes.get(i).removeParent(this);
    }
    this.processes.clear();
  }

  @Override
  public CfgModelNode deepcopy() {
    final Module copy = new Module(this.name);
    copy.setInstanceName(this.instanceName);
    return copy;
  }


  /**
   * Adds child node.
   *
   * @param node child node
   *
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.Process} type
   * and not of {@link ru.ispras.retrascope.model.cfg.Module} type.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  @Override
  public void addChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof Process) {
      if (!this.processes.contains(node)) {
        this.processes.add((Process) node);
      }
    } else if (node instanceof Module) {
      if (!this.modules.contains(node)) {
        this.modules.add((Module) node);
      }
    }
    else
      throw new IllegalArgumentException(
        String.format("Illegal child kind: %s.", node.getType()));

    if (!node.hasParents() || !node.getParents().contains(this)) {
      node.addParent(this);
    }
  }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.IllegalArgumentException when the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.Module} type.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  @Override
  public void addParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    switch (node.getType()) {
      case MODULE:
        this.parent = node;

        if (!node.hasChildren() || !node.getChildren().contains(this)) {
          node.addChild(this);
        }
        break;
      default:
        throw new IllegalArgumentException(
          String.format("Illegal parent kind: %s.", node.getType()));
    }
  }

  /**
   * Removes child node.
   *
   * @param node child node
   *
   * @throws java.lang.IllegalArgumentException when: 1) the specified node
   * is not of {@link ru.ispras.retrascope.model.cfg.Process} type
   * and not of {@link ru.ispras.retrascope.model.cfg.Module} type;
   * 2) this object does not contain the specified node as child.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  @Override
  public void removeChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.hasChildren() && this.getChildren().contains(node)) {
      if (node instanceof Process) {
        this.processes.remove(node);
      } else if (node instanceof Module) {
        this.modules.remove(node);
      } else throw new IllegalArgumentException(
            String.format("Illegal node kind: %s.", node.getType()));
    } else throw new IllegalArgumentException(
        String.format("No such child node: %s.", node.getId()));
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.IllegalArgumentException when this object
   * does not contain the specified node as parent node.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  @Override
  public void removeParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent == node) {
      this.parent = null;
      node.removeChild(this);
    } else {
      throw new IllegalArgumentException(
          String.format("No such parent node: %s.", node.getId()));
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() { return Collections.singleton(this.parent); }

  @Override
  public CfgModelNode getOnlyChild() { return getChildren().iterator().next(); }

  @Override
  public Collection<CfgModelNode> getChildren() {
    Set<CfgModelNode> children = new LinkedHashSet<>();
    children.addAll(this.getProcesses());
    children.addAll(this.getModules());
    return children;
  }

  @Override
  public String getDescription(ExprTreePrinter printer) {
    return this.getType().name().toLowerCase() + " " + this.getName();
  }

  @Override
  public CfgNodeType getType() { return CfgNodeType.MODULE; }

  @Override
  public boolean hasParents() { return this.parent != null; }

  @Override
  public boolean hasChildren() {
    return !this.getProcesses().isEmpty()
      || !this.getModules().isEmpty();
  }

  @Override
  public CfgModelNode getOnlyParent() { return this.parent; }

  @Override
  public Set<NodeVariable> getUses() {

    Set<NodeVariable> uses = new LinkedHashSet<>();

    List<Process> processes = this.getProcesses();
    for (Process process : processes) {
      Iterable<NodeVariable> processUses = process.getUses();

      for (NodeVariable processUsed : processUses) {
        if (!uses.contains(processUsed)) {
          uses.add(processUsed);
        }
      }
    }

    List<Module> modules = this.getModules();
    for (Module module : modules) {
      Iterable<NodeVariable> moduleUses = module.getUses();

      for (NodeVariable mUsed : moduleUses) {
        if (!uses.contains(mUsed)) {
          uses.add(mUsed);
        }
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    Set<NodeVariable> defines = new LinkedHashSet<>();

    List<Process> processes = this.getProcesses();
    for (Process process : processes) {
      Iterable<NodeVariable> processDefines = process.getDefines();

      for (NodeVariable processDefined : processDefines) {
        if (!defines.contains(processDefined)) {
          defines.add(processDefined);
        }
      }
    }

    List<Module> modules = this.getModules();
    for (Module module : modules) {
      Iterable<NodeVariable> mDefines = module.getDefines();

      for (NodeVariable mDefined : mDefines) {
        if (!defines.contains(mDefined)) {
          defines.add(mDefined);
        }
      }
    }
    return defines;
  }

//  @Override
//  public boolean equals(Object o) {
//    if (this == o)
//      return true;
//    if (o == null || getClass() != o.getClass())
//      return false;
//
//    Module module = (Module) o;
//
//    return !(instanceName != null ?
//            !instanceName.equals(module.instanceName) :
//            module.instanceName != null) &&
//        modules.equals(module.modules) &&
//        name.equals(module.name) &&
//        processes.equals(module.processes);
//
//  }
//
//  @Override
//  public int hashCode() {
//    int result = name.hashCode();
//    result = 31 * result + (instanceName != null ? instanceName.hashCode() : 0);
//    result = 31 * result + processes.hashCode();
//    result = 31 * result + modules.hashCode();
//    return result;
//  }
}
