/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Set;

/**
 * HDL module process.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Process extends CfgVariableContainerNode implements UseDef {

  /* The parent module. */
  private Module parent;

  /* The embedded control flow graph. */
  private Cfg cfg;

  /* The sensitivity list. */
  private final EventList eventList;

  /* The initialization flag. */
  private final boolean initial;

  /**
   * The default constructor.
   *
   * Constructs non-initial process.
   */
  public Process() {
    this.initial = false;
    this.eventList = new EventList();
  }

  /**
   * Constructs new process with the specified value of "initial" flag
   * and empty collection of events.
   *
   * @param initial initial flag
   */
  public Process(boolean initial) {
    this.initial = initial;
    this.eventList = new EventList();
  }

  /**
   * Checks rather process is of initial kind.
   *
   * @return {@code true} if process is of initial kind, {@code false} otherwise.
   */
  public boolean isInitial() { return this.initial; }

  /**
   * Returns process control flow graph.
   *
   * @return control flow graph of the process
   */
  public Cfg getCfg() { return this.cfg; }

  /**
   * Returns {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @return event list
   */
  public EventList getSensitivityList() { return this.eventList; }

  /**
   * Sets new {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @param eventList event list
   */
  public void setSensitivityList(EventList eventList) {

    if (this.initial && !eventList.isEmpty())
      throw new IllegalStateException(
          String.format(
              "Cannot set to initial process an event collection with size: %d.",
              eventList.eventNum()));

    this.eventList.clear();
    this.eventList.addEvents(eventList.getEvents());
  }

  /** Applies bindings to process internals.
   *
   * @param bindings "variable name - expression" binding
   */
  public void applyBindings(Map<String, Node> bindings) {
    this.cfg.applyToCfg(bindings);
    this.eventList.applyToEventList(bindings);
  }

  @Override
  public CfgModelNode deepcopy() {
    final Process copy = new Process(this.initial);
    copy.setSensitivityList(this.eventList);
    return copy;
  }

  @Override
  public void addChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.CFG) {
      throw new IllegalArgumentException(
          String.format("Incorrect node kind: %s.", node.getType()));
    }
    if (this.cfg == null) {
      this.cfg = (Cfg) node;
      if (!node.hasParents() || !node.getParents().contains(this)) {
        node.addParent(this);
      }
    } else {
      throw new IllegalStateException(
          String.format("Child node for %s re-assignment with: %s.",
              this.getType(), node.getType()));
    }
  }

  @Override
  public void addParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.MODULE) {
      throw new IllegalArgumentException(
          String.format("Incorrect node kind: %s.", node.getType()));
    }
    if (this.parent == null) {
      this.parent = (Module) node;
      if (!node.hasChildren() || node.getChildren().contains(this)) {
        node.addChild(this);
      }
    } else {
      throw new IllegalStateException(
          String.format("Parent node for %s re-assignment with: %s.",
              this.getType(), node.getType()));
    }
  }

  @Override
  public void removeChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.CFG) {
      throw new IllegalArgumentException(
          String.format("Incorrect node kind: %s.", node.getType()));
    }
    if (this.cfg.equals(node)) {

      this.cfg = null;

      if (node.hasParents() && node.getParents().contains(this)) {
        node.removeParent(this);
      }
    } else {
      throw new IllegalStateException(
          String.format("No such child node: %s.", node.getId()));
    }
  }

  @Override
  public void removeParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.MODULE) {
      throw new IllegalArgumentException(
          String.format("Incorrect node kind: %s.", node.getType()));
    }

    if (this.parent.equals(node)) {
      this.parent = null;

      if (node.hasChildren() && node.getChildren().contains(this)) {
        node.removeChild(this);
      }
    } else {
      throw new IllegalStateException(
        String.format("No such parent node: %s.", node.getId()));
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton((CfgModelNode) this.parent);
  }

  @Override
  public CfgModelNode getOnlyChild() { return this.cfg; }

  @Override
  public Collection<CfgModelNode> getChildren() {
    return Collections.singleton((CfgModelNode) this.cfg);
  }

  @Override
  public String getDescription(ExprTreePrinter printer) {
    return this.getType().name().toLowerCase() + ((!this.eventList.isEmpty()) ?
      " " + this.eventList.toString() : "");
  }

  @Override
  public CfgNodeType getType() { return CfgNodeType.PROCESS; }

  @Override
  public boolean hasParents() { return this.parent != null; }

  @Override
  public boolean hasChildren() { return this.cfg != null; }

  @Override
  public CfgModelNode getOnlyParent() { return this.parent; }

  @Override
  public Set<NodeVariable> getUses() { return cfg.getUses(); }

  @Override
  public Set<NodeVariable> getDefines() { return cfg.getDefines(); }

//  @Override
//  public boolean equals(Object o) {
//    if (this == o)
//      return true;
//    if (o == null || getClass() != o.getClass())
//      return false;
//
//    Process process = (Process) o;
//
//    return initial == process.initial &&
//        cfg.equals(process.cfg) &&
//        eventList.equals(process.eventList);
//  }
//
//  @Override
//  public int hashCode() {
//    int result = cfg.hashCode();
//    result = 31 * result + eventList.hashCode();
//    result = 31 * result + (initial ? 1 : 0);
//    return result;
//  }
}
