/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Source (first) node of control flow graph.
 *
 * Source node has only one parent node that is of higher level (not control flow graph node)
 * and only one child node.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Source extends CfgNode {

  /* The parent node. */
  private CfgModelNode parent;

  /**
   * Constructs empty object.
   */
  public Source() {}

  /**
   * Constructs object with the specified parent node.
   *
   * @param parent parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Source(CfgModelNode parent) {
    InvariantChecks.checkNotNull(parent);

    this.parent = parent;
  }

  @Override
  public CfgModelNode deepcopy() { return new Source(); }

  /**
   * Throws exception.
   *
   * @param node parent node
   *
   * @throws java.lang.UnsupportedOperationException every time it is called.
   */
  @Override
  public void addParent(CfgModelNode node) {
    throw new UnsupportedOperationException(
      String.format("No more parents for: %s.", this.getType()));
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.IllegalStateException when the internal parent node
   * is not equal to the specified node.
   */
  @Override
  public void removeParent(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent.equals(node)) {

      this.parent = null;

      if (node.hasChildren() && node.getChildren().contains(this)) {
        node.removeChild(this);
      }
    } else {
      throw new IllegalStateException(
        String.format("No such parent node: %s.", node.getId()));
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() { return Collections.singleton(this.parent); }

  @Override
  public CfgModelNode getOnlyChild() { return this.childNode; }

  @Override
  public CfgNodeType getType() { return CfgNodeType.SOURCE; }

  @Override
  public boolean hasParents() { return this.parent != null; }

  @Override
  public CfgModelNode getOnlyParent() { return this.parent; }

  @Override
  public Set<NodeVariable> getUses() { return new LinkedHashSet<>(); }

  @Override
  public Set<NodeVariable> getDefines() { return new LinkedHashSet<>(); }

//  @Override
//  public boolean equals(Object o) {
//    return this == o || !(o == null || getClass() != o.getClass());
//  }
//
//  @Override
//  public int hashCode() { return 0; }
}
