/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.RangedVariable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Switch node of control flow graph. When executed,
 * only one output node will be active.
 *
 * This node represents conditional block. It's children contain possible values of the
 * condition expression. Also this block may be of event type, when it's condition
 * is {@code true}, but it can be executed only when the specified event
 * {@link ru.ispras.retrascope.model.basis.Event} happens.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Switch extends CfgNode {

  /* The condition expression. */
  private Node condition;

  /* The signal event. */
  private final Event event;

  /* The child nodes. */
  private List<Case> children;

  /**
   * Constructs an object with the specified expression.
   *
   * @param condition switch expression
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Switch(Node condition) {
    InvariantChecks.checkNotNull(condition);

    this.condition = condition;
    this.event = null;
    this.children = new ArrayList<>();
  }

  /**
   * Constructs an object with the specified event
   * and trivial expression.
   *
   * @param event event
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Switch(Event event) {
    InvariantChecks.checkNotNull(event);

    this.condition = NodeValue.newBoolean(true);
    this.event = event;
    this.children = new ArrayList<>();
  }

  /**
   * Returns condition expression.
   *
   * @return condition
   */
  public Node getCondition() { return this.condition; }

  /**
   * Sets new expression.
   *
   * @param expr expression
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setExpression(Node expr) {
    InvariantChecks.checkNotNull(expr);

    this.condition = expr;
  }

  /**
   * Returns event.
   *
   * @return event
   */
  public Event getEvent() { return this.event; }

  @Override
  public CfgModelNode deepcopy() {
    final Switch copy = new Switch(this.event);
    copy.setExpression(this.condition);
    return copy;
  }

  /**
   * Returns a string description in the format of the specified printer.
   *
   * @param printer printer that provides an output string format
   *
   * @return object string description
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  @Override
  public String getDescription(ExprTreePrinter printer) {
    InvariantChecks.checkNotNull(printer);

    String description = "";
    if (this.event != null) {
      description += "[" + this.event.toString() + "] ";
    }
    description += printer.toString(this.condition);
    return description;
  }

  @Override
  public CfgNodeType getType() { return CfgNodeType.SWITCH; }

  @Override
  public boolean hasChildren() { return !this.children.isEmpty(); }

  @Override
  public CfgModelNode getOnlyParent() { return this.parentNode; }

  @Override
  public CfgModelNode getOnlyChild() { return getChildren().iterator().next(); }

  /**
   * Adds the specified child node.
   *
   * @param node child node
   *
   * @throws java.lang.IllegalArgumentException when argument
   * is not of {@link ru.ispras.retrascope.model.cfg.Case} type.
   */
  @Override
  public void addChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof Case) {

      this.children.add((Case) node);
      if (!node.hasParents() || !node.getParents().contains(this)) {
        node.addParent(this);
      }
    } else {
      throw new IllegalArgumentException(
        String.format("Incorrect type of child node: %s.", node.getId()));
    }
  }

  /**
   * Removes child node.
   *
   * @param node child node
   *
   * @throws java.lang.IllegalArgumentException when argument
   * is not of {@link ru.ispras.retrascope.model.cfg.Case} type.
   */
  @Override
  public void removeChild(CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof Case) {
      if (hasChildren() && getChildren().contains(node)) {
        this.children.remove(node);
        if (node.hasParents() && node.getParents().contains(this)) {
          node.removeParent(this);
        }
      }
    } else {
      throw new IllegalArgumentException(
          String.format("Incorrect type of child node: %s.", node.getId()));
    }
  }

  @Override
  public Collection<CfgModelNode> getChildren() {
    Collection<CfgModelNode> children = new LinkedHashSet<>();
    children.addAll(this.children);
    return children;
  }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> uses = new LinkedHashSet<>();

    /* Get variables that are used in expression. */
    Node expr = getCondition();
    Collection<NodeVariable> usedExprVariables = ExprUtils.getVariables(expr);
    uses.addAll(usedExprVariables);

    /* Get variables that are used in event. */
    if (getEvent() != null) {
      RangedVariable rVariable = getEvent().getRangedVariable();
      NodeVariable eventVariable = rVariable.getVariable();

      uses.add(eventVariable);

      if (rVariable.isRanged()) {
        Node youngNum = rVariable.getRange().getYoung();
        Node oldNum = rVariable.getRange().getOld();

        uses.addAll(ExprUtils.getVariables(youngNum));
        uses.addAll(ExprUtils.getVariables(oldNum));
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() { return new LinkedHashSet<>(); }

//  @Override
//  public boolean equals(Object o) {
//    if (this == o)
//      return true;
//    if (o == null || getClass() != o.getClass())
//      return false;
//
//    Switch aSwitch = (Switch) o;
//
//    return !(event != null ? !event.equals(aSwitch.event) : aSwitch.event != null) &&
//        !(condition != null ? !condition.equals(aSwitch.condition) : aSwitch.condition != null);
//  }
//
//  @Override
//  public int hashCode() {
//    int result = condition != null ? condition.hashCode() : 0;
//    result = 31 * result + (event != null ? event.hashCode() : 0);
//    return result;
//  }
}
