/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * Class for keeping the variable declaration data.
 * The data include own variable or the expression is bound to
 * and possibly an invariant - a condition, to that at every moment variable value must meet.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class VariableDeclaration {

  /* The variable or an expression that is bound to. */
  private Node varExpr;

  /* The variable invariant. */
  private Node invariant;

  /**
   * Constructs a variable declaration object from the specified expression
   * and {@code null} invariant.
   *
   * @param varExpr variable or it's bound expression
   *
   * @throws java.lang.NullPointerException when argument is {@code null}
   */
  public VariableDeclaration(Node varExpr) {
    InvariantChecks.checkNotNull(varExpr);

    this.varExpr = varExpr;
    this.invariant = null;
  }

  /**
   * Constructs a variable declaration object
   * from the specified variable/expression and it's invariant.
   *
   * @param varExpr variable or it's bound expression
   * @param invariant invariant
   *
   * @throws java.lang.NullPointerException when any of the arguments is {@code null}.
   */
  public VariableDeclaration(Node varExpr, Node invariant)
  {
    InvariantChecks.checkNotNull(varExpr);
    InvariantChecks.checkNotNull(invariant);

    this.varExpr = varExpr;
    this.invariant = invariant;
  }

  /**
   * Returns the declared variable or it's bound expression.
   *
   * @return variable or binding expression
   */
  public Node getVarExpression() { return this.varExpr; }

  /**
   * Checks rather variable also has it's own invariant.
   *
   * @return <code>true</code> if there is not-null invariant for this variable,
   * <code>false</code> otherwise
   */
  public boolean hasInvariant() { return this.invariant != null; }

  /**
   * Returns variable invariant.
   *
   * @return variable invariant.
   */
  public Node getInvariant() { return this.invariant; }
}
