/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.EventList;

/**
 * Wait node of control flow graph model.
 *
 * The object contains a {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list
 * and a {@link ru.ispras.fortress.expression.Node} "until" condition.
 * While entering in this node to model simulator should stop the execution until at least
 * one event from the sensitivity list. When the sensitivity list is empty or
 * at least one event appears, the "until" condition should be checked rather it is {@code true}.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Wait extends CfgNode {

  /* The events sensitivity list. */
  private final EventList list;

  /* The "until" condition. */
  private final Node untilCondition;

  /**
   * Constructs node object with the specified {@link ru.ispras.retrascope.model.basis.EventList}
   * that is non-empty collection of {@link ru.ispras.retrascope.model.basis.Event} events
   * and the specified {@link ru.ispras.fortress.expression.Node} condition.
   *
   * @param list event list
   * @param node condition
   *
   * @throws java.lang.NullPointerException
   * when {@link ru.ispras.retrascope.model.basis.EventList} argument is {@code null}.
   * @throws java.lang.IllegalArgumentException
   * when {@link ru.ispras.retrascope.model.basis.EventList} is empty.
   * @throws java.lang.IllegalArgumentException when condition data type
   * is not equal to {@code ru.ispras.fortress.data.DataTypeId.BOOLEAN}.
   *
   */
  public Wait(EventList list, Node node) {
    InvariantChecks.checkNotNull(list);
    if (list.getEvents().isEmpty())
      throw new IllegalArgumentException(String.format("Empty %s.", list.getClass()));
    if (!node.isType(DataType.BOOLEAN))
      throw new IllegalArgumentException(
          String.format(
              "Incorrect condition data type: %s; only %s is allowed.",
              node.getDataType(),
              DataType.BOOLEAN));

    this.list = list;
    this.untilCondition = node;
  }

  /**
   * Constructs object with the specified {@link ru.ispras.retrascope.model.basis.EventList}
   * that is non-empty collection of {@link ru.ispras.retrascope.model.basis.Event} events.
   *
   * @param list event list
   *
   *@throws java.lang.NullPointerException
   * when {@link ru.ispras.retrascope.model.basis.EventList} argument is {@code null}.
   * @throws java.lang.IllegalArgumentException
   * when {@link ru.ispras.retrascope.model.basis.EventList} is empty.
   */
  public Wait(EventList list) {
    InvariantChecks.checkNotNull(list);
    if (list.getEvents().isEmpty())
      throw new IllegalArgumentException(String.format("Empty %s.", list.getClass()));

    this.list = list;
    this.untilCondition = null;
  }

  /**
   * Constructs node object with the specified {@link ru.ispras.fortress.expression.Node} condition.
   *
   * @param node condition
   *
   * @throws java.lang.IllegalArgumentException when condition data type
   * is not equal to {@code ru.ispras.fortress.data.DataTypeId.BOOLEAN}.
   */
  public Wait(Node node) {
    if (!node.isType(DataType.BOOLEAN))
      throw new IllegalArgumentException(
        String.format(
          "Incorrect condition data type: %s; only %s is allowed.",
          node.getDataType(),
          DataType.BOOLEAN));

    this.list = new EventList();
    this.untilCondition = node;
  }

  /**
   * Returns the {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @return event list
   */
  public EventList getEventList() { return this.list; }

  /**
   * Checks rather object contains condition,
   *
   * @return {@code true} when this object contains not-{@code null} condition,
   * {@code false} otherwise.
   */
  public boolean isConditioned() { return this.untilCondition != null; }

  /**
   * Returns the {@link ru.ispras.fortress.expression.Node} condition.
   *
   * @return condition
   */
  public Node getCondition() { return this.untilCondition; }

  @Override
  public CfgModelNode deepcopy() {
    return new Wait(this.list.deepcopy(), this.untilCondition.deepCopy());
  }

  @Override
  public CfgNodeType getType() { return CfgNodeType.WAIT; }

  @Override
  public CfgModelNode getOnlyParent() { return this.parentNode; }

  @Override
  public CfgModelNode getOnlyChild() { return this.childNode; }

  @Override
  public String getDescription(ExprTreePrinter printer) {
    return ((!this.list.isEmpty()) ? " " + this.list.toString() : "") +
        (this.untilCondition != null ? " until " + printer.toString(this.untilCondition) : "");
  }
}
