/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.UseDef;
import ru.ispras.retrascope.model.cfg.BasicBlock;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The "sequence of code statements" action.
 * 
 * These statements must be executed with the same order, as they are presented in this class.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Action implements UseDef {

  /* String representation constants. */

  private static final String LEFT_STRING_BOUND = "{";
  private static final String ASSIGNMENT_STRING_HEADER = "assignment: ";
  private static final String ASSIGNMENT_STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  /* Sequence of assignments. */
  private final List<Assignment> assignments;

  /**
   * Constructs an assignment action containing only
   * the specified {@link ru.ispras.retrascope.model.basis.Assignment} assignment.
   *
   * @param statement - the assign statement.
   * 
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Action(Assignment statement) {
    InvariantChecks.checkNotNull(statement);

    this.assignments = new ArrayList<>(1);
    this.assignments.add(statement);
  }

  /**
   * Constructs object from the specified list of assignments.
   *
   * @param list list of statements
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Action(List<Assignment> list) {
    InvariantChecks.checkNotNull(list);

    this.assignments = list;
  }

  /**
   * Returns all code statements.
   *
   * @return list of statements
   */
  public List<Assignment> getStatements() { return this.assignments; }

  /**
   * Returns action as one-element list of basic blocks of non-concurrent type.
   *
   * @return one-element list of basic blocks
   */
  public BasicBlock toBasicBlock() { return new BasicBlock(this.assignments, false); }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> uses = new LinkedHashSet<>();
    for (Assignment stmt : this.assignments) {
      Set<NodeVariable> stmtUses = stmt.getUses();
      if (!stmtUses.isEmpty()) {
        uses = CollectionUtils.uniteSets(uses, stmtUses);
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    Set<NodeVariable> defines = new LinkedHashSet<>();
    for (Assignment stmt : this.assignments) {
      Set<NodeVariable> stmtDefines = stmt.getDefines();
      if (!stmtDefines.isEmpty()) {
        defines = CollectionUtils.uniteSets(defines, stmtDefines);
      }
    }
    return defines;
  }

  /**
   * Returns mappings between modified variables and variables which affects corresponding
   * modifications.
   * <br>
   * For example, if this action consists of two subsequent assignments {@code x = y} and
   * {@code z = x + i}, the resulting map contains mappings from {@code x} to {@code y} and from
   * {@code z} to {@code y} and {@code i}.
   *
   * @return a map containing modified variables and variables which affect corresponding
   * modifications
   */
  public Map<NodeVariable, Set<NodeVariable>> getResultingDependencies() {
    Map<NodeVariable, Set<NodeVariable>> substitutions = new HashMap<>(this.assignments.size());

    for (Assignment assignment: this.getStatements()) {
      Set<NodeVariable> usedVariables =
          new HashSet<>(ExprUtils.getVariables(assignment.getExpression()));
      Set<NodeVariable> affectingVariables;

      if (usedVariables.isEmpty()) {
        affectingVariables = Collections.emptySet();
      } else {
        affectingVariables = new HashSet<>(usedVariables.size());

        for (NodeVariable variable: usedVariables) {
          if (substitutions.containsKey(variable)) {
            affectingVariables.addAll(substitutions.get(variable));
          } else {
            affectingVariables.add(variable);
          }
        }
      }

      substitutions.put(assignment.getRangedVariable().getVariable(), affectingVariables);
    }

    return substitutions;
  }

  /**
   * Returns mappings between variables and corresponding substitutions representing total variable
   * modifications during execution of this action. That is, execution of this action is equal to
   * parallel evaluation of the substitutions and subsequent assignment of derived values to the
   * corresponding variables.
   *
   * @return a map containing modified variables and the substitutions representing these
   * modifications
   */
  public Map<NodeVariable, Node> getResultingSubstitutions() {
    LinkedHashMap<NodeVariable, Node> substitutions = new LinkedHashMap<>();

    for (Assignment assignment: this.getStatements()) {
      Node expression = assignment.getExpression();

      for (NodeVariable variable: ExprUtils.getVariables(expression)) {
        if (substitutions.containsKey(variable)) {
          expression =
              Transformer.substitute(expression, variable.getName(), substitutions.get(variable));
        }
      }

      RangedVariable rangedVariable = assignment.getRangedVariable();
      NodeVariable nodeVariable = rangedVariable.getVariable();

      if (rangedVariable.isRanged()) {
        DataType dataType = nodeVariable.getDataType();

        if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          throw new IllegalArgumentException("Unsupported data type of ranged variable: " + dataType);
        }

        Range range = rangedVariable.getRange();

        if (!(range.getOld() instanceof NodeValue) || !(range.getYoung() instanceof NodeValue)) {
          throw new IllegalArgumentException("Variable range bounds are currently unsupported");
        }

        NodeValue leftBound = (NodeValue) range.getYoung();
        NodeValue rightBound = (NodeValue) range.getOld();

        int leftBoundInteger = leftBound.getInteger().intValue();
        int rightBoundInteger = rightBound.getInteger().intValue();

        int variableSize = dataType.getSize();

        if (leftBoundInteger < rightBoundInteger) {
          throw new IllegalArgumentException("This type of sub-vectors is currently unsupported");
        }

        NodeOperation mostSignificantPart = null;
        NodeOperation lessSignificantPart = null;
        
        if (leftBoundInteger < variableSize - 1) {
          NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), 10));
          NodeValue rightIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBoundInteger + 1), 10));

          mostSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
              nodeVariable);
        }

        if (rightBoundInteger > 0) {
          NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBoundInteger - 1), 10));
          NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", 10));

          lessSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
              nodeVariable);
        }

        if ((mostSignificantPart != null) && (lessSignificantPart != null)) {
          expression = new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression,
              lessSignificantPart);
        } else {
          if (mostSignificantPart != null) {
            expression =
                new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression);
          } else {
            if (lessSignificantPart != null) {
              expression =
                  new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignificantPart);
            }
          }
        }
      }

      if (substitutions.containsKey(nodeVariable)) {
        substitutions.remove(nodeVariable);
      }

      substitutions.put(nodeVariable, expression);
    }

    return substitutions;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.assignments.hashCode();

    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if (obj == null) {
      return false;
    }

    if (!(obj instanceof Action)) {
      return false;
    }

    Action actionToCompare = (Action) obj;

    return this.assignments.equals(actionToCompare.assignments);
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(Action.LEFT_STRING_BOUND);

    int initialStringBuilderLength = stringBuilder.length();

    for (Assignment assignment: this.assignments) {
      if (stringBuilder.length() > initialStringBuilderLength) {
        stringBuilder.append(Action.ASSIGNMENT_STRING_DELIMITER);
      }

      stringBuilder.append(Action.ASSIGNMENT_STRING_HEADER);
      stringBuilder.append(assignment);
    }

    stringBuilder.append(Action.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
