/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.Model;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;

/**
 * The representation of an extended finite state machine.
 * 
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class Efsm extends Model {
  private final String name;

  private final Set<EfsmState> states;
  private final Set<EfsmTransition> transitions;

  private EfsmState initialState;
  private GuardedAction resetGuardedAction;

  private final Map<String, DataType> inputs;
  private final Map<String, DataType> outputs;
  private final Map<String, DataType> innerVariables;

  /* The mapping of internal variable/signal names to their value invariants. */
  private final Map<String, Node> invariants;

  /**
   * Constructs an empty EFSM with the specified name.
   * 
   * @param name the EFSM name
   * 
   * @throws NullPointerException if {@code name} is {@code null}
   */
  public Efsm(String name) {
    InvariantChecks.checkNotNull(name);

    this.name = name;

    this.states = new LinkedHashSet<>();
    this.transitions = new LinkedHashSet<>();

    this.initialState = null;

    this.inputs = new LinkedHashMap<>();
    this.outputs = new LinkedHashMap<>();
    this.innerVariables = new LinkedHashMap<>();
    this.invariants = new LinkedHashMap<>();
  }

  /**
   * Creates a new state based on the specified logical expression and adds it into this EFSM.
   *
   * @param expression the expression representing the state
   * 
   * @return the created state
   */
  public EfsmState createState(NodeOperation expression) {
    EfsmState state = new EfsmState(expression);
    this.states.add(state);
    this.processAddedExpression(expression);

    return state;
  }

  /**
   * Adds the specified state to this EFSM.
   *
   * @param state the state to be added
   *
   * @return {@code true} if state is successfully added, otherwise returns {@code false}
   */
  public boolean addState(EfsmState state) {
    InvariantChecks.checkNotNull(state);

    this.processAddedExpression(state.getExpression());

    return this.states.add(state);
  }

  /**
   * Creates a new transition with the specified guarded action between the specified states and
   * adds it into this EFSM.
   * 
   * @param sourceState a source state
   * @param destinationState a destination state
   * @param guardedAction the guarded action
   * 
   * @return the created transition
   * 
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM
   */
  public EfsmTransition createTransition(EfsmState sourceState, EfsmState destinationState,
      GuardedAction guardedAction) {
    if (!this.states.contains(sourceState)) {
      throw new IllegalArgumentException(
          "The specified source state of a new transition is not found in this EFSM: " +
              sourceState);
    }

    if (!this.states.contains(destinationState)) {
      throw new IllegalArgumentException(
          "The specified destination state of a new transition is not found in this EFSM: " +
              destinationState);
    }

    EfsmTransition transition = new EfsmTransition(sourceState, destinationState, guardedAction);
    this.transitions.add(transition);
    this.processAddedExpression(guardedAction.getGuard().toNode());

    for (Assignment assignment: guardedAction.getAction().getStatements()) {
      this.processAddedExpression(assignment.getExpression());
      this.processAddedExpression(assignment.getRangedVariable().getVariable());
    }

    return transition;
  }

  /**
   * Returns an invariant condition which is related to the specified variable. Typically, such
   * conditions represent constraints from the variable declarations.
   *
   * @param variableName a name of the variable
   *
   * @return the invariant condition. {@code null} is returned if there is no condition which is
   * related to the specified variable
   *
   * @see Efsm#getInvariantCondition(NodeVariable)
   */
  public Node getInvariantCondition(String variableName) {
    return this.invariants.get(variableName);
  }

  /**
   * Returns an invariant condition which is related to the specified variable. Typically, such
   * conditions represent constraints from the variable declarations.
   *
   * @param variable the variable
   *
   * @return the invariant condition. {@code null} is returned if there is no condition which is
   * related to the specified variable
   *
   * @see Efsm#getInvariantCondition(String)
   */
  public Node getInvariantCondition(NodeVariable variable) {
    return this.invariants.get(variable.getName());
  }

  /**
   * Returns a name of this EFSM.
   * 
   * @return the EFSM name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Checks if the specified state is in this EFSM.
   * 
   * @param state the state whose presence in the EFSM to be tested
   * 
   * @return {@code true} if this EFSM contains the specified state. Otherwise returns {@code false}
   */
  public boolean containsState(EfsmState state) {
    return this.states.contains(state);
  }

  @Override
  public void addInput(String name, DataType type) {
    this.inputs.put(name, type);
  }

  @Override
  public void addInnerVariable(String name, DataType type) {
    this.innerVariables.put(name, type);
  }

  @Override
  public void addOutput(String name, DataType type) {
    this.outputs.put(name, type);
  }

  /**
   * Adds the specified invariant condition on the value of the specified variable.
   *
   * @param name a name of the variable
   * @param invariant the condition
   *
   * @throws NullPointerException if any of the arguments is {@code null}
   */
  public void addInvariant(String name, Node invariant) {
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(invariant);

    this.invariants.put(name, invariant);
  }

  /**
   * Marks the specified state of this EFSM as an initial state.
   * 
   * @param state the new initial state of this EFSM
   * 
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public void setInitialState(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException("The specified state is not found in this EFSM: " + state);
    }

    this.initialState = state;
  }

  /**
   * Returns an initial state of this EFSM.
   * 
   * @return the initial state of this EFSM
   */
  @Override
  public EfsmState getInitialState() {
    return this.initialState;
  }

  /**
   * Adds the specified guarded action to this EFSM. This guarded action represents a precondition
   * of the initial state and an action which is performed during initialisation of this EFSM.
   *
   * @param resetGuardedAction the reset guarded action
   *
   * @throws NullPointerException if the {@code resetGuardedAction} is {@code null}
   */
  public void setResetGuardedAction(GuardedAction resetGuardedAction) {
    InvariantChecks.checkNotNull(resetGuardedAction);

    this.resetGuardedAction = resetGuardedAction;
  }

  /**
   * Returns a reset guarded action of this EFSM, that is, such a guarded action which represents a
   * precondition of the initial state and an action which is performed during initialisation of
   * this EFSM.
   *
   * @return the reset guarded action
   */
  public GuardedAction getResetGuardedAction() {
    return this.resetGuardedAction;
  }

  /**
   * Returns all states of this EFSM.
   * 
   * @return a set containing the states of this EFSM
   */
  public Set<EfsmState> getStates() {
    return new LinkedHashSet<>(this.states);
  }

  /**
   * Returns all transitions of this EFSM.
   * 
   * @return a set containing the transitions of this EFSM
   */
  public Set<EfsmTransition> getTransitions() {
    return new LinkedHashSet<>(this.transitions);
  }

  @Override
  public DataType getInputDataType(String name) {
    if (this.inputs.containsKey(name)) {
      return this.inputs.get(name);
    } else {
      throw new IllegalArgumentException("The variable " + name
          + " is not found among the EFSM inputs");
    }
  }

  @Override
  public Map<String, DataType> getInputDataTypes(Collection<String> names) {
    LinkedHashMap<String, DataType> result = new LinkedHashMap<>(names.size());

    for (String name : names) {
      if (this.inputs.containsKey(name)) {
        result.put(name, this.inputs.get(name));
      } else {
        throw new IllegalArgumentException("The variable " + name
            + " is not found among the EFSM inputs");
      }
    }

    return result;
  }

  @Override
  public Map<String, DataType> getInputDataTypes() {
    return new LinkedHashMap<>(this.inputs);
  }

  @Override
  public DataType getOutputDataType(String name) {
    if (this.outputs.containsKey(name)) {
      return this.outputs.get(name);
    } else {
      throw new IllegalArgumentException("The variable " + name
          + " is not found among the EFSM outputs");
    }
  }

  @Override
  public Map<String, DataType> getOutputDataTypes(Collection<String> names) {
    LinkedHashMap<String, DataType> result = new LinkedHashMap<>(names.size());

    for (String name : names) {
      if (this.outputs.containsKey(name)) {
        result.put(name, this.outputs.get(name));
      } else {
        throw new IllegalArgumentException("The variable " + name
            + " is not found among the EFSM outputs");
      }
    }

    return result;
  }

  @Override
  public Map<String, DataType> getOutputDataTypes() {
    return new LinkedHashMap<>(this.outputs);
  }

  @Override
  public DataType getInnerDataType(String name) {
    if (this.innerVariables.containsKey(name)) {
      return this.innerVariables.get(name);
    } else {
      throw new IllegalArgumentException("The variable " + name
          + " is not found among the EFSM inner variables");
    }
  }

  @Override
  public Map<String, DataType> getInnerDataTypes(Collection<String> names)
      throws IllegalArgumentException {
    LinkedHashMap<String, DataType> result = new LinkedHashMap<>(names.size());

    for (String name : names) {
      if (this.innerVariables.containsKey(name)) {
        result.put(name, this.innerVariables.get(name));
      } else {
        throw new IllegalArgumentException("The variable " + name
            + " is not found among the EFSM inner variables");
      }
    }

    return result;
  }

  @Override
  public Map<String, DataType> getInnerDataTypes() {
    return new LinkedHashMap<>(this.innerVariables);
  }

  /**
   * Checks if this EFSM is synchronous, that is, contains at least one event-sensible guarded
   * action.
   *
   * @return {@code true} if this EFSM is synchronous. Otherwise returns {@code false}
   */
  public boolean isSynchronous() {
    for (EfsmTransition transition: this.transitions) {
      if (!transition.getGuardedAction().getClocks().isEmpty()) {
        return true;
      }
    }

    return false;
  }

  /**
   * Checks if there is a path from the specified source state to the specified destination state.
   *
   * @param sourceState the source state
   * @param destinationState the destination state
   *
   * @return {@code true} if there is a path from the source state to the destination state.
   * Otherwise returns {@code false}
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   * in this EFSM
   */
  public boolean isReachable(EfsmState sourceState, EfsmState destinationState) {
    if (!this.states.contains(sourceState)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!this.states.contains(destinationState)) {
      throw new IllegalArgumentException(
          "The specified destination state is not found in this EFSM");
    }

    if (sourceState.equals(destinationState)) {
      return true;
    }

    LinkedHashSet<EfsmState> visitedStates = new LinkedHashSet<>(this.states.size());
    LinkedList<EfsmState> queue = new LinkedList<>();
    visitedStates.add(sourceState);
    queue.add(sourceState);

    while (!queue.isEmpty()) {
      for (EfsmState nextState : this.getDestinationStates(queue.poll())) {
        if (nextState.equals(destinationState)) {
          return true;
        }

        if (!visitedStates.contains(nextState)) {
          queue.add(nextState);
          visitedStates.add(nextState);
        }
      }
    }

    return false;
  }

  /**
   * Looks for the shortest path from the specified source state to the specified destination state
   * with the specified principle of transition weight defining. This method is based on the
   * Dijkstra's algorithm. <br>
   * Currently CONST and GUARD_REGISTER_AMOUNT TransitionWeightBasis types are supported.
   *
   * @param sourceState the source state
   * @param destinationState the destination state
   * @param weightBasis the principle used for transition weight computing
   *
   * @return a list containing the transitions of found path. It's empty if the source state matches
   * to the destination state. If there is no path from the source state to the destination
   * state {@code null} is returned
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   * in this EFSM or if unsupported {@link TransitionWeightBasis} is passed
   */
  public List<EfsmTransition> getShortestPath(EfsmState sourceState, EfsmState destinationState,
      TransitionWeightBasis weightBasis) {
    if (!this.states.contains(sourceState)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!this.states.contains(destinationState)) {
      throw new IllegalArgumentException(
          "The specified destination state is not found in this EFSM");
    }

    if (sourceState.equals(destinationState)) {
      return Collections.emptyList();
    }

    if (!this.isReachable(sourceState, destinationState)) {
      return null;
    }

    final LinkedHashMap<EfsmState, Integer> distances = new LinkedHashMap<>(this.states.size());
    LinkedHashMap<EfsmState, EfsmTransition> previousNodes =
        new LinkedHashMap<>(this.states.size());
    LinkedHashSet<EfsmState> visitedStates = new LinkedHashSet<>(this.states.size());

    for (EfsmState state : this.states) {
      distances.put(state, Integer.MAX_VALUE);
      previousNodes.put(state, null);
    }

    distances.put(sourceState, 0);
    final LinkedHashMap<EfsmTransition, Integer> transitionWeights = new LinkedHashMap<>();

    for (EfsmTransition transition : this.transitions) {
      switch (weightBasis) {
        case CONSTANT:
          transitionWeights.put(transition, 1);

          break;
        case GUARD_REGISTER_AMOUNT:
          int weight = 0;

          for (NodeVariable variable : transition.getGuardedAction().getGuard().getUses()) {
            if (this.innerVariables.containsKey(variable.getName())) {
              weight++;
            }
          }

          transitionWeights.put(transition, weight);

          break;
        default:
          throw new IllegalArgumentException("Unsupported weight defining principle: "
              + weightBasis);
      }
    }

    Comparator<EfsmTransition> transitionComparator = new Comparator<EfsmTransition>() {
      @Override
      public int compare(EfsmTransition transition1, EfsmTransition transition2) {
        return transitionWeights.get(transition1) - transitionWeights.get(transition2);
      }
    };

    Comparator<EfsmState> stateComparator = new Comparator<EfsmState>() {
      @Override
      public int compare(EfsmState state1, EfsmState state2) {
        return distances.get(state1) - distances.get(state2);
      }
    };

    visitedStates.add(sourceState);
    LinkedList<EfsmState> queue = new LinkedList<>();
    queue.add(sourceState);

    while (!queue.isEmpty()) {
      EfsmState currentState = queue.poll();
      int currentDistance = distances.get(currentState);
      LinkedList<EfsmTransition> nextTransitions =
          new LinkedList<>(this.getOutgoingTransitions(currentState));
      Collections.sort(nextTransitions, transitionComparator);

      for (EfsmTransition nextTransition : nextTransitions) {
        EfsmState nextState = nextTransition.getDestinationState();
        int newDistance = currentDistance + transitionWeights.get(nextTransition);

        if (newDistance < distances.get(nextState)) {
          previousNodes.put(nextState, nextTransition);
          distances.put(nextState, newDistance);

          if (nextState.equals(destinationState)) {
            break;
          }

          if (!visitedStates.contains(nextState)) {
            visitedStates.add(nextState);
            queue.add(nextState);
          }
        }
      }

      Collections.sort(queue, stateComparator);
    }

    LinkedList<EfsmTransition> result = new LinkedList<>();
    EfsmState currentState = destinationState;

    while (previousNodes.get(currentState) != null) {
      result.addFirst(previousNodes.get(currentState));
      currentState = previousNodes.get(currentState).getSourceState();
    }

    return result;
  }

  /**
   * Returns transitions which ends at the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing incoming transitions for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmTransition> getIncomingTransitions(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException("The specified state is not found in this EFSM: " + state);
    }

    Set<EfsmTransition> incomingTransitions = new LinkedHashSet<>();

    for (EfsmTransition transition : this.transitions) {
      if (transition.getDestinationState().equals(state))
        incomingTransitions.add(transition);
    }

    return incomingTransitions;
  }

  /**
   * Returns transitions which starts at the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing outgoing transitions for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmTransition> getOutgoingTransitions(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException("The specified state is not found in this EFSM: " + state);
    }

    Set<EfsmTransition> outgoingTransitions = new LinkedHashSet<>();

    for (EfsmTransition transition : this.transitions) {
      if (transition.getSourceState().equals(state)) {
        outgoingTransitions.add(transition);
      }
    }

    return outgoingTransitions;
  }

  /**
   * Returns states which are start points of incoming transitions of the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing the source states for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmState> getSourceStates(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException("The specified state is not found in this EFSM: " + state);
    }

    Set<EfsmTransition> incomingTransitions = this.getIncomingTransitions(state);

    if (incomingTransitions.isEmpty()) {
      return Collections.emptySet();
    }

    Set<EfsmState> sourceStates = new LinkedHashSet<>(incomingTransitions.size());

    for (EfsmTransition transition : incomingTransitions) {
      sourceStates.add(transition.getSourceState());
    }

    return sourceStates;
  }

  /**
   * Returns states which are end points of outgoing transitions of the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing the destination states for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmState> getDestinationStates(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException("The specified state is not found in this EFSM: " + state);
    }

    Set<EfsmTransition> outgoingTransitions = this.getOutgoingTransitions(state);

    if (outgoingTransitions.isEmpty()) {
      return Collections.emptySet();
    }

    Set<EfsmState> destinationStates = new LinkedHashSet<>(outgoingTransitions.size());

    for (EfsmTransition transition : outgoingTransitions) {
      destinationStates.add(transition.getDestinationState());
    }

    return destinationStates;
  }

  @Override
  public boolean containsInput(String name) {
    return this.inputs.containsKey(name);
  }

  @Override
  public boolean containsInnerVariable(String name) {
    return this.innerVariables.containsKey(name);
  }

  @Override
  public boolean containsOutput(String name) {
    return this.outputs.containsKey(name);
  }

  /**
   * Checks if this EFSM contains an invariant condition on the specified variable.
   *
   * @param name a name of the variable
   *
   * @return {@code true} if this EFSM contains an invariant on the variable with the specified
   * name, {@code false} otherwise
   *
   * @throws NullPointerException if {@code name} is {@code null}
   */
  public boolean containsInvariant(String name) {
    InvariantChecks.checkNotNull(name);

    return this.invariants.containsKey(name);
  }

  /**
   * Constructs a synchronous composition from a list of EFSMs.
   *
   * @param efsmList the list of EFSMs
   *
   * @return the synchronous composition of the EFSMs
   *
   * @throws IllegalArgumentException if the specified list of EFSMs is {@code null} or contains
   *         less than two models
   */
  public static Efsm product(Efsm... efsmList) {

    if ((efsmList == null) || (efsmList.length < 2)) {
      throw new IllegalArgumentException("The number of EFSMs should be greater than 1!");
    }

    // This data structure represents a state with its outgoing transitions.

    class StateWithTransitions {
      public EfsmState state;
      public Collection<EfsmTransition> transitions;

      public StateWithTransitions(EfsmState state, Collection<EfsmTransition> transitions) {
        this.state = state;
        this.transitions = transitions;
      }
    }

    // Constructing a new EFSM name.

    String productName = efsmList[0].getName();
    for (int i = 1; i < efsmList.length; i++) {
      productName += (" || " + efsmList[i].getName());
    }
    Efsm product = new Efsm(productName);

    // Each EFSM is represented as a list of states with transitions
    List<List<StateWithTransitions>> oldStates = new ArrayList<>(efsmList.length);
    List<EfsmState> initialStates = new ArrayList<>();

    for (Efsm efsm : efsmList) {

      // Collecting initial states.

      initialStates.add(efsm.getInitialState());

      // Collecting input, output and inner signals.

      for (Map.Entry<String, DataType> inputSignal : efsm.getInputDataTypes().entrySet()) {
        product.addInput(inputSignal.getKey(), inputSignal.getValue());
      }
      for (Map.Entry<String, DataType> outputSignal : efsm.getOutputDataTypes().entrySet()) {
        product.addOutput(outputSignal.getKey(), outputSignal.getValue());
      }
      for (Map.Entry<String, DataType> innerSignal : efsm.getInnerDataTypes().entrySet()) {
        product.addInnerVariable(innerSignal.getKey(), innerSignal.getValue());
      }

      ArrayList<StateWithTransitions> swt = new ArrayList<>(efsm.getStates().size());

      for (EfsmState state : efsm.getStates()) {
        Collection<EfsmTransition> trans = efsm.getOutgoingTransitions(state);
        if (!trans.isEmpty()) {
          List<Node> guards = new ArrayList<>();

          // If the guards of the outgoing transitions are not complete (a situation is
          // possible when no transition is enabled), a negation of disjunction of guards
          // is constructed.

          for (EfsmTransition item : trans) {
            guards.add(item.getGuardedAction().getGuard().toNode());
          }
          if (!areComplete(trans)) {
            trans.add(new EfsmTransition(state, state, new GuardedAction(new Guard(ExprUtils
                .getComplement(guards.toArray(new Node[guards.size()]))), new Action(
                new Assignment()))));
          }
        } else {

          // If a state has no outgoing transitions, a new transition is constructed
          // (guard = true, action = null). This is needed by the construction algorithm.

          trans.add(new EfsmTransition(state, state, new GuardedAction(new Guard(NodeValue
              .newBoolean(true)), new Action(new Assignment()))));
        }
        swt.add(new StateWithTransitions(state, new ArrayList<>(trans)));

      }
      oldStates.add(swt);

    }

    // Constructing a new initial state.

    Node initialStateNode = null;
    for (EfsmState initState : initialStates) {
      if (initialStateNode == null) {
        initialStateNode = initState.getExpression();
      } else {
        initialStateNode = Node.AND(initialStateNode, initState.getExpression());
      }
    }

    EfsmState newInitialState = new EfsmState(initialStateNode);

    // Combining old EFSM states to construct new EFSM states.

    List<List<StateWithTransitions>> stateCombinations = combineLists(oldStates);

    LinkedHashMap<EfsmState, List<List<EfsmTransition>>> transitionsMapping = new LinkedHashMap<>();

    // Combining transitions of new state combinations.

    for (List<StateWithTransitions> combinedStateList : stateCombinations) {
      Node stateNode = null;
      List<List<EfsmTransition>> newStateTransitions = new ArrayList<>();
      List<List<EfsmTransition>> combinedTransitions;

      for (StateWithTransitions item : combinedStateList) {
        if (stateNode == null) {
          stateNode = item.state.getExpression();
        } else {
          stateNode = Node.AND(stateNode, item.state.getExpression());
        }
        newStateTransitions.add((ArrayList<EfsmTransition>) item.transitions);
      }

      combinedTransitions = combineLists(newStateTransitions);

      // Creating a new state.

      EfsmState newState = product.createState((NodeOperation) stateNode);
      transitionsMapping.put(newState, combinedTransitions);

      if (newState.equals(newInitialState)) {
        product.setInitialState(newState);
      }

    }

    // Constructing new transitions.

    for (Map.Entry<EfsmState, List<List<EfsmTransition>>> entry : transitionsMapping.entrySet()) {
      for (List<EfsmTransition> item : entry.getValue()) {
        EfsmState destState = getNewDestinationState(item);
        if (transitionsMapping.keySet().contains(destState)) {
          GuardedAction ga = constructGuardedAction(item);
          if (ga != null) {
            product.createTransition(entry.getKey(), destState, ga);
          }
        }
      }
    }

    return product;

  }

  /**
   * Checks whether the guards of the specified transitions are complete.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   *
   * @return {@code true} if the guards are complete, {@code false} otherwise
   */
  private static boolean areComplete(Collection<EfsmTransition> transitions) {

    Node constraint = null;
    List<Node> guards = new ArrayList<>();

    for (EfsmTransition item : transitions) {
      guards.addAll(item.getGuardedAction().getGuard().getGuardExpr());
      if (constraint == null) {
        constraint = item.getGuardedAction().getGuard().toNode();
      } else {
        constraint = Node.AND(constraint, item.getGuardedAction().getGuard().toNode());
      }
    }

    return ExprUtils.areComplete(guards.toArray(new Node[guards.size()]));
  }

  /**
   * Performs a Cartesian product of several lists (sets) of elements.
   *
   * @param lists {@link List} containing the lists to product
   *
   * @return the new {@link List} containing a Cartesian product of the input lists
   */
  private static <T> List<List<T>> combineLists(List<List<T>> lists) {
    InvariantChecks.checkNotNull(lists);

    int combinations = 1;
    for (List<T> item : lists) {
      combinations = combinations * item.size();
    }

    List<List<T>> product = new ArrayList<>();

    for (int i = 0; i < combinations; i++) {
      List<T> combination = new ArrayList<>();
      int j = 1;
      for (List<T> item : lists) {
        combination.add(item.get((i / j) % item.size()));
        j = j * item.size();
      }
      product.add(combination);
    }

    return product;

  }

  /**
   * Constructs a new EfsmState according to the destinations of transitions, contained in the
   * specified collection.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   *
   * @return a new {@link EfsmState}
   */
  private static EfsmState getNewDestinationState(Collection<EfsmTransition> transitions) {
    List<EfsmState> result = new ArrayList<>();
    Node constraint = null;

    for (EfsmTransition item : transitions) {
      result.add(item.getDestinationState());
    }

    for (EfsmState item : result) {
      if (constraint == null) {
        constraint = item.getExpression();
      } else {
        constraint = Node.AND(constraint, item.getExpression());
      }
    }

    return new EfsmState(constraint);

  }

  /**
   * Constructs a single {@link GuardedAction} from the guarded actions of the specified transitions.
   *
   * @param transitions a collection of the transitions
   *
   * @return the new guarded action
   */
  private static GuardedAction constructGuardedAction(Collection<EfsmTransition> transitions) {
    List<Node> guardPredicates = new ArrayList<>();
    List<Assignment> assignments = new ArrayList<>();
    List<Event> events = new ArrayList<>();

    for (EfsmTransition item : transitions) {
      guardPredicates.addAll(item.getGuardedAction().getGuard().getGuardExpr());
      events.addAll(item.getGuardedAction().getClocks());
      if (item.getGuardedAction().getAction() != null) {
        Action action = item.getGuardedAction().getAction();
        for (Assignment atomic : action.getStatements()) {
          if (atomic.getExpression() != null) {
            assignments.add(atomic);
          }
        }
      }
    }

    if (!ExprUtils.areCompatible(guardPredicates.toArray(new Node[guardPredicates.size()]))) {
      return null;
    }

    GuardedAction ga =
        new GuardedAction(new Guard(guardPredicates), new Action(assignments));
    ga.setClocks(events);

    return ga;
  }

  /**
   * Checks if this EFSM contains states or not.
   *
   * @return {@code true} if this EFSM contains states, {@code false} otherwise
   */
  public boolean containsStates() {
    return !this.states.isEmpty();
  }

  private void processAddedExpression(Node expression) {
    if (expression != null) {
      for (NodeVariable variable: ExprUtils.getVariables(expression)) {
        Object variableData = variable.getUserData();

        if (variableData instanceof VariableData) {
          VariableType variableType = ((VariableData) variableData).getVariableType();

          if (variableType == VariableType.REG) {
            this.addInnerVariable(variable.getName(), variable.getDataType());
          } else {
            if ((variableType == VariableType.IN) || (variableType == VariableType.INOUT)) {
              this.addInput(variable.getName(), variable.getDataType());
            }

            if ((variableType == VariableType.OUT) || (variableType == VariableType.INOUT)) {
              this.addOutput(variable.getName(), variable.getDataType());
            }
          }
        }
      }
    }
  }

  public static enum TransitionWeightBasis {
    /**
     * All transitions are considered as having one and the same weight
     */
    CONSTANT,
    /**
     * A transition's weight is equal to amount of registers used in the transition's guard
     */
    GUARD_REGISTER_AMOUNT
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.name.hashCode();
    result = prime * result + ((this.initialState == null) ? 0 : this.initialState.hashCode());
    result = prime * result + this.inputs.hashCode();
    result = prime * result + this.outputs.hashCode();
    result = prime * result + this.innerVariables.hashCode();
    result = prime * result + this.states.hashCode();
    result = prime * result + this.transitions.hashCode();

    return result;
  }

  /**
   * Compares the specified object with this EFSM for equality. The object is considered as equal to
   * this EFSM if and only if it is also an EFSM, it has the same name and initial state, it has the
   * same states, transitions and variables as this EFSM does, or more formally:
   * <ul>
   * <li>it is an instance of the {@code Efsm} class or its subclass;
   * <li>its name and initial state are equal to this EFSM's name and initial state correspondingly;
   * <li>it has the same amount of states, transitions, input, output, and inner variables as this
   * EFSM does;
   * <li>for each its state and transition there is an equal one in this EFSM;
   * <li>for each its input, output, and inner variable there is a one with equal name and data type
   * in this EFSM.
   * </ul>
   *
   * @param obj the object to be compared for equality with this EFSM
   *
   * @return {@code true} if the specified object is equal to this EFSM. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((!(obj instanceof Efsm)) || (this.hashCode() != obj.hashCode())) {
      return false;
    }

    Efsm efsmForComparison = (Efsm) obj;

    return (this.name.equals(efsmForComparison.name)
        && this.states.equals(efsmForComparison.states)
        && this.transitions.equals(efsmForComparison.transitions)
        && this.initialState.equals(efsmForComparison.initialState)
        && this.inputs.equals(efsmForComparison.inputs)
        && this.outputs.equals(efsmForComparison.outputs) && this.innerVariables
          .equals(efsmForComparison.innerVariables));
  }

  @Override
  public String toString() {
    return this.name;
  }
}
