/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;

/**
 * Entity representation of EFSM collection.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class EfsmModel extends Entity {
  /**
   * The entity type of all {@code EfsmModel} entities.
   */
  public static final EntityType TYPE = EntityType.get("efsm");

  private List<Efsm> efsmList;

  /**
   * Creates an empty EfsmModel.
   */
  public EfsmModel() {
    super(EfsmModel.TYPE);
    efsmList = new ArrayList<>();
  }

  /**
   * Creates an EfsmModel above the specified EFSM.
   *
   * @param efsm the EFSM
   */
  public EfsmModel(Efsm efsm) {
    super(EfsmModel.TYPE);
    this.efsmList = new ArrayList<>();
    this.efsmList.add(efsm);
  }

  /**
   * Creates an EfsmModel above the specified EFSMs.
   *
   * @param efsms the {@link List} of the EFSMs.
   */
  public EfsmModel(List<Efsm> efsms) {
    super(EfsmModel.TYPE);
    efsmList = efsms;
  }

  /**
   * Returns the set of EFSMs containing in the model.
   *
   * @return the {@link List} of EFSMs.
   */
  public List<Efsm> getEfsmList() {
    return efsmList;
  }

  @Override
  public void add(Entity entity) {
    if (entity instanceof EfsmModel) {
      this.addEfsm(((EfsmModel) entity).getEfsmList());
    } else {
      throw new IllegalArgumentException(String.format("Unknown node kind: %s.", entity
          .getEntityType().getId()));
    }
  }

  /**
   * Adds a specified collection to the set.
   *
   * @param efsms a specified {@link Efsm} collection.
   */
  public void addEfsm(Collection<Efsm> efsms) {
    this.efsmList.addAll(efsms);
  }

  /**
   * Adds a specified EFSM to the set.
   *
   * @param efsm a specified {@link Efsm}.
   */
  public void addEfsm(Efsm efsm) {
    efsmList.add(efsm);
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + ((this.efsmList == null) ? 0 : this.efsmList.hashCode());

    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if (!super.equals(obj) || !(obj instanceof EfsmModel)) {
      return false;
    }

    EfsmModel efsmModelToCompare = (EfsmModel) obj;

    if (this.efsmList == null) {
      if (efsmModelToCompare.efsmList != null) {
        return false;
      }
    } else {
      if (!this.efsmList.equals(efsmModelToCompare.efsmList)) {
        return false;
      }
    }

    return true;
  }
}
