/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * The guarded action-based representation of a transition of an extended finite state machine.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmTransition {
  private static final String LEFT_STRING_BOUND = "{";
  private static final String SOURCE_STATE_STRING_HEADER = "source state: ";
  private static final String DESTINATION_STATE_STRING_HEADER = "destination state: ";
  private static final String GUARDED_ACTION_STRING_HEADER = "guarded action: ";
  private static final String STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  private final EfsmState sourceState;
  private final EfsmState destinationState;

  private final GuardedAction guardedAction;

  /**
   * Constructs a transition between the specified states with the specified guarded action.
   *
   * @param sourceState a source state
   * @param destinationState a destination state
   * @param guardedAction the guarded action
   *
   * @throws NullPointerException if at least one argument is {@code null}
   */
  public EfsmTransition(EfsmState sourceState, EfsmState destinationState,
      GuardedAction guardedAction) {
    InvariantChecks.checkNotNull(sourceState);
    InvariantChecks.checkNotNull(destinationState);
    InvariantChecks.checkNotNull(guardedAction);

    this.sourceState = sourceState;
    this.destinationState = destinationState;
    this.guardedAction = guardedAction;
  }

  /**
   * Returns a state which is the start point of this transition.
   *
   * @return the source state of this transition
   */
  public EfsmState getSourceState() {
    return this.sourceState;
  }

  /**
   * Returns a state which is the end point of this transition.
   *
   * @return the destination state of this transition
   */
  public EfsmState getDestinationState() {
    return this.destinationState;
  }

  /**
   * Returns a guarded action of this transition.
   *
   * @return the guarded action of this transition
   */
  public GuardedAction getGuardedAction() {
    return this.guardedAction;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.guardedAction.hashCode();
    result = prime * result + this.destinationState.hashCode();
    result = prime * result + this.sourceState.hashCode();

    return result;
  }

  /**
   * Compares the specified object with this transition for equality. The object is considered as
   * equal to this transition if and only if it is also an EFSM transition, it connects the same
   * states with the same direction, and it has the same guarded action as this transition does, or
   * more formally:
   * <ul>
   *    <li>it is an instance of the {@code EfsmTransition} class or its subclass;
   *    <li>its source and destination states are equal to this transition ones;
   *    <li>its guarded action is equal to this transition's one.
   * </ul>
   *
   * @param obj the object to be compared for equality with this EFSM state
   *
   * @return {@code true} if the specified object is equal to this EFSM state. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if (!(obj instanceof EfsmTransition) || (this.hashCode() != obj.hashCode())) {
      return false;
    }

    EfsmTransition transitionForComparison = (EfsmTransition) obj;

    return (this.sourceState.equals(transitionForComparison.sourceState) &&
        this.destinationState.equals(transitionForComparison.destinationState) &&
        this.guardedAction.equals(transitionForComparison.guardedAction));
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(EfsmTransition.LEFT_STRING_BOUND);
    stringBuilder.append(EfsmTransition.SOURCE_STATE_STRING_HEADER);
    stringBuilder.append(this.sourceState);
    stringBuilder.append(EfsmTransition.STRING_DELIMITER);
    stringBuilder.append(EfsmTransition.DESTINATION_STATE_STRING_HEADER);
    stringBuilder.append(this.destinationState);
    stringBuilder.append(EfsmTransition.STRING_DELIMITER);
    stringBuilder.append(EfsmTransition.GUARDED_ACTION_STRING_HEADER);
    stringBuilder.append(this.guardedAction);
    stringBuilder.append(EfsmTransition.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
