/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.UseDef;

/**
 * Guard is a boolean condition, that consists of some predicates.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Guard extends MetaInfo implements UseDef {

  /* String representation constants. */

  private static final String LEFT_STRING_BOUND = "{";
  private static final String PREDICATE_STRING_HEADER = "predicate: ";
  private static final String PREDICATE_STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  /* Boolean predicates collection. */
  private final List<Node> predicates;

  /**
   * The default constructor.
   */
  public Guard() { this.predicates = new ArrayList<>(); }

  /**
   * Constructs guard object from the specified predicate.
   *
   * @param node guard predicate
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Guard(Node node) {
    InvariantChecks.checkNotNull(node);

    this.predicates = new ArrayList<>();

    this.predicates.add(node);
  }

  /**
   * Constructs guard object from the specified list of predicates.
   *
   * @param list list of guard predicates
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Guard(Collection<Node> list) {
    InvariantChecks.checkNotNull(list);

    this.predicates = new ArrayList<>();

    this.predicates.addAll(list);
  }

  /**
   * Returns list of guard predicates.
   *
   * @return guard predicates list
   */
  public List<Node> getGuardExpr() { return this.predicates; }

  /**
   * Convert guard expression into {@link ru.ispras.fortress.expression.Node} expression.
   *
   * @return guard expression transformed to {@link ru.ispras.fortress.expression.Node} expression.
   */
  public Node toNode() {
    return ExprUtils.getConjunction(this.predicates.toArray(new Node[this.predicates.size()]));
  }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> union = new LinkedHashSet<>();

    for (Node predicate : this.getGuardExpr()) {

      Set<NodeVariable> prUses = new LinkedHashSet<>(ExprUtils.getVariables(predicate));

      if (union.size() == 0) {
        union.addAll(prUses);
      } else {
        union = CollectionUtils.uniteSets(union, prUses);
      }
    }
    return union;
  }

  @Override
  public Set<NodeVariable> getDefines() { return new LinkedHashSet<>(); }

  @Override
  public int hashCode() { return this.predicates.hashCode(); }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;

    if ((!(obj instanceof Guard)) || (this.hashCode() != obj.hashCode()))
      return false;

    Guard givenCBGuard = (Guard) obj;
    return this.predicates.equals(givenCBGuard.getGuardExpr());
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(Guard.LEFT_STRING_BOUND);

    int initialStringBuilderLength = stringBuilder.length();

    for (Node predicate: this.predicates) {
      if (stringBuilder.length() > initialStringBuilderLength) {
        stringBuilder.append(Guard.PREDICATE_STRING_DELIMITER);
      }

      stringBuilder.append(Guard.PREDICATE_STRING_HEADER);
      stringBuilder.append(predicate);
    }

    stringBuilder.append(Guard.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }

}
