/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.Collection;
import java.util.Set;

/**
 * Guarded action is a triple: a collection of clock
 * {@link ru.ispras.retrascope.model.basis.Event} events, a guard
 * {@link ru.ispras.retrascope.model.efsm.Guard} predicate,
 * and an associated action {@link ru.ispras.retrascope.model.efsm.Action}.
 *
 * When clock event happens, if guard becomes {@code true} then an action proceeds.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class GuardedAction extends MetaInfo implements UseDef {

  /* The delimiter for object's string representation. */
  private static final String GA_DELIMITER = "->";

  /* The guard. */
  private final Guard guard;

  /* The action. */
  private final Action action;

  /* The clock events. */
  private final EventList clocks;

  /**
   * Constructs guarded action object from the specified guard and action.
   *
   * @param guard guard
   * @param action action
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public GuardedAction(Guard guard, Action action) {
    InvariantChecks.checkNotNull(guard);
    InvariantChecks.checkNotNull(action);

    this.guard = guard;
    this.action = action;
    this.clocks = new EventList();
  }

  /**
   * Returns action.
   *
   * @return action
   */
  public Action getAction() { return this.action; }

  /**
   * Returns guard.
   *
   * @return guard
   */
  public Guard getGuard() { return this.guard; }

  /**
   * Returns list of object's clock events.
   *
   * @return clocks list
   */
  public EventList getEventList() { return this.clocks; }

  /**
   * Returns guarded action clocks.
   *
   * @return clocks collection
   */
  public Collection<Event> getClocks() { return this.clocks.getEvents(); }

  /**
   * Associates this guarded action with the specified {@link Event} as a clock.
   *
   * @param clock clock event
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setClock(Event clock) {
    InvariantChecks.checkNotNull(clock);

    this.clocks.add(clock);
  }

  /**
   * Sets guarded action clocks.
   *
   * @param clocks clocks collection
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setClocks(Collection<Event> clocks) {
    InvariantChecks.checkNotNull(clocks);

    this.clocks.addEvents(clocks);
  }

  /**
   * Returns variables that are used by guard.
   *
   * @return sequence of variables
   */
  public Iterable<NodeVariable> getUsesByGuard() { return this.getGuard().getUses(); }

  /**
   * Returns variables that are used by action.
   *
   * @return sequence of variables
   */
  public Iterable<NodeVariable> getUsesByAction() { return this.getAction().getUses(); }

  /**
   * Checks if the specified event is in a sensitivity list of this guarded action.
   *
   * @param event the event to be tested
   *
   * @return {@code true} if the specified event is in the sensitivity list of this guarded action
   * or the list is empty. Otherwise returns {@code false}
   */
  public boolean isSensible(Event event) {
    return this.clocks.isTriggeredBy(event);
  }

  /**
   * Checks if this guarded action is sensible to the specified events.
   *
   * @param eventList an event list to be tested
   *
   * @return {@code true} if at least one of the specified events is in the sensitivity list of
   * this guarded action or the sensitivity list is empty. Otherwise returns {@code false}
   */
  public boolean isSensible(EventList eventList) {
    for (Event event: eventList.getEvents()) {
      if (this.isSensible(event)) {
        return true;
      }
    }

    return false;
  }

  @Override
  public Set<NodeVariable> getUses() {
    return CollectionUtils.uniteSets(this.getGuard().getUses(), this.getAction().getUses());
  }

  @Override
  public Set<NodeVariable> getDefines() { return this.getAction().getDefines(); }

  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder();
    builder.append("{");
    builder.append(this.clocks).append(": ");
    builder.append(this.getGuard());
    builder.append(GA_DELIMITER);
    builder.append("{");

    if (this.getAction() != null)
      builder.append(this.getAction());
    builder.append("}");
    builder.append("}");
    return builder.toString();
  }
}
