/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.engine.cfg.CfgDfsWrVisitorBackend;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.ArrayList;
import java.util.List;

/**
 * This backend transforms basic block {@link ru.ispras.retrascope.model.cfg.BasicBlock} nodes
 * of concurrent (non-blocking) type into basic blocks of non-concurrent (blocking) type.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgConcurrentBasicBlockBackend extends CfgDfsWrVisitorBackend {

  /* The prefix for new variables names (for those that are defined at blocking assignments). */
  private static final String BLOCK = "BLOCK_";

  /* The current module. */
  private Module module;

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) { this.module = module; }

  @Override
  public void onModuleEnd(Module module) { this.module = null; }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { /* Empty. */ }

  @Override
  public void onCfgEnd(Cfg cfg) { /* Empty. */ }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) {
    /* Transform concurrent block into sequential (blocking). */
    if (basicBlock.isConcurrent()) {
      if (assignsVariables(basicBlock)) {
        List<Assignment> sequential = makeBlocking(basicBlock.getAssignments());
        basicBlock.clearStatements();
        basicBlock.appendAssignments(sequential);
      }
      basicBlock.setConcurrent(false);
    }

  }

  /* Checks rather basic block contains at least one assignment that assigns a variable
*  to the target. */
  private boolean assignsVariables(BasicBlock basicBlock) {
    for (Assignment assignment : basicBlock.getAssignments()) {
      if (!ExprUtils.getVariables(assignment.getExpression()).isEmpty()) {
        return true;
      }
    }
    return false;
  }

  private List<Assignment> makeBlocking(List<Assignment> list) {

    /* Create a sequence of assignments like "new variable := used expression". */
    List<Assignment> blocking = new ArrayList<>();

    /* Create a sequence of assignments like "defined variables := new variable". */
    List<Assignment> oldToNewList = new ArrayList<>();

    for (Assignment assignment : list) {

      RangedVariable rVariable = assignment.getRangedVariable();

      /* Create internal variable with unique name. */
      String defName = rVariable.getName();
      String varName = BLOCK + defName + "_" + blocking.size();

      DataType dataType;

      /* Determine temporary variable data type.*/

      /* Try to get assigned expression's data type. */
      dataType = assignment.getExpression().getDataType();

      if (dataType.equals(DataType.UNKNOWN)) {
        //TODO: remove it when data type calculation in Fortress will be improved
        if (rVariable.isRanged()) {

          NodeVariable defVariable = rVariable.getVariable();
          Range defRange = rVariable.getRange();

          if (defVariable.isType(DataTypeId.LOGIC_INTEGER) ||
            defVariable.isType(DataTypeId.LOGIC_REAL) ||
            defVariable.isType(DataTypeId.LOGIC_BOOLEAN)) {

            /* The boolean or integer or real can't be ranged. */

            throw new IllegalStateException(
              String.format("Cannot parse: %s.", rVariable.toString()));
          } else if (defVariable.isType(DataTypeId.MAP)) {

            /* The array. */

            /* Array element. */
            if (defRange.getYoung().equals(defRange.getOld())) {
              /* Get data type of array element. */
              dataType = ((DataType)defVariable.getData().getType().getParameters()[1]);
            } else {
              /* NOTE: It is rather strange to extract sub-array. isn't it?. */
              throw new IllegalStateException(
                String.format("Cannot parse: %s.", rVariable.toString()));
            }

          } else if (defVariable.isType(DataTypeId.BIT_VECTOR)) {
            /* The bit vector. */
            dataType =
              DataType.BIT_VECTOR(
                ((NodeValue) Transformer.reduce(
                  ReduceOptions.NEW_INSTANCE,
                  new NodeOperation(
                    StandardOperation.SUB,
                    defRange.getOld(),
                    defRange.getYoung()))).getInteger().intValue());
          } else
            throw new IllegalArgumentException(
              String.format(
                "Cannot parse: %s.",
                defVariable.getDataTypeId()));
        } else {
          dataType = rVariable.getVariable().getDataType();
        }
      }

      NodeVariable var = new NodeVariable(varName, dataType);
      if (!this.module.containsVariable(var)) {
        this.module.addVariable(var);
      }
      var.setUserData(new VariableData(VariableType.REG));
      blocking.add(new Assignment(var, assignment.getExpression()));

      oldToNewList.add(new Assignment(rVariable, var));
    }
    blocking.addAll(oldToNewList);
    return blocking;
  }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) { /* Empty. */ }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { /* Empty. */ }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
