/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.engine.basis.TransformUtils;
import ru.ispras.retrascope.engine.cfg.CfgDfsVisitorBackend;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgVariableContainerNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * This backend detects loops or cycles in {@link ru.ispras.retrascope.model.cfg.Cfg}
 * control flow graphs and transforms them into acyclic graphs.
 *
 * The loops that this class is able to transform can be of the following structure only:
 *
 *                     |
 *           counter := init-value (LOOP_INIT)
 *                     |
 *    _______________merge (LOOP_ENTRY)
 *   |                 |
 *   |         switch(expression-upon-counter) (LOOP_COND)
 *   |                 |              |
 *   |            true-case       false-case
 *   |                 |              |
 *   |      <loop-body-sub-graph>   merge
 *   |                 |              |
 *   |                 |            <...>
 *   |                 |
 *   |    <counter := increment-function> (LOOP_INC)
 *   |_________________|
 *
 * When the expression is equal to {@code true}, the counter-incrementing basic block
 * is unnecessary, because all the loop-body-sub-graph will be executed in the dependant separate
 * process. Otherwise the loop is transformed to the sequence of loop-body-sub-graphs. The amount
 * of the sequence is calculated as how many times the expression-upon-counter will stay
 * satisfiable when counter will be incremented on calling the increment-function again and again.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgLoopBackend extends CfgDfsVisitorBackend {

  /* The current module. */
  private Module module;

  /* The processes that are extracted from infinite loops. */
  private List<Process> loopProcesses;

  /* The current control flow graph. */
  private Cfg cfg;

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) {
    this.module = module;
    this.loopProcesses = new ArrayList<>();
  }

  @Override
  public void onModuleEnd(Module module) {
    this.module = null;

    if (!this.loopProcesses.isEmpty()) {
      for (Process process : this.loopProcesses) {
        module.addChild(process);
      }
      this.loopProcesses = null;
    }
  }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { this.cfg = cfg; }

  @Override
  public void onCfgEnd(Cfg cfg) { this.cfg = null; }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) {

    if (merge.hasMetaInfo(MetaInfoType.LOOP_NODE) &&
        merge.getMetaInfo(MetaInfoType.LOOP_NODE).equals(MetaInfoValue.LOOP_ENTRY.name())) {

      /* The loop has been found, try to detect other loop components. */

      Collection<CfgModelNode> parents = merge.getParents();

      /* The node for "i := 0" initialization assignment keeping. */
      BasicBlock initBlock = null;

      /* The node for "i++" incrementation assignment keeping. */
      BasicBlock incrementBlock = null;

      /* The last of the loop nodes. */
      CfgNode loopLastNode = null;

      /* The "to-be-removed" node. */
      Merge toBeRemoved = null;

      for (CfgModelNode parent : parents) {
        if (parent.hasMetaInfo(MetaInfoType.LOOP_NODE)) {

          String loopNodeInfo = parent.getMetaInfo(MetaInfoType.LOOP_NODE);

          if (loopNodeInfo.equals(MetaInfoValue.LOOP_INIT.name())) {

            /* The counter initialization block has been found. */
            initBlock = (BasicBlock) parent;

          } else if (loopNodeInfo.equals(MetaInfoValue.LOOP_INC.name())) {

            /* The counter incrementation block has been found. */
            incrementBlock = (BasicBlock) parent;
          } else if (loopNodeInfo.equals(MetaInfoValue.LOOP_FIN.name())) {

            /* The final node of the loop iteration has been found. */
            loopLastNode = (CfgNode) parent;
          }
        }
        if (isSinglePathMerge(parent) && isIncrementBlock(parent.getOnlyParent())) {

          incrementBlock = (BasicBlock) parent.getOnlyParent();
          toBeRemoved = (Merge) parent;
        }
      }

      /* Remove the single path merge node
       * (that has only one parent increment block and only one child node). */
      if (toBeRemoved != null && incrementBlock != null) {
        merge.removeParent(toBeRemoved);
        incrementBlock.removeChild(toBeRemoved);
        incrementBlock.addChild(merge);
        this.cfg.removeNode(toBeRemoved);
      }

      CfgModelNode child = merge.getOnlyChild();

      Switch condition = null;

      if (child.hasMetaInfo(MetaInfoType.LOOP_NODE) &&
          child.getMetaInfo(MetaInfoType.LOOP_NODE).equals(MetaInfoValue.LOOP_COND.name())) {
        condition = (Switch) child;
      }

      /* Check the condition. */
      assert condition != null;
      final Node conditionExpr = condition.getCondition();

      /* Detect TRUE/FALSE case child for loop condition switch node. */
      final Collection<CfgModelNode> condChildren = condition.getChildren();
      final Iterator<CfgModelNode> iterator = condChildren.iterator();
      final Case firstCase = (Case) iterator.next();
      final Case secondCase = (Case) iterator.next();
      final Case trueCase = getValueCase(firstCase, secondCase, true);
      final Case falseCase = getValueCase(firstCase, secondCase, false);

      /* Remove old links, make merge block child-free & false-case-child parent-free. */
      merge.removeChild(condition);
      final CfgNode falseCaseChild = (CfgNode) falseCase.getOnlyChild();
      falseCase.removeChild(falseCaseChild);

      if (initBlock != null && incrementBlock != null) {
        /* Finite iteration number loop.*/

        final List<CfgNode> loopPath = TransformUtils.getPath(trueCase, incrementBlock);

        /* Build a sequence of iterations. */
        CfgNode unlinked = merge;
        final Node incrementExpr = incrementBlock.getAssignments().get(0).getExpression();
        final Assignment initAssignment = initBlock.getAssignments().get(0);
        String varName = initAssignment.getRangedVariable().getVariable().getName();

        Node currentIncrementValue = initAssignment.getExpression();
        Node whileExpr = Transformer.substitute(conditionExpr, varName, currentIncrementValue);

        while (ExprUtils.isSAT(whileExpr)) {

          /* Create a copy of path. */
          final List<CfgNode> loopPathCopy = new ArrayList<>(loopPath.size());
          for (CfgNode origNode : loopPath) {
            CfgModelNode copy = origNode.deepcopy();
            if (!loopPathCopy.isEmpty()) {
              copy.addParent(loopPathCopy.get(loopPathCopy.size() - 1));
            }
            loopPathCopy.add((CfgNode) copy);
          }

          /* Substitute iteration variable by it's incrementation expression. */
          List<CfgNode> substituted =
              TransformUtils.substitute(loopPathCopy, varName, currentIncrementValue);

          unlinked.addChild(substituted.get(0));

          this.cfg.addNodes(substituted);
          unlinked = substituted.get(substituted.size() - 1);
          currentIncrementValue =
              Transformer.substitute(incrementExpr, varName, currentIncrementValue);
          whileExpr = Transformer.substitute(conditionExpr, varName, currentIncrementValue);
        }

        /* Remove internal loop nodes. */
        for (CfgNode loopNode : loopPath) {
          this.cfg.removeNode(loopNode);
        }

        unlinked.addChild(falseCaseChild);

      } else if (ExprUtils.getVariables(conditionExpr).isEmpty()) {

        if (ExprUtils.isSAT(conditionExpr)) {

          /* Infinite loop, create an additional process. */
          final List<CfgNode> loopPath = TransformUtils.getPath(trueCase, merge);

          final Process process = new Process();

          /* Create empty control flow graph for process, */
          final Cfg processCfg = new Cfg();
          process.addChild(processCfg);
          Source source = processCfg.getSource();
          final Sink sink = processCfg.getSink();
          source.removeChild(sink);

          final NodeVariable loopVariable =
              new NodeVariable(
                  CfgVariableContainerNode.NEW_VAR_PREFIX + this.module.getDeclarations().size(),
                  DataType.BOOLEAN);

          this.module.addVariable(loopVariable);

          process.setSensitivityList(
            new EventList(
              Collections.singleton(new Event(new RangedVariable(loopVariable)))));

          /* Init loop variable in the block node & add it to the parent process. */
          final BasicBlock loopVariableBlock =
              new BasicBlock(new Assignment(loopVariable, NodeValue.newBoolean(true)));
          this.cfg.addNode(loopVariableBlock);
          merge.addChild(loopVariableBlock);
          loopVariableBlock.addChild(falseCaseChild);

          /* Construct the sub-process control flow graph. */

          processCfg.addNodes(loopPath);

          CfgNode first = loopPath.get(0);
          CfgNode last = loopPath.get(loopPath.size() - 1);

          if (first.hasParents() || last.hasChildren()) {
            throw new IllegalArgumentException("Wrong path has been extracted");
          } else {
            source.addChild(first);
            last.addChild(sink);
          }
          this.loopProcesses.add(process);

          /* Remove internal loop nodes. */
          for (CfgNode loopNode : loopPath) {
            this.cfg.removeNode(loopNode);
          }
        } else {

          /* Dead code. */
          merge.addChild(falseCaseChild);
        }
      } else
          throw new IllegalArgumentException(
              String.format("Cannot parse loop: %s.", condition.toString()));

      /* No more loop merge here. */
      merge.removeMetaInfo(MetaInfoType.LOOP_NODE);

      /* Remove outdated nodes. */
      this.cfg.removeNode(condition);
      this.cfg.removeNode(trueCase);
      this.cfg.removeNode(falseCase);

      if (incrementBlock != null) {
        this.cfg.removeNode(incrementBlock);
        merge.removeParent(incrementBlock);
      } else if (loopLastNode != null) {
        merge.removeParent(loopLastNode);
      }
    }
  }

  private boolean isSinglePathMerge(CfgModelNode parent) {
    return parent.getChildren().size() == 1 &&
      parent.getParents().size() == 1 &&
      parent instanceof Merge;
  }

  private boolean isIncrementBlock(CfgModelNode node) {
    return node.hasMetaInfo(MetaInfoType.LOOP_NODE) &&
      node.getMetaInfo(MetaInfoType.LOOP_NODE).equals(MetaInfoValue.LOOP_INC.name());
  }

  private static Case getValueCase(Case firstCase, Case secondCase, boolean value) {
    return firstCase.getValues().contains(NodeValue.newBoolean(value)) ? firstCase : secondCase;
  }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { /* Empty. */ }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
