/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.retrascope.engine.cfg.CfgDfsWrVisitorBackend;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.Collection;

/**
 * This backend merges {@link ru.ispras.retrascope.model.cfg.Merge} nodes,
 * following each other (i.e. sequences of merge nodes).
 * Also it removes {@link ru.ispras.retrascope.model.cfg.Merge} nodes,
 * which have exactly one parent node and exactly one child node.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgMergeSequenceBackend extends CfgDfsWrVisitorBackend {

  /* The control flow graph that is to be processed. */
  private Cfg cfg;

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) { /* Empty. */ }

  @Override
  public void onModuleEnd(Module module) { /* Empty. */ }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { this.cfg = cfg; }

  @Override
  public void onCfgEnd(Cfg cfg) { this.cfg = null; }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) {
    while(hasOnlyMergeChild(merge)) {
      mergeMerges(merge, (Merge) merge.getOnlyChild());
    }
  }

  private void mergeMerges(Merge merge, Merge child) {
    merge.removeChild(child);
    if (child.hasChildren()) {
      Collection<CfgModelNode> childParents = child.getParents();
      for (CfgModelNode childParent : childParents) {
        if (!childParent.equals(merge)) {
          childParent.removeChild(child);
          childParent.addChild(merge);
        }
      }
      if (child.hasChildren()) {
        Collection<CfgModelNode> grandChildren = child.getChildren();
        for (CfgModelNode grandChild : grandChildren) {
          grandChild.removeParent(child);
          grandChild.addParent(merge);
        }
      } else
        throw new IllegalArgumentException(
          String.format("Has no children: %s.", child.getType()));
    }

    /* Remove "hanging" nodes. */
    if (this.cfg != null) {
      this.cfg.removeNode(child);
    }
  }

  private boolean hasOnlyMergeChild(Merge merge) {
    if (merge.hasChildren()) {
      Collection<CfgModelNode> children = merge.getChildren();
      if (children.size() == 1 && merge.getOnlyChild() instanceof Merge) {
        return true;
      }
    } else
      throw new IllegalStateException(String.format("Has no children: %s.", merge));
    return false;
  }

  @Override
  public void onMergeEnd(Merge merge) {
    if (merge.getParents().size() == 1 && merge.getChildren().size() == 1) {
      CfgModelNode parent = merge.getOnlyParent();
      CfgModelNode child = merge.getOnlyChild();
      parent.removeChild(merge);
      child.removeParent(merge);
      parent.addChild(child);
      this.cfg.removeNode(merge);
    }
  }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { /* Empty. */ }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
