/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.basis.backend;


import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.engine.cfg.CfgDfsWrVisitorBackend;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * This backend merges {@link ru.ispras.retrascope.model.cfg.Process} processes,
 * which contain only one {@link ru.ispras.retrascope.model.cfg.BasicBlock} basic block node
 * of concurrent type, with only one {@link ru.ispras.retrascope.model.basis.Assignment} assignment,
 * and equal {@link ru.ispras.retrascope.model.basis.EventList} event lists.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgOneAssignmentProcessBackend extends CfgDfsWrVisitorBackend {

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) {

    Set<Process> merged = new LinkedHashSet<>();
    List<Process> processes = module.getProcesses();

    if (!processes.isEmpty()) {
      for (Process process : processes) {
        if (oneBlockCfg(process)) {

          if (!merged.isEmpty()) {

            boolean isMerged = false;
            for (Process mProcess : merged) {

              /* If the target process can be merged into "merged" process. */
              if (oneBlockCfg(mProcess) &&
                !CollectionUtils.areIntersectedSets(mProcess.getDefines(), process.getUses())) {
                mergeProcesses(mProcess, process);
                isMerged = true;
                break;
              }
            }
            if (!isMerged) {
              merged.add(process);
            }
          } else {
            merged.add(process);
          }
        } else {
          merged.add(process);
        }
      }

      /* Remove "old" processes. */
      module.removeProcesses();

      /* Add new "merged" processes. */
      for (Process process : merged) {
        module.addChild(process);
      }
    }
  }

  private void mergeProcesses(Process result, Process target) {

    /* Merge sensitivity lists */
    EventList rList = result.getSensitivityList();
    Collection<Event> tList = target.getSensitivityList().getEvents();

    for (Event event : tList) {
      rList.add(event);
    }

    /* Merge control flow graph.s */

    /* Get basic bock node at the result process. */
    Cfg resultCfg = result.getCfg();
    CfgNode resultSrc = resultCfg.getSource();
    BasicBlock resultBasicBlock = (BasicBlock) resultSrc.getOnlyChild();

    /* Get basic block at the target process. */
    Cfg targetCfg = target.getCfg();
    CfgNode targetSrc = targetCfg.getSink();
    BasicBlock targetBasicBlock = (BasicBlock) targetSrc.getOnlyChild();

    resultBasicBlock.appendAssignments(targetBasicBlock.getAssignments());
  }

  private boolean oneBlockCfg(Process process) {
    if (process.hasChildren()) {
      Collection<CfgModelNode> children = process.getChildren();
      if (children.size() == 1) {
        CfgModelNode child = children.iterator().next();
        return isConcurrentAssignmentBlockCfg(child);
      } else return false;
    }
    else return false;
  }

  private boolean isConcurrentAssignmentBlockCfg(CfgModelNode node) {
    if (node instanceof Cfg) {
      Collection<CfgModelNode> children = node.getChildren();
      if (children.size() == 1) {
        Collection<CfgModelNode> grandChildren = children.iterator().next().getChildren();
        if (grandChildren.size() == 1) {
          CfgModelNode grandChild = grandChildren.iterator().next();

          return (grandChild instanceof BasicBlock &&
              grandChild.getChildren().iterator().next() instanceof Sink &&
              ((BasicBlock) grandChild).isConcurrent());

        } else throw new IllegalArgumentException(String.format("Node has more than one child"));
      } else
        throw new IllegalArgumentException(
            String.format("Graph with more than one source node"));

    } else return false;
  }

  @Override
  public void onModuleEnd(Module module) { /* Empty. */ }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { /* Empty. */ }

  @Override
  public void onCfgEnd(Cfg cfg) { /* Empty. */ }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) { /* Empty. */ }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { /* Empty. */ }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
