/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.engine.cfg.CfgDfsWrVisitorBackend;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.Collection;

/** HDL parser backend that optimises sequences
 * of control flow graph model {@link ru.ispras.retrascope.model.cfg.Switch} switch nodes.
 * Switch nodes must contain conditions of the following form:
 * "x == const_1 || ... || x == const_n", and only the "x" must be used
 * in every sequence node.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgSwitchSequenceBackend extends CfgDfsWrVisitorBackend {

  /* The control flow graph that is to be processed.*/
  private Cfg cfg;

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) { /* Empty. */ }

  @Override
  public void onModuleEnd(Module module) { /* Empty. */ }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { this.cfg = cfg; }

  @Override
  public void onCfgEnd(Cfg cfg) { this.cfg = null; }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) {

    Switch grandChild;
    while ((grandChild = getSwitchGrandChildWithEqualExpr(aSwitch)) != null) {
      collapseSwitchSubTree(aSwitch, grandChild);
    }
  }

  private void collapseSwitchSubTree(Switch aSwitch, Switch grandChild) {

    Collection<CfgModelNode> children = grandChild.getChildren();

    /* Remove links between next switch and it's children. */
    for (CfgModelNode child : children) {
      grandChild.removeChild(child);
      aSwitch.addChild(child);
    }

    /* Remove links from and to middle node. */
    CfgModelNode middle = grandChild.getOnlyParent();
    middle.removeChild(grandChild);
    middle.removeParent(aSwitch);

    this.cfg.removeNode((CfgNode) middle);
    this.cfg.removeNode(grandChild);
  }

  private Switch getSwitchGrandChildWithEqualExpr(Switch aSwitch) {
    if (aSwitch.hasChildren()) {
      Collection<CfgModelNode> switchChildren = aSwitch.getChildren();
      for (CfgModelNode child : switchChildren) {
        if (child.hasChildren()) {
          Collection<CfgModelNode> grandChildren = child.getChildren();
          if (grandChildren.size() == 1) {
            for (CfgModelNode grandChild : grandChildren) {

              if (grandChild instanceof Switch) {
                Node grandChildExpr = ((Switch) grandChild).getCondition();
                Node kernel = getMultiOrKernel(grandChildExpr);

                if (aSwitch.getCondition().equals(grandChildExpr)
                  || (kernel != null && aSwitch.getCondition().equals(kernel))) {
                  return (Switch) grandChild;
                }
              }
            }
          } else
            throw new IllegalArgumentException(
                String.format("More than one child for: %s.", child.getType()));
        }
      }
      return null;
    } else
      throw new IllegalStateException(
          String.format("Node has no children: %s.", aSwitch.getType()));
  }

  private Node getMultiOrKernel(Node expr) {
    if (expr instanceof NodeOperation) {
      if (((NodeOperation) expr).getOperationId().equals(StandardOperation.OR)) {
        Node kernel = null;
        int valIndex = -1;
        int varIndex = -1;

        for (int i = 0; i < ((NodeOperation) expr).getOperandCount(); i++) {
          Node exprOperand = ((NodeOperation) expr).getOperand(i);
          if (exprOperand instanceof NodeOperation) {
            if (((NodeOperation) exprOperand).getOperationId().equals(StandardOperation.EQ)) {
              for (int j = 0; j < ((NodeOperation) exprOperand).getOperandCount(); j++) {
                Node operand = ((NodeOperation) exprOperand).getOperand(j);
                if (operand instanceof NodeVariable) {
                  varIndex = j;
                  if (kernel != null && !kernel.equals(operand))
                    return null;
                  kernel = operand;
                }
                else if (operand instanceof NodeValue) {
                  valIndex = j;
                }
              }
              if (varIndex == -1 || valIndex == -1) return null;
            }
          } else return null;
        }
        return kernel;
      } else return null;

    } else return null;
  }

  @Override
  public void onSwitchEnd(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onMergeBegin(Merge merge) { /* Empty. */ }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { /* Empty. */ }

  @Override
  public void onWaitBegin(Wait wait) { /* Empty. */ }

  @Override
  public void onWaitEnd(Wait wait) { /* Empty. */ }
}
