/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.engine.cfg.CfgDfsVisitorBackend;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgVariableContainerNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgWaitBackend extends CfgDfsVisitorBackend {

  /* The current module. */
  private Module module;

  /* The current parent wait node. */
  private Wait waitNode;

  /* The mapping for wait nodes & nodes of their processes. */
  private Map<Wait, Set<CfgNode>> map;

  /* The control flow graph nodes and their related nearest parent wait nodes. */
  private Map<CfgNode, Wait> nearestParentWaitMap;

  /* The wait nodes parents & corresponding to them new variables. */
  private Map<CfgNode, NodeVariable> waitParentNewVarMap;

  /* The wait nodes & corresponding new variables. */
  private Map<Wait, NodeVariable> waitNewVarMap;

  /* The nodes that are reachable non through wait. */
  private Set<CfgNode> nonWaitReachableNodes;

  /* The collection for wait-processes. */
  private Map<Module, List<Process>> waitProcessMap;

  @Override
  public Status getStatus() { return Status.OK; }

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { this.waitProcessMap = new LinkedHashMap<>(); }

  @Override
  public void onCfgModelEnd() {
    for (Module module : this.waitProcessMap.keySet()) {
      final List<Process> processes = this.waitProcessMap.get(module);
      for (Process process : processes) {
        module.addChild(process);
      }
    }

    this.waitProcessMap.clear();
  }

  @Override
  public void onModuleBegin(Module module) {
    this.module = module;
    this.waitProcessMap.put(module, new ArrayList<Process>());
  }

  @Override
  public void onModuleEnd(Module module) {
    this.module = null;
  }

  @Override
  public void onProcessBegin(Process process) {
    this.map = new LinkedHashMap<>();
    this.nonWaitReachableNodes = new LinkedHashSet<>();
    this.nearestParentWaitMap = new LinkedHashMap<>();
    this.waitParentNewVarMap = new LinkedHashMap<>();
    this.waitNewVarMap = new LinkedHashMap<>();
  }

  @Override
  public void onProcessEnd(Process process) {

    Cfg cfg = process.getCfg();

    final Map<CfgModelNode, CfgModelNode> oldNewMap = new LinkedHashMap<>();

    for (Wait waitNode : this.map.keySet()) {

      /* Create new process. */
      cfg.removeNode(waitNode);

      final Process waitProcess = new Process();
      waitProcess.setSensitivityList(getSensitivityList(waitNode));

      /*
       * Create & fill process's control flow graph.
       */

      final Cfg waitCfg = new Cfg();
      final CfgNode src = waitCfg.getSource();
      final CfgNode sink = waitCfg.getSink();
      src.removeChild(sink);
      waitProcess.addChild(waitCfg);

      /* Start CFG building from creating the switch-case pair
       * that checks the related new variable. */

      final NodeVariable newVariable = this.waitNewVarMap.get(waitNode);
      final Node variableCondition =
        new NodeOperation(StandardOperation.EQ, newVariable, NodeValue.newBoolean(true));
      final Node condition =
          waitNode.getCondition() != null ?
            new NodeOperation(StandardOperation.AND, waitNode.getCondition(), variableCondition) :
            variableCondition;

      Switch processCfgMainSwitch = new Switch(condition);
      waitCfg.addNode(processCfgMainSwitch);
      src.addChild(processCfgMainSwitch);

      final Case aCase = new Case(NodeValue.newBoolean(true));
      waitCfg.addNode(aCase);
      processCfgMainSwitch.addChild(aCase);

      /*
       * Remove the link with wait node parent, & wait node itself.
       */

      final CfgNode waitNodeParent = (CfgNode) waitNode.getOnlyParent();

      if (waitNodeParent != null) {
        finishCfgPathWithTriggerBlock(cfg, waitNode, newVariable);
      }

      /*
       * Add nodes to the process control flow graph.
       */

      Set<CfgNode> nodes = this.map.get(waitNode);

      if (nodes.isEmpty()) {
        final NodeVariable waitVariable = this.waitParentNewVarMap.get(waitNode);
        final BasicBlock waitVarBlock = new BasicBlock(new Assignment(waitVariable, NodeValue.newBoolean(true)));
        waitCfg.addNode(waitVarBlock);
        waitVarBlock.addParent(aCase);
        waitVarBlock.addChild(sink);
      } else if (hasOnlySink(nodes)) {
        aCase.addChild(sink);
      } else {
        for (CfgNode node : nodes) {

          final Collection<CfgModelNode> nodeParents = node.getParents();

          if (this.nonWaitReachableNodes.contains(node)) {

          /* Create the copy & put it to the new process. */
            final CfgModelNode copy = node.deepcopy();
            waitCfg.addNode(copy);

            oldNewMap.put(node, copy);

          /* Try to find parent nodes that are already added to the new process. */

            for (CfgModelNode nodeParent : nodeParents) {
              if (oldNewMap.containsKey(nodeParent)) {
                copy.addParent(oldNewMap.get(nodeParent));
              } else if (nodeParent instanceof CfgNode && waitCfg.getNodes().contains(nodeParent)) {
                copy.addParent(nodeParent);
              }
            }

          } else {
          /* Move node to the new wait process. */
            cfg.removeNode(node);

            waitCfg.addNode(node);

            if (node.getOnlyParent().equals(waitNode)) {
              node.removeParent(waitNode);
              node.addParent(aCase);
            }
          }

          if (this.waitParentNewVarMap.containsKey(node)) {

          /* This node is parent for other wait, finish related sub-path with trigger block. */
            finishCfgPathWithTriggerBlock(
              waitCfg, (Wait) node.getOnlyChild(), this.waitParentNewVarMap.get(node));
          }
        }
      }

      this.waitProcessMap.get(this.module).add(waitProcess);
      cfg = waitProcess.getCfg();
    }
    this.map.clear();
    this.nonWaitReachableNodes.clear();
    this.nearestParentWaitMap.clear();
    this.waitParentNewVarMap.clear();
    this.waitNewVarMap.clear();
  }

  private EventList getSensitivityList(Wait waitNode) {
    if (!waitNode.getEventList().getEvents().isEmpty()) {
      return waitNode.getEventList().deepcopy();
    } else if (waitNode.getCondition() != null) {

      final Set<Event> events = new LinkedHashSet<>();
      Collection<NodeVariable> variables = ExprUtils.getVariables(waitNode.getCondition());

      for (NodeVariable variable : variables) {
        events.add(new Event(new RangedVariable(variable), EventType.ANY_EDGE));
      }
      return new EventList(events);
    } else
        throw new IllegalArgumentException(
            String.format("Empty node of type: %s.", waitNode.getClass()));
  }

  private boolean hasOnlySink(Set<CfgNode> nodes) {
    return nodes.size() == 1 && nodes.iterator().next() instanceof Sink;
  }

  private void finishCfgPathWithTriggerBlock(Cfg cfg, Wait waitNode, NodeVariable newVariable) {

    final CfgNode cfgSink = cfg.getSink();

    final CfgNode nodeParent = (CfgNode) waitNode.getOnlyParent();

    final BasicBlock trigger =
        new BasicBlock(new Assignment(newVariable, NodeValue.newBoolean(true)));
    cfg.addNode(trigger);

    if (nodeParent.hasChildren()) {
      nodeParent.removeChild(waitNode);
    }
    nodeParent.addChild(trigger);

    /*
     * Link trigger block with pre-sink merge node & create it if needed.
     */
    Merge preSinkMerge;
    final CfgNode sinkParent = (CfgNode) cfgSink.getOnlyParent();

    if (sinkParent != null && cfgSink.getOnlyParent() instanceof Merge) {

    /* Appropriate merge node has been found. */
      preSinkMerge = (Merge) cfgSink.getOnlyParent();
    } else {

    /* Insert merge that is a parent for sink node. */
      preSinkMerge = new Merge();

      if (sinkParent != null) {
        sinkParent.removeChild(cfgSink);
        sinkParent.addChild(preSinkMerge);
      }
      preSinkMerge.addChild(cfgSink);
      trigger.addChild(preSinkMerge);
    }
    cfg.addNode(preSinkMerge);
  }

  @Override
  public void onCfgBegin(Cfg cfg) { /* Empty. */ }

  @Override
  public void onCfgEnd(Cfg cfg) { /* Empty. */ }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { processNodeBegin(basicBlock); }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { processNodeEnd(basicBlock); }

  @Override
  public void onCaseBegin(Case aCase) { processNodeBegin(aCase); }

  @Override
  public void onCaseEnd(Case aCase) { processNodeEnd(aCase); }

  @Override
  public void onSwitchBegin(Switch aSwitch) { processNodeBegin(aSwitch); }

  @Override
  public void onSwitchEnd(Switch aSwitch) { processNodeEnd(aSwitch); }

  @Override
  public void onMergeBegin(Merge merge) { processNodeBegin(merge); }

  @Override
  public void onMergeEnd(Merge merge) { processNodeEnd(merge); }

  @Override
  public void onSrcBegin(Source source) { processNodeBegin(source); }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { processNodeBegin(sink); }

  @Override
  public void onWaitBegin(Wait wait) {

    processNodeBegin(wait);

    this.waitNode = wait;

    this.map.put(wait, new LinkedHashSet<CfgNode>());
  }

  @Override
  public void onWaitEnd(Wait wait) { this.waitNode = null; }

  private void processNodeBegin(CfgNode node) {
    if (this.waitNode == null) {
      this.nonWaitReachableNodes.add(node);
    } else  if (!(node instanceof Wait)){

      /* Put to wait-related nodes all children except wait children. */
      this.map.get(this.waitNode).add(node);
    }
    this.nearestParentWaitMap.put(node, this.waitNode);

    if (node.getChildren().size() == 1) {
      checkOneChildNode(node);
    }
  }

  private void checkOneChildNode(CfgNode node) {
    final CfgNode waitChild = (CfgNode) node.getOnlyChild();
    if (waitChild instanceof Wait) {

      /* Associate a new variable with this node. */
      if (!this.waitParentNewVarMap.containsKey(node)) {

        /* Create new variable that will be the trigger
         * for process extracted from wait-children sub-graph. */
        final int varNum = this.module.getVariables().size();
        final NodeVariable variable =
          new NodeVariable(CfgVariableContainerNode.NEW_VAR_PREFIX + varNum, DataType.BOOLEAN);

        variable.setUserData(new VariableData(VariableType.REG));
        this.module.addVariable(variable);

        this.waitParentNewVarMap.put(node, variable);
        this.waitNewVarMap.put((Wait) waitChild, variable);
      }
    }
  }

  private void processNodeEnd(CfgNode node) {
    this.waitNode = this.nearestParentWaitMap.get(node);
  }
}
