/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.parser.vhdl;

import java.util.LinkedHashMap;

import org.zamia.instgraph.IGOperationBinary;

import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * Mapping between IG binary operation identifiers and standard operation identifiers that are
 * supported.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VhdlBinaryMapping {
  private final static LinkedHashMap<IGOperationBinary.BinOp, Enum<?>> map;

  static {
    map = new LinkedHashMap<>();

    map.put(IGOperationBinary.BinOp.ADD, StandardOperation.ADD);
    map.put(IGOperationBinary.BinOp.AND, StandardOperation.AND);
    map.put(IGOperationBinary.BinOp.CONCAT, StandardOperation.BVCONCAT);
    map.put(IGOperationBinary.BinOp.DIV, StandardOperation.DIV);
    map.put(IGOperationBinary.BinOp.EQUAL, StandardOperation.EQ);
    map.put(IGOperationBinary.BinOp.GREATER, StandardOperation.GREATER);
    map.put(IGOperationBinary.BinOp.GREATEREQ, StandardOperation.GREATEREQ);
    map.put(IGOperationBinary.BinOp.LESS, StandardOperation.LESS);
    map.put(IGOperationBinary.BinOp.LESSEQ, StandardOperation.LESSEQ);
    map.put(IGOperationBinary.BinOp.MAX, StandardOperation.MAX);
    map.put(IGOperationBinary.BinOp.MIN, StandardOperation.MIN);
    map.put(IGOperationBinary.BinOp.MOD, StandardOperation.MOD);
    map.put(IGOperationBinary.BinOp.MUL, StandardOperation.MUL);
    map.put(IGOperationBinary.BinOp.NAND, StandardOperation.BVNAND);
    map.put(IGOperationBinary.BinOp.NEQUAL, StandardOperation.NOTEQ);
    map.put(IGOperationBinary.BinOp.NOR, StandardOperation.BVNOR);
    map.put(IGOperationBinary.BinOp.OR, StandardOperation.OR);
    map.put(IGOperationBinary.BinOp.POWER, StandardOperation.POWER);
    map.put(IGOperationBinary.BinOp.REM, StandardOperation.REM);
    map.put(IGOperationBinary.BinOp.ROL, StandardOperation.BVROL);
    map.put(IGOperationBinary.BinOp.ROR, StandardOperation.BVROR);
    map.put(IGOperationBinary.BinOp.SLA, StandardOperation.BVASHL);
    map.put(IGOperationBinary.BinOp.SLL, StandardOperation.BVLSHL);
    map.put(IGOperationBinary.BinOp.SRA, StandardOperation.BVASHR);
    map.put(IGOperationBinary.BinOp.SRL, StandardOperation.BVLSHR);
    map.put(IGOperationBinary.BinOp.SUB, StandardOperation.SUB);
    map.put(IGOperationBinary.BinOp.XNOR, StandardOperation.BVXNOR);
    map.put(IGOperationBinary.BinOp.XOR, StandardOperation.XOR);
  }

  /**
   * Checks rather operation is supported.
   *
   * If operation is supported, then it's identifier is in this mapping.
   *
   * @param operation IG operation identifier
   *
   * @return {@code true} if operation is supported, {@code false} otherwise.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public static boolean hasOperation(IGOperationBinary.BinOp operation) {
    InvariantChecks.checkNotNull(operation);

    return map.containsKey(operation);
  }

  /**
   * Gets identifier of supported operation.
   *
   * @param operation IG operation identifier
   *
   * @return operation identifier
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public static Enum<?> getOperation(IGOperationBinary.BinOp operation) {
    InvariantChecks.checkNotNull(operation);

    return map.get(operation);
  }
}
