/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.vhdl;

import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping for transforming operation codes for logic types
 * into operation codes for bit vector.
 * Mapping is used at VHDL expressions elaboration in
 * {@link VhdlIgOperationParser} component.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VhdlBitVectorOpcodeMapping {

  private static Map<Enum<?>, Enum<?>> map;

  static {
    map = new LinkedHashMap<>();

    map.put(StandardOperation.EQ, StandardOperation.EQ);
    map.put(StandardOperation.NOTEQ, StandardOperation.NOTEQ);
    map.put(StandardOperation.ADD, StandardOperation.BVADD);
    map.put(StandardOperation.SUB, StandardOperation.BVSUB);
    map.put(StandardOperation.AND, StandardOperation.BVAND);
    map.put(StandardOperation.OR, StandardOperation.BVOR);
    map.put(StandardOperation.XOR, StandardOperation.BVXOR);
    map.put(StandardOperation.NOT, StandardOperation.BVNOT);
    map.put(StandardOperation.MUL, StandardOperation.BVMUL);
    map.put(StandardOperation.MINUS, StandardOperation.BVNEG);
    map.put(StandardOperation.REM, StandardOperation.BVSREM);
    map.put(StandardOperation.MOD, StandardOperation.BVSMOD);
  }

  /**
   * Checks rather operation is supported.
   *
   * If operation is supported, then it's identifier is in this mapping.
   *
   * @param operation operation identifier
   *
   * @return {@code true} if operation is supported, {@code false} otherwise.
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public static boolean hasOperation(Enum<?> operation) {
    InvariantChecks.checkNotNull(operation);

    return map.containsKey(operation);
  }

  /**
   * Gets bit vector analogue operation code of the specified operation code.
   *
   * @param operation operation code
   *
   * @return corresponding operation code
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public static Enum<?> getOperation(Enum<?> operation) {
    InvariantChecks.checkNotNull(operation);

    return map.get(operation);
  }
}
