/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.ZamiaException;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAlias;
import org.zamia.instgraph.IGOperationAllocate;
import org.zamia.instgraph.IGOperationArrayAggregate;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGOperationBinary;
import org.zamia.instgraph.IGOperationDeref;
import org.zamia.instgraph.IGOperationError;
import org.zamia.instgraph.IGOperationIndex;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationLiteral;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGOperationPhi;
import org.zamia.instgraph.IGOperationRange;
import org.zamia.instgraph.IGOperationRecordAggregate;
import org.zamia.instgraph.IGOperationRecordField;
import org.zamia.instgraph.IGOperationTypeConversion;
import org.zamia.instgraph.IGOperationTypeQualification;
import org.zamia.instgraph.IGOperationUnary;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.Radix;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.Module;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * The VHDL parser {@link ru.ispras.retrascope.parser.vhdl.VhdlParser} component
 * for Zamia Instantiation Graph internal IGOperation objects processing.
 *
 * The component of VHDL descriptions parser {@link ru.ispras.retrascope.parser.vhdl.VhdlParser}
 * that elaborates {@link org.zamia.instgraph.IGOperation} objects.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VhdlIgOperationParser {

  /* The parent module. */
  private Module module;

  /**
   * Returns parent module object for the IG part
   * is elaborated by this object.
   *
   * @return parent module
   */
  public Module getParentModule() { return this.module; }

  /**
   * Sets parent module.
   *
   * @param module module
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setParentModule(Module module) {
    InvariantChecks.checkNotNull(module);

    this.module = module;
  }

  /**
   * Extracts {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list
   * from the specified {@link org.zamia.instgraph.IGSequentialWait}
   * wait statement object.
   *
   * @param stmt wait statement
   *
   * @return {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list
   */
  public EventList getEventList(IGSequentialWait stmt) {

    final EventList list = new EventList();

    ArrayList<IGOperation> sensList = stmt.getSensitivityList();

    if (sensList != null) {
      for (IGOperation current : sensList) {
        list.addEvents(getEvents(current));
      }
      IGOperation conditionClause = stmt.getConditionClause();
      if (conditionClause != null) {
        list.addEvents(getEvents(conditionClause));
      }
    }
    return list;
  }

  private Collection<Event> getEvents(IGOperation current) {

    Set<Event> events = new LinkedHashSet<>();
    RangedVariable rVar = this.getRangedVariable(current);

    if (current instanceof IGOperationInvokeSubprogram) {
      if (current.toHRString().equalsIgnoreCase(EventType.POSITIVE_EDGE.name())) {
        Event event = new Event(rVar, EventType.POSITIVE_EDGE);
        events.add(event);
      } else if (current.toHRString().equalsIgnoreCase(EventType.NEGATIVE_EDGE.name())) {
        Event event = new Event(rVar, EventType.NEGATIVE_EDGE);
        events.add(event);
      }
      /* Both edges are in sensitivity list. */
      else {
        events.add(new Event(rVar));
      }
      /* Both edges are in sensitivity list. */
    } else {
      events.add(new Event(rVar));
    }
    return events;
  }


  /**
   * Performs IGOperation into syntax expression node.
   *
   * @param op object to be performed
   *
   * @return syntax expression node
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   * @throws java.lang.UnsupportedOperationException when argument is of unsupported data type.
   */
  public VhdlIgOperationParserResult performIGOperation(IGOperation op) {
    InvariantChecks.checkNotNull(op);

    VhdlIgOperationParserResult result;

    if (op instanceof IGOperationAlias) {

      /* The only one operand is here. */
      result = performIGOperation(op.getOperand(0));
    } else if (op instanceof IGOperationAllocate) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
    } else if (op instanceof IGOperationArrayAggregate) {
      result = performArrayAggregate((IGOperationArrayAggregate) op);
    } else if (op instanceof IGOperationAttribute) {
      result = performAttribute((IGOperationAttribute) op);
    } else if (op instanceof IGOperationBinary) {
      result = performBinary((IGOperationBinary) op);
    } else if (op instanceof IGOperationDeref) {

      /* The only one operand is here. */
      result = performIGOperation(op.getOperand(0));
    } else if (op instanceof IGOperationError) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
    } else if (op instanceof IGOperationIndex) {
      result = performIndex((IGOperationIndex) op);
    } else if (op instanceof IGOperationInvokeSubprogram) {
      result = performSubprogram((IGOperationInvokeSubprogram) op);
    } else if (op instanceof IGOperationLiteral) {
      result = performLiteral((IGOperationLiteral) op);
    } else if (op instanceof IGOperationObject) {
      result = performObject((IGOperationObject) op);
    } else if (op instanceof IGOperationPhi) {
      result = performPhi((IGOperationPhi) op);
    } else if (op instanceof IGOperationRange) {
      result = performRange((IGOperationRange) op);
    } else if (op instanceof IGOperationUnary) {
      result = performUnaryOperation((IGOperationUnary) op);
    } else if (op instanceof IGStaticValue) {
      result = performStaticValue((IGStaticValue) op);
    } else if (op instanceof IGOperationRecordAggregate) {
      result = performRecordAggregate((IGOperationRecordAggregate) op);
    } else if (op instanceof IGOperationRecordField) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
    } else if (op instanceof IGOperationTypeConversion) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
    } else if (op instanceof IGOperationTypeQualification) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
    } else {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
    }
    return result;
  }

  /**
   * Extracts ranged variable from IGOperation object.
   *
   * @param op object that contains ranged variable
   *
   * @return ranged variable
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public RangedVariable getRangedVariable(IGOperation op) {
    InvariantChecks.checkNotNull(op);

    RangedVariable var = new RangedVariable();
    if (op instanceof IGOperationInvokeSubprogram) {
      int opNum = op.getNumOperands();
      if (opNum == 1) {
        var = getRangedVariable(op.getOperand(0));
      } else {
        throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
      }
    } else if (op instanceof IGOperationArrayAggregate) {
      throw new UnsupportedOperationException(
        String.format("Cannot parse: %s.", op.toHRString()));
    } else if (op instanceof IGOperationObject) {

      VariableType variableType;

      switch (((IGOperationObject) op).getObject().getCat()) {
        case SIGNAL:
          variableType = getVariableType(op);
          break;
        case VARIABLE:
          variableType = VariableType.REG;
          break;
        case CONSTANT:
        default:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", op.toHRString()));
      }

      DataType variableValueType;
      if (op.getType() != null) {

        variableValueType = getDataType(op.getType());

        NodeVariable variable =
            new NodeVariable(((IGOperationObject) op).getObject().getId(), variableValueType);
        VariableData variableData = new VariableData(variableType);
        variableData.addMetaInfo(MetaInfoType.HDL_TYPE_STR,
            ((IGOperationObject) op).getObject().getType().toHRString());
        variable.setUserData(variableData);

        Node invariant = getInvariant(variable, op.getType());

        Module parent = getParentModule();

        /* Add the variable declaration to parent module. */
        if (!parent.containsVariable(variable)) {
          parent.addVariable(variable, invariant);
        } else {
          variable = parent.getVariable(variable.getName());
        }

        var.setVariable(variable);
      }
    } else if (op instanceof IGOperationRecordField) {
      throw new UnsupportedOperationException(
        String.format("Cannot parse: %s.", op.toHRString()));
    } else if (op instanceof IGOperationIndex) {
      IGOperation igOperand = ((IGOperationIndex) op).getOperand();
      Node operand = performIGOperation(igOperand).getExpr();

      VariableType varType = null;
      try {
        switch (((IGOperationIndex) op).getOperand().getDirection()) {
          case IN:
            varType = VariableType.IN;
            break;
          case OUT:
            varType = VariableType.OUT;
            break;
          case INOUT:
            varType = VariableType.INOUT;
            break;
          default:
            varType = VariableType.REG;
            break;
        }
      } catch (ZamiaException e) {
        e.printStackTrace();
      }
      VariableData operandData = new VariableData(varType);
      operandData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, igOperand.getType().toHRString());
      operand.setUserData(operandData);

      Module parent = getParentModule();
      if (!parent.containsVariable((NodeVariable) operand)) {
        parent.addVariable((NodeVariable) operand, null);
      } else {
        operand = parent.getVariable(((NodeVariable) operand).getName());
      }

      var.setVariable((NodeVariable) operand);
      Node index = performIGOperation(((IGOperationIndex) op).getIndex()).getExpr();

      switch (index.getKind()) {
        case OPERATION:
        throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.toHRString()));
        case VARIABLE:
          var.setRange(new Range(index, index));
        case VALUE:
          var.setRange(new Range(index, index));
          break;
        default:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", op.toHRString()));
      }
    } else if (op instanceof IGOperationRange) {
      Node operand = performIGOperation(((IGOperationRange) op).getOperand()).getExpr();

      String name;
      DataType dataType;
      IGOperation opr;
      switch (operand.getKind()) {
        case OPERATION:
        case VALUE:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", op.toHRString()));
        case VARIABLE:
          name = ((NodeVariable) operand).getName();
          dataType = ((NodeVariable) operand).getData().getType();
          opr = ((IGOperationRange) op).getOperand();
          IGOperation range = ((IGOperationRange) op).getRange();

          Range operandRange;
          if (range instanceof IGRange) {
            IGOperation leftIndex = range.getOperand(0);
            IGOperation rightIndex = range.getOperand(2);
            Node left = performIGOperation(leftIndex).getExpr();
            Node right = performIGOperation(rightIndex).getExpr();
            operandRange = new Range(left, right);
          } else
            throw new UnsupportedOperationException(
              String.format("Cannot parse: %s.", range.toHRString()));
          var.setRange(operandRange);
          break;
        default:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", op.toHRString()));
      }
      NodeVariable variable = new NodeVariable(name, dataType);
      VariableData varData = new VariableData(getVarType(opr));
      varData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, op.getType().toHRString());
      variable.setUserData(varData);

      Module parent = getParentModule();
      if (!parent.containsVariable(variable)) {
        parent.addVariable(variable, null);
      } else {
        variable = parent.getVariable(variable.getName());
      }
      var.setVariable(variable);

    } else {
      throw new UnsupportedOperationException(
        String.format("Cannot parse: %s.", op.toHRString()));
    }
    return var;
  }

  /**
   * Extracts {@link ru.ispras.retrascope.model.basis.VariableType} variable type.
   *
   * @param op object to be analyzed
   * @return {@link ru.ispras.retrascope.model.basis.VariableType} variable type if succeeds,
   * {@code null} otherwise
   */
  public VariableType getVariableType(IGOperation op) {
    try {
      IGObject.OIDir direction = op.getDirection();

      switch (direction) {
        case NONE:
          return VariableType.REG;
        case IN:
        case LINKAGE:
          return VariableType.IN;
        case INOUT:
          return VariableType.INOUT;
        case OUT:
          return VariableType.OUT;
        default:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", op.toHRString()));
      }
    } catch (ZamiaException e) {
      e.printStackTrace();
    }
    return null;
  }

  /**
   * Extracts {@link ru.ispras.fortress.data.DataType} data type from IGType object.
   *
   * @param type object that contains data type
   *
   * @return data type
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public DataType getDataType(IGType type) {
    InvariantChecks.checkNotNull(type);

    DataType dataType;
    switch (type.getCat()) {
      case INTEGER:
        dataType = DataType.INTEGER;
        break;
      case REAL:
        dataType = DataType.REAL;
        break;
      case ENUM:
        if (type.isBit()) {
          dataType = DataType.BIT_VECTOR(1);
        } else if (type.isBool()) {
          dataType = DataType.BOOLEAN;
        } else if (type.isLogic()) {
          dataType = DataType.BOOLEAN;
        } else
            throw new UnsupportedOperationException(
                String.format("Cannot parse: %s.", type.toHRString()));
        break;
      case ARRAY:
        IGType elemType = type.getElementType();
        IGType indexType = type.getIndexType();

        if (elemType.isBit()) {

          /* Bit vector, calculate it's size. */
          if (indexType instanceof IGTypeStatic) {
            int size = getSize((IGTypeStatic) indexType);
            dataType = DataType.BIT_VECTOR(size);
          } else
            throw new UnsupportedOperationException(
              String.format("Cannot parse: %s.", type.toHRString()));

        } else if (elemType.isArray()) {
          dataType = DataType.MAP(DataType.INTEGER, getDataType(elemType));
        } else if (elemType.isInteger()) {
          dataType = DataType.MAP(DataType.INTEGER, DataType.INTEGER);
        } else if (elemType.isLogic()) {
          //TODO: add support 9-sign VHDL logic
          dataType = DataType.MAP(DataType.INTEGER, DataType.BOOLEAN);
        } else {
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", type.toHRString()));
        }
        break;
      default:
        throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", type.toHRString()));
    }
    return dataType;
  }

  /**
   * Returns size of the specified index type.
   *
   * @param indexType index type
   *
   * @throws java.lang.UnsupportedOperationException when index left/right borders
   * cannot be transformed into {@link ru.ispras.fortress.expression.NodeValue} objects.
   * @throws java.lang.UnsupportedOperationException when index left/right borders
   * are not of integer types.
   *
   * @return size
   */
  public int getSize(IGTypeStatic indexType) {
    IGStaticValue range = indexType.getStaticRange();
    Node left = performStaticValue(range.getLeft()).getExpr();
    Node right = performStaticValue(range.getRight()).getExpr();
    if (left.getKind() != Node.Kind.VALUE || right.getKind() != Node.Kind.VALUE) {
      throw new UnsupportedOperationException(
        String.format("Unknown types: %s, %s.",
          left.getKind().name(), right.getKind().name()));
    }
    if (((NodeValue) left).getData().getType() != DataType.INTEGER
      || ((NodeValue) right).getData().getType() != DataType.INTEGER) {
      throw new UnsupportedOperationException(
        String.format("Unknown types: %s, %s.",
          ((NodeValue) left).getData().getType().toString(),
          ((NodeValue) right).getData().getType().toString()));
    }

    return Math.abs(((BigInteger)((NodeValue) left).getValue()).intValue()
      - ((BigInteger) ((NodeValue) right).getValue()).intValue() + 1);
  }

  /**
   * Extracts invariant from IGType object.
   *
   * @param type object that contains type
   *
   * @return invariant
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Node getInvariant(NodeVariable variable, IGType type) {
    InvariantChecks.checkNotNull(type);

    /* One-bit-size vector. */
    if (type.isBit()) {
      return new NodeOperation(
          StandardOperation.OR,
          new NodeOperation(
              StandardOperation.EQ,
              variable,
              NodeValue.newBitVector(BitVector.TRUE)),
          new NodeOperation(
              StandardOperation.EQ,
              variable,
              NodeValue.newBitVector(BitVector.FALSE)));
    }

    /* Boolean. */
    if (type.isBool()) {
      return new NodeOperation(
          StandardOperation.OR,
          new NodeOperation(StandardOperation.EQ, variable, NodeValue.newBoolean(true)),
          new NodeOperation(StandardOperation.EQ, variable, NodeValue.newBoolean(false))
      );
    }

    IGOperation range = type.getRange();
    if (range == null) {
      return null;
    }
    if (range instanceof IGStaticValue) {

      /* The left/right borders. */
      Node left = performStaticValue(((IGStaticValue) range).getLeft()).getExpr();
      Node right = performStaticValue(((IGStaticValue) range).getRight()).getExpr();

      /* The flag that shows rather range is ascending or descending. */
      boolean isAscending = ((IGStaticValue) range).getAscending().isTrue();

      return getRangeInvariant(variable, left, right, isAscending);

    } else if (range instanceof IGRange) {
      Node left = performIGOperation(((IGRange) range).getLeft()).getExpr();
      Node right = performIGOperation(((IGRange) range).getRight()).getExpr();

      Node ascending = performIGOperation(((IGRange) range).getAscending()).getExpr();

      /* Determine rather ascending node is true or false. */

      NodeValue trueNode;

      if (ascending.isType(DataTypeId.BIT_VECTOR)) {
        trueNode = NodeValue.newBitVector(BitVector.TRUE);
      } else {
        trueNode = NodeValue.newBoolean(true);
      }

      NodeOperation checkAscending = new NodeOperation(StandardOperation.EQ, ascending, trueNode);

      boolean isAscending = ExprUtils.isSAT(checkAscending);

      return getRangeInvariant(variable, left, right, isAscending);
    } else {
      throw new ClassCastException(String.format("Cannot parse: %s.", range.getClass().toString()));
    }
  }

  private VariableType getVariableType(IGObject obj) {
    final  IGObject.OIDir direction = obj.getDirection();
    switch (direction) {
      case NONE:
        return VariableType.REG;
      case IN:
      case LINKAGE:
        return VariableType.IN;
      case INOUT:
        return VariableType.INOUT;
      case OUT:
        return VariableType.OUT;
      default:
        throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", obj.toString()));
    }
  }

  private Node getRangeInvariant(Node variable, Node left, Node right, boolean isAscending) {

    /* The identifiers for operation codes. */
    Enum<?> greaterEqOp, lessEqOp;

    if (left.isType(DataTypeId.BIT_VECTOR) && right.isType(DataTypeId.BIT_VECTOR)) {
      greaterEqOp = StandardOperation.BVUGE;
      lessEqOp = StandardOperation.BVULE;
    } else {
      greaterEqOp = StandardOperation.GREATEREQ;
      lessEqOp = StandardOperation.LESSEQ;
    }

    Node greater, lesser;

      /* Detect which border is greater. */
    if (isAscending) {
      lesser = left;
      greater = right;
    } else {
      lesser = right;
      greater = left;
    }

    return new NodeOperation(
      StandardOperation.AND,
      new NodeOperation(greaterEqOp, variable, lesser),
      new NodeOperation(lessEqOp, variable, greater));
  }

  /**
   * Extracts variable type from IGOperation object.
   *
   * @param op object that contains variable type
   *
   * @return variable type
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public VariableType getVarType(IGOperation op) {
    InvariantChecks.checkNotNull(op);

    try {
      switch (op.getDirection()) {
        case APPEND:
        case BUFFER:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", op.getDirection().toString()));
        case IN:
        case LINKAGE:
          return VariableType.IN;
        case OUT:
          return VariableType.OUT;
        case INOUT:
          return VariableType.INOUT;
        case NONE:
          return VariableType.REG;
        default:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", op.getDirection().toString()));
      }
    } catch (ZamiaException e) {
      e.printStackTrace();
    }
    throw new UnsupportedOperationException(
      String.format("Cannot parse: %s.", op.toHRString()));
  }

  /**
   * Performs the specified static value.
   *
   * @param staticValue static value
   *
   * @return the processing result
   */
  public VhdlIgOperationParserResult performStaticValue(IGStaticValue staticValue) {

    NodeValue value = null;
    IGType type = staticValue.getType();

    if (type.isBit()) {

      /* One-bit-size vector value. */
      value = NodeValue.newBitVector(BitVector.valueOf(staticValue.isTrue()));
    } else if (type.isBool()) {

      /* Boolean value. */
      value = NodeValue.newBoolean(staticValue.isTrue());
    } else if (type.isInteger()) {

      /* Integer value. */
      value = NodeValue.newInteger(staticValue.getInt());
    } else if (type.isReal()) {

      /* Real value. */
      value = NodeValue.newReal(staticValue.getReal().doubleValue());
    } else if (type.isLogic()) {

      //TODO: implement 9-sign VHDL logic
      value = NodeValue.newBoolean(staticValue.isTrue());
    }

    if (value == null)
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", staticValue.toHRString()));

    return new VhdlIgOperationParserResult(value);
  }

  /**
   * Performs the specified object into node variable.
   *
   * @param obj object
   * @return node variable
   */
  public VhdlIgOperationParserResult performIGObject(IGObject obj) {

    NodeVariable var = new NodeVariable(obj.getId(), getDataType(obj.getType()));
    var.setUserData(new VariableData(getVariableType(obj)));
    return new VhdlIgOperationParserResult(var);
  }

  private VhdlIgOperationParserResult performIndex(IGOperationIndex op) {

    Node resultNode;
    Node index = performIGOperation(op.getIndex()).getExpr();
    Node operand = performIGOperation(op.getOperand()).getExpr();

    if (operand.isType(DataTypeId.BIT_VECTOR)) {

      /* For parametric operations their parameters should go earlier than operands. */
      resultNode = new NodeOperation(StandardOperation.BVEXTRACT, index, index, operand);
    } else if (operand.isType(DataTypeId.MAP)) {
      resultNode = new NodeOperation(StandardOperation.SELECT, operand, index);
    } else
      throw new IllegalArgumentException(
          String.format("Cannot parse: %s.", operand.toString()));
    return new VhdlIgOperationParserResult(resultNode);
  }

  private VhdlIgOperationParserResult performRecordAggregate(IGOperationRecordAggregate agg) {

    throw new UnsupportedOperationException(
        String.format("Cannot parse: %s.", agg.toHRString()));
  }

  private VhdlIgOperationParserResult performArrayAggregate(IGOperationArrayAggregate agg) {

    DataType keyType = getDataType(agg.getType().getIndexType());
    if (keyType != DataType.INTEGER)
      throw new IllegalArgumentException();
    DataType valueType = getDataType(agg.getType().getElementType());
    DataMap dataMap = new DataMap(keyType, valueType);

    /* Check 'others' element. */
    int offset = 0;
    if (agg.getOperand(0) == null) {
      offset = 1;
    }

    for (int i = offset; i < agg.getNumOperands(); i++) {
        Node value =
          Transformer.reduce(
            ReduceOptions.NEW_INSTANCE,
            performIGOperation(agg.getOperand(i)).getExpr());

      if (value instanceof NodeValue) {
        dataMap.put(Data.newInteger(i - 1), ((NodeValue) value).getData());
      } else
        throw new IllegalArgumentException(String.format("Cannot parse: %s", value.toString()));
    }

    return new VhdlIgOperationParserResult(new NodeValue(Data.newArray(dataMap)));
  }

  private VhdlIgOperationParserResult performAttribute(IGOperationAttribute attribute) {

    switch (attribute.getAttrOp()) {
      case EVENT:
        IGItem item = attribute.getItem();

        if (item instanceof IGOperation) {
          RangedVariable var = getRangedVariable((IGOperation) item);

          Event event = new Event(var);

          return new VhdlIgOperationParserResult(event);
        }
        break;
      default:
        throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", attribute.toHRString()));
    }
    throw new UnsupportedOperationException(
        String.format("Cannot parse: %s.", attribute.toHRString()));
  }

  private VhdlIgOperationParserResult performLiteral(IGOperationLiteral operation) {
    NodeValue value = null;
    try {
      IGStaticValue opStaticValue = operation.computeStaticValue(null);
      switch (opStaticValue.getStaticType().getCat()) {
        case INTEGER:
          value = new NodeValue(Data.newInteger(opStaticValue.getInt()));
          break;
        case REAL:
          value =  new NodeValue(Data.newReal(opStaticValue.getReal().doubleValue()));
          break;
        case ARRAY:

          // TODO: change when 'Z' values will be supported
          String strVector = correct(opStaticValue.toString());

          value = new NodeValue(Data.newBitVector(strVector, Radix.BIN.value(), opStaticValue
              .toHRString().length()));
          break;
        default:
          throw new UnsupportedOperationException(
              String.format("Cannot parse: %s.", operation.toHRString()));
      }
    } catch (ZamiaException e) {
      e.printStackTrace();
    }
    return new VhdlIgOperationParserResult(value);
  }

  private VhdlIgOperationParserResult performObject(IGOperationObject operation) {

    Node node;
    IGObject.IGObjectCat category = operation.getObject().getCat();
    IGType opType = operation.getType();

    if ((category.equals(IGObject.IGObjectCat.SIGNAL)) ||
      (category.equals(IGObject.IGObjectCat.VARIABLE))) {
      DataType objectType = null;

      switch (opType.getCat()) {
        case INTEGER:
          objectType = DataType.INTEGER;
          break;
        case ENUM:
          if (opType.isBit()) {
            objectType = DataType.BIT_VECTOR(1);
          } else if (opType.isBool()) {
            objectType = DataType.BOOLEAN;
          }
          break;
        case REAL:
          objectType = DataType.REAL;
          break;
        case ARRAY:

          IGType elementType = opType.getElementType();

          if (elementType.isBit()) {

            /* Bit vector is detected. */
            int vectorSize = getSize((IGTypeStatic) opType.getIndexType());
            objectType = DataType.BIT_VECTOR(vectorSize);
          } else {

            /*Array is detected. */
            objectType = DataType.MAP(DataType.INTEGER, getDataType(elementType));
          }
          break;
        default:
          throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", operation.toHRString()));
      }
      VariableType type = getVarType(operation);
      node = new NodeVariable(operation.getObject().getId(), objectType);
      VariableData nodeData = new VariableData(type);
      nodeData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, opType.toHRString());
      node.setUserData(nodeData);

      /* Perform object invariant, if exists. */
      Node invariant = getInvariant((NodeVariable) node, opType);

      Module parent = getParentModule();
      if (!parent.containsVariable((NodeVariable) node)) {
        parent.addVariable((NodeVariable) node, invariant);
      }

    } else if (category.equals(IGObject.IGObjectCat.CONSTANT)) {
      IGOperation initValue = operation.getObject().getInitialValue();
      node = performIGOperation(initValue).getExpr();
    } else {
      throw new IllegalArgumentException(
          String.format("Cannot parse: %s.", operation.toHRString()));
    }
    return new VhdlIgOperationParserResult(node);
  }

  private VhdlIgOperationParserResult performUnaryOperation(IGOperationUnary operation) {
    Enum<?> opId;
    switch (operation.getUnaryOp()) {
      case ABS:
        opId = StandardOperation.ABS;
        break;
      case NEG:
        opId = StandardOperation.MINUS;
        break;
      case BUF:
        opId = StandardOperation.ADD;
        break;
      case NOT:
        opId = StandardOperation.NOT;
        break;
      default:
        throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", operation.toHRString()));
    }
    Node operandElem = performIGOperation(operation.getA()).getExpr();
    return new VhdlIgOperationParserResult(new NodeOperation(opId, operandElem));
  }

  private VhdlIgOperationParserResult performRange(IGOperationRange operation) {
    Node operand = performIGOperation(operation.getOperand()).getExpr();
    InvariantChecks.checkNotNull(operand);

    IGOperation range = operation.getRange();

    if (range instanceof IGRange) {
      IGOperation isAscending = ((IGRange) range).getAscending();

      Node left = performIGOperation(((IGRange) range).getLeft()).getExpr();
      Node right = performIGOperation(((IGRange) range).getRight()).getExpr();

      Node firstIdx, secondIdx; // convenience variables for making code shorter
      if (((IGStaticValue) isAscending).isTrue()) {
        firstIdx = right;
        secondIdx = left;
      } else {
        firstIdx = left;
        secondIdx = right;
      }
      return new VhdlIgOperationParserResult(
          new NodeOperation(StandardOperation.BVEXTRACT, firstIdx, secondIdx, operand));
    } else {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", operation.toHRString()));
    }
  }

  private VhdlIgOperationParserResult performPhi(IGOperationPhi operation) {
    Node condition = performIGOperation(operation.getOperand(0)).getExpr();
    Node firstVariant = performIGOperation(operation.getOperand(1)).getExpr();
    Node secondVariant = performIGOperation(operation.getOperand(2)).getExpr();

    return new VhdlIgOperationParserResult(
        new NodeOperation(StandardOperation.ITE, condition, firstVariant, secondVariant));
  }

  private VhdlIgOperationParserResult performSubprogram(IGOperationInvokeSubprogram operation) {
    int i;
    String subProgramName = operation.getSub().getId();

    int operandNum = operation.getNumOperands();

    Enum<?> opCode = getBuiltinOpType(subProgramName);

    List<Node> nodes = new ArrayList<>();
    for (i = 0; i < operandNum; i++) {
      Node node = performIGOperation(operation.getOperand(i)).getExpr();
      if (node != null) {
        nodes.add(node);
      }
    }

    Node[] resultTemplate = new Node[nodes.size()];
    if (opCode == null) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", operation.toHRString()));
    }

    /* HACK: substitute binary operation identifiers to unary if needed. */
    if (nodes.size() == 1) {
      if (opCode.equals(StandardOperation.ADD)) {
        opCode = StandardOperation.PLUS;
      } else if (opCode.equals(StandardOperation.SUB)) {
        opCode = StandardOperation.MINUS;
      }
    }

    /* Substitute operation to bit-vector-analogue if at least one operand is bit vector. */
    boolean hasBvOperand = false;
    for (Node node : nodes) {
      if (node.isType(DataTypeId.BIT_VECTOR)) {
        hasBvOperand = true;
        break;
      }
    }
    if (hasBvOperand) {
      if (VhdlBitVectorOpcodeMapping.hasOperation(opCode)) {
        opCode = VhdlBitVectorOpcodeMapping.getOperation(opCode);
      }
    }

    return new VhdlIgOperationParserResult(
        new NodeOperation(opCode, nodes.toArray(resultTemplate)));
  }

  private VhdlIgOperationParserResult performBinary(IGOperationBinary operation) {

    Enum<?> type = getBinaryOpType(operation);

    Node left = performIGOperation(operation.getA()).getExpr();
    Node right = performIGOperation(operation.getB()).getExpr();

    return new VhdlIgOperationParserResult(new NodeOperation(type, left, right));
  }

  private Enum<?> getBinaryOpType(IGOperationBinary op) {

    if (!VhdlBinaryMapping.hasOperation(op.getBinOp())) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", op.getBinOp().name()));
    } else {
      return VhdlBinaryMapping.getOperation(op.getBinOp());
    }
  }

  private Enum<?> getBuiltinOpType(String opName) {

    char doubleQuote = '"';
    int dqIndex = opName.indexOf(doubleQuote);
    String substring =
        (dqIndex < 0) ? opName : opName.substring(dqIndex + 1, opName.lastIndexOf(doubleQuote));

    if (VhdlSubprogramMapping.hasOperation(substring)) {
      return VhdlSubprogramMapping.getOperation(substring);
    }
    throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", opName));
  }

  private String correct(String s) {

    char[] array = s.toCharArray();
    for (int i = 0; i < s.length(); i++) {
      if ((array[i] == '-') || (array[i] == 'Z')) {
        array[i] = '0';
      }
    }
    return new String(array);
  }
}
