/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.ZamiaProject;
import org.zamia.instgraph.IGConcurrentStatement;
import org.zamia.instgraph.IGInstantiation;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGMapping;
import org.zamia.instgraph.IGModule;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStructure;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Parser of IG (Instantiation Graph) representation of VHDL.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VhdlIgParser {

  private VhdlIgSequentialStatementParser seqParser; // sequential statement sub-parser

  /**
   * Default constructor
   */
  public VhdlIgParser() { this.seqParser = new VhdlIgSequentialStatementParser(); }

  private final static String DEFAULT_MODULE_NAME = ""; // default module name

  /**
   * Extracts module from Zamia IG representation.
   *
   * @param fZPrj Zamia project
   * @param igModule Zamia IG module
   *
   * @throws java.lang.NullPointerException when any of arguments is {@code null}.
   *
   * @return module
   */
  public Module parseModule(ZamiaProject fZPrj, IGModule igModule) {
    InvariantChecks.checkNotNull(fZPrj);
    InvariantChecks.checkNotNull(igModule);

    Module module = new Module(igModule.getDUUID().toCompactString());
    module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    this.seqParser.setModule(module);

    IGStructure structure = igModule.getStructure();

    if (structure != null) {
      for (int i = 0; i < structure.getNumStatements(); i++) {
        IGConcurrentStatement concurrent = structure.getStatement(i);
        module.addChild(
            getGraphNode(
                fZPrj,
                structure,
                concurrent,
                getConcurrentStatementType(concurrent)));
      }
    }
    return module;
  }

  private CfgNodeType getConcurrentStatementType(IGConcurrentStatement stmt) {

    if (stmt instanceof IGInstantiation)
      return CfgNodeType.INSTANCE;

    else if (stmt instanceof IGProcess)
      return CfgNodeType.PROCESS;

    else if (stmt instanceof IGStructure)
      return CfgNodeType.MODULE;

    else {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toString()));
    }
  }

  private CfgModelNode getGraphNode(ZamiaProject fZPrj, IGStructure structure,
      IGConcurrentStatement stmt, CfgNodeType nodeKind) {
    switch (nodeKind) {
      case INSTANCE:

        /* Extract the module name and the current instance name. */
        IGManager manager = fZPrj.getIGM();
        IGModule instModule = manager.findModule(((IGInstantiation) stmt).getSignature());
        Module module = parseModule(fZPrj, instModule);
        this.seqParser.setModule(module);

        /* Extract bindings of the current instance. */
        Map<String, Node> bindings = getBindings((IGInstantiation) stmt);

        /* Instantiate the extracted module. */
        String instanceName = stmt.getLabel();
        module.instantiate(instanceName, bindings);
        return module;
      case PROCESS:
        Process process = new Process();
        this.seqParser.setProcess(process);

        /* Extract process sensitivity list events. */
        EventList sensList = getProcessEventList((IGProcess) stmt);
        process.setSensitivityList(sensList);

        IGSequenceOfStatements sos = ((IGProcess) stmt).getSequenceOfStatements();
        Cfg cfg = this.seqParser.getCfg(sos);
        process.addChild(cfg);
        return process;
      case MODULE:
        Module blockModule = new Module(DEFAULT_MODULE_NAME);
        blockModule.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
        this.seqParser.setModule(blockModule);

        for (int i = 0; i < structure.getNumStatements(); i++) {
          IGConcurrentStatement concurrent = structure.getStatement(i);
          blockModule.addChild(getGraphNode(fZPrj, structure, concurrent, nodeKind));
        }
        return blockModule;
      default:
        throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", nodeKind.toString()));
    }
  }

  private EventList getProcessEventList(IGProcess stmt) {

    EventList result = new EventList();

    IGSequenceOfStatements sos = stmt.getSequenceOfStatements();

    int sosStmtNum = sos.getNumStatements();
    int i = 0;

    while (i < sosStmtNum) {

      IGSequentialStatement sosStmt = sos.getStatement(i);

      if (sosStmt instanceof IGSequentialWait) {

        result.addEvents(this.seqParser.getOpParser().getEventList((IGSequentialWait) sosStmt).getEvents());
      }
      i++;
    }
    return result;
  }

  private Map<String, Node> getBindings(IGInstantiation stmt) {
    Map<String, Node> map = new LinkedHashMap<>();

    int mappingNum = stmt.getNumMappings();
    int i = 0;

    while (i < mappingNum) {
      IGMapping mapping = stmt.getMapping(i);

      /* Extract the variable binding. */
      IGOperation formal = mapping.getFormal();

      String varName;

      if (formal instanceof IGOperationObject) {
        varName = ((IGOperationObject) formal).getObject().getId();
      } else {
        throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", formal.toHRString()));
      }

      /* Extract the binding's target expression. */
      Node expression =
          this.seqParser.getOpParser().performIGOperation(mapping.getActual()).getExpr();

      map.put(varName, expression);
      i++;
    }
    return map;
  }
}
