/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialAssert;
import org.zamia.instgraph.IGSequentialAssignment;
import org.zamia.instgraph.IGSequentialExit;
import org.zamia.instgraph.IGSequentialIf;
import org.zamia.instgraph.IGSequentialLoop;
import org.zamia.instgraph.IGSequentialNext;
import org.zamia.instgraph.IGSequentialProcedureCall;
import org.zamia.instgraph.IGSequentialReport;
import org.zamia.instgraph.IGSequentialRestart;
import org.zamia.instgraph.IGSequentialReturn;
import org.zamia.instgraph.IGSequentialSetConnection;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgVariableContainerNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.Collection;

/**
 * The VHDL sequential statements parser is a component
 * of {@link ru.ispras.retrascope.parser.vhdl.VhdlParser} parser.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VhdlIgSequentialStatementParser {

  /* The IGOperation objects sub-parser. */
  private VhdlIgOperationParser opParser;

  /* The parent process. */
  private Process process;

  /**
   * Constructs parser object.
   */
  public VhdlIgSequentialStatementParser() { this.opParser = new VhdlIgOperationParser(); }

  /**
   * Sets parent process.
   *
   * The specified process will be set as parent also for operation sub-parser.
   *
   * @param process process
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public void setProcess(Process process) {
    InvariantChecks.checkNotNull(process);

    this.process = process;
  }

  public void setModule(Module module) {
    InvariantChecks.checkNotNull(module);

    this.opParser.setParentModule(module);
  }

  /**
   *  Returns operation parser.
   *
   * @return parser of internal Zamia Instantiation Graph IGOperation object.
   */
  public VhdlIgOperationParser getOpParser() { return this.opParser; }

  /**
   * Returns control flow graph of the specified sequence of statements.
   *
   * @param sos sequence of statements
   *
   * @return control flow graph
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public Cfg getCfg(IGSequenceOfStatements sos) {
    InvariantChecks.checkNotNull(sos);

    Cfg cfg = new Cfg();

    int sosSize = sos.getNumStatements();
    int i = 0;
    CfgModelNode node;
    while (i < sosSize) {

      node = getNode(sos.getStatement(i));

      if (node != null) {
        addNode(cfg, node);
      }
      i++;
    }
    return cfg;
  }

  private void addNode(Cfg cfg, CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      CfgNode cfgSink = cfg.getSink();
      CfgModelNode cfgSinkParent = cfgSink.getOnlyParent();

      /*
       * If the extracted basic block node has basic block as parent,
       * and both of them are concurrent, then they will be merged into one node
       */
      if ((cfgSinkParent.getType().equals(CfgNodeType.BASIC_BLOCK))
        && (node.getType().equals(CfgNodeType.BASIC_BLOCK))
        && ((BasicBlock)node).isConcurrent() && ((BasicBlock)cfgSinkParent).isConcurrent()) {
        ((BasicBlock)cfgSinkParent).appendAssignments(((BasicBlock) node).getAssignments());

      } else {
        cfgSink.removeParent(cfgSinkParent);
        cfgSinkParent.addChild(node);
        node.addChild(cfgSink);
        cfg.addNode(node);
      }
    } else if (node instanceof Cfg) {
      cfg.append((Cfg)node);
    } else
      throw new IllegalArgumentException(
          String.format("Incorrect node type: %s.", node.getType()));
  }

  /**
   * Extracts control flow graph model node from the specified sequence of statements.
   *
   * @param stmt sequential statement
   * @return control flow graph model node
   */
  public CfgModelNode getNode(IGSequentialStatement stmt) {

    CfgModelNode node;

    if (stmt instanceof IGSequenceOfStatements) {
      node = getCfg((IGSequenceOfStatements) stmt);
    } else if (stmt instanceof IGSequentialIf) {
      node = getIfNode((IGSequentialIf) stmt);
    } else if (stmt instanceof IGSequentialSetConnection) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toHRString()));
    } else if (stmt instanceof IGSequentialExit) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toHRString()));
    } else if (stmt instanceof IGSequentialReport) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toHRString()));
    } else if (stmt instanceof IGSequentialAssert) {
      node = getAssert((IGSequentialAssert) stmt);
    } else if (stmt instanceof IGSequentialAssignment) {
      node = getBasicBlock((IGSequentialAssignment) stmt);
    } else if (stmt instanceof IGSequentialNext) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toHRString()));
    } else if (stmt instanceof IGSequentialLoop) {
      node = getLoop((IGSequentialLoop) stmt);
    } else if (stmt instanceof IGSequentialProcedureCall) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toHRString()));
    } else if (stmt instanceof IGSequentialReturn) {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toHRString()));
    } else if (stmt instanceof IGSequentialWait) {
      node = getWait((IGSequentialWait)stmt);
    } else if (stmt instanceof IGSequentialRestart) {
      node = null;
    } else {
      throw new UnsupportedOperationException(
          String.format("Cannot parse: %s.", stmt.toHRString()));
    }
    return node;
  }

  private CfgModelNode getWait(IGSequentialWait stmt) {

//    Node condition = null;
//
//    final EventList sensitivityList = this.opParser.getEventList(stmt);
//
//    if (stmt.getConditionClause() != null) {
//      condition = this.opParser.performIGOperation(stmt.getConditionClause()).getExpr();
//
//      if (condition == null)
//        throw new UnsupportedOperationException(
//            String.format("Cannot parse: %s.", stmt.toHRString()));
//    }
//
//    return new Wait(sensitivityList, condition);
    return null;
  }

  private CfgModelNode getAssert(IGSequentialAssert stmt) {

    IGOperation op = stmt.getOp();

    /* Check the severity level - NOTE, WARNING or FAILURE, ERROR. */
    String sevString = stmt.getSeverity().toHRString();

    String varSeverity;
    if (sevString.equalsIgnoreCase("NOTE"))
      varSeverity = "WARN";
    else if (sevString.equalsIgnoreCase("WARNING"))
      varSeverity = "WARN";
    else if (sevString.equalsIgnoreCase("ERROR"))
      varSeverity = "ERROR";
    else if (sevString.equalsIgnoreCase("FAILURE"))
      varSeverity = "ERROR";
    else
      throw new IllegalArgumentException(
        String.format("Severity level: %s.", sevString));
    varSeverity += "_";

    /* Depending on the severity level the new variable is created. */
    NodeVariable errorVariable =
        new NodeVariable(CfgVariableContainerNode.NEW_VAR_PREFIX + varSeverity + op.toHRString(),
            DataType.BIT_VECTOR(1));
    VariableData errorVarData = new VariableData(VariableType.REG);

    errorVarData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, "BIT");
    errorVariable.setUserData(errorVarData);

    /* Declare new variable to the parent module. */
    Collection<CfgModelNode> processParents = this.process.getParents();

    int pSize = processParents.size();
    if (pSize == 1) {
      Module parent = (Module) this.process.getOnlyParent();
      parent.addVariable(errorVariable);
    } else
      throw new IllegalStateException(
          String.format("The size of parents collection for process: %d.", pSize));

    /* Insert top-level checking to the parent process control flow graph. */
    Source src = this.process.getCfg().getSource();
    Sink sink = this.process.getCfg().getSink();

    /* Take the source node child. */
    CfgModelNode srcChild = src.getOnlyChild();
    src.removeChild(srcChild);

    NodeValue isUnInitialized = errorVariable.getData().hasValue() ?
        NodeValue.newBoolean(false) :
        NodeValue.newBoolean(true);

    Switch errorSwitch = new Switch(
        new NodeOperation(
            StandardOperation.OR,
            new NodeOperation(StandardOperation.EQ, errorVariable, NodeValue.newBoolean(false)),
            isUnInitialized));

    Case errorTrueCase = new Case(NodeValue.newBoolean(true));
    errorSwitch.addChild(errorTrueCase);
    errorTrueCase.addChild(srcChild);
    Case errorFalseCase = new Case(NodeValue.newBoolean(false));
    errorSwitch.addChild(errorFalseCase);
    errorFalseCase.addChild(sink);

    /* Transform assert node into sub-graph. */
    Cfg assertCfg = new Cfg();
    CfgNode asSrc = assertCfg.getSource();
    CfgNode asSink = assertCfg.getSink();
    asSrc.removeChild(asSink);

    Node opExpr = this.opParser.performIGOperation(op).getExpr();
    Switch sw = new Switch(opExpr);
    Case falseCase = new Case();
    falseCase.setDefault();
    BasicBlock errorBlock = new BasicBlock(
        new Assignment(
            errorVariable,
            NodeValue.newBoolean(true)));
    falseCase.addChild(errorBlock);
    Case trueCase = new Case(NodeValue.newBoolean(true));

    asSrc.addChild(sw);
    sw.addChild(falseCase);
    sw.addChild(trueCase);
    trueCase.addChild(asSink);
    errorBlock.addChild(asSink);

    return assertCfg;
  }

  private CfgModelNode getBasicBlock(IGSequentialAssignment stmt) {

    CfgModelNode node = null;

    IGOperation target = stmt.getTarget();

    VariableType targetType = this.opParser.getVariableType(target);

    if (targetType != null) {

      IGOperation value = stmt.getValue();
      Node valueExpr = this.opParser.performIGOperation(value).getExpr();
      node = getBasicBlock(target, valueExpr);
    }

    return node;
  }

  private BasicBlock getBasicBlock(IGOperation target, Node valueExpr) {
    RangedVariable rangedVariable = this.opParser.getRangedVariable(target);

    Assignment assignment = new Assignment(rangedVariable, valueExpr);

    /* If the variable is of signal type, then assignment is concurrent. */
    boolean isConcurrent = VariableType.isSignal(
      ((VariableData) rangedVariable.getVariable().getUserData()).getVariableType());

    return new BasicBlock(assignment, isConcurrent);
  }

  private CfgModelNode getLoop(IGSequentialLoop stmt) {

    /* The range iteration variable. */
    NodeVariable rangeVariable =
        (NodeVariable) this.opParser.performIGObject((IGObject)stmt.getChild(1)).getExpr();

    /* The loop condition. */
    IGRange igRange = (IGRange) stmt.getChild(0);
    Node left = opParser.performIGOperation(igRange.getLeft()).getExpr();
    Node right = opParser.performIGOperation(igRange.getRight()).getExpr();
    boolean isAscending =
        opParser.performIGOperation(
          igRange.getAscending()).getExpr().equals(NodeValue.newBoolean(true));

    Node range = createRange(rangeVariable, left, right, isAscending);

    IGSequenceOfStatements body = (IGSequenceOfStatements) stmt.getChild(4);

    /* Create the loop sub-graph. */
    Cfg loopCfg = new Cfg();
    Source loopCfgSrc = loopCfg.getSource();
    Sink loopCfgSink = loopCfg.getSink();
    loopCfgSrc.removeChild(loopCfgSink);

    /* Create the "i := 0" initialization block. */
    BasicBlock initBlock = new BasicBlock(new Assignment(rangeVariable, left));
    initBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INIT);
    loopCfgSrc.addChild(initBlock);

    /* Create the 'entry' merge block. */
    Merge merge = new Merge();
    merge.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_ENTRY);
    initBlock.addChild(merge);

    /* Create the condition switch block. */
    Switch switchBlock = new Switch(range);
    switchBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_COND);

    Case caseBlock, negCaseBlock;

    caseBlock = new Case(NodeValue.newBoolean(true));
    switchBlock.addChild(caseBlock);
    switchBlock.addParent(merge);

    Cfg subCfg = getCfg(body);
    subCfg.addAsSource(caseBlock);

    /* The incrementing (i++) basic block. */

    Enum<?> addOp;
    NodeValue one;

    if (rangeVariable.isType(DataTypeId.LOGIC_INTEGER)) {
      addOp = StandardOperation.ADD;
      one = NodeValue.newInteger(1);
    } else
        throw new IllegalArgumentException(
            String.format(
                "Non-synthesizable loop parameter type: %s.",
                rangeVariable.getDataTypeId().toString()));

    BasicBlock incrementBlock =
        new BasicBlock(
            new Assignment(
                rangeVariable,
                new NodeOperation(addOp, rangeVariable, one)));
    incrementBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INC);
    incrementBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_FIN);
    subCfg.addAsSink(incrementBlock);
    incrementBlock.addChild(merge);

    negCaseBlock = new Case(NodeValue.newBoolean(false));
    switchBlock.addChild(negCaseBlock);

    Merge finalMerge = new Merge();
    negCaseBlock.addChild(finalMerge);
    finalMerge.addChild(loopCfgSink);

    /* Add nodes to loop control flow sub-graph. */

    loopCfg.addNode(negCaseBlock);
    loopCfg.addNode(incrementBlock);
    loopCfg.addNode(switchBlock);
    loopCfg.addNode(merge);
    loopCfg.addNodes(subCfg.getNodes());
    loopCfg.addNode(caseBlock);

    return loopCfg;
  }

  private Node createRange(NodeVariable var, Node left, Node right, boolean isAscending) {

    /* Select operation ids depending on data types. */
    Enum<?> greaterEqOp = StandardOperation.GREATEREQ;
    Enum<?> lessEqOp = StandardOperation.LESSEQ;
    if (left.isType(DataTypeId.BIT_VECTOR) && right.isType(DataTypeId.BIT_VECTOR)) {
      greaterEqOp = StandardOperation.BVUGE;
      lessEqOp = StandardOperation.BVULE;
    }

    Node expr;
    if (isAscending) {

      /* The result is: left <= var <= right. */
      expr =
        new NodeOperation(
          StandardOperation.AND,
          new NodeOperation(greaterEqOp, var, left),
          new NodeOperation(lessEqOp, var, right));
    } else {

      /* The result is: left >= var >= right. */
      expr =
        new NodeOperation(
          StandardOperation.AND,
          new NodeOperation(lessEqOp, var, left),
          new NodeOperation(greaterEqOp, var, right));
    }

    return expr;
  }

  private CfgModelNode getIfNode(IGSequentialIf stmt) {

    Cfg ifCfg = new Cfg();
    Switch switchBlock;
    Merge merge = new Merge();

    IGOperation ifCondition = stmt.getCond();
    IGSequenceOfStatements elseSOS = stmt.getElseSOS();
    IGSequenceOfStatements thenSos = stmt.getThenSOS();

    Node condition;

    /* Elaborate conditions like "x'event && x='0'". */

    int eventIdx = -1;
    int opNum = ifCondition.getNumOperands();
    IGOperation[] operands = new IGOperation[opNum];

    for (int i = 0; i < opNum; i++) {
      operands[i] = ifCondition.getOperand(i);
      if (operands[i] instanceof IGOperationAttribute) {
        eventIdx = i;
      }
    }

    if (eventIdx >= 0) {

      /* There is an event in condition. */
      RangedVariable eventVariable = null;
      NodeOperation varExpr = null;
      EventType type = EventType.ANY_EDGE;

      if (opNum == 2) {
        for (int i = 0; i < opNum; i++) {
          if (operands[i] instanceof IGOperationAttribute) {
            eventVariable =
                ((Event) this.opParser.
                    performIGOperation(operands[i]).getData()).getRangedVariable();
          } else {
            varExpr = (NodeOperation) this.opParser.performIGOperation(operands[i]).getExpr();
          }
        }
      } else
        throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", ifCondition.toHRString()));

      if (eventVariable != null && varExpr != null) {


        Enum<?> opId = varExpr.getOperationId();
        if (opId.equals(StandardOperation.EQ)) {
          int opCount = varExpr.getOperandCount();
          for (int i = 0; i < opCount; i++) {
            Node operand = varExpr.getOperand(i);
            if (operand instanceof NodeValue) {
              if (operand.equals(NodeValue.newBitVector(BitVector.FALSE))) {
                type = EventType.NEGATIVE_EDGE;
              } else if (operand.equals(NodeValue.newBitVector(BitVector.TRUE))) {
                type = EventType.POSITIVE_EDGE;
              }
            } else if (!(operand instanceof NodeVariable) ||
                (!operand.equals(eventVariable.getVariable())))
              throw new UnsupportedOperationException(
                  String.format("Cannot parse: %s.", ifCondition.toHRString()));
          }
        } else
          throw new UnsupportedOperationException(
              String.format("Cannot parse: %s.", ifCondition.toHRString()));
      } else
        throw new UnsupportedOperationException(
            String.format("Cannot parse: %s.", ifCondition.toHRString()));
      switchBlock = new Switch(new Event(eventVariable, type));

      /* Elaborate 'then' branch. */
      processThenBranch(ifCfg, switchBlock, merge, thenSos);

      if (elseSOS != null) {
        Log.getLogger().log(
            LogLevel.WARNING,
            "DEAD CODE: Non-empty else-branch for switch: %s.",
            ifCondition.toHRString());
      }
    } else {

      /* No events are found in condition. */

      condition = this.opParser.performIGOperation(ifCondition).getExpr();
      switchBlock = new Switch(condition);

      /* Elaborate 'then' branch. */
      processThenBranch(ifCfg, switchBlock, merge, thenSos);

      /* Elaborate 'else' branch. */
      processElseBranch(ifCfg, switchBlock, merge, elseSOS);

    }
    /* Remove old link between SOURCE and SINK
     * and add link between SOURCE and top condition
     * and between merge node and SINK.
     */
    CfgNode source = ifCfg.getSource();
    CfgNode sink = ifCfg.getSink();
    source.removeChild(sink);
    source.addChild(switchBlock);
    merge.addChild(sink);

    ifCfg.addNode(switchBlock);

    return ifCfg;
  }

  private void processElseBranch(
      Cfg ifCfg, Switch switchBlock, Merge merge, IGSequenceOfStatements elseSOS) {
    Case negationBlock;
    Cfg elseCfg;
    negationBlock = new Case();
    switchBlock.addChild(negationBlock);
    ifCfg.addNode(negationBlock);


    if (elseSOS != null) {

      negationBlock.addValue(NodeValue.newBoolean(false));
      elseCfg = getCfg(elseSOS);
      elseCfg.addAsSource(negationBlock);
      elseCfg.addAsSink(merge);
      ifCfg.addNodes(elseCfg.getNodes());
    } else {
      negationBlock.setDefault();
      negationBlock.addChild(merge);
    }

    ifCfg.addNode(negationBlock);
    ifCfg.addNode(merge);
  }

  private void processThenBranch(
      Cfg ifCfg, Switch switchBlock, Merge merge, IGSequenceOfStatements thenSos) {
    Case caseBlock;
    Cfg thenCfg;
    if (thenSos != null) {

      caseBlock = new Case(NodeValue.newBoolean(true));
      switchBlock.addChild(caseBlock);
      thenCfg = getCfg(thenSos);
      thenCfg.addAsSource(caseBlock);
      thenCfg.addAsSink(merge);

      ifCfg.addNode(caseBlock);
      ifCfg.addNodes(thenCfg.getNodes());
    }
  }
}
