/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.vhdl;

import org.apache.commons.cli.ParseException;
import org.apache.log4j.Level;
import org.zamia.BuildPath;
import org.zamia.SourceFile;
import org.zamia.ZamiaException;
import org.zamia.ZamiaLogger;
import org.zamia.ZamiaProject;
import org.zamia.ZamiaProjectBuilder;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGModule;
import org.zamia.zdb.ZDBException;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.parser.basis.HDLParser;
import ru.ispras.retrascope.util.LogLevel;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Map;

/**
 * The parser for VHDL hardware models.
 *
 * Uses ZamiaCAD toolkit for making an intermediate representation (Instantiation Graph, IG)
 * before control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} building.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class VhdlParser extends HDLParser {

  /* The engine identifier. */
  public static final String ENGINE_ID = "vhdl-parser";

  /* The engine output entity type. */
  public static final EntityType OUT_TYPE = EntityType.get("vhd");

  /* The VHDL toplevel module name. */
  private static final Parameter TOP_LEVEL =
      new Parameter("toplevel", true, "Toplevel VHDL modules names");

  /**
   * Default constructor.
   */
  public VhdlParser() {
    super(VhdlParser.ENGINE_ID);

    addInputType(VhdlParser.OUT_TYPE);
    setOutputType(CfgModel.TYPE);
    
    addParameter(TOP_LEVEL);
  }

  @Override
  public Entity parse(Map<EntityType, Entity> inputs) {
    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);
    final FileNames fileNames = (FileNames)inputs.get(VhdlParser.OUT_TYPE);

    final CfgModel cfgModel = new CfgModel();

    try {
      String topLevelNames;
      
      /* Get top level modules names from command line. */
      Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      if (engineArgs.contains(TOP_LEVEL)) {
        topLevelNames = engineArgs.getValue(TOP_LEVEL);
  
        String[] names = topLevelNames.split(File.pathSeparator);
  
        if (names.length > 0) {
          for (int i = 0; i < names.length; i++) {
  
            String fileName = fileNames.getList().get(i);
  
            if (!((new File(fileName)).exists())) {
              this.getLogger().log(LogLevel.ERROR, "No such file: " + fileName);
            } else {

             /*
              * Create BuildPath.txt
              */

              /* Get full path to VHDL file directory. */
              String sourceDirPath;
              int separatorIndex = fileName.lastIndexOf(File.separatorChar);

              if (separatorIndex >= 0) {
                sourceDirPath = fileName.substring(0, separatorIndex);
              } else {

                /* source file is in the current directory. */
                sourceDirPath = System.getProperty("user.dir");
              }

              File buildFile =
                FileCreator.newFile(sourceDirPath + File.separator + "BuildPath.txt");

              FileWriter writer = new FileWriter(buildFile);

              /* Clean the file. */
              writer.flush();

              /* Write "local "WORK path/to/source/file"". */
              writer.write("local WORK " + "\"" + sourceDirPath + "\"\n");

              /* write "toplevel UPPER_CASE_MODULE_NAME". */
              writer.write("toplevel " + names[i].toUpperCase() + "\n");
              writer.close();

              /*
               * Create Zamia IG representation of the source code.
               */
              ZamiaLogger.setup(Level.OFF);
              ZamiaProject zamiaProject = new ZamiaProject(
                  "", sourceDirPath, new SourceFile(buildFile));

              ZamiaProjectBuilder builder = zamiaProject.getBuilder();
              builder.build(true, true, null);

              /*
               * Parse Zamia IG, create CfgModel.
               */

              IGManager igm = zamiaProject.getIGM();

              BuildPath bp = zamiaProject.getBuildPath();

              /* Elaborate top level modules. */
              int topLevelNum = bp.getNumToplevels();

              VhdlIgParser igParser = new VhdlIgParser();

              if (topLevelNum > 0) {
                for (int j = 0; j < topLevelNum; j++) {
                  IGModule module = igm.findModule(bp.getToplevel(j));
                  cfgModel.addModule(igParser.parseModule(zamiaProject, module));
                }
              }
              zamiaProject.shutdown();
            }
          }
        }
      }
    } catch (ParseException | ZamiaException | IOException | ZDBException e) {
      e.printStackTrace();
    }

    return cfgModel;
  }
}
