/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between IG subprogram identifiers and standard operation identifiers that are supported.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VhdlSubprogramMapping {

  private static Map<String, Enum<?>> map;

  static {
    map = new LinkedHashMap<>();

    map.put("=", StandardOperation.EQ);
    map.put("<=", StandardOperation.LESSEQ);
    map.put("<", StandardOperation.LESS);
    map.put(">=", StandardOperation.GREATEREQ);
    map.put(">", StandardOperation.GREATER);
    map.put("+", StandardOperation.ADD);
    map.put("-", StandardOperation.SUB);
    map.put("AND", StandardOperation.AND);
    map.put("OR", StandardOperation.OR);
    map.put("NOT", StandardOperation.NOT);
    map.put("/=", StandardOperation.NOTEQ);
    map.put("&", StandardOperation.BVCONCAT);
    map.put("XOR", StandardOperation.BVXOR);
    map.put("*", StandardOperation.MUL);
    map.put("MOD", StandardOperation.MOD);
    map.put("/", StandardOperation.DIV);
    map.put("**", StandardOperation.POWER);
    map.put("NOR", StandardOperation.BVNOR);
    map.put("BV_NEGATE", StandardOperation.BVNEG);
  }

  /**
   * Checks rather operation is supported.
   *
   * If operation is supported, then it's identifier is in this mapping.
   *
   * @param operation operation identifier
   *
   * @return {@code true} if operation is supported, {@code true} otherwise
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public static boolean hasOperation(String operation) {
    InvariantChecks.checkNotNull(operation);

    return map.containsKey(operation);
  }

  /**
   * Gets identifier of supported operation.
   *
   * @param operation operation identifier
   *
   * @return operation identifier
   *
   * @throws java.lang.NullPointerException when argument is {@code null}.
   */
  public static Enum<?> getOperation(String operation) {
    InvariantChecks.checkNotNull(operation);

    return map.get(operation);
  }
}
