/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.graphml;

import java.io.PrintWriter;

/**
 * GraphML utils.
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class GraphMlUtils {
  public final static String FORM_TRIANGLE = "triangle";
  public final static String FORM_ELLIPSE = "ellipse";
  public final static String FORM_DIAMOND = "diamond";
  public final static String FORM_PARALLELOGRAM = "parallelogram";
  public final static String FORM_RECTANGLE = "rectangle";
  public final static String FORM_OCTAGON = "octagon";
  public final static String FORM_TRAPEZOID = "trapezoid";
  public final static String FORM_TRAPEZOID2 = "trapezoid2";

  public final static String COLOR_WHITE = "#ffffff";
  public final static String COLOR_YELLOW = "#ffcc00";
  public final static String COLOR_PURPLE = "#ccccff";
  public final static String COLOR_GREEN = "#008000";
  public final static String COLOR_RED = "#ff0000";
  public final static String COLOR_BLUE = "#0000ff";
  public final static String COLOR_ORANGE = "#ff8400";

  // Cropping text in the node, if label size of text > node size
  public final static Boolean CROPPING_TEXT = true;

  /**
   * Print the GraphML title tag to GraphML file.
   * 
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphMlBegin(PrintWriter printWriter) {
    // Write xml data
    printWriter.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    // Open GraphML tag
    printWriter
        .println("<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns\" xmlns:y=\"http://www.yworks.com/xml/graphml\">");
    //
    printWriter.println("<key for=\"node\" id=\"d0\" yfiles.type=\"nodegraphics\"/>");
    printWriter.println("<key for=\"edge\" id=\"d1\" yfiles.type=\"edgegraphics\"/>");
  }

  /**
   * Print the GraphML close tag to GraphML file.
   * 
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphMlEnd(PrintWriter printWriter) {
    // Close GraphML tag
    printWriter.format("</graphml>%n");
  }

  /**
   * Print the graph open tag.
   * 
   * @param printWriter the GraphML file writer.
   * @param modelId the model identifier.
   */
  public static void onGraphBegin(PrintWriter printWriter, String modelId) {
    printWriter.format("<graph id=\"model" + modelId + "\" edgedefault=\"undirected\">%n");
  }

  /**
   * Print the graph open tag.
   * 
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphBegin(PrintWriter printWriter) {
    onGraphBegin(printWriter, "");
  }

  /**
   * Print the graph open tag.
   * 
   * @param printWriter the GraphML file writer.
   */
  public static void onGraphEnd(PrintWriter printWriter) {
    printWriter.format("</graph>%n");
  }

  /**
   * Print node.
   * 
   * @param printWriter the GraphML file writer.
   * @param nodeName the node identifier.
   */
  public static void printNode(PrintWriter printWriter, String nodeName) {
    printWriter.format("<node id=\"%s\"/>%n", nodeName);
  }

  /**
   * Print node with this label, form, color to GraphML file.
   * 
   * @param printWriter the GraphML file writer.
   * @param nodeName the node identifier.
   * @param nodeLabel the Label for node.
   * @param form the node form.
   * @param color the node color.
   */
  public static void printNode(PrintWriter printWriter, String nodeName, String nodeLabel,
      String form, String color) {
    printNode(printWriter, nodeName, nodeLabel, form, color, false);
  }

  public static void printNode(PrintWriter printWriter, String nodeName, String nodeLabel,
      String form, String color, boolean croppingText) {
    String croppingLabel = "";
    if (croppingText) {
      croppingLabel = " autoSizePolicy=\"node_size\" configuration=\"CroppingLabel\"";
    }
    printWriter.format("<node id=\"%s\">%n", nodeName);
    printWriter.format("   <data key=\"d0\">%n");
    printWriter.format("      <y:ShapeNode>%n");
    printWriter.format("           <y:Fill color=\"%s\"/>%n", color);
    printWriter.format("      <y:NodeLabel%s> <![CDATA[%s]]> </y:NodeLabel>%n", croppingLabel,
        nodeLabel);
    printWriter.format("           <y:Shape type=\"%s\"/>%n", form);
    printWriter.format("       </y:ShapeNode>%n");
    printWriter.format("   </data>%n");
    printWriter.format("</node>%n");
  }

  /**
   * Print edge to GraphML file.
   * 
   * @param printWriter the GraphML file writer.
   * @param nodeSourceName the source node identifier.
   * @param nodeTargetName the target node identifier.
   */
  public static void printEdge(PrintWriter printWriter, String nodeSourceName, String nodeTargetName) {
    printWriter.format("<edge source=\"%s\" target=\"%s\"/>%n", nodeSourceName, nodeTargetName);
  }

  /**
   * Print edge with label to GraphML file.
   * 
   * @param printWriter the GraphML file writer.
   * @param nodeSourceName the source node identifier.
   * @param nodeLabel the label of edge
   * @param nodeTargetName the target node identifier.
   */
  public static void printEdge(PrintWriter printWriter, String nodeSourceName, String nodeLabel,
      String nodeTargetName) {
    printWriter.format("<edge source=\"%s\" target=\"%s\">%n", nodeSourceName, nodeTargetName);
    printWriter.format("	<data key=\"d1\">%n");
    printWriter.format("	<y:PolyLineEdge>%n");
    printWriter.format("      <y:EdgeLabel> <![CDATA[%s]]> </y:EdgeLabel>%n", nodeLabel);
    printWriter.format("      <y:Arrows source=\"none\" target=\"standard\"/>%n");
    printWriter.format("	</y:PolyLineEdge>%n");
    printWriter.format("	</data>%n");
    printWriter.format("</edge>%n");
  }

  /**
   * Close graphml file.
   * 
   * @param printWriter the GraphML file writer.
   */
  public static void closeGraphMlFile(PrintWriter printWriter) {
    printWriter.close();
  }
}
