/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.testbench;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;

/**
 * The representation of a test. A test consisted of a list of input sequences.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Sequence
 */
@XmlType(name = "test")
@XmlRootElement
public class Test extends Entity {
  /**
   * The entity type ID of all {@code Test} instances.
   */
  public static final String ENTITY_TYPE_ID = "test".intern();

  private static final String LEFT_STRING_BOUND = "{";
  private static final String SEQUENCE_STRING_HEADER = "sequence: ";
  private static final String SEQUENCE_STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  @XmlElement(name = "testSequence")
  private List<Sequence> sequences;

  /**
   * Constructs an empty test.
   */
  public Test() {
    super(EntityType.get(Test.ENTITY_TYPE_ID));
    this.sequences = new ArrayList<>();
  }

  /**
   * Constructs a test containing the specified list of sequences. An original order of the
   * sequences is preserved.
   *
   * @param sequences the list of sequences
   */
  public Test(List<Sequence> sequences) {
    super(EntityType.get(Test.ENTITY_TYPE_ID));
    this.sequences = new ArrayList<>(sequences);
  }

  /**
   * Adds the specified sequence to this test. The sequence is appended to the end of this test.
   *
   * @param sequence the sequence to be added.
   */
  public void addSequence(Sequence sequence) {
    this.sequences.add(sequence);
  }

  /**
   * Adds the specified sequences to this test. The sequences is appended to the end of this test,
   * an original order is preserved.
   *
   * @param sequences a list of the sequences to be added
   */
  public void addSequences(List<Sequence> sequences) {
    this.sequences.addAll(sequences);
  }

  /**
   * Returns a list of sequences which this test consisted of.
   *
   * @return the list of sequences
   */
  public List<Sequence> getSequences() {
    return new ArrayList<>(this.sequences);
  }

  /**
   * Returns an amount of sequences which this test consisted of.
   *
   * @return the amount of sequences in this test.
   */
  public int getSequencesAmount() {
    return this.sequences.size();
  }

  /**
   * Adds sequences of the specified test to this test. The sequences is appended to the end of
   * this test, an original order is preserved.
   *
   * @param entity the test to be added.
   *
   * @throws IllegalArgumentException if the specified entity is not an instance of the
   * {@link Test} class.
   */
  @Override
  public void add(final Entity entity) {
    if (entity instanceof Test) {
      this.sequences.addAll(((Test) entity).getSequences());
    } else {
      throw new IllegalArgumentException("Unsupported entity type: " +
          entity.getClass().getSimpleName());
    }
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + ((this.sequences == null) ? 0 : this.sequences.hashCode());

    return result;
  }

  /**
   * Compares the specified object with this test for equality. The object is considered as equal to
   * this test if and only if it is also a test and it contains the same sequences in the same order
   * as this test does, or more formally:
   * <ul>
   *    <li>it is an instance of the {@code Test} class or its subclass;
   *    <li>it contains the same amount of sequences as this test does;
   *    <li>each its sequence is equal to the corresponding sequence in this test.
   * </ul>
   *
   * @param obj the object to be compared for equality with this test
   *
   * @return {@code true} if the specified object is equal to this test. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof Test)) {
      return false;
    }

    Test testToCompare = (Test) obj;

    return this.sequences.equals(testToCompare.sequences);
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(Test.LEFT_STRING_BOUND);

    int initialStringBuilderLength = stringBuilder.length();

    for (Sequence sequence : this.sequences) {
      if (stringBuilder.length() > initialStringBuilderLength) {
        stringBuilder.append(Test.SEQUENCE_STRING_DELIMITER);
      }

      stringBuilder.append(Test.SEQUENCE_STRING_HEADER);
      stringBuilder.append(sequence);
    }

    stringBuilder.append(Test.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
