/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.testbench;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

import ru.ispras.fortress.expression.Node;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;

/**
 * The representation of an input/output vector of variables. A vector is consisted of a set of
 * transactions along with a list of attendant events (for example clock or reset signal change).
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Transaction
 * @see EventList
 */
@XmlType(propOrder = {"events", "transactions"}, name = "vector")
@XmlRootElement
public class Vector {
  private static final String LEFT_STRING_BOUND = "{";
  private static final String EVENTS_STRING_HEADER = "events: ";
  private static final String EVENTS_STRING_DELIMITER = "; ";
  private static final String TRANSACTION_STRING_HEADER = "transaction: ";
  private static final String TRANSACTION_STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  @XmlElement
  private EventList events;
  @XmlElement(name = "transaction")
  private Set<Transaction> transactions;

  /**
   * Constructs an empty vector.
   */
  public Vector() {
    this.transactions = new HashSet<>();
    this.events = new EventList();
  }

  /**
   * Constructs a vector containing the specified transactions and events.
   * 
   * @param transactions    a collection of the transactions
   * @param events          the events
   */
  public Vector(Collection<Transaction> transactions, EventList events) {
    this.transactions = new HashSet<>(transactions);
    this.events = events;
  }

  /**
   * Adds the specified transaction to this vector.
   * 
   * @param transaction the transaction to be added
   */
  public void addTransaction(Transaction transaction) {
    this.transactions.add(transaction);
  }

  /**
   * Adds the specified event to the vector.
   * 
   * @param event   the event to be added
   */
  public void addEvent(Event event) {
    this.events.add(event);
  }

  /**
   * Adds the specified events to this vector.
   * 
   * @param events   the events to be added
   */
  public void addEvents(Collection<Event> events) {
    for (Event event: events) {
      this.events.add(event);
    }
  }

  /**
   * Returns an {@link EventList} associated with this vector.
   *
   * @return the event list
   */
  public EventList getEventList() {
    return this.events;
  }

  /**
   * Returns all transactions included in this vector.
   * 
   * @return a set of the transactions representing this vector
   */
  public Set<Transaction> getTransactions() {
    return new HashSet<>(this.transactions);
  }

  /**
   * Checks if this vector is empty.
   * 
   * @return {@code true} if this vector contains neither any non-empty transactions nor any
   * events. Otherwise returns {@code false}
   */
  public boolean isEmpty() {
    if (!this.events.isEmpty()) {
      return false;
    } else {
      for (Transaction transaction: this.transactions) {
        if (!transaction.isEmpty()) {
          return false;
        }
      }

      return true;
    }
  }
  
  /**
   * Checks if the specified variable is in this vector.
   * 
   * @param name    a name of the variable.
   * 
   * @return {@code true} if this vector contains the specified variable. Otherwise returns
   * {@code false}
   */
  public boolean containsVariable(String name) {
    for (Transaction transaction: this.transactions) {
      if (transaction.containsVariable(name)) {
        return true;
      }
    }

    return false;
  }

  /**
   * Returns a value of the specified variable from this vector.
   * 
   * @param variableName    a name of the variable
   * 
   * @return the variable's value if the specified variable is in this vector; otherwise returns
   * {@code null}
   */
  public Node getValue(String variableName) {
    for (Transaction transaction: this.transactions) {
      if (transaction.containsVariable(variableName)) {
        return transaction.getValue(variableName);
      }
    }

    return null;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.events.hashCode();
    result = prime * result + this.transactions.hashCode();

    return result;
  }

  /**
   * Compares the specified object with this vector for equality. The object is considered as
   * equal to this vector if and only if it is also a vector and it contains the same events
   * and transactions as this vector does, or more formally:
   * <ul>
   *    <li>it is an instance of the {@code Vector} class or its subclass;
   *    <li>its {@code EventList} is equal to this vector's one;
   *    <li>its contains the same amount of transactions as this vector does;
   *    <li>for each its transaction there is a equal one in this vector.
   * </ul>
   * 
   * @param obj the object to be compared for equality with this vector
   * 
   * @return {@code true} if the specified object is equal to this vector. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof Vector)) {
      return false;
    }

    Vector vectorToCompare = (Vector) obj;

    return this.events.equals(vectorToCompare.events) &&
        this.transactions.equals(vectorToCompare.transactions);
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(Vector.LEFT_STRING_BOUND);
    stringBuilder.append(Vector.EVENTS_STRING_HEADER);
    stringBuilder.append(this.events);
    stringBuilder.append(Vector.EVENTS_STRING_DELIMITER);

    int preVectorsStringBuilderLength = stringBuilder.length();

    for (Transaction transaction : this.transactions) {
      if (stringBuilder.length() > preVectorsStringBuilderLength) {
        stringBuilder.append(Vector.TRANSACTION_STRING_DELIMITER);
      }

      stringBuilder.append(Vector.TRANSACTION_STRING_HEADER);
      stringBuilder.append(transaction);
    }

    stringBuilder.append(Vector.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
