/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.ast.printer.smv;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.result.ast.Ast;
import ru.ispras.retrascope.result.ast.AstModel;
import ru.ispras.retrascope.result.ast.AstNode;
import ru.ispras.retrascope.result.ast.AstVisitor;

/**
 * This class represents a visitor of an {@link AstModel} that builds its SMV representation.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class AstSmvVisitor implements AstVisitor {

  private PrintWriter printWriter;
  private String indentation = "  ";
  private HashMap<Switch, Boolean> trueFalseSwitches;
  private StringBuilder declarations;
  private StringBuilder body;
  private HashMap<Node, String> maps;
  private int mapCounter;

  public static final int INTEGER_SIZE = 32;


  /**
   * Constructs a new visitor object using the specified *.smv file name.
   * 
   * @param fileName - the specified file name.
   */
  public AstSmvVisitor(String fileName) {
    File smvFile = FileCreator.newFile(fileName);
    indentation = "  ";
    trueFalseSwitches = new HashMap<Switch, Boolean>();
    declarations = new StringBuilder();
    body = new StringBuilder();
    maps = new HashMap<Node, String>();
    mapCounter = 0;
    try {
      printWriter = new PrintWriter(new BufferedWriter(new FileWriter(smvFile)));
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  @Override
  public void onRootNodeBegin() {
    /* Empty. */
  }

  @Override
  public void onRootNodeEnd() {
    /* Empty. */
  }

  @Override
  public void onNodeBegin(AstNode node) {
    /* Empty. */
  }

  @Override
  public void onNodeEnd(AstNode node) {
    /* Empty. */
  }

  @Override
  public void onBasicBlockBegin(AstNode node) {

    for (Assignment assignment : ((BasicBlock) node.getCfgNode()).getAssignments()) {
      body.append(indentation);
      printAssignment(assignment);
    }
  }

  @Override
  public void onBasicBlockEnd(AstNode node) {
    /* Empty. */
  }

  @Override
  public void onSwitchBegin(AstNode node) {
    Switch aSwitch = (Switch) node.getCfgNode();
    Collection<CfgModelNode> children = aSwitch.getChildren();
    ArrayList<Case> cases = new ArrayList<Case>();
    HashSet<NodeValue> values = new HashSet<NodeValue>();
    values.add(new NodeValue(Data.newBoolean(true)));
    values.add(new NodeValue(Data.newBoolean(false)));

    for (CfgModelNode child : children) {
      if (child.getClass() == Case.class) {
        cases.add((Case) child);
      }
    }

    if (cases.size() < 2) {
      trueFalseSwitches.put(aSwitch, Boolean.TRUE);
      printSwitchDescription(aSwitch.getExpression(), true);
    } else {
      if (cases.size() == 2) {
        HashSet<NodeValue> caseValues = new HashSet<NodeValue>();
        Set<NodeValue> case1Values = cases.get(0).getValues();
        Set<NodeValue> case2Values = cases.get(1).getValues();
        caseValues.addAll(case1Values);
        caseValues.addAll(case2Values);
        if (caseValues.equals(values)) {
          trueFalseSwitches.put(aSwitch, Boolean.FALSE);
        } else {
          if (((case1Values.size() == 1) && (caseValues.contains(case1Values.iterator().next())) && (cases
              .get(1).isDefault()))
              || ((case2Values.size() == 1) && (caseValues.contains(case2Values.iterator().next())) && (cases
                  .get(0).isDefault()))) {
            trueFalseSwitches.put(aSwitch, Boolean.FALSE);
          }
        }
      } else {
        printSwitchDescription(aSwitch.getExpression(), false);
        increaseIndentation();
      }
    }
  }

  @Override
  public void onSwitchEnd(AstNode node) {
    if (!trueFalseSwitches.containsKey((Switch) node.getCfgNode())) {
      decreaseIndentation();
    }

    body.append(indentation + "}\n");
  }

  @Override
  public void onCaseBegin(AstNode node) {
    Case aCase = (Case) node.getCfgNode();
    CfgModelNode child = aCase.getOnlyChild();
    Switch parent = (Switch) aCase.getOnlyParent();

    // Ignore empty cases.
    if ((child.getClass() != Merge.class) && (child.getClass() != Sink.class)) {

      // Check if is default.
      if (aCase.isDefault()) {
        body.append(indentation + "default: \n");
      } else {

        if (trueFalseSwitches.containsKey(parent)) {

          if (!trueFalseSwitches.get(parent).booleanValue()) {
            NodeValue val = aCase.getValues().iterator().next();
            if ((Boolean) val.getValue() == Boolean.TRUE) {
              printSwitchDescription(parent.getExpression(), true);
              trueFalseSwitches.put(parent, Boolean.TRUE);
            } else {
              printSwitchDescription(ExprUtils.getNegation(parent.getExpression()), true);
              trueFalseSwitches.put(parent, Boolean.TRUE);
            }
          } else {
            if ((!aCase.isDefault()) && (parent.getChildren().size() == 2)) {
              body.append(indentation + "} else {\n");
            }
          }
        } else {

          // Print values
          for (NodeValue value : aCase.getValues()) {
            String valueString = value.toString();
            valueString = valueString.replaceAll("true", "1");
            valueString = valueString.replaceAll("false", "0");
            body.append(indentation + "case " + valueString + " : \n");
          }
        }
      }
      increaseIndentation();
    }
  }

  @Override
  public void onCaseEnd(AstNode node) {
    CfgModelNode child = ((Case) node.getCfgNode()).getOnlyChild();
    if ((child.getClass() != Merge.class) && (child.getClass() != Sink.class)) {
      decreaseIndentation();
    }
  }

  @Override
  public void onTreeBegin(Ast tree) {
    Collection<CfgModelNode> parents = tree.getParentProcess().getParents();
    declarations = new StringBuilder();
    body = new StringBuilder();
    maps = new HashMap<Node, String>();
    for (CfgModelNode node : parents) {
      if (((Module) node).getProcesses().size() > 1) {
        printModuleHeader((Module) node, tree.getParentProcess().getMetaInfo("name"));
      }
    }

  }

  @Override
  public void onTreeEnd(Ast tree) {
    Collection<CfgModelNode> parents = tree.getParentProcess().getParents();
    for (CfgModelNode node : parents) {
      if (((Module) node).getProcesses().size() > 1) {
        body.append("}\n\n");
      }
    }
    for (Node map : maps.keySet()) {
      DataMap dataMap = (DataMap) ((NodeValue) map).getValue();
      DataType keyType = dataMap.getKeyType();
      declarations.append("  " + maps.get(map) + " : array " + (dataMap.size() - 1)
          + "..0 of array " + (INTEGER_SIZE - 1) + "..0 of boolean;\n\n");
      declarations.append("  " + maps.get(map) + " := [");
      for (int i = 0; i < dataMap.size(); i++) {
        Data data = dataMap.get(keyType.valueOf(String.valueOf(i), 10));
        Object value = data.getValue();
        if (i != 0) {
          declarations.append(", ");
        }
        if (dataMap.getValueType().getTypeId() == DataTypeId.BIT_VECTOR) {
          declarations.append(((BitVector) value).bigIntegerValue().toString());
        } else {
          declarations.append(value.toString());
        }
      }
      declarations.append("];\n");
    }
    printWriter.print(declarations.toString());
    printWriter.print(body.toString());
  }

  @Override
  public void onModelBegin(AstModel model) {
    /* Empty. */
  }

  @Override
  public void onModelEnd(AstModel model) {
    printWriter.close();

  }

  @Override
  public void onModuleBegin(Module module) {
    printModuleHeader(module, "main"/* moduleName */);
    if (module.getProcesses().size() > 1) {
      printProcessInstances(module);
      declarations.append("}\n\n");
    }
    printWriter.println(declarations.toString());
  }

  @Override
  public void onModuleEnd(Module module) {
    if (module.getProcesses().size() <= 1) {
      printWriter.println("}");
      printWriter.println();
    }

    printWriter.println();

  }

  /**
   * Prints a header of a SMV module, parameters and variables.
   * 
   * @param module - the specified Cfg {@link Module}.
   * @param name - the specified name of a module.
   */
  private void printModuleHeader(Module module, String name) {
    List<NodeVariable> inputs = module.getInputs();
    List<NodeVariable> outputs = module.getOutputs();
    List<NodeVariable> internal = module.getVariables();
    internal.removeAll(inputs);
    internal.removeAll(outputs);

    declarations.append("module " + name);
    printParameters(inputs, outputs);
    declarations.append("\n{\n");

    printVariables(inputs, VariableType.IN);
    printVariables(outputs, VariableType.OUT);
    if (!internal.isEmpty()) {
      printVariables(internal, VariableType.REG);
    }
  }

  /**
   * Prints processes of a Cfg module as SMV module instances.
   * 
   * @param module - the specified Cfg {@link Module}.
   */
  private void printProcessInstances(Module module) {
    List<NodeVariable> inputs = module.getInputs();
    List<NodeVariable> outputs = module.getOutputs();
    int processIndex = 1;
    for (Process proc : module.getProcesses()) {
      declarations.append("  processInst_" + processIndex + " : process_" + processIndex);
      printParameters(inputs, outputs);
      declarations.append(";\n");
      proc.addMetaInfo("name", "process_" + processIndex);
      processIndex++;
    }
  }

  /**
   * Prints formal parameters of a SMV module.
   * 
   * @param inputs - the list of input variables.
   * @param outputs - the list of output variables.
   */
  private void printParameters(List<NodeVariable> inputs, List<NodeVariable> outputs) {
    declarations.append("(");
    for (int i = 0; i < inputs.size(); i++) {
      if (i != 0) {
        declarations.append(", ");
      }
      declarations.append(inputs.get(i).getName().toLowerCase());
    }

    for (int j = 0; j < outputs.size(); j++) {
      declarations.append(", " + outputs.get(j).getName().toLowerCase());
    }

    declarations.append(")");
  }

  /**
   * Prints variables of the specified type.
   * 
   * @param variables - the list of variables.
   * @param type - the specified type (IN, OUT, REG).
   */
  private void printVariables(List<NodeVariable> variables, VariableType type) {

    for (NodeVariable var : variables) {
      DataType dataType = var.getDataType();

      switch (type) {
        case IN:
          declarations.append("  input " + var.getName().toLowerCase() + " : ");
          break;
        case OUT:
          declarations.append("  output " + var.getName().toLowerCase() + " : ");
          break;
        case REG:
          declarations.append("  " + var.getName().toLowerCase() + " : ");
          break;
        default:
          break;
      }

      switch (dataType.getTypeId()) {
        case BIT_VECTOR:
          if (dataType.getSize() == 1) {
            declarations.append("boolean;\n");
          } else {
            declarations.append("array " + (dataType.getSize() - 1) + "..0 of boolean;\n");
          }
          break;
        case LOGIC_BOOLEAN:
          declarations.append("boolean;\n");
          break;
        case LOGIC_INTEGER:
          declarations.append("array " + (INTEGER_SIZE - 1) + "..0 of boolean;\n");
          break;
        default:
          break;
      }
    }

    declarations.append("\n");

    if (type == VariableType.REG) {
      for (NodeVariable var : variables) {
        body.insert(0, "  init(" + var.getName().toLowerCase() + ") := 0;\n");
      }
      body.append("\n");
    }
  }

  /**
   * Adds two whitespaces to the specified indentation string.
   * 
   * @return A {@link String} with two whitespaces added.
   */
  private void increaseIndentation() {
    indentation = indentation.concat("  ");
  }

  /**
   * Shortens the specified indentation string by two characters.
   * 
   * @return A {@link String} with the last two characters deleted.
   */
  private void decreaseIndentation() {
    StringBuilder builder = new StringBuilder(indentation);
    if (builder.length() >= 2) {
      indentation = builder.delete(builder.length() - 2, builder.length()).toString();
    } else {
      indentation = new String("  ");
    }
  }

  /**
   * Prints an assignment.
   * 
   * @param assignment - the specified {@link Assignment}.
   */
  private void printAssignment(Assignment assignment) {
    RangedVariable variable = assignment.getRangedVariable();
    ExprTreePrinter exprPrinter = new SmvExprPrinter();

    // Print left-hand side.
    body.append("next(" + variable.getName().toLowerCase() + ") := ");

    // Print right-hand side.
    if (assignment.getExpression().getKind() == Kind.VALUE) {
      NodeValue value = (NodeValue) assignment.getExpression();

      // Print bit vector.
      if ((value.isType(DataTypeId.BIT_VECTOR))
          && (value.getDataType().getSize() > 1)) {
        body.append(/* getBitVectorDescription */((BitVector) value.getValue()).bigIntegerValue()
            .toString() + ";\n");
      } else {
        // Print other values.
        body.append(exprPrinter.toString(makeSubstitutions(assignment.getExpression())) + ";\n");
      }

    } else {
      Node expression = assignment.getExpression();
      expression = makeSubstitutions(expression);
      body.append(exprPrinter.toString(expression).toLowerCase() + ";\n");
    }

  }

  /**
   * Prints a switch.
   * 
   * @param node - the specified switch condition node.
   * @param isIf - shows whether the switch can be represented as an if/else construction.
   */
  private void printSwitchDescription(Node node, boolean isIf) {

    ExprTreePrinter printer = new SmvExprPrinter();
    HashMap<String, String> bitVectors = new HashMap<String, String>();

    node = makeSubstitutions(node);

    if (node.getKind() == Kind.OPERATION) {
      NodeOperation operation = (NodeOperation) node;
      for (int i = operation.getOperandCount() - 1; i >= 0; i--) {
        if (operation.getOperand(i).getKind() == Kind.VALUE) {
          NodeValue value = (NodeValue) operation.getOperand(i);
          if ((value.isType(DataTypeId.BIT_VECTOR))
              && (value.getDataType().getSize() > 1)) {
            bitVectors.put(value.toString(), /* getBitVectorDescription */
                ((BitVector) value.getValue()).bigIntegerValue().toString());
          }
        }
      }
    }

    String switchDescription = printer.toString(node);
    for (String bv : bitVectors.keySet()) {
      switchDescription = switchDescription.replaceAll(bv, bitVectors.get(bv));
    }

    if (isIf) {
      body.append(indentation + "if (" + switchDescription.toLowerCase() + ") {\n");
    } else {
      body.append(indentation + "switch (" + switchDescription.toLowerCase() + ") {\n");
    }
  }

  // /**
  // * Returns a {@link String} containing a SMV representation of a bit vector (for example, "101"
  // is
  // * converted to "[1,0,1]").
  // *
  // * @param bv - the specified {@link BitVector}.
  // * @return The String representation of the specified vector.
  // */
  // private String getBitVectorDescription(BitVector bv) {
  // StringBuilder description = new StringBuilder();
  // description.append("[");
  // for (int i = bv.getBitSize() - 1; i >= 0; i--) {
  // if (i > 0) {
  // description.append((bv.getBit(i) ? "1" : "0") + ",");
  // } else {
  // description.append((bv.getBit(i) ? "1" : "0") + "]");
  // }
  // }
  // return description.toString();
  // }

  /**
   * Performs the following substitutions on the specified Node:
   * <ul>
   * <li>Substitutes a power operation by an arithmetic left shift operation.
   * <li>If a map value is found, a corresponding variable is created in the declarations section
   * and the value is replaced by this variable.
   * <li>If a BVEXTRACT operation extracts only one bit it is replaced by a SELECT operation.
   * </ul>
   * 
   * @param node - the specified node.
   * @return The node with the substitutions made.
   */
  private Node makeSubstitutions(Node node) {

    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule powerRule = new TransformerRule() {

      @Override
      public boolean isApplicable(Node arg0) {
        if (arg0.getKind() == Node.Kind.OPERATION) {
          return ((NodeOperation) arg0).getOperationId() == StandardOperation.POWER;
        }
        return false;
      }

      @Override
      public Node apply(Node arg0) {
        return new NodeOperation(StandardOperation.BVASHL, new NodeValue(Data.newInteger(1)),
            ((NodeOperation) arg0).getOperand(1));
      }
    };

    transformer.addRule(StandardOperation.POWER, powerRule);

    final TransformerRule mapRule = new TransformerRule() {

      @Override
      public boolean isApplicable(Node arg0) {
        if (arg0.getKind() == Node.Kind.VALUE) {
          return ((NodeValue) arg0).isType(DataTypeId.MAP);
        }
        return false;
      }

      @Override
      public Node apply(Node arg0) {
        if (!maps.containsKey(arg0)) {
          maps.put(arg0, "map" + mapCounter);
          mapCounter++;
        }
        DataMap map = (DataMap) ((NodeValue) arg0).getValue();
        DataType type = DataType.BIT_VECTOR(map.size());
        Variable var = new Variable(maps.get(arg0), type);
        return new NodeVariable(var);
      }

    };

    transformer.addRule(Node.Kind.VALUE, mapRule);

    final TransformerRule extractRule = new TransformerRule() {

      @Override
      public boolean isApplicable(Node arg0) {
        if (arg0.getKind() == Node.Kind.OPERATION) {
          if (((NodeOperation) arg0).getOperationId() == StandardOperation.BVEXTRACT) {
            return ((NodeOperation) arg0).getOperand(0) == ((NodeOperation) arg0).getOperand(1);
          }
        }
        return false;
      }

      @Override
      public Node apply(Node arg0) {
        return new NodeOperation(StandardOperation.SELECT, ((NodeOperation) arg0).getOperand(2),
            ((NodeOperation) arg0).getOperand(0));
      }
    };

    transformer.addRule(StandardOperation.BVEXTRACT, extractRule);

    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

}
