/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.ast;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Stack;

import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.result.ast.Ast;
import ru.ispras.retrascope.result.ast.AstModel;
import ru.ispras.retrascope.result.ast.AstNode;

/**
 * This class represents a visitor that is used to transform {@link CfgModel} to {@link AstModel}.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class CfgAstVisitor implements CfgVisitor {

  private Map<Module, Collection<Ast>> ast;
  private Collection<Ast> trees;
  private Ast processTree;
  private HashMap<CfgModelNode, AstNode> cfgAstNodes;
  private HashMap<Merge, Stack<AstNode>> mergeMap;
  private Stack<AstNode> nodeStack;

  public CfgAstVisitor() {
    initialize();
  }
  
  public void initialize() {
    ast = new LinkedHashMap<Module, Collection<Ast>>();
    cfgAstNodes = new HashMap<CfgModelNode, AstNode>();
    mergeMap = new HashMap<Merge, Stack<AstNode>>();
    nodeStack = new Stack<AstNode>();
  }

  public Map<Module, Collection<Ast>> getAst() {
    return ast;
  }

  @Override
  public void onRootBegin() {
    /* Empty. */
  }

  @Override
  public void onRootEnd() {
    /* Empty. */
  }

  @Override
  public void onCfgModelBegin() {
    /* Empty. */
  }

  @Override
  public void onCfgModelEnd() {
    /* Empty. */
  }

  @Override
  public void onModuleBegin(Module module) {
    trees = new LinkedHashSet<Ast>();
  }

  @Override
  public void onModuleEnd(Module module) {
    ast.put(module, trees);

  }

  @Override
  public void onProcessBegin(Process process) {
    processTree = new Ast();
    processTree.setParentProcess(process);

  }

  @Override
  public void onProcessEnd(Process process) {
    trees.add(processTree);

  }

  @Override
  public void onCfgBegin(Cfg cfg) {
    /* Empty. */
  }

  @Override
  public void onCfgEnd(Cfg cfg) {
    /* Empty. */
  }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) {
    AstNode node = new AstNode(basicBlock, nodeStack.peek());
    cfgAstNodes.put(basicBlock, node);
    processTree.addNode(node);
    nodeStack.push(node);
  }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) {
    nodeStack.pop();

  }

  @Override
  public void onCaseBegin(Case aCase) {
    AstNode node = new AstNode(aCase, nodeStack.peek());
    cfgAstNodes.put(aCase, node);
    processTree.addNode(node);
    nodeStack.push(node);
  }

  @Override
  public void onCaseEnd(Case aCase) {
    nodeStack.pop();

  }

  @Override
  public void onSwitchBegin(Switch aSwitch) {
    AstNode node = new AstNode(aSwitch, nodeStack.peek());
    cfgAstNodes.put(aSwitch, node);
    processTree.addNode(node);
    nodeStack.push(node);
  }

  @Override
  public void onSwitchEnd(Switch aSwitch) {
    nodeStack.pop();

  }

  @Override
  public void onMergeBegin(Merge merge) {
    Stack<AstNode> mergeStack = new Stack<AstNode>();

    while ((nodeStack.peek().getType() != CfgNodeType.SWITCH)
        && (nodeStack.peek().getType() != CfgNodeType.MERGE) && (!nodeStack.peek().isRoot())) {
      mergeStack.push(nodeStack.pop());
    }

    if (nodeStack.peek().getType() == CfgNodeType.SWITCH) {
      mergeStack.push(nodeStack.pop());
    }

    mergeMap.put(merge, mergeStack);

    AstNode node = new AstNode(merge, nodeStack.peek());
    cfgAstNodes.put(merge, node);
    processTree.addNode(node);
    nodeStack.push(node);
  }

  @Override
  public void onMergeEnd(Merge merge) {
    nodeStack.pop();

    Stack<AstNode> mergeStack = mergeMap.get(merge);
    while (!mergeStack.isEmpty()) {
      nodeStack.push(mergeStack.pop());
    }

  }

  @Override
  public void onSrcBegin(Source source) {
    AstNode root = new AstNode(source);
    root.setRoot(true);
    cfgAstNodes.put(source, root);
    processTree.setRootNode(root);
    nodeStack.push(root);

  }

  @Override
  public void onSrcEnd(Source source) {
    nodeStack.pop();

  }

  @Override
  public void onSink(Sink sink) {
    AstNode node = new AstNode(sink, nodeStack.peek());
    cfgAstNodes.put(sink, node);
    processTree.addNode(node);
  }


}
