/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.petri.deadlock;

import java.util.ArrayList;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintBuilder;
import ru.ispras.fortress.solver.constraint.ConstraintKind;
import ru.ispras.fortress.solver.constraint.Formulas;
import ru.ispras.retrascope.model.petri.*;

/**
 * 
 */

/**
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 * 
 */
public class ZakrevskijMethod {


  private ZakrevskijMethod() {
    
  }

  /**
   * Constructs the disjunction of the output variables for Zakrevskij method
   * 
   * @param builder the constraint builder
   * @return NodeOperation containing the disjunction
   */
  public static Node getDisjunction(ConstraintBuilder builder, ArrayList<Integer> ids) {
    // TODO Can ArrayList be empty?
    if (ids.size() == 1) {
      return new NodeVariable(builder.addVariable("x" + ids.get(0).toString(), DataType.BOOLEAN));
    } else {
      NodeOperation outputExpr = null;
      for (int i = 1; i < ids.size(); i++) {
        if (outputExpr == null) {
          outputExpr =
              new NodeOperation(StandardOperation.OR, new NodeVariable(builder.addVariable("x"
                  + ids.get(i - 1).toString(), DataType.BOOLEAN)), new NodeVariable(
                  builder.addVariable("x" + ids.get(i).toString(), DataType.BOOLEAN)));
        } else {
          outputExpr =
              new NodeOperation(StandardOperation.OR, outputExpr, new NodeVariable(builder.addVariable(
                  "x" + ids.get(i).toString(), DataType.BOOLEAN)));
        }
      }
      return outputExpr;
    }
  }

  /**
   * Constructs the implication of output and input variable disjunctions for Zakrevskij method
   * 
   * @param builder the constraint builder
   * @return NodeOperation containing the implication
   */
  public static NodeOperation generateDeadlockFormula(ConstraintBuilder builder, Transition transition) {
    NodeOperation deadlockFormula =
        new NodeOperation(StandardOperation.IMPL, getDisjunction(builder, transition.getOutputPlaces()),
            getDisjunction(builder, transition.getInputPlaces()));
    return deadlockFormula;
  }

  /**
   * Constructs the implication of input and output variable disjunctions for Zakrevskij method
   * 
   * @param builder the constraint builder
   * @return NodeOperation containing the implication
   */
  public static NodeOperation generateTrapFormula(ConstraintBuilder builder, Transition transition) {
    NodeOperation trapFormula =
        new NodeOperation(StandardOperation.IMPL, getDisjunction(builder, transition.getInputPlaces()),
            getDisjunction(builder, transition.getOutputPlaces()));
    return trapFormula;
  }

  /**
   * Generates deadlock equations used in Zakrevskij method
   * 
   * @return constraint containing the equations
   */
  public static Constraint generateDeadlockEquations(PetriNet petriNet) {
    ConstraintBuilder builder = new ConstraintBuilder();
    builder.setName("DeadlockEquations");
    builder.setDescription("Deadlock Equations of a Petri Net");
    builder.setKind(ConstraintKind.FORMULA_BASED);

    Formulas formulas = new Formulas();
    builder.setInnerRep(formulas);

    ArrayList<NodeOperation> expressions = new ArrayList<NodeOperation>();

    for (Transition item : petriNet.getAllTransitions()) {
      expressions.add(generateDeadlockFormula(builder, item));
    }

    NodeOperation equations = expressions.get(0);
    for (int i = 1; i < expressions.size(); i++) {
      equations = (NodeOperation) Node.AND(equations, expressions.get(i));
    }

    Data data = new Data(DataType.BOOLEAN, true);
    formulas.add(new NodeOperation(StandardOperation.EQ, equations, new NodeValue(data)));

    return builder.build();
  }

  /**
   * Generates trap equations used in Zakrevskij method
   * 
   * @return constraint containing the equations
   */
  public static Constraint generateTrapEquations(PetriNet petriNet) {
    ConstraintBuilder builder = new ConstraintBuilder();
    builder.setName("TrapEquations");
    builder.setDescription("Trap Equations of a Petri Net");
    builder.setKind(ConstraintKind.FORMULA_BASED);

    Formulas formulas = new Formulas();
    builder.setInnerRep(formulas);

    ArrayList<NodeOperation> expressions = new ArrayList<NodeOperation>();

    for (Transition item : petriNet.getAllTransitions()) {
      expressions.add(generateTrapFormula(builder, item));
    }

    NodeOperation equations = expressions.get(0);
    for (int i = 1; i < expressions.size(); i++) {
      equations = (NodeOperation) Node.AND(equations, expressions.get(i));
    }

    Data data = new Data(DataType.BOOLEAN, true);
    formulas.add(new NodeOperation(StandardOperation.EQ, equations, new NodeValue(data)));

    return builder.build();
  }

  /**
   * Excludes zero solutions of the equations
   * 
   * @param constraint Deadlock or trap equations
   */
  public static void excludeZeroSolutions(Constraint constraint) {
    ArrayList<Variable> vars = (ArrayList<Variable>) constraint.getUnknownVariables();
    NodeOperation expr = null;
    for (int i = 1; i < vars.size(); i++) {
      if (expr == null) {
        expr =
            new NodeOperation(StandardOperation.OR, new NodeVariable(vars.get(i - 1)), new NodeVariable(
                vars.get(i)));
      } else {
        expr = new NodeOperation(StandardOperation.OR, expr, new NodeVariable(vars.get(i)));
      }
    }
    Formulas formulas = (Formulas) constraint.getInnerRep();
    formulas.add(new NodeOperation(StandardOperation.NOTEQ, expr, new NodeValue(new Data(
        DataType.BOOLEAN, false))));
  }

}
