/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.petri;

import ru.ispras.retrascope.model.basis.MetaInfo;

import java.util.ArrayList;

/**
 * This class represents a model of a Petri net.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 * 
 */
public class PetriNet extends MetaInfo {
  private ArrayList<Place> places;
  private ArrayList<Transition> transitions;

  /**
   * Constructs a PetriNet object using the provided place and transition lists.
   * 
   * @param places - The list of places.
   * @param transitions - The list of transitions.
   */
  public PetriNet(ArrayList<Place> places, ArrayList<Transition> transitions) {
    this.places = new ArrayList<Place>(places);
    this.transitions = new ArrayList<Transition>(transitions);
  }

  /**
   * Constructs a PetriNet with empty place and transition lists.
   */
  public PetriNet() {
    this.places = new ArrayList<Place>();
    this.transitions = new ArrayList<Transition>();
  }

  /**
   * Returns all places of the net.
   * 
   * @return A list of places.
   */
  public ArrayList<Place> getAllPlaces() {
    return places;
  }

  /**
   * Returns the place with the specified id.
   * 
   * @param id - The specified id.
   * @return The place with the specified id.
   */
  public Place getPlace(Integer id) {
    return places.get(id);
  }

  /**
   * Sets all places in the net.
   * 
   * @param places - The list of places to set.
   */
  public void setAllPlaces(ArrayList<Place> places) {
    this.places = places;
  }

  /**
   * Replaces the specified place by the new place with the same id.
   * 
   * @param place - The new place.
   * @return The old place.
   */
  public Place setPlace(Place place) {
    return places.set(place.getId(), place);
  }

  /**
   * Adds a new place to the net.
   * 
   * @param place - The specified place.
   * @return {@code true} if the specified place has been added to the net, {@code false} otherwise.
   */
  public boolean addPlace(Place place) {
    return places.add(place);
  }

  /**
   * Returns all transitions of the net.
   * 
   * @return A list of transitions.
   */
  public ArrayList<Transition> getAllTransitions() {
    return transitions;
  }

  /**
   * Returns the transition with the specified id.
   * 
   * @param id - the specified id.
   * @return The transition with the specified id.
   */
  public Transition getTransition(Integer id) {
    return transitions.get(id);
  }

  /**
   * Sets all transitions in the net.
   * 
   * @param transitions - The specified list of transitions to set.
   */
  public void setAllTransitions(ArrayList<Transition> transitions) {
    this.transitions = transitions;
  }

  /**
   * Replaces the specified transition by the new transition with the same id.
   * 
   * @param transition - the new transition.
   * @return The old transition.
   */
  public Transition setTransition(Transition transition) {
    return transitions.set(transition.getId(), transition);
  }

  /**
   * Adds a new transition to the net.
   * 
   * @param transition - The specified transition.
   * @return {@code true} if the specified transition has been added to the net, {@code false}
   *         otherwise.
   */
  public boolean addTransition(Transition transition) {
    return transitions.add(transition);
  }

  /**
   * Connects a place and a transition via its output and input respectively.
   * 
   * @param place - The specified place.
   * @param transition - The specified transition.
   * @return {@code true} in case of success, {@code false} otherwise.
   */
  public static boolean connect(Place place, Transition transition) {
    boolean r1 = place.addOutputTransiton(transition.getId());
    boolean r2 = transition.addInputPlace(place.getId());
    return (r1 & r2);
  }

  /**
   * Connects a transition and a place via its output and input respectively.
   * 
   * @param transition - The specified transition.
   * @param place - The specified place.
   * @return {@code true} in case of success, {@code false} otherwise.
   */
  public static boolean connect(Transition transition, Place place) {
    boolean r1 = transition.addOutputPlace(place.getId());
    boolean r2 = place.addInputTransition(transition.getId());
    return (r1 & r2);
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    return "PetriNet\n\nplaces:\n" + places + ",\n\ntransitions:\n" + transitions + "\n";
  }

}
