/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.petri;

import ru.ispras.retrascope.model.basis.MetaInfo;

import java.util.ArrayList;

/**
 * This class represents a place of a Petri net.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class Place extends MetaInfo {
  private int id;
  private ArrayList<Integer> inputTransitions;
  private ArrayList<Integer> outputTransitions;

  /**
   * Constructs a Place object using the provided id value and input and output transitions lists.
   * 
   * @param id - The specified id.
   * @param inputTransitions - The list of input transitions.
   * @param outputTransitions - The list of output transitions.
   */
  public Place(int id, ArrayList<Integer> inputTransitions, ArrayList<Integer> outputTransitions) {
    this.id = id;
    this.inputTransitions = new ArrayList<Integer>(inputTransitions);
    this.outputTransitions = new ArrayList<Integer>(outputTransitions);
  }

  /**
   * Constructs a Place object using the provided id value.
   * 
   * @param id - The specified id.
   */
  public Place(int id) {
    this.id = id;
    this.inputTransitions = new ArrayList<Integer>();
    this.outputTransitions = new ArrayList<Integer>();
  }

  /**
   * Returns the place id.
   * 
   * @return The id.
   */
  public int getId() {
    return this.id;
  }

  /**
   * Sets the place id.
   * 
   * @param newId - The specified id.
   */
  public void setId(int newId) {
    this.id = newId;
  }

  /**
   * Returns the list of input transition ids.
   * 
   * @return A list of ids.
   */
  public ArrayList<Integer> getInputTransitions() {
    return this.inputTransitions;
  }

  /**
   * Sets the list of input transition ids.
   * 
   * @param inputTransitions - The specified list.
   */
  public void setInputTransitions(ArrayList<Integer> inputTransitions) {
    this.inputTransitions = inputTransitions;
  }

  /**
   * Adds the specified transition id to the list of input transition ids.
   * 
   * @param transitionId - The specified input transition.
   * @return {@code true} if the specified transition has been added to the list or already contains
   *         in it, {@code false} otherwise.
   */
  public boolean addInputTransition(Integer transitionId) {
    if (inputTransitions.contains(transitionId)) {
      return true;
    } else {
      return inputTransitions.add(transitionId);
    }
  }

  /**
   * Returns the list of output transition ids.
   * 
   * @return A list of ids.
   */
  public ArrayList<Integer> getOutputTransitions() {
    return this.outputTransitions;
  }

  /**
   * Sets the list of output transition ids.
   * 
   * @param outputTransitions - The specified list.
   */
  public void setOutputTransitions(ArrayList<Integer> outputTransitions) {
    this.outputTransitions = outputTransitions;
  }

  /**
   * Adds the specified transition id to the list of output transition ids.
   * 
   * @param transitionId - The specified output transition.
   * @return {@code true} if the specified transition has been added to the list or already contains
   *         in it, {@code false} otherwise.
   */
  public boolean addOutputTransiton(Integer transitionId) {
    if (outputTransitions.contains(transitionId)) {
      return true;
    } else {
      return outputTransitions.add(transitionId);
    }
  }

  /**
   * Removes the specified input transition id from the list of input transition ids.
   * 
   * @param removeId - The specified id to remove.
   * @return {@code true} in case of success, {@code false} otherwise.
   */
  public boolean removeInputTransition(Integer removeId) {
    return inputTransitions.remove(removeId);
  }

  /**
   * Removes the specified output transition id from the list of output transition ids.
   * 
   * @param removeId - The specified id to remove.
   * @return {@code true} in case of success, {@code false} otherwise.
   */
  public boolean removeOutputTransition(Integer removeId) {
    return outputTransitions.remove(removeId);
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    return "\nPlace id=" + id + ",\n    inputTransitions=" + inputTransitions
        + ",\n    outputTransitions=" + outputTransitions + "";
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + id;
    return result;
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    Place other = (Place) obj;
    if (id != other.id)
      return false;
    return true;
  }

}
