/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.petri;

import ru.ispras.retrascope.model.basis.MetaInfo;

import java.util.ArrayList;



/**
 * This class represents a transition of a Petri net.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 * 
 */
public class Transition extends MetaInfo {
  private int id;
  private ArrayList<Integer> inputPlaces;
  private ArrayList<Integer> outputPlaces;

  /**
   * Constructs a Transition object using the provided id value and input and output places lists.
   * 
   * @param id - The specified id.
   * @param inputPlaces - The specified list of input places.
   * @param outputPlaces - The specified list of output places.
   */
  public Transition(int id, ArrayList<Integer> inputPlaces, ArrayList<Integer> outputPlaces) {
    this.id = id;
    this.inputPlaces = inputPlaces;
    this.outputPlaces = outputPlaces;

  }

  /**
   * Constructs a Transition object using the provided id value.
   * 
   * @param id - The specified id.
   */
  public Transition(int id) {
    this.id = id;
    this.inputPlaces = new ArrayList<Integer>();
    this.outputPlaces = new ArrayList<Integer>();
  }



  /**
   * Returns the transition id.
   * 
   * @return The id.
   */
  public int getId() {
    return id;
  }

  /**
   * Sets the transition id.
   * 
   * @param id - The new id.
   */
  public void setId(int id) {
    this.id = id;
  }

  /**
   * Returns the list of input place ids.
   * 
   * @return A list of input ids.
   */
  public ArrayList<Integer> getInputPlaces() {
    return inputPlaces;
  }

  /**
   * Sets the list of input place ids.
   * 
   * @param inputPlaces - The specified list.
   */
  public void setInputPlaces(ArrayList<Integer> inputPlaces) {
    this.inputPlaces = inputPlaces;
  }

  /**
   * Adds the specified place id to the list of input place ids.
   * 
   * @param placeId - The specified id.
   * @return {@code true} if the specified place has been added to the list or already contains in
   *         it, {@code false} otherwise.
   */
  public boolean addInputPlace(Integer placeId) {
    if (inputPlaces.contains(placeId)) {
      return true;
    } else {
      return inputPlaces.add(placeId);
    }
  }

  /**
   * Returns the list of output place ids.
   * 
   * @return A list of output ids.
   */
  public ArrayList<Integer> getOutputPlaces() {
    return outputPlaces;
  }

  /**
   * Sets the list of output place ids.
   * 
   * @param outputPlaces - The specified list.
   */
  public void setOutputPlaces(ArrayList<Integer> outputPlaces) {
    this.outputPlaces = outputPlaces;
  }

  /**
   * Adds the specified place id to the list of output place ids.
   * 
   * @param placeId - The specified id.
   * @return {@code true} if the specified place has been added to the list or already contains in
   *         it, {@code false} otherwise.
   */
  public boolean addOutputPlace(Integer placeId) {
    if (outputPlaces.contains(placeId)) {
      return true;
    } else {
      return outputPlaces.add(placeId);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    return "\nTransition id=" + id + ",\n    inputPlaces=" + inputPlaces + ",\n    outputPlaces="
        + outputPlaces + "";
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + id;
    return result;
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    Transition other = (Transition) obj;
    if (id != other.id)
      return false;
    return true;
  }

}
