/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.engine.cfg.CfgDfsWrVisitorBackend;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Cfg;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;

/**
 * This component merges sub-graphs of {@link ru.ispras.retrascope.model.cfg.Cfg} control flow
 * graphs which have the following structure:
 *         switch
 *        /      \
 *     case1    case2
 *       |       |
 *    block1   block2 (or nothing)
 *       \      /
 *        merge
 * Every {@link ru.ispras.retrascope.model.cfg.BasicBlock} basic block in such sub-graph
 * contains exactly one {@link ru.ispras.retrascope.model.basis.Assignment} assignment.
 * If the hammock sub-graph contains two blocks, the define same variable.
 *
 * The component merges such sub-graph.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgAtomicHammockBackend extends CfgDfsWrVisitorBackend {

  private Cfg cfg;

  @Override
  public void onRootBegin() { /* Empty. */ }

  @Override
  public void onRootEnd() { /* Empty. */ }

  @Override
  public void onCfgModelBegin() { /* Empty. */ }

  @Override
  public void onCfgModelEnd() { /* Empty. */ }

  @Override
  public void onModuleBegin(Module module) { /* Empty. */ }

  @Override
  public void onModuleEnd(Module module) { /* Empty. */ }

  @Override
  public void onProcessBegin(Process process) { /* Empty. */ }

  @Override
  public void onProcessEnd(Process process) { /* Empty. */ }

  @Override
  public void onCfgBegin(Cfg cfg) { this.cfg = cfg; }

  @Override
  public void onCfgEnd(Cfg cfg) { this.cfg = null; }

  @Override
  public void onBasicBlockBegin(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onBasicBlockEnd(BasicBlock basicBlock) { /* Empty. */ }

  @Override
  public void onCaseBegin(Case aCase) { /* Empty. */ }

  @Override
  public void onCaseEnd(Case aCase) { /* Empty. */ }

  @Override
  public void onSwitchBegin(Switch aSwitch) { /* Empty. */ }

  @Override
  public void onSwitchEnd(Switch aSwitch) {
    /* If this node is parent for atomic hammock,
    * collapse the hammock. */
    collapseAtomicHammock(aSwitch);
  }

  private void collapseAtomicHammock(Switch aSwitch) {
    Collection<CfgModelNode> cases = aSwitch.getChildren();
    if (cases.size() == 2) {
      Iterator<CfgModelNode> iterator = cases.iterator();
      Case case1 = (Case) iterator.next();
      CfgModelNode block1 = case1.getOnlyChild();
      Case case2 = (Case) iterator.next();
      CfgModelNode block2 = case2.getOnlyChild();

      if (block1.hasChildren() &&
        block2.hasChildren() &&
        block1.getChildren().equals(block2.getChildren())) {

        /* It is a hammock. */
        BasicBlock bBlock1 = null, bBlock2 = null;
        Merge merge1 = null, merge2 = null;
        if (block1 instanceof BasicBlock) {
          bBlock1 = (BasicBlock) block1;
        } else if (block1 instanceof Merge) {
          merge1 = (Merge) block1;
        }
        if (block2 instanceof BasicBlock) {
          bBlock2 = (BasicBlock) block2;
        } else if (block2 instanceof Merge) {
          merge2 = (Merge) block2;
        }

        BasicBlock newBlock = null;
        NodeOperation expr = getExpr(aSwitch, case1, case2);

        if (bBlock1 != null && bBlock2 != null) {
          /* Both grand-children are blocks. */
          if (bBlock1.getAssignments().size() == 1 &&
            bBlock2.getAssignments().size() == 1 &&
            bBlock1.getDefines().equals(bBlock2.getDefines()) &&
            bBlock1.isConcurrent() == bBlock2.isConcurrent()) {
            /* Each block contains only one assignment, both assignments define the same variable. */
            Assignment bBlock1Assignment = bBlock1.getAssignments().get(0);
            Assignment bBlock2Assignment = bBlock2.getAssignments().get(0);
            RangedVariable variable = bBlock1Assignment.getRangedVariable();
            Assignment result = new Assignment(
              variable,
              new NodeOperation(
                StandardOperation.ITE,
                expr,
                case1.isDefault() ?
                  bBlock2Assignment.getExpression() : bBlock1Assignment.getExpression(),
                case1.isDefault() ?
                  bBlock1Assignment.getExpression() : bBlock2Assignment.getExpression()));
            newBlock = new BasicBlock(result, bBlock1.isConcurrent());
          }
        } else if (merge1 != null && bBlock2 != null) {
          newBlock = getBasicBlock(case1, bBlock2, expr);
        } else if (bBlock1 != null && merge2 != null) {
          newBlock = getBasicBlock(case1, bBlock1, expr);
        }
        if (newBlock != null) {

          /* Add new node. */
          CfgModelNode parent = aSwitch.getOnlyParent();
          parent.removeChild(aSwitch);
          newBlock.addParent(parent);
          CfgModelNode child;
          if (block1 instanceof BasicBlock) {
            child = block1.getOnlyChild();
            child.removeParent(block1);
          } else {
            child = block2.getOnlyChild();
            child.removeParent(block2);
          }
          newBlock.addChild(child);
          this.cfg.addNode(newBlock);

          /* Remove non-needed nodes. */
          this.cfg.removeNode(aSwitch);
          this.cfg.removeNode(case1);
          this.cfg.removeNode(case2);
          if (block1 instanceof BasicBlock) {
            this.cfg.removeNode((CfgNode) block1);
          }
          if (block2 instanceof BasicBlock) {
            this.cfg.removeNode((CfgNode) block2);
          }
        }
      }
    }
  }

  private BasicBlock getBasicBlock(Case aCase, BasicBlock block, NodeOperation expr) {
    BasicBlock newBlock = null;
    if (block.getAssignments().size() == 1) {
      Assignment bBlock2Assignment = block.getAssignments().get(0);
      RangedVariable variable = block.getAssignments().get(0).getRangedVariable();
      Assignment result = new Assignment(variable, new NodeOperation(
        StandardOperation.ITE,
        expr,
        aCase.isDefault() ?
          bBlock2Assignment.getExpression() : variable.getVariable(),
        aCase.isDefault() ?
          variable.getVariable() : bBlock2Assignment.getExpression()));
      newBlock = new BasicBlock(result, block.isConcurrent());
    }
    return newBlock;
  }

  private NodeOperation getExpr(Switch aSwitch, Case case1, Case case2) {
    NodeOperation expr;
    if (case1.isDefault()) {
      Set<NodeValue> case2Values = case2.getValues();
      Iterator<NodeValue> case2ValIterator = case2Values.iterator();
      Node[] nodes = new Node[case2Values.size()];
      for (int i = 0; i < nodes.length; i++) {
        nodes[i] = new NodeOperation(
          StandardOperation.EQ,
          aSwitch.getExpression(),
          case2ValIterator.next());
      }
      expr = new NodeOperation(StandardOperation.OR, nodes);
    } else {
      Set<NodeValue> case1Values = case1.getValues();
      Iterator<NodeValue> case1ValIterator = case1Values.iterator();
      Node[] nodes = new Node[case1Values.size()];
      for (int i = 0; i < nodes.length; i++) {
        nodes[i] = new NodeOperation(
          StandardOperation.EQ,
          aSwitch.getExpression(),
          case1ValIterator.next());
      }
      if (nodes.length > 1) {
        expr = new NodeOperation(StandardOperation.OR, nodes);
      } else {
        return (NodeOperation) nodes[0];
      }
    }
    return expr;
  }

  @Override
  public void onMergeBegin(Merge merge) { /* Empty. */ }

  @Override
  public void onMergeEnd(Merge merge) { /* Empty. */ }

  @Override
  public void onSrcBegin(Source source) { /* Empty. */ }

  @Override
  public void onSrcEnd(Source source) { /* Empty. */ }

  @Override
  public void onSink(Sink sink) { /* Empty. */ }
}
