/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.ast;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.cfg.Module;

/**
 * This class represents the entity of the Ast model. It contains a mapping of a Cfg Module to a
 * collection of {@link Ast}.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class AstModel extends Entity {

  public static final EntityType TYPE = EntityType.get("ast");

  private Map<Module, Collection<Ast>> moduleMap;

  /**
   * Creates a new abstract syntax tree model object.
   */
  public AstModel() {
    super(TYPE);
    this.moduleMap = new LinkedHashMap<Module, Collection<Ast>>();
  }

  /**
   * Creates a new abstract syntax tree model object with the specified mapping.
   * 
   * @param ast - the specified Map<Module, Collection<Ast>>.
   */
  public AstModel(Map<Module, Collection<Ast>> ast) {
    super(TYPE);
    this.moduleMap = new LinkedHashMap<Module, Collection<Ast>>();
    this.moduleMap.putAll(ast);
  }

  @Override
  public void add(Entity entity) {
    if (entity instanceof AstModel) {
      this.addAllTrees(((AstModel) entity).getModuleMap());
    } else {
      throw new IllegalArgumentException();
    }
  }

  /**
   * Adds the specified mapping to the map.
   * 
   * @param module - the specified {@link Module}.
   * @param nodeCollection - the specified collection of {@link Ast}.
   */
  public void addTrees(Module module, Collection<Ast> nodeCollection) {
    InvariantChecks.checkNotNull(nodeCollection);
    InvariantChecks.checkNotNull(module);

    moduleMap.put(module, nodeCollection);
  }

  /**
   * Adds the specified map to current mapping.
   * 
   * @param nodeMap - the specified map.
   */
  public void addAllTrees(Map<Module, Collection<Ast>> nodeMap) {
    InvariantChecks.checkNotNull(nodeMap);

    moduleMap.putAll(nodeMap);
  }


  /**
   * Returns the mapping.
   * 
   * @return The Map<Module, Collection<Ast>> contained in this model.
   */
  public Map<Module, Collection<Ast>> getModuleMap() {
    return this.moduleMap;
  }

  /**
   * Returns the modules contained in the mapping.
   * 
   * @return The Set of modules.
   */
  public Set<Module> getModules() {
    return moduleMap.keySet();
  }

  /**
   * Returns the Ast collection mapped to the specified module.
   * 
   * @param module - the specified module.
   * @return The collection mapped to the module.
   */
  public Collection<Ast> getAstCollection(Module module) {
    return moduleMap.get(module);
  }

}
