/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.ast;

import java.util.Collection;
import java.util.LinkedHashSet;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;

/**
 * This class represents a node of an abstract syntax tree. Based on a {@link CfgNode}
 * representation.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class AstNode extends MetaInfo {

  private CfgNode cfgNode;
  private AstNode parent;
  private Collection<AstNode> children;
  private boolean isRoot;

  /**
   * Creates a new AstNode object using the specified {@link CfgNode}.
   * 
   * @param node - the specified {@link CfgNode}.
   */
  public AstNode(CfgNode node) {
    InvariantChecks.checkNotNull(node);

    this.cfgNode = node;
    children = new LinkedHashSet<AstNode>();
    isRoot = false;
  }

  /**
   * Creates a new AstNode object using the specified {@link CfgNode} and Ast parent node.
   * 
   * @param node - the specified {@link CfgNode}.
   * @param parent - the specified parent {@link AstNode}.
   */
  public AstNode(CfgNode node, AstNode parent) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(parent);

    this.cfgNode = node;
    children = new LinkedHashSet<AstNode>();
    isRoot = false;
    setParent(parent);
  }

  /**
   * Sets the specified {@link AstNode} as a parent of this node.
   * 
   * @param node - the specified parent {@link AstNode}.
   */
  public void setParent(AstNode node) {
    InvariantChecks.checkNotNull(node);

    this.parent = node;
    node.addChild(this);
  }

  /**
   * Returns the parent of this node.
   * 
   * @return The parent {@link AstNode}.
   */
  public AstNode getParent() {
    return this.parent;
  }

  /**
   * Sets a {@link CfgNode} associated with this node.
   * 
   * @param node - the specified {@link CfgNode}.
   */
  public void setCfgNode(CfgNode node) {
    InvariantChecks.checkNotNull(node);

    this.cfgNode = node;
  }

  /**
   * Returns the {@link CfgNode} associated with this node.
   * 
   * @return The associated {@link CfgNode}.
   */
  public CfgNode getCfgNode() {
    return this.cfgNode;
  }

  /**
   * Adds a child to this node.
   * 
   * @param child - the specified {@link AstNode}.
   */
  public void addChild(AstNode child) {
    InvariantChecks.checkNotNull(child);

    this.children.add(child);
  }

  /**
   * Returns the children of this node.
   * 
   * @return The collection of child {@link AstNode}.
   */
  public Collection<AstNode> getChildren() {
    return this.children;
  }

  /**
   * Sets the status of this node.
   * 
   * @param root - {@code true} if is root node, {@code false} otherwise.
   */
  public void setRoot(boolean root) {
    this.isRoot = root;
  }

  /**
   * Returns the status of this node.
   * 
   * @return {@code true} if this node is a root node, {@code false} otherwise.
   */
  public boolean isRoot() {
    return this.isRoot;
  }

  /**
   * Returns the type of the associated {@link CfgNode}.
   * 
   * @return The {@link CfgNodeType} of the associated CfgNode.
   */
  public CfgNodeType getType() {
    return this.cfgNode.getType();
  }

}
