/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.ast;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.Module;

/**
 * This class represents a walker of an abstract syntax tree.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class AstWalker {

  private AstVisitor visitor;
  private Set<AstNode> visitedNodes;

  /**
   * Constructs a new walker object using the specified {@link AstVisitor}.
   * 
   * @param visitor - the specified {@link AstVisitor}.
   */
  public AstWalker(AstVisitor visitor) {
    this.visitor = visitor;
    visitedNodes = new HashSet<AstNode>();
  }

  /**
   * Start walking a model.
   * 
   * @param ast - the specified {@link AstModel}.
   */
  public void start(AstModel ast) {
    InvariantChecks.checkNotNull(ast);
    
    visitor.onModelBegin(ast);
    for (Module module : ast.getModules()) {
      visitor.onModuleBegin(module);
      visitTrees(ast.getAstCollection(module));
      visitor.onModuleEnd(module);
    }
    visitor.onModelEnd(ast);

  }

  /**
   * Visits the root node.
   * 
   * @param node - the specified AstNode.
   */
  private void visitRoot(AstNode node) {
    InvariantChecks.checkNotNull(node);

    visitor.onRootNodeBegin();
    visit(node.getChildren());
    visitor.onRootNodeEnd();
  }

  /**
   * Visits the specified collection of abstract syntax trees.
   * 
   * @param trees - the specified collection of {@link Ast}.
   */
  private void visitTrees(Collection<Ast> trees) {
    for (Ast tree : trees) {
      visitor.onTreeBegin(tree);
      visitRoot(tree.getRootNode());
      visitor.onTreeEnd(tree);
    }
  }


  /**
   * Visits the specified collection of nodes.
   * 
   * @param nodes - the specified {@link AstNode} collection.
   */
  public void visit(Collection<AstNode> nodes) {
    for (AstNode node : nodes) {
      if (!visitedNodes.contains(node)) {
        visitedNodes.add(node);
        visitNode(node);
      }
    }
  }

  /**
   * Visits the specified node.
   * 
   * @param node - the specified {@link AstNode}.
   */
  public void visitNode(AstNode node) {
    InvariantChecks.checkNotNull(node);

    switch (node.getCfgNode().getType()) {

      case BASIC_BLOCK:
        visitor.onBasicBlockBegin(node);
        visit(node.getChildren());
        visitor.onBasicBlockEnd(node);
        break;

      case SWITCH:
        visitor.onSwitchBegin(node);
        visit(node.getChildren());
        visitor.onSwitchEnd(node);
        break;

      case CASE:
        visitor.onCaseBegin(node);
        visit(node.getChildren());
        visitor.onCaseEnd(node);
        break;

        //TODO: Add more cases?
        
      default:
        visitor.onNodeBegin(node);
        visit(node.getChildren());
        visitor.onNodeEnd(node);
        break;
    }
  }

}
