/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.testbench;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import ru.ispras.retrascope.testbench.message.MessageField;
import ru.ispras.retrascope.testbench.message.MessageType;

/**
 * The class representing an interface of a hardware device and a corresponding model.
 * 
 */
public class Interface implements Xmlable {
  private final String name;
  private final InterfaceType type;
  private final MessageType messageType;
  private final Adapter adapter;
  private final HashMap<String, Signal> signals;

  public enum InterfaceType {
    INPUT, OUTPUT
  }

  private Interface(String interfaceName, InterfaceType interfaceType,
      MessageType interfaceMessageType, Adapter interfaceAdapter,
      HashMap<String, Signal> interfaceSignals) {
    this.name = interfaceName;
    this.type = interfaceType;
    this.messageType = interfaceMessageType;
    this.adapter = interfaceAdapter;
    this.signals = interfaceSignals;
  }

  /**
   * Constructs an interface based on an XML file with the specified file name.
   * 
   * @param interfaceFileName - an input file name.
   * @return a constructed interface.
   */
  public Interface generateInterfaceFromXML(String interfaceFileName) throws SAXException,
      IOException, ParserConfigurationException {
    InterfaceFileHandler interfaceXMLHandler = new InterfaceFileHandler();
    SAXParserFactory
        .newInstance()
        .newSAXParser()
        .parse(new File(System.getProperty("user.dir") + "/" + interfaceFileName),
            interfaceXMLHandler);

    return interfaceXMLHandler.generatedInterface;
  }

  /**
   * Returns a name of the interface.
   * 
   * @return a name of the interface.
   */
  public String getName() {
    return name;
  }

  /**
   * Returns a type of the interface.
   * 
   * @return a type of the interface.
   */
  public InterfaceType getType() {
    return type;
  }

  /**
   * Returns a message type related to the interface.
   * 
   * @return a message type related to the interface.
   */
  public MessageType getMessageType() {
    return messageType;
  }

  /**
   * Returns a signal which the specified field mapped to.
   * 
   * @return a signal which the specified message field mapped to
   */
  public Signal getMapping(MessageField field) {
    return this.adapter.getMapping(field);
  }

  /**
   * Returns a message field which the specified field mapped to.
   * 
   * @return a message field which the specified signal mapped to
   */
  public MessageField getMapping(Signal signal) {
    return this.adapter.getMapping(signal);
  }

  /**
   * Returns an XML representation of the adapter.
   * 
   * @return an XML representation of the adapter.
   */
  @Override
  public String toXML() {

    /*
     * xml.append("<interface name = \"" + name + "\" type = \"" + type.toString().toLowerCase() +
     * "\">"); xml.append(Interface.lineSeparator); xml.append(messageType.toXML());
     * xml.append(Interface.lineSeparator); xml.append("</interface>");
     */

    return "";
  }

  private class InterfaceFileHandler extends DefaultHandler {
    private Interface generatedInterface = null;

    private String interfaceName = null;
    private InterfaceType interfaceType = null;
    private final MessageType interfaceMessageType = null;
    private Adapter interfaceAdapter = null;
    private final HashMap<String, Signal> interfaceSignals = new HashMap<>();

    @Override
    public void endDocument() throws SAXException {
      this.generatedInterface =
          new Interface(interfaceName, interfaceType, interfaceMessageType, interfaceAdapter,
              interfaceSignals);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes)
        throws SAXException {
      if (qName.toLowerCase().equals("interface")) {
        if (attributes != null) {
          this.interfaceName = attributes.getValue("name");
          if (this.interfaceName == null)
            throw new SAXException("No name specified for an interface");

          if (attributes.getValue("type") == null)
            throw new SAXException("No type specified for an interface");
          this.interfaceType = InterfaceType.valueOf(attributes.getValue("type"));
        }
      } else if (qName.toLowerCase().equals("messagetype")) {
        if (attributes != null) {
          String messageName = attributes.getValue("name");
          if (messageName == null)
            throw new SAXException("No name specified for an interface");
        }
      } else if (qName.toLowerCase().equals("messagefield")) {
        String name = attributes.getValue("name");
        if (name == null)
          throw new SAXException("No name specified for a message field");

        if (attributes.getValue("capacity") == null)
          throw new SAXException("No capacity specified for a message field");
        int capacity = Integer.parseInt(attributes.getValue("capacity"));

        Integer mask = null;
        if (attributes.getValue("mask") != null)
          mask = Integer.parseInt(attributes.getValue("mask").substring(2), 16);

        this.interfaceMessageType.addField(new MessageField(name, capacity, mask));
      } else if (qName.toLowerCase().equals("signal")) {
        String name = attributes.getValue("name");
        if (name == null)
          throw new SAXException("No name specified for a signal");

        if (attributes.getValue("capacity") == null)
          throw new SAXException("No capacity specified for a signal");
        int capacity = Integer.parseInt(attributes.getValue("capacity"));

        this.interfaceSignals.put(name, new Signal(name, capacity));
      } else if (qName.toLowerCase().equals("adapter"))
        this.interfaceAdapter = new Adapter();
      else if (qName.toLowerCase().equals("mapping")) {
        String fieldName = attributes.getValue("fieldName");
        if (fieldName == null)
          throw new SAXException("No field specified for a mapping");

        String signalName = attributes.getValue("signalName");
        if (signalName == null)
          throw new SAXException("No signal specified for a mapping");

        this.interfaceAdapter.addMapping(this.interfaceSignals.get(signalName),
            this.interfaceMessageType.getFields().get(fieldName));
      }
    }
  }
}
