/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.engine.basis;

import org.junit.Test;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.printer.ExprPrinter;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.List;

/**
 * The test for {@link ru.ispras.retrascope.engine.basis.TransformUtils} class.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class TransFormUtilsTestCase {

  private static final NodeValue ZERO = NodeValue.newInteger(0);
  private static final NodeValue ONE = NodeValue.newInteger(1);
  private static final NodeValue TWO = NodeValue.newInteger(2);
  private static final NodeValue THREE = NodeValue.newInteger(3);
  private static final NodeValue TRUE = NodeValue.newBoolean(true);
  private static final NodeValue FALSE = NodeValue.newBoolean(false);

  @Test
  public void runTest() {

    /* The tests for class methods. */

    runSubstituteBackwardTest();
    runGetSubGraphTest();
  }

  /* The method for {@code TransformUtils.getSubGraph} method testing. */
  private void runGetSubGraphTest() {

    /**
     * Create the following graph for the sub-graph to be extracted:
     *
     *                       switch(x > 0)
     *                       |           |
     *                     true        false
     *                       |           |
     *                       |       block(x := 1)
     *                       |           |
     *                       |      switch(y > 0)
     *                       |      |           |
     *                       |    true        false
     *                       |      |           |
     *                       | block(x := 2) block(x := 3)
     *                       |      |           |
     *                       | block(y := 1) block(y := 2)
     *                       |      |           |
     *                    merge_____|___________|
     *
     *
     * The sub-graph should be:
     *     block(x := 2)
     *          |
     *     block(y := 1)
     */

    final NodeVariable x = new NodeVariable("x", DataType.INTEGER);
    final NodeVariable y = new NodeVariable("y", DataType.INTEGER);

    /* Create graph nodes. */
    final Switch xSwitch = new Switch(new NodeOperation(StandardOperation.GREATER, x, ZERO));
    final Case xTrueCase = new Case(TRUE);
    final Case xFalseCase = new Case(FALSE);
    final BasicBlock x1Block = new BasicBlock(new Assignment(x, ONE));
    final Switch ySwitch = new Switch(new NodeOperation(StandardOperation.GREATER, y, ZERO));
    final Case yTrueCase = new Case(TRUE);
    final Case yFalseCase = new Case(FALSE);
    final BasicBlock x2Block = new BasicBlock(new Assignment(x, TWO));
    final BasicBlock y1Block = new BasicBlock(new Assignment(y, ONE));
    final BasicBlock x3Block = new BasicBlock(new Assignment(x, THREE));
    final BasicBlock y2Block = new BasicBlock(new Assignment(y, TWO));
    final Merge merge = new Merge();

    /* Create links between graph nodes. */
    xSwitch.addChild(xTrueCase);
    xSwitch.addChild(xFalseCase);
    xTrueCase.addChild(merge);
    xFalseCase.addChild(x1Block);
    x1Block.addChild(ySwitch);
    ySwitch.addChild(yTrueCase);
    ySwitch.addChild(yFalseCase);
    yTrueCase.addChild(x2Block);
    x2Block.addChild(y1Block);
    y1Block.addChild(merge);
    yFalseCase.addChild(x3Block);
    x3Block.addChild(y2Block);
    y2Block.addChild(merge);

    /* Create the sub-graph. */
    final BasicBlock x2SubGraphBlock = new BasicBlock(new Assignment(x, TWO));
    final BasicBlock y1SubGraphBlock = new BasicBlock(new Assignment(y, ONE));

    final List<CfgNode> result = new ArrayList<>();
    result.add(x2SubGraphBlock);
    result.add(y1SubGraphBlock);
    x2SubGraphBlock.addChild(y1SubGraphBlock);

    /* Get the sub-graph & compare it. */
    List<CfgNode> path = TransformUtils.getPath(yTrueCase, merge);

    if (/*!path.equals(result)*/path.size() != result.size()) {
      throw new IllegalStateException("Wrong sub-graph has been extracted");
    }

  }

  /* The method for {@code TransformUtils.transformBackward} method testing. */
  private void runSubstituteBackwardTest() {

    NodeVariable x = new NodeVariable("x", DataType.INTEGER);
    NodeVariable y = new NodeVariable("y", DataType.INTEGER);
    NodeVariable z = new NodeVariable("z", DataType.BIT_VECTOR(3));
    NodeVariable v = new NodeVariable("v", DataType.MAP(DataType.INTEGER, DataType.INTEGER));
    List<BasicBlock> blocks = new ArrayList<>();

    runSBdSimpleTest(x, y, blocks);
    runSBNoSSATest(x, y, blocks);
    runSBBitVectorTest(z, blocks);
    runSBArrayTest(x, v, blocks);
  }

  private void runSBdSimpleTest(
    NodeVariable x, NodeVariable y, List<BasicBlock> blocks) {

    /**
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      x := 1;
     *        |
     *      y := 1;
     *        |
     * switch(x + y <= 2)
     *
     * The result should be: true
     *
     */
    blocks.clear();
    blocks.add(new BasicBlock(new Assignment(x, ONE)));
    blocks.add(new BasicBlock(new Assignment(y, ONE)));

    NodeOperation switchExpr =
        new NodeOperation(
            StandardOperation.LESSEQ,
            new NodeOperation(StandardOperation.ADD, x, y), TWO);

    NodeValue correctResult = NodeValue.newBoolean(true);

    Node substituteResult = TransformUtils.substituteBackward(blocks, switchExpr);
    String result = ExprPrinter.VHDL.toString(substituteResult);

    if (!substituteResult.equals(correctResult))
        throw new IllegalArgumentException(
            "Incorrect substitution result: " + result);
  }

  private void runSBNoSSATest(
    NodeVariable x, NodeVariable y, List<BasicBlock> blocks) {

    /**
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      y := 1;
     *        |
     *      x := y;
     *        |
     *      y := 2;
     *        |
     * switch(x + y <= 3)
     *
     * The result should be: 1 + 2 <= 3
     *
     */

    blocks.clear();
    blocks.add(new BasicBlock(new Assignment(y, ONE)));
    blocks.add(new BasicBlock(new Assignment(x, y)));
    blocks.add(new BasicBlock(new Assignment(y, TWO)));

    NodeOperation switchExpr =
      new NodeOperation(
        StandardOperation.LESSEQ,
        new NodeOperation(StandardOperation.ADD, x, y), THREE);

    NodeValue correctResult = NodeValue.newBoolean(true);

    Node substituteResult = TransformUtils.substituteBackward(blocks, switchExpr);
    String result = ExprPrinter.VHDL.toString(substituteResult);

    if (!substituteResult.equals(correctResult))
      throw new IllegalArgumentException(
        "Incorrect substitution result: " + result);
  }

  private void runSBBitVectorTest(NodeVariable z, List<BasicBlock> blocks) {

    /**
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      z[0] := '1';
     *        |
     *      z[1] := '1';
     *        |
     *      z[2] := '0';
     *        |
     * switch((z - '011') == '000')
     *
     * The result should be: true
     *
     */

    blocks.clear();
    blocks.add(
      new BasicBlock(
        new Assignment(
          new RangedVariable(z, new Range(ZERO, ZERO)),
          NodeValue.newBitVector(BitVector.valueOf(1, 1)))));
    blocks.add(
      new BasicBlock(
        new Assignment(
          new RangedVariable(z, new Range(ONE, ONE)),
          NodeValue.newBitVector(BitVector.valueOf(1, 1)))));
    blocks.add(
      new BasicBlock(
        new Assignment(
          new RangedVariable(z, new Range(TWO, TWO)),
          NodeValue.newBitVector(BitVector.valueOf(0, 1)))));

    NodeValue constVector = NodeValue.newBitVector(BitVector.valueOf("011", 2, 3));
    NodeOperation switchExpr =
      new NodeOperation(
        StandardOperation.EQ,
        new NodeOperation(StandardOperation.BVSUB,
          z,
          constVector),
        NodeValue.newBitVector(BitVector.valueOf(0, 3)));

    Node transformedExpr = TransformUtils.substituteBackward(blocks, switchExpr);

    Node resultNode =
      new NodeOperation(
        StandardOperation.AND,
        transformedExpr,
        new NodeOperation(StandardOperation.EQ, z, constVector));

    /* Check rather the transformation result is satisfiable. */
    if (!ExprUtils.isSAT(resultNode)) {
      throw new IllegalArgumentException("Incorrect substitution result: " + transformedExpr);
    }
  }

  private void runSBArrayTest(NodeVariable x, NodeVariable v, List<BasicBlock> blocks) {
    /**
     * The CFG part (written in VHDL notation for example) to which method is applied is:
     *
     *      x := 0;
     *        |
     *      v[x] := 2;
     *        |
     *      v[1] := 3;
     *        |
     * switch(x[0] == 2)
     *
     * The result should be: true
     *
     */

    blocks.clear();
    blocks.add(new BasicBlock(new Assignment(x, ZERO)));
    blocks.add(new BasicBlock(new Assignment(new RangedVariable(v, new Range(x, x)), TWO)));
    blocks.add(new BasicBlock(new Assignment(new RangedVariable(v, new Range(ONE, ONE)), THREE)));

    NodeOperation switchExpr =
        new NodeOperation(
            StandardOperation.EQ, new NodeOperation(StandardOperation.SELECT, v, ZERO), TWO);

    Node transformedExpr = TransformUtils.substituteBackward(blocks, switchExpr);

    /* Check rather the transformation result is satisfiable. */
    if (!ExprUtils.isSAT(transformedExpr)) {
      throw new IllegalArgumentException("Incorrect substitution result: " + transformedExpr);
    }
  }
}
