/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import org.junit.Test;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * The test constructs control graph model {@link ru.ispras.retrascope.model.cfg.CfgModel}
 * of the module that is described in the following VHDL files:

driver.vhd:
------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity Driver is
port( clk: in std_logic;
reset: in std_logic;
z: out std_logic
);
end Driver;

architecture behav of Driver is
begin
  process(clk)
  begin
    if (clk'event and clk='1') then
      z <= '1';
    else
      assert(reset='1')
      severity ERROR;
    end if;
  end process;
end behav;
-----------------------------------

test.vhd
-----------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity test is
port (clk : in std_logic;
reset :in std_logic
);
end test;

architecture Behavioral of test is
signal c : std_logic_vector(3 downto 0);
begin
  process(clk,reset)
  begin
    if(clk'event and clk='1') then
      if(reset = '0') then
        c(0) <='1';
      end if;
    end if;
    if(reset='1') then
      c(0) <='0';
    end if;
  end process;
end Behavioral;
-----------------------------------

example.vhd
-----------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity example is
port(   clk: in std_logic;
reset: in std_logic;
z: out std_logic;
w: out std_logic
);
end example;

architecture behavior of example is

component driver
port(clk: in std_logic;
reset: in std_logic;
z: out std_logic
);
end component;

component test
port (clk: in std_logic;
reset :in std_logic
);
end component;

begin
  test1: test port map(
  clk => clk,
  reset => reset
  );

  driver1: driver port map(
  clk => clk,
  reset => reset,
  z => z
  );

  process(clk)
  begin
    if (reset='1') then
      w <= reset;
    else
      w <= not reset;
    end if;
  end process;
end behavior;

 @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>

 **/
public class ExampleCfgModelTestCase {
  private static final String STD_LOGIC = "std_logic";
  private static final String STD_3_LOGIC_VECTOR = "std_logic_vector(3 downto 0)";

  @Test
  public void runTest() {

    CfgModel model = new CfgModel();

    Module example = new Module("example");
    example.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);

    // module variables declaration
    NodeVariable clk = new NodeVariable("clk", DataType.BOOLEAN);
    VariableData clkData = new VariableData(VariableType.IN);
    clkData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    clk.setUserData(clkData);

    NodeVariable reset = new NodeVariable("reset", DataType.BOOLEAN);
    VariableData resetData = new VariableData(VariableType.IN);
    resetData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    reset.setUserData(resetData);

    NodeVariable z = new NodeVariable("z", DataType.BOOLEAN);
    VariableData zData = new VariableData(VariableType.OUT);
    zData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    z.setUserData(zData);

    NodeVariable w = new NodeVariable("w", DataType.BOOLEAN);
    VariableData wData = new VariableData(VariableType.OUT);
    wData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    w.setUserData(wData);

    example.addVariable(clk, null);
    example.addVariable(reset, null);
    example.addVariable(z, null);
    example.addVariable(w, null);

    /*
     * "example" module process
     */

    Process exampleProcess = new Process();

    Cfg exampleCfg = new Cfg();
    Switch eSwitch = new Switch(reset);
    Case c1 = new Case(new NodeValue(Data.newBoolean(true)));

    BasicBlock bb1 = new BasicBlock(true);
    bb1.addAssignment(new Assignment(w, reset));

    Case c2 = new Case(new NodeValue(Data.newBoolean(false)));

    BasicBlock bb2 = new BasicBlock(true);
    bb2.addAssignment(
      new Assignment(
        w,
        new NodeOperation(StandardOperation.NOT, reset)));

    Merge eMerge = new Merge();

    Source exampleSource = exampleCfg.getSource();
    Sink exampleSink = exampleCfg.getSink();

    // remove automatically created link
    exampleSource.removeChild(exampleSink);

    // add links
    exampleSource.addChild(eSwitch);
    eSwitch.addChild(c1);
    eSwitch.addChild(c2);
    c1.addChild(bb1);
    c2.addChild(bb2);
    bb1.addChild(eMerge);
    bb2.addChild(eMerge);
    eMerge.addChild(exampleSink);

    // add nodes
    exampleCfg.addNode(eSwitch);
    exampleCfg.addNode(c1);
    exampleCfg.addNode(c2);
    exampleCfg.addNode(bb1);
    exampleCfg.addNode(bb2);
    exampleCfg.addNode(eMerge);

    // add process
    exampleProcess.addChild(exampleCfg);
    example.addChild(exampleProcess);

    /*
     * "test" module instance
     */

    Module test1 = new Module("test");
    test1.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    test1.setInstanceName("test1");

    NodeVariable testClk = new NodeVariable("clk", DataType.BOOLEAN);
    VariableData testClkData = new VariableData(VariableType.IN);
    testClkData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    testClk.setUserData(testClkData);

    NodeVariable testReset = new NodeVariable("reset", DataType.BOOLEAN);
    VariableData testResetData = new VariableData(VariableType.IN);
    testResetData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    testReset.setUserData(testResetData);

    NodeVariable test1c = new NodeVariable("c", DataType.BIT_VECTOR(4));
    VariableData test1cData = new VariableData(VariableType.REG);
    test1cData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_3_LOGIC_VECTOR);
    test1c.setUserData(test1cData);

    test1.addVariable(testClk, null);
    test1.addVariable(testReset, null);
    test1.addVariable(test1c, null);

    Map<String, Node> test1Bindings = new LinkedHashMap<>();
    test1Bindings.put("clk", clk);
    test1Bindings.put("reset", reset);
    test1.setBindings(test1Bindings);

    Process t1Process = new Process();
    EventList t1List = new EventList();
    t1List.add(new Event(new RangedVariable(testClk)));
    t1List.add(new Event(new RangedVariable(testReset)));
    t1Process.setSensitivityList(t1List);

    Cfg t1Cfg = new Cfg();
    t1Process.addChild(t1Cfg);

    Switch test1s1 = new Switch(testClk);
    Case test1c1 = new Case(new NodeValue(Data.newBoolean(true)));
    Case test1c2 = new Case(new NodeValue(Data.newBoolean(false)));

    Switch test1s2 = new Switch(testReset);
    Case test1c3 = new Case(new NodeValue(Data.newBoolean(false)));

    Case test1c4 = new Case(new NodeValue(Data.newBoolean(true)));
    BasicBlock test1bb1 = new BasicBlock(true);
    NodeValue zero = NodeValue.newInteger(0);
    test1bb1.addAssignment(
      new Assignment(
        new RangedVariable(test1c, new Range(zero, zero)),
        new NodeValue(Data.newBoolean(true))));
    Merge test1m1 = new Merge();
    Merge test1m2 = new Merge();
    Switch test1s3 = new Switch(testReset);
    Case test1c5 = new Case(new NodeValue(Data.newBoolean(true)));
    Case test1c6 = new Case(new NodeValue(Data.newBoolean(false)));
    BasicBlock test1bb2 = new BasicBlock(true);
    test1bb2.addAssignment(
      new Assignment(
        new RangedVariable(test1c, new Range(zero, zero)),
        new NodeValue(Data.newBoolean(false))));
    Merge test1m3 = new Merge();

    // add links
    test1s1.addChild(test1c1);
    test1s1.addChild(test1c2);
    test1c1.addChild(test1s2);
    test1c2.addChild(test1m2);
    test1s2.addChild(test1c3);
    test1s2.addChild(test1c4);
    test1c3.addChild(test1bb1);
    test1c4.addChild(test1m1);
    test1bb1.addChild(test1m1);
    test1m1.addChild(test1m2);
    test1m2.addChild(test1s3);
    test1s3.addChild(test1c5);
    test1s3.addChild(test1c6);
    test1c5.addChild(test1bb2);
    test1bb2.addChild(test1m3);
    test1c6.addChild(test1m3);

    // add nodes to control flow graph
    CfgNode t1CfgSource = t1Cfg.getSource();
    CfgNode t1CfgSink = t1Cfg.getSink();
    t1CfgSource.removeChild(t1CfgSink);
    t1CfgSource.addChild(test1s1);
    test1m3.addChild(t1CfgSink);
    t1Cfg.addNode(test1s1);
    t1Cfg.addNode(test1c1);
    t1Cfg.addNode(test1c2);
    t1Cfg.addNode(test1s2);
    t1Cfg.addNode(test1c3);
    t1Cfg.addNode(test1c4);
    t1Cfg.addNode(test1bb1);
    t1Cfg.addNode(test1m1);
    t1Cfg.addNode(test1m2);
    t1Cfg.addNode(test1s3);
    t1Cfg.addNode(test1c5);
    t1Cfg.addNode(test1c6);
    t1Cfg.addNode(test1bb2);
    t1Cfg.addNode(test1m3);

    test1.addChild(t1Process);
    example.addChild(test1);

    /*
     * "driver" module instance
     */
    Module d1 = new Module("driver");
    d1.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    d1.setInstanceName("driver1");

    // variables declaration
    NodeVariable d1Clk = new NodeVariable("clk", DataType.BOOLEAN);
    VariableData d1ClkData = new VariableData(VariableType.IN);
    d1ClkData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    d1Clk.setUserData(d1ClkData);

    NodeVariable d1Reset = new NodeVariable("reset", DataType.BOOLEAN);
    VariableData d1ResetData = new VariableData(VariableType.IN);
    d1ResetData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    d1Reset.setUserData(d1ResetData);

    NodeVariable d1Z = new NodeVariable("z", DataType.BOOLEAN);
    VariableData d1ZData = new VariableData(VariableType.OUT);
    d1ZData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_LOGIC);
    d1Z.setUserData(d1ZData);

    d1.addVariable(d1Clk, null);
    d1.addVariable(d1Reset, null);
    d1.addVariable(d1Z, null);

    // bindings
    Map<String, Node> d1b = new LinkedHashMap<>();
    d1b.put("clk", clk);
    d1b.put("reset", reset);
    d1b.put("reset", z);
    d1.setBindings(d1b);

    Process d1p = new Process();
    d1.addChild(d1p);
    EventList d1pEvents = new EventList();
    d1pEvents.add(new Event(new RangedVariable(d1Clk)));

    Cfg d1pCfg = new Cfg();
    d1p.addChild(d1pCfg);

    // control flow graph nodes
    Switch d1s1 = new Switch(d1Clk);
    Case d1c1 = new Case(new NodeValue(Data.newBoolean(true)));
    Case d1c2 = new Case();
    d1c2.setDefault();
    BasicBlock d1bb1 = new BasicBlock(true);
    d1bb1.addAssignment(
      new Assignment(
        d1Z,
        new NodeValue(Data.newBoolean(true))));
    Switch d1s2 = new Switch(d1Reset);
    Case d1c3 = new Case(new NodeValue(Data.newBoolean(true)));
    Case d1c4 = new Case(new NodeValue(Data.newBoolean(false)));
    Merge d1m1 = new Merge();

    // add links
    CfgNode d1src = d1pCfg.getSource();
    CfgNode d1sink = d1pCfg.getSink();
    d1src.removeChild(d1sink);
    d1src.addChild(d1s1);
    d1s1.addChild(d1c1);
    d1s1.addChild(d1c2);
    d1c1.addChild(d1bb1);
    d1c2.addChild(d1s2);
    d1bb1.addChild(d1m1);
    d1s2.addChild(d1c4);
    d1c4.addChild(d1sink);
    d1c3.addChild(d1m1);

    d1pCfg.addNode(d1s1);
    d1pCfg.addNode(d1c1);
    d1pCfg.addNode(d1c2);
    d1pCfg.addNode(d1bb1);
    d1pCfg.addNode(d1s2);
    d1pCfg.addNode(d1c3);
    d1pCfg.addNode(d1c4);
    d1pCfg.addNode(d1m1);

    example.addChild(d1);

    model.addModule(example);
  }
}
