/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm.examples;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.constraint.Formulas;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.cfginterface.CfgSignalInterface;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class B04 {
  public static final Efsm EFSM = new Efsm("B04");
  public static final CfgInterface CFG_INTERFACE = new CfgInterface();

  /* VHDL string representations of variable types. */
  private static final String BIT_TYPE = "bit";
  private static final String INT_127_TYPE = "integer range 127 downto -128";
  private static final String INT_2_TYPE = "integer range 2 downto 0";
  private static final String INT_TYPE = "integer";

  static {
    String restartName = "RESTART";
    String averageName = "AVERAGE";
    String enableName = "ENABLE";
    String dataInName = "DATA_IN";

    String statoName = "stato";
    String rMaxName = "RMAX";
    String rMinName = "RMIN";
    String rLastName = "RLAST";
    String reg1Name = "REG1";
    String reg2Name = "REG2";
    String reg3Name = "REG3";
    String reg4Name = "REG4";
    String regdName = "REGD";
    String tempName = "temp";
    String resName = "RES";
    String aveName = "AVE";
    String enaName = "ENA";

    String dataOutName = "DATA_OUT";

    String clockName = "CLOCK";
    String resetName = "RESET";

    DataType restartType = DataType.BIT_VECTOR(1);
    DataType averageType = DataType.BIT_VECTOR(1);
    DataType enableType = DataType.BIT_VECTOR(1);
    DataType dataInType = DataType.INTEGER;

    DataType statoType = DataType.INTEGER;
    DataType rMaxType = DataType.INTEGER;
    DataType rMinType = DataType.INTEGER;
    DataType rLastType = DataType.INTEGER;
    DataType reg1Type = DataType.INTEGER;
    DataType reg2Type = DataType.INTEGER;
    DataType reg3Type = DataType.INTEGER;
    DataType reg4Type = DataType.INTEGER;
    DataType regdType = DataType.INTEGER;
    DataType tempType = DataType.INTEGER;
    DataType resType = DataType.BIT_VECTOR(1);
    DataType aveType = DataType.BIT_VECTOR(1);
    DataType enaType = DataType.BIT_VECTOR(1);

    DataType dataOutType = DataType.INTEGER;

    DataType clockType = DataType.BIT_VECTOR(1);
    DataType resetType = DataType.BIT_VECTOR(1);

    B04.EFSM.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    B04.EFSM.addInput(clockName, clockType);

    NodeVariable restart = new NodeVariable(restartName, restartType);
    VariableData restartData = new VariableData(VariableType.IN);
    restartData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    restart.setUserData(restartData);
    NodeVariable average = new NodeVariable(averageName, averageType);
    VariableData averageData = new VariableData(VariableType.IN);
    averageData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    average.setUserData(averageData);
    NodeVariable enable = new NodeVariable(enableName, enableType);
    VariableData enableData = new VariableData(VariableType.IN);
    enableData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    enable.setUserData(enableData);
    NodeVariable dataIn = new NodeVariable(dataInName, dataInType);
    VariableData dataInData = new VariableData(VariableType.IN);
    dataInData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    dataIn.setUserData(dataInData);

    NodeVariable stato = new NodeVariable(statoName, statoType);
    VariableData statoData = new VariableData(VariableType.REG);
    statoData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_2_TYPE);
    stato.setUserData(statoData);
    NodeVariable rmax = new NodeVariable(rMaxName, rMaxType);
    VariableData rmaxData = new VariableData(VariableType.REG);
    rmaxData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    rmax.setUserData(rmaxData);
    NodeVariable rmin = new NodeVariable(rMinName, rMinType);
    VariableData rminData = new VariableData(VariableType.REG);
    rminData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    rmin.setUserData(rminData);
    NodeVariable rlast = new NodeVariable(rLastName, rLastType);
    VariableData rlastData = new VariableData(VariableType.REG);
    rlastData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    rlast.setUserData(rlastData);
    NodeVariable reg1 = new NodeVariable(reg1Name, reg1Type);
    VariableData reg1Data = new VariableData(VariableType.REG);
    reg1Data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    reg1.setUserData(reg1Data);
    NodeVariable reg2 = new NodeVariable(reg2Name, reg2Type);
    VariableData reg2Data = new VariableData(VariableType.REG);
    reg2Data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    reg2.setUserData(reg2Data);
    NodeVariable reg3 = new NodeVariable(reg3Name, reg3Type);
    VariableData reg3Data = new VariableData(VariableType.REG);
    reg3Data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    reg3.setUserData(reg3Data);
    NodeVariable reg4 = new NodeVariable(reg4Name, reg4Type);
    VariableData reg4Data = new VariableData(VariableType.REG);
    reg4Data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    reg4.setUserData(reg4Data);
    NodeVariable regd = new NodeVariable(regdName, regdType);
    VariableData regdData = new VariableData(VariableType.REG);
    regdData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    regd.setUserData(regdData);
    NodeVariable temp = new NodeVariable(tempName, tempType);
    VariableData tempData = new VariableData(VariableType.REG);
    tempData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_TYPE);
    temp.setUserData(tempData);
    NodeVariable res = new NodeVariable(resName, resType);
    VariableData resData = new VariableData(VariableType.REG);
    resData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    res.setUserData(resData);
    NodeVariable ave = new NodeVariable(aveName, aveType);
    VariableData aveData = new VariableData(VariableType.REG);
    aveData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    ave.setUserData(aveData);
    NodeVariable ena = new NodeVariable(enaName, enaType);
    VariableData enaData = new VariableData(VariableType.REG);
    enaData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    ena.setUserData(enaData);

    NodeVariable dataOut = new NodeVariable(dataOutName, dataOutType);
    VariableData dataOutData = new VariableData(VariableType.OUT);
    dataOutData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_127_TYPE);
    dataOut.setUserData(dataOutData);

    NodeVariable clock = new NodeVariable(clockName, clockType);
    VariableData clockData = new VariableData(VariableType.IN);
    clockData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    clock.setUserData(clockData);
    NodeVariable reset = new NodeVariable(resetName, resetType);
    VariableData resetData = new VariableData(VariableType.IN);
    resetData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    reset.setUserData(resetData);

    RangedVariable restartRangedVariable = new RangedVariable();
    RangedVariable averageRangedVariable = new RangedVariable();
    RangedVariable enableRangedVariable = new RangedVariable();
    RangedVariable dataInRangedVariable = new RangedVariable();

    RangedVariable statoRangedVariable = new RangedVariable();

    RangedVariable clockRangedVariable = new RangedVariable();
    RangedVariable resetRangedVariable = new RangedVariable();

    restartRangedVariable.setVariable(restart);
    averageRangedVariable.setVariable(average);
    enableRangedVariable.setVariable(enable);
    dataInRangedVariable.setVariable(dataIn);

    statoRangedVariable.setVariable(stato);

    clockRangedVariable.setVariable(clock);
    resetRangedVariable.setVariable(reset);

    EfsmState sA = B04.EFSM.createState(new NodeOperation(StandardOperation.EQ, stato,
        new NodeValue(DataType.INTEGER.valueOf("0", 10))));
    B04.EFSM.setInitialState(sA);

    EfsmState sB = B04.EFSM.createState(new NodeOperation(StandardOperation.EQ, stato,
        new NodeValue(DataType.INTEGER.valueOf("1", 10))));

    EfsmState sC = B04.EFSM.createState(new NodeOperation(StandardOperation.EQ, stato,
        new NodeValue(DataType.INTEGER.valueOf("2", 10))));

    Node guardPredicate17 = new NodeOperation(StandardOperation.GREATER, dataIn, rmax);

    Formulas formulas = new Formulas();
    formulas.add(new NodeOperation(StandardOperation.LESS, dataIn, rmin));
    formulas.add(new NodeOperation(StandardOperation.LESSEQ, dataIn, rmax));
    Node guardPredicate18 = formulas.asSingleExpr();

    Node guardPredicate15 = new NodeOperation(StandardOperation.GREATEREQ,
        new NodeOperation(StandardOperation.ADD, dataIn, reg4),
        new NodeValue(DataType.INTEGER.valueOf("0", 10)));

    Node guardPredicate16 = new NodeOperation(StandardOperation.LESS,
        new NodeOperation(StandardOperation.ADD, dataIn,reg4),
        new NodeValue(DataType.INTEGER.valueOf("0", 10)));

    Node guardPredicate6 = new NodeOperation(StandardOperation.EQ, enable,
        new NodeValue(enableType.valueOf("1", 10)));

    Node guardPredicate7 = new NodeOperation(StandardOperation.EQ, enable,
        new NodeValue(enableType.valueOf("0", 10)));

    Node guardPredicate9 = new NodeOperation(StandardOperation.EQ, restart,
        new NodeValue(DataType.BIT_VECTOR(1).valueOf("1", 10)));

    formulas = new Formulas();
    formulas.add(new NodeOperation(StandardOperation.NOTEQ, restart,
        new NodeValue(DataType.BIT_VECTOR(1).valueOf("1", 10))));
    formulas.add(new NodeOperation(StandardOperation.NOTEQ, enable,
        new NodeValue(DataType.BIT_VECTOR(1).valueOf("1", 10))));
    Node guardPredicate8 = formulas.asSingleExpr();

    formulas = new Formulas();
    formulas.add(new NodeOperation(StandardOperation.NOTEQ, restart,
        new NodeValue(DataType.BIT_VECTOR(1).valueOf("1", 10))));
    formulas.add(new NodeOperation(StandardOperation.EQ, enable,
        new NodeValue(DataType.BIT_VECTOR(1).valueOf("1", 10))));
    Node guardPredicate10 = formulas.asSingleExpr();

    Node guardPredicate11 = new NodeOperation(StandardOperation.EQ, average,
        new NodeValue(DataType.BIT_VECTOR(1).valueOf("11", 10)));

    Node guardPredicate12 = new NodeOperation(StandardOperation.NOTEQ, average,
        new NodeValue(DataType.BIT_VECTOR(1).valueOf("1", 10)));

    Node guardPredicate13 = new NodeOperation(StandardOperation.GREATEREQ,
        new NodeOperation(StandardOperation.ADD, rmax, rmin),
        new NodeValue(DataType.INTEGER.valueOf("0", 10)));

    Node guardPredicate14 = new NodeOperation(StandardOperation.LESS,
        new NodeOperation(StandardOperation.ADD, rmax, rmin),
        new NodeValue(DataType.INTEGER.valueOf("0", 10)));

    Node guardPredicateTrue = NodeValue.newBoolean(true);

    ArrayList<Node> guardPredicateList1 = new ArrayList<>();
    guardPredicateList1.add(guardPredicate6);
    guardPredicateList1.add(guardPredicate10);
    guardPredicateList1.add(guardPredicate12);
    guardPredicateList1.add(guardPredicate15);
    guardPredicateList1.add(guardPredicate18);

    ArrayList<Node> guardPredicateList2 = new ArrayList<>();
    guardPredicateList2.add(guardPredicate7);
    guardPredicateList2.add(guardPredicate8);
    guardPredicateList2.add(guardPredicate17);

    ArrayList<Node> guardPredicateList3 = new ArrayList<>();
    guardPredicateList3.add(guardPredicate7);
    guardPredicateList3.add(guardPredicate7);
    guardPredicateList3.add(guardPredicate9);
    guardPredicateList3.add(guardPredicate13);
    guardPredicateList3.add(guardPredicate17);

    ArrayList<Node> guardPredicateList4 = new ArrayList<>();
    guardPredicateList4.add(guardPredicate7);
    guardPredicateList4.add(guardPredicate9);
    guardPredicateList4.add(guardPredicate14);
    guardPredicateList4.add(guardPredicate17);

    ArrayList<Node> guardPredicateList8 = new ArrayList<>();
    guardPredicateList8.add(guardPredicate7);
    guardPredicateList8.add(guardPredicate8);
    guardPredicateList8.add(guardPredicate18);

    ArrayList<Node> guardPredicateList9 = new ArrayList<>();
    guardPredicateList9.add(guardPredicate7);
    guardPredicateList9.add(guardPredicate9);
    guardPredicateList9.add(guardPredicate13);
    guardPredicateList9.add(guardPredicate18);

    ArrayList<Node> guardPredicateList10 = new ArrayList<>();
    guardPredicateList10.add(guardPredicate7);
    guardPredicateList10.add(guardPredicate9);
    guardPredicateList10.add(guardPredicate14);
    guardPredicateList10.add(guardPredicate18);

    ArrayList<Node> guardPredicateList22 = new ArrayList<>();
    guardPredicateList22.add(guardPredicateTrue);

    ArrayList<Node> guardPredicateList23 = new ArrayList<>();
    guardPredicateList23.add(guardPredicateTrue);

    ArrayList<Node> guardPredicateList25 = new ArrayList<>();
    guardPredicateList25.add(guardPredicate6);
    guardPredicateList25.add(guardPredicate9);
    guardPredicateList25.add(guardPredicate13);
    guardPredicateList25.add(guardPredicate17);

    ArrayList<Node> guardPredicateList26 = new ArrayList<>();
    guardPredicateList26.add(guardPredicate6);
    guardPredicateList26.add(guardPredicate9);
    guardPredicateList26.add(guardPredicate14);
    guardPredicateList26.add(guardPredicate17);

    ArrayList<Node> guardPredicateList27 = new ArrayList<>();
    guardPredicateList27.add(guardPredicate6);
    guardPredicateList27.add(guardPredicate10);
    guardPredicateList27.add(guardPredicate11);
    guardPredicateList27.add(guardPredicate17);

    ArrayList<Node> guardPredicateList28 = new ArrayList<>();
    guardPredicateList28.add(guardPredicate6);
    guardPredicateList28.add(guardPredicate10);
    guardPredicateList28.add(guardPredicate12);
    guardPredicateList28.add(guardPredicate15);
    guardPredicateList28.add(guardPredicate17);

    ArrayList<Node> guardPredicateList29 = new ArrayList<>();
    guardPredicateList29.add(guardPredicate6);
    guardPredicateList29.add(guardPredicate10);
    guardPredicateList29.add(guardPredicate12);
    guardPredicateList29.add(guardPredicate16);
    guardPredicateList29.add(guardPredicate17);

    ArrayList<Node> guardPredicateList31 = new ArrayList<>();
    guardPredicateList31.add(guardPredicate6);
    guardPredicateList31.add(guardPredicate9);
    guardPredicateList31.add(guardPredicate13);
    guardPredicateList31.add(guardPredicate18);

    ArrayList<Node> guardPredicateList32 = new ArrayList<>();
    guardPredicateList32.add(guardPredicate6);
    guardPredicateList32.add(guardPredicate9);
    guardPredicateList32.add(guardPredicate14);
    guardPredicateList32.add(guardPredicate18);

    ArrayList<Node> guardPredicateList33 = new ArrayList<>();
    guardPredicateList33.add(guardPredicate6);
    guardPredicateList33.add(guardPredicate10);
    guardPredicateList33.add(guardPredicate11);
    guardPredicateList33.add(guardPredicate18);

    ArrayList<Assignment> statementsB = new ArrayList<>(3);
    statementsB.add(new Assignment(res, restart));
    statementsB.add(new Assignment(ena, enable));
    statementsB.add(new Assignment(ave, average));

    ArrayList<Assignment> statementsD = new ArrayList<>(3);
    statementsD.add(new Assignment(rmax, dataIn));
    statementsD.add(new Assignment(rmin, dataIn));
    statementsD.add(new Assignment(reg1, new NodeValue(reg1Type.valueOf("0", 10))));
    statementsD.add(new Assignment(reg2, new NodeValue(reg2Type.valueOf("0", 10))));
    statementsD.add(new Assignment(reg3, new NodeValue(reg3Type.valueOf("0", 10))));
    statementsD.add(new Assignment(reg4, new NodeValue(reg4Type.valueOf("0", 10))));
    statementsD.add(new Assignment(rlast, new NodeValue(rLastType.valueOf("0", 10))));
    statementsD.add(new Assignment(dataOut,
        new NodeValue(dataOutType.valueOf("0", 10))));

    ArrayList<Assignment> statementsE = new ArrayList<>(3);
    statementsE.add(new Assignment(rlast, dataIn));

    ArrayList<Assignment> statementsG = new ArrayList<>(3);
    statementsG.add(new Assignment(regd,
        new NodeOperation(StandardOperation.MOD,
            new NodeOperation(StandardOperation.ADD, rmax, rmin),
            new NodeValue(DataType.INTEGER.valueOf("128", 10)))));
    statementsG.add(new Assignment(temp,
        new NodeOperation(StandardOperation.ADD, rmax, rmin)));

    ArrayList<Assignment> statementsJ = new ArrayList<>(3);
    statementsJ.add(new Assignment(dataOut,
        new NodeOperation(StandardOperation.DIV,
            regd,
            new NodeValue(DataType.INTEGER.valueOf("2", 10)))));

    ArrayList<Assignment> statementsK = new ArrayList<>(3);
    statementsK.add(new Assignment(dataOut,
        new NodeOperation(StandardOperation.MINUS,
            new NodeOperation(StandardOperation.DIV,
                new NodeOperation(StandardOperation.MINUS, regd),
                new NodeValue(DataType.INTEGER.valueOf("2", 10))))));

    ArrayList<Assignment> statementsH = new ArrayList<>(3);
    statementsH.add(new Assignment(dataOut, reg4));

    ArrayList<Assignment> statementsI = new ArrayList<>(3);
    statementsI.add(new Assignment(regd,
        new NodeOperation(StandardOperation.MOD,
            new NodeOperation(StandardOperation.ADD, dataIn, reg4),
            new NodeValue(DataType.INTEGER.valueOf("128", 10)))));
    statementsI.add(new Assignment(temp,
        new NodeOperation(StandardOperation.ADD, dataIn, reg4)));

    ArrayList<Assignment> statementsF = new ArrayList<>(3);
    statementsF.add(new Assignment(dataOut, rlast));

    ArrayList<Assignment> statementsN = new ArrayList<>(3);
    statementsN.add(new Assignment(rmax, dataIn));

    ArrayList<Assignment> statementsO = new ArrayList<>(3);
    statementsO.add(new Assignment(rmin, dataIn));

    ArrayList<Assignment> statementsP = new ArrayList<>(3);
    statementsP.add(new Assignment(reg4, reg3));
    statementsP.add(new Assignment(reg3, reg2));
    statementsP.add(new Assignment(reg2, reg1));
    statementsP.add(new Assignment(reg1, dataIn));

    ArrayList<Assignment> action1Statements = new ArrayList<>();
    action1Statements.addAll(statementsB);
    action1Statements.addAll(statementsE);
    action1Statements.addAll(statementsI);
    action1Statements.addAll(statementsJ);
    action1Statements.addAll(statementsO);
    action1Statements.addAll(statementsP);
    Action action1 = new Action(action1Statements);

    ArrayList<Assignment> action2Statements = new ArrayList<>();
    action2Statements.addAll(statementsB);
    action2Statements.addAll(statementsF);
    action2Statements.addAll(statementsN);
    action2Statements.addAll(statementsP);
    Action action2 = new Action(action2Statements);

    ArrayList<Assignment> action3Statements = new ArrayList<>();
    action3Statements.addAll(statementsB);
    action3Statements.addAll(statementsG);
    action3Statements.addAll(statementsJ);
    action3Statements.addAll(statementsN);
    action3Statements.addAll(statementsP);
    Action action3 = new Action(action3Statements);

    ArrayList<Assignment> action4Statements = new ArrayList<>();
    action4Statements.addAll(statementsB);
    action4Statements.addAll(statementsG);
    action4Statements.addAll(statementsK);
    action4Statements.addAll(statementsN);
    action4Statements.addAll(statementsP);
    Action action4 = new Action(action4Statements);

    ArrayList<Assignment> action8Statements = new ArrayList<>();
    action8Statements.addAll(statementsB);
    action8Statements.addAll(statementsF);
    action8Statements.addAll(statementsO);
    action8Statements.addAll(statementsP);
    Action action8 = new Action(action8Statements);

    ArrayList<Assignment> action9Statements = new ArrayList<>();
    action9Statements.addAll(statementsB);
    action9Statements.addAll(statementsG);
    action9Statements.addAll(statementsJ);
    action9Statements.addAll(statementsO);
    action9Statements.addAll(statementsP);
    Action action9 = new Action(action9Statements);

    ArrayList<Assignment> action10Statements = new ArrayList<>();
    action10Statements.addAll(statementsB);
    action10Statements.addAll(statementsG);
    action10Statements.addAll(statementsK);
    action10Statements.addAll(statementsO);
    action10Statements.addAll(statementsP);
    Action action10 = new Action(action10Statements);

    ArrayList<Assignment> action22Statements = new ArrayList<>();
    action22Statements.addAll(statementsB);
    Action action22 = new Action(action22Statements);

    ArrayList<Assignment> action23Statements = new ArrayList<>();
    action23Statements.addAll(statementsB);
    action23Statements.addAll(statementsD);
    Action action23 = new Action(action23Statements);

    ArrayList<Assignment> action25Statements = new ArrayList<>();
    action25Statements.addAll(statementsB);
    action25Statements.addAll(statementsE);
    action25Statements.addAll(statementsG);
    action25Statements.addAll(statementsJ);
    action25Statements.addAll(statementsN);
    action25Statements.addAll(statementsP);
    Action action25 = new Action(action25Statements);

    ArrayList<Assignment> action26Statements = new ArrayList<>();
    action26Statements.addAll(statementsB);
    action26Statements.addAll(statementsE);
    action26Statements.addAll(statementsG);
    action26Statements.addAll(statementsK);
    action26Statements.addAll(statementsN);
    action26Statements.addAll(statementsP);
    Action action26 = new Action(action26Statements);

    ArrayList<Assignment> action27Statements = new ArrayList<>();
    action27Statements.addAll(statementsB);
    action27Statements.addAll(statementsE);
    action27Statements.addAll(statementsH);
    action27Statements.addAll(statementsN);
    action27Statements.addAll(statementsP);
    Action action27 = new Action(action27Statements);

    ArrayList<Assignment> action28Statements = new ArrayList<>();
    action28Statements.addAll(statementsB);
    action28Statements.addAll(statementsE);
    action28Statements.addAll(statementsI);
    action28Statements.addAll(statementsJ);
    action28Statements.addAll(statementsN);
    action28Statements.addAll(statementsP);
    Action action28 = new Action(action28Statements);

    ArrayList<Assignment> action29Statements = new ArrayList<>();
    action29Statements.addAll(statementsB);
    action29Statements.addAll(statementsE);
    action29Statements.addAll(statementsI);
    action29Statements.addAll(statementsK);
    action29Statements.addAll(statementsN);
    action29Statements.addAll(statementsP);
    Action action29 = new Action(action29Statements);

    ArrayList<Assignment> action31Statements = new ArrayList<>();
    action31Statements.addAll(statementsB);
    action31Statements.addAll(statementsE);
    action31Statements.addAll(statementsG);
    action31Statements.addAll(statementsJ);
    action31Statements.addAll(statementsO);
    action31Statements.addAll(statementsP);
    Action action31 = new Action(action31Statements);

    ArrayList<Assignment> action32Statements = new ArrayList<>();
    action32Statements.addAll(statementsB);
    action32Statements.addAll(statementsE);
    action32Statements.addAll(statementsG);
    action32Statements.addAll(statementsK);
    action32Statements.addAll(statementsO);
    action32Statements.addAll(statementsP);
    Action action32 = new Action(action32Statements);

    ArrayList<Assignment> action33Statements = new ArrayList<>();
    action33Statements.addAll(statementsB);
    action33Statements.addAll(statementsE);
    action33Statements.addAll(statementsH);
    action33Statements.addAll(statementsO);
    action33Statements.addAll(statementsP);
    Action action33 = new Action(action33Statements);

    GuardedAction ga1 = new GuardedAction(new Guard(guardPredicateList1), action1);
    GuardedAction ga2 = new GuardedAction(new Guard(guardPredicateList2), action2);
    GuardedAction ga3 = new GuardedAction(new Guard(guardPredicateList3), action3);
    GuardedAction ga4 = new GuardedAction(new Guard(guardPredicateList4), action4);
    GuardedAction ga8 = new GuardedAction(new Guard(guardPredicateList8), action8);
    GuardedAction ga9 = new GuardedAction(new Guard(guardPredicateList9), action9);
    GuardedAction ga10 = new GuardedAction(new Guard(guardPredicateList10), action10);
    GuardedAction ga22 = new GuardedAction(new Guard(guardPredicateList22), action22);
    GuardedAction ga23 = new GuardedAction(new Guard(guardPredicateList23), action23);
    GuardedAction ga25 = new GuardedAction(new Guard(guardPredicateList25), action25);
    GuardedAction ga26 = new GuardedAction(new Guard(guardPredicateList26), action26);
    GuardedAction ga27 = new GuardedAction(new Guard(guardPredicateList27), action27);
    GuardedAction ga28 = new GuardedAction(new Guard(guardPredicateList28), action28);
    GuardedAction ga29 = new GuardedAction(new Guard(guardPredicateList29), action29);
    GuardedAction ga31 = new GuardedAction(new Guard(guardPredicateList31), action31);
    GuardedAction ga32 = new GuardedAction(new Guard(guardPredicateList32), action32);
    GuardedAction ga33 = new GuardedAction(new Guard(guardPredicateList33), action33);

    Event clockEvent = new Event(clockRangedVariable, EventType.POSITIVE_EDGE);

    Event resetEvent = new Event(resetRangedVariable, EventType.POSITIVE_EDGE);
    ArrayList<Event> resetEventList = new ArrayList<>(1);
    resetEventList.add(resetEvent);
    
    ga22.setClock(clockEvent);
    ga23.setClock(clockEvent);
    ga1.setClock(clockEvent);
    ga2.setClock(clockEvent);
    ga3.setClock(clockEvent);
    ga4.setClock(clockEvent);
    ga8.setClock(clockEvent);
    ga9.setClock(clockEvent);
    ga10.setClock(clockEvent);
    ga25.setClock(clockEvent);
    ga26.setClock(clockEvent);
    ga27.setClock(clockEvent);
    ga28.setClock(clockEvent);
    ga29.setClock(clockEvent);
    ga31.setClock(clockEvent);
    ga32.setClock(clockEvent);
    ga33.setClock(clockEvent);

    B04.EFSM.createTransition(sA, sB, ga22);
    B04.EFSM.createTransition(sB, sC, ga23);
    B04.EFSM.createTransition(sC, sC, ga1);
    B04.EFSM.createTransition(sC, sC, ga2);
    B04.EFSM.createTransition(sC, sC, ga3);
    B04.EFSM.createTransition(sC, sC, ga4);
    B04.EFSM.createTransition(sC, sC, ga8);
    B04.EFSM.createTransition(sC, sC, ga9);
    B04.EFSM.createTransition(sC, sC, ga10);
    B04.EFSM.createTransition(sC, sC, ga25);
    B04.EFSM.createTransition(sC, sC, ga26);
    B04.EFSM.createTransition(sC, sC, ga27);
    B04.EFSM.createTransition(sC, sC, ga28);
    B04.EFSM.createTransition(sC, sC, ga29);
    B04.EFSM.createTransition(sC, sC, ga31);
    B04.EFSM.createTransition(sC, sC, ga32);
    B04.EFSM.createTransition(sC, sC, ga33);
  }

  static {
    List<NodeVariable> inputs = new ArrayList<>(B04.EFSM.getInputDataTypes().size());

    for (Map.Entry<String,DataType> input: B04.EFSM.getInputDataTypes().entrySet()) {
      inputs.add(new NodeVariable(input.getKey(), input.getValue()));
    }

    List<NodeVariable> outputs = new ArrayList<>(B04.EFSM.getInputDataTypes().size());

    for (Map.Entry<String,DataType> output: B04.EFSM.getOutputDataTypes().entrySet()) {
      outputs.add(new NodeVariable(output.getKey(), output.getValue()));
    }

    CfgSignalInterface cfgSignalInterface = new CfgSignalInterface(inputs, outputs);
    B04.CFG_INTERFACE.addInterface(B04.EFSM.getName(), cfgSignalInterface);
  }

  private B04() {
  }
}
