/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm.examples;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.ArrayList;

/**
 * EFSM model of B13.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class B13 {

  private static EfsmModel efsmModel = new EfsmModel();
  private static Efsm process1 = new Efsm("B13_process1");
  private static Efsm process2 = new Efsm("B13_process2");
  private static Efsm process3 = new Efsm("B13_process3");
  private static Efsm process4 = new Efsm("B13_process4");
  private static Efsm process5 = new Efsm("B13_process5");

  /* VHDL string representations of variable types. */
  private static final String BIT_TYPE = "bit";
  private static final String INT_8_TYPE = "integer range 8 downto 0";
  private static final String BITVECTOR_7_0_TYPE = "bit_vector( 7 downto 0 )";
  private static final String INT_7_TYPE = "integer range 7 downto 0";
  private static final String INT_3_TYPE = "integer range 3 downto 0";
  private static final String INT_9_TYPE = "integer range 9 downto 0";
  private static final String INT_512_TYPE = "integer range 512 downto 0";


  public static EfsmModel getModel() {

    return efsmModel;
  }

  static {

    /* Add meta-info about HDL from that EFSMs are extracted. */
    process1.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    process2.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    process3.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    process4.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
    process5.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);

    efsmModel.addEfsm(process1);
    efsmModel.addEfsm(process2);
    efsmModel.addEfsm(process3);
    efsmModel.addEfsm(process4);
    efsmModel.addEfsm(process5);

    final int GP001 = 0;
    final int GP010 = 1;
    final int GP011 = 2;
    final int GP100 = 3;
    final int GP100w = 4;
    final int GP101 = 5;
    final int GP110 = 6;
    final int GP111 = 7;

    final int GP01 = 0;
    final int GP10 = 1;
    final int GP11 = 2;
    final int GP11w = 3;

    final int START_BIT = 0;
    final int STOP_BIT = 1;
    final int BIT0 = 2;
    final int BIT1 = 3;
    final int BIT2 = 4;
    final int BIT3 = 5;
    final int BIT4 = 6;
    final int BIT5 = 7;
    final int BIT6 = 8;
    final int BIT7 = 9;

    final int G_IDLE = 0;
    final int G_LOAD = 1;
    final int G_SEND = 2;
    final int G_WAIT_END = 3;

    final String clockName = "clock";
    final String resetName = "reset";
    final String eocName = "eoc";
    final String socName = "soc";
    final String loadDatoName = "load_dato";
    final String addMpx2Name = "add_mpx2";
    final String canaleName = "canale";
    final String muxEnName = "mux_en";
    final String dataInName = "data_in";
    final String dataOutName = "data_out";
    final String dsrName = "dsr";
    final String errorName = "error";

    final String s1Name = "S1";
    final String s2Name = "S2";
    final String mpxName = "mpx";
    final String rdyName = "rdy";
    final String sendDataName = "";
    final String confirmName = "confirm";
    final String shotName = "shot";
    final String sendEnName = "send_en";
    final String treName = "tre";
    final String outRegName = "out_reg";
    final String nextBitName = "next_bit";
    final String txEndName = "tx_end";
    final String itfcStateName = "itfc_state";
    final String sendName = "send";
    final String loadName = "load";
    final String txContaName = "tx_conta";

    final DataType clockDataType = DataType.BIT_VECTOR(1);
    final DataType resetDataType = DataType.BIT_VECTOR(1);
    final DataType eocDataType = DataType.BIT_VECTOR(1);
    final DataType socDataType = DataType.BIT_VECTOR(1);
    final DataType loadDatoDataType = DataType.BIT_VECTOR(1);
    final DataType addMpx2DataType = DataType.BIT_VECTOR(1);
    final DataType canaleDataType = DataType.BIT_VECTOR(9);
    final DataType muxEnDataType = DataType.BIT_VECTOR(1);
    final DataType dsrDataType = DataType.BIT_VECTOR(1);
    final DataType errorDataType = DataType.BIT_VECTOR(1);
    final DataType dataInDataType = DataType.BIT_VECTOR(8);
    final DataType dataOutDataType = DataType.BIT_VECTOR(1);

    final DataType s1DataType = DataType.BIT_VECTOR(8);
    final DataType s2DataType = DataType.BIT_VECTOR(4);
    final DataType mpxDataType = DataType.BIT_VECTOR(1);
    final DataType rdyDataType = DataType.BIT_VECTOR(1);
    final DataType sendDataDataType = DataType.BIT_VECTOR(1);
    final DataType confirmDataType = DataType.BIT_VECTOR(1);
    final DataType shotDataType = DataType.BIT_VECTOR(1);
    final DataType sendEnDataType = DataType.BIT_VECTOR(1);
    final DataType treDataType = DataType.BIT_VECTOR(1);
    final DataType outRegDataType = DataType.BIT_VECTOR(8);
    final DataType nextBitDataType = DataType.BIT_VECTOR(10);
    final DataType txEndDataType = DataType.BIT_VECTOR(1);
    final DataType itfcStateDataType = DataType.BIT_VECTOR(4);
    final DataType sendDataType = DataType.BIT_VECTOR(1);
    final DataType loadDataType = DataType.BIT_VECTOR(1);
    final DataType txContaDataType = DataType.BIT_VECTOR(513);

    final NodeVariable clock = new NodeVariable(clockName, clockDataType);
    final NodeVariable reset = new NodeVariable(resetName, resetDataType);
    final NodeVariable eoc = new NodeVariable(eocName, eocDataType);
    final NodeVariable soc = new NodeVariable(socName, socDataType);
    final NodeVariable loadDato = new NodeVariable(loadDatoName, loadDatoDataType);
    final NodeVariable addMpx2 = new NodeVariable(addMpx2Name, addMpx2DataType);
    final NodeVariable canale = new NodeVariable(canaleName, canaleDataType);
    final NodeVariable muxEn = new NodeVariable(muxEnName, muxEnDataType);
    final NodeVariable dsr = new NodeVariable(dsrName, dsrDataType);
    final NodeVariable error = new NodeVariable(errorName, errorDataType);
    final NodeVariable dataIn = new NodeVariable(dataInName, dataInDataType);
    final NodeVariable dataOut = new NodeVariable(dataOutName, dataOutDataType);


    final NodeVariable s1 = new NodeVariable(s1Name, s1DataType);
    final NodeVariable s2 = new NodeVariable(s2Name, s2DataType);
    final NodeVariable mpx = new NodeVariable(mpxName, mpxDataType);
    final NodeVariable rdy = new NodeVariable(rdyName, rdyDataType);
    final NodeVariable sendData = new NodeVariable(sendDataName, sendDataDataType);
    final NodeVariable confirm = new NodeVariable(confirmName, confirmDataType);
    final NodeVariable shot = new NodeVariable(shotName, shotDataType);
    final NodeVariable sendEn = new NodeVariable(sendEnName, sendEnDataType);
    final NodeVariable tre = new NodeVariable(treName, treDataType);
    final NodeVariable outReg = new NodeVariable(outRegName, outRegDataType);
    final NodeVariable nextBit = new NodeVariable(nextBitName, nextBitDataType);
    final NodeVariable txEnd = new NodeVariable(txEndName, txEndDataType);
    final NodeVariable itfcState = new NodeVariable(itfcStateName, itfcStateDataType);
    final NodeVariable send = new NodeVariable(sendName, sendDataType);
    final NodeVariable load = new NodeVariable(loadName, loadDataType);
    final NodeVariable txConta = new NodeVariable(txContaName, txContaDataType);


    final RangedVariable clockRangedVariable = new RangedVariable(clock);
    final RangedVariable resetRangedVariable = new RangedVariable(reset);

    Efsm[] efsms = {process1, process2, process3, process4, process5};
    for (Efsm efsm : efsms) {
      efsm.addInput(clockName, clockDataType);
      efsm.addInput(resetName, resetDataType);
      efsm.addInput(eocName, eocDataType);
      efsm.addInput(dataInName, dataInDataType);
      efsm.addInput(dsrName, dsrDataType);

      efsm.addOutput(socName, socDataType);
      efsm.addOutput(loadDatoName, loadDatoDataType);
      efsm.addOutput(addMpx2Name, addMpx2DataType);
      efsm.addOutput(canaleName, canaleDataType);
      efsm.addOutput(muxEnName, muxEnDataType);
      efsm.addOutput(errorName, errorDataType);
      efsm.addOutput(dataOutName, dataOutDataType);

      efsm.addInnerVariable(s1Name, s1DataType);
      efsm.addInnerVariable(s2Name, s2DataType);
      efsm.addInnerVariable(mpxName, mpxDataType);
      efsm.addInnerVariable(rdyName, rdyDataType);
      efsm.addInnerVariable(sendDataName, sendDataDataType);
      efsm.addInnerVariable(confirmName, confirmDataType);
      efsm.addInnerVariable(shotName, shotDataType);
      efsm.addInnerVariable(sendEnName, sendEnDataType);
      efsm.addInnerVariable(treName, treDataType);
      efsm.addInnerVariable(outRegName, outRegDataType);
      efsm.addInnerVariable(nextBitName, nextBitDataType);
      efsm.addInnerVariable(txEndName, txEndDataType);
      efsm.addInnerVariable(itfcStateName, itfcStateDataType);
      efsm.addInnerVariable(sendName, sendDataType);
      efsm.addInnerVariable(loadName, loadDataType);
      efsm.addInnerVariable(txContaName, txContaDataType);
    }

    /* Add meta-information & type. */

    /* Signals */
    VariableData clockUData = new VariableData(VariableType.IN);
    clockUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    clock.setUserData(clockUData);

    VariableData resetUData = new VariableData(VariableType.IN);
    resetUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    reset.setUserData(resetUData);

    VariableData eocUData = new VariableData(VariableType.IN);
    eocUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    eoc.setUserData(eocUData);

    VariableData dataInUData = new VariableData(VariableType.IN);
    dataInUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BITVECTOR_7_0_TYPE);
    dataIn.setUserData(dataInUData);

    VariableData dsrUData = new VariableData(VariableType.IN);
    dsrUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    dsr.setUserData(dsrUData);

    VariableData socUData = new VariableData(VariableType.OUT);
    socUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    soc.setUserData(socUData);

    VariableData loadDatoUData = new VariableData(VariableType.OUT);
    loadDatoUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    loadDato.setUserData(loadDatoUData);

    VariableData addMpx2UData = new VariableData(VariableType.OUT);
    addMpx2UData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    addMpx2.setUserData(addMpx2UData);

    VariableData canaleUData = new VariableData(VariableType.OUT);
    canaleUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_8_TYPE);
    canale.setUserData(canaleUData);

    VariableData muxEnUData = new VariableData(VariableType.OUT);
    muxEnUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    muxEn.setUserData(muxEnUData);

    VariableData errorUData = new VariableData(VariableType.OUT);
    errorUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    error.setUserData(errorUData);

    VariableData dataOutUData = new VariableData(VariableType.OUT);
    dataOutUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    dataOut.setUserData(dataOutUData);

    /* Registers */
    VariableData s1UData = new VariableData(VariableType.REG);
    s1UData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_7_TYPE);
    s1.setUserData(s1UData);

    VariableData s2UData = new VariableData(VariableType.REG);
    s2UData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_3_TYPE);
    s2.setUserData(s2UData);

    VariableData mpxUData = new VariableData(VariableType.REG);
    mpxUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    mpx.setUserData(new VariableData(VariableType.REG));

    VariableData rdyUData = new VariableData(VariableType.REG);
    rdyUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    rdy.setUserData(rdyUData);

    VariableData sendDataUData = new VariableData(VariableType.REG);
    sendDataUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    sendData.setUserData(sendDataUData);

    VariableData confirmUData = new VariableData(VariableType.REG);
    confirmUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    confirm.setUserData(confirmUData);

    VariableData shotUData = new VariableData(VariableType.REG);
    shotUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    shot.setUserData(shotUData);

    VariableData sendEnUData = new VariableData(VariableType.REG);
    sendEnUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    sendEn.setUserData(new VariableData(VariableType.REG));

    VariableData treUData = new VariableData(VariableType.REG);
    treUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    tre.setUserData(treUData);

    VariableData outRegUData = new VariableData(VariableType.REG);
    outRegUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BITVECTOR_7_0_TYPE);
    outReg.setUserData(outRegUData);

    VariableData nextBitUData = new VariableData(VariableType.REG);
    nextBitUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_9_TYPE);
    nextBit.setUserData(nextBitUData);

    VariableData txEndUData = new VariableData(VariableType.REG);
    txEndUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    txEnd.setUserData(txEndUData);

    VariableData itfcStateUData = new VariableData(VariableType.REG);
    itfcStateUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_3_TYPE);
    itfcState.setUserData(itfcStateUData);

    VariableData sendUData = new VariableData(VariableType.REG);
    sendUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    send.setUserData(sendUData);

    VariableData loadUData = new VariableData(VariableType.REG);
    loadUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, BIT_TYPE);
    load.setUserData(loadUData);

    VariableData txContaUData = new VariableData(VariableType.REG);
    txContaUData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, INT_512_TYPE);
    txConta.setUserData(txContaUData);

    Guard guardTrue = new Guard(NodeValue.newBoolean(true));

    // Clocks
    Event clockEvent = new Event(clockRangedVariable, EventType.POSITIVE_EDGE);
    Event resetEvent = new Event(resetRangedVariable, EventType.POSITIVE_EDGE);

    // process1 efsm

    final String contaTmpName = "conta_tmp";
    final DataType contaTmpDataType = DataType.BIT_VECTOR(9);
    final NodeVariable contaTmpVariable = new NodeVariable(contaTmpName, contaTmpDataType);
    contaTmpVariable.setUserData(new VariableData(VariableType.REG));
    process1.addInnerVariable(contaTmpName, contaTmpDataType);

    // States

    EfsmState gp001State =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP001), 16))));
    EfsmState gp010State =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP010), 16))));
    EfsmState gp011State =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP011), 16))));
    EfsmState gp101State =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP101), 16))));
    EfsmState gp110State =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP110), 16))));
    EfsmState gp111State =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP111), 16))));
    EfsmState gp100wState =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP100w), 16))));
    EfsmState gp100State =
        process1.createState(new NodeOperation(StandardOperation.EQ, s1, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP100), 15))));

    process1.setInitialState(gp001State);

    // Guards

    Node p1GuardPredicate1 =
        new NodeOperation(StandardOperation.EQ, rdy, new NodeValue(rdyDataType.valueOf("0",
            2)));
    Node p1GuardPredicate2 =
        new NodeOperation(StandardOperation.EQ, contaTmpVariable, new NodeValue(
            contaTmpDataType.valueOf("8", 16)));
    Node p1GuardPredicate3 =
        new NodeOperation(StandardOperation.EQ, eoc, new NodeValue(eocDataType.valueOf("1",
            2)));

    // Assignments

    ArrayList<Assignment> p1Statements1 = new ArrayList<>();
    p1Statements1
        .add(new Assignment(soc, new NodeValue(socDataType.valueOf("0", 2))));
    p1Statements1.add(new Assignment(canale, new NodeValue(canaleDataType.valueOf("0", 2))));
    p1Statements1.add(
        new Assignment(contaTmpVariable, new NodeValue(contaTmpDataType.valueOf("0", 2))));
    p1Statements1.add(new Assignment(sendData, new NodeValue(sendDataDataType.valueOf("0", 2))));
    p1Statements1.add(new Assignment(loadDato, new NodeValue(loadDatoDataType.valueOf("0", 2))));
    p1Statements1.add(new Assignment(muxEn, new NodeValue(muxEnDataType.valueOf("0", 2))));
    p1Statements1.add(new Assignment(s1, new NodeValue(s1DataType.valueOf("0", 16))));

    ArrayList<Assignment> p1Statements2 = new ArrayList<>();
    p1Statements2.add(new Assignment(muxEn, new NodeValue(muxEnDataType.valueOf("1", 2))));
    p1Statements2.add(new Assignment(s1, new NodeValue(s1DataType.valueOf("1", 16))));

    Assignment p1Statement3 = new Assignment(s1, new NodeValue(s1DataType.valueOf("2", 16)));

    ArrayList<Assignment> p1Statements4 = new ArrayList<>();
    p1Statements4
        .add(new Assignment(soc, new NodeValue(socDataType.valueOf("1", 2))));
    p1Statements4.add(new Assignment(s1, new NodeValue(s1DataType.valueOf("5", 16))));

    ArrayList<Assignment> p1Statements5 = new ArrayList<>();
    p1Statements5.add(new Assignment(loadDato, new NodeValue(loadDatoDataType.valueOf("0", 2))));
    p1Statements5
        .add(new Assignment(soc, new NodeValue(socDataType.valueOf("0", 2))));


    ArrayList<Assignment> p1Statements6 = new ArrayList<>();
    p1Statements6.add(new Assignment(sendData, new NodeValue(sendDataDataType.valueOf("1", 2))));
    p1Statements6.add(new Assignment(s1, new NodeValue(s1DataType.valueOf("4", 16))));

    ArrayList<Assignment> p1Statements7 = new ArrayList<>();
    p1Statements7.add(new Assignment(contaTmpVariable, new NodeValue(contaTmpDataType
        .valueOf("0", 16))));

    Assignment p1Statement8 = new Assignment(s1, new NodeValue(s1DataType.valueOf("3", 16)));

    ArrayList<Assignment> p1Statements9 = new ArrayList<>();
    p1Statements9.add(new Assignment(canale, contaTmpVariable));
    p1Statements9.add(new Assignment(s1, new NodeValue(s1DataType.valueOf("7", 16))));

    ArrayList<Assignment> p1Statements10 = new ArrayList<>();
    p1Statements10.add(new Assignment(loadDato, new NodeValue(loadDatoDataType.valueOf("1", 2))));
    p1Statements10.add(new Assignment(muxEn, new NodeValue(muxEnDataType.valueOf("0", 2))));
    p1Statements10.add(new Assignment(s1, new NodeValue(s1DataType.valueOf("6", 16))));

    Assignment p1Statement11 = new Assignment(s1, new NodeValue(s1DataType.valueOf("5", 16)));

    ArrayList<Assignment> p1Statements12 = new ArrayList<>();
    p1Statements12.add(new Assignment(sendData, new NodeValue(sendDataDataType.valueOf("0", 2))));
    p1Statements12.add(new Assignment(s1, new NodeValue(s1DataType.valueOf("0", 16))));

    Assignment p1Statement13 = new Assignment(s1, new NodeValue(s1DataType.valueOf("3", 16)));

    // Guarded Actions

    Action p1Action1 = new Action(p1Statements1);
    GuardedAction p1Reset = new GuardedAction(guardTrue, p1Action1);
    p1Reset.setClock(resetEvent);

    Action p1Action2 = new Action(p1Statements2);
    GuardedAction p1ga2 = new GuardedAction(guardTrue, p1Action2);
    p1ga2.setClock(clockEvent);

    GuardedAction p1ga3 = new GuardedAction(guardTrue, new Action(p1Statement3));
    p1ga3.setClock(clockEvent);

    Action p1Action4 = new Action(p1Statements4);
    GuardedAction p1ga4 = new GuardedAction(guardTrue, p1Action4);
    p1ga4.setClock(clockEvent);

    GuardedAction p1ga5 =
        new GuardedAction(new Guard(p1GuardPredicate3), new Action(p1Statement11));
    p1ga5.setClock(clockEvent);

    Action p1Action6 = new Action(p1Statements10);
    GuardedAction p1ga6 = new GuardedAction(new Guard(Node.NOT(p1GuardPredicate3)), p1Action6);
    p1ga6.setClock(clockEvent);

    ArrayList<Assignment> p1Action7Statements = new ArrayList<>();
    p1Action7Statements.addAll(p1Statements5);
    p1Action7Statements.addAll(p1Statements7);
    p1Action7Statements.addAll(p1Statements9);
    Action p1Action7 = new Action(p1Action7Statements);
    GuardedAction p1ga7 = new GuardedAction(new Guard(p1GuardPredicate2), p1Action7);
    p1ga7.setClock(clockEvent);

    p1Statements5
        .add(new Assignment(contaTmpVariable, new NodeOperation(StandardOperation.ADD,
            contaTmpVariable, new NodeValue(contaTmpDataType.valueOf("1", 16)))));

    ArrayList<Assignment> p1Action8Statements = new ArrayList<>();
    p1Action8Statements.addAll(p1Statements5);
    p1Action8Statements.addAll(p1Statements9);
    Action p1Action8 = new Action(p1Action8Statements);
    GuardedAction p1ga8 = new GuardedAction(new Guard(Node.NOT(p1GuardPredicate2)), p1Action8);
    p1ga8.setClock(clockEvent);

    Action p1Action9 = new Action(p1Statements6);
    GuardedAction p1ga9 = new GuardedAction(guardTrue, p1Action9);
    p1ga9.setClock(clockEvent);

    GuardedAction p1ga10 = new GuardedAction(guardTrue, new Action(p1Statement8));
    p1ga10.setClock(clockEvent);

    Action p1Action11 = new Action(p1Statements12);
    GuardedAction p1ga11 = new GuardedAction(new Guard(Node.NOT(p1GuardPredicate1)), p1Action11);
    p1ga11.setClock(clockEvent);

    GuardedAction p1ga12 =
        new GuardedAction(new Guard(p1GuardPredicate1), new Action(p1Statement13));
    p1ga12.setClock(clockEvent);

    // Transitions

    process1.createTransition(gp001State, gp001State, p1Reset);
    process1.createTransition(gp010State, gp001State, p1Reset);
    process1.createTransition(gp011State, gp001State, p1Reset);
    process1.createTransition(gp100State, gp001State, p1Reset);
    process1.createTransition(gp100wState, gp001State, p1Reset);
    process1.createTransition(gp101State, gp001State, p1Reset);
    process1.createTransition(gp110State, gp001State, p1Reset);
    process1.createTransition(gp111State, gp001State, p1Reset);

    process1.createTransition(gp001State, gp010State, p1ga2);
    process1.createTransition(gp010State, gp011State, p1ga3);
    process1.createTransition(gp011State, gp101State, p1ga4);
    process1.createTransition(gp101State, gp101State, p1ga5);
    process1.createTransition(gp101State, gp110State, p1ga6);
    process1.createTransition(gp110State, gp111State, p1ga7);
    process1.createTransition(gp110State, gp111State, p1ga8);
    process1.createTransition(gp111State, gp100wState, p1ga9);
    process1.createTransition(gp100wState, gp100State, p1ga10);
    process1.createTransition(gp100State, gp100State, p1ga12);
    process1.createTransition(gp100State, gp001State, p1ga11);


    // process2 efsm

    // States

    EfsmState gp01State =
        process2.createState(new NodeOperation(StandardOperation.EQ, s2, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP01), 16))));
    EfsmState gp10State =
        process2.createState(new NodeOperation(StandardOperation.EQ, s2, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP10), 16))));
    EfsmState gp11State =
        process2.createState(new NodeOperation(StandardOperation.EQ, s2, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP11), 16))));
    EfsmState gp11wState =
        process2.createState(new NodeOperation(StandardOperation.EQ, s2, new NodeValue(
            DataType.INTEGER.valueOf(String.valueOf(GP11w), 16))));

    process2.setInitialState(gp01State);

    // Guards

    Node p2GuardPredicate1 =
        new NodeOperation(StandardOperation.EQ, confirm, new NodeValue(
            confirmDataType.valueOf("0", 2)));
    Node p2GuardPredicate2 =
        new NodeOperation(StandardOperation.EQ, sendData, new NodeValue(
            sendDataDataType.valueOf("1", 2)));
    Node p2GuardPredicate3 =
        new NodeOperation(StandardOperation.EQ, mpx, new NodeValue(mpxDataType.valueOf("0",
            2)));



    // Assignments

    ArrayList<Assignment> p2Statements1 = new ArrayList<>();
    p2Statements1.add(new Assignment(rdy, new NodeValue(rdyDataType.valueOf("0", 2))));
    p2Statements1.add(new Assignment(addMpx2, new NodeValue(addMpx2DataType.valueOf("0", 2))));
    p2Statements1.add(new Assignment(mpx, new NodeValue(mpxDataType.valueOf("0", 2))));
    p2Statements1.add(new Assignment(shot, new NodeValue(shotDataType.valueOf("0", 2))));
    p2Statements1.add(new Assignment(s2, new NodeValue(s2DataType.valueOf("0", 16))));

    ArrayList<Assignment> p2Statements2 = new ArrayList<>();
    p2Statements2.add(new Assignment(shot, new NodeValue(shotDataType.valueOf("1", 2))));
    p2Statements2.add(new Assignment(s2, new NodeValue(s2DataType.valueOf("2", 16))));

    ArrayList<Assignment> p2Statements3 = new ArrayList<>();
    p2Statements3.add(new Assignment(shot, new NodeValue(shotDataType.valueOf("0", 2))));
    p2Statements3.add(new Assignment(s2, new NodeValue(s2DataType.valueOf("2", 16))));

    ArrayList<Assignment> p2Statements4 = new ArrayList<>();
    p2Statements4.add(new Assignment(rdy, new NodeValue(rdyDataType.valueOf("1", 2))));
    p2Statements4.add(new Assignment(s2, new NodeValue(s2DataType.valueOf("1", 16))));

    Assignment p2Statement5 = new Assignment(s2, new NodeValue(s2DataType.valueOf("0", 16)));

    ArrayList<Assignment> p2Statements6 = new ArrayList<>();
    p2Statements6.add(new Assignment(addMpx2, new NodeValue(addMpx2DataType.valueOf("1", 2))));
    p2Statements6.add(new Assignment(mpx, new NodeValue(mpxDataType.valueOf("1", 2))));
    p2Statements6.add(new Assignment(s2, new NodeValue(s2DataType.valueOf("1", 16))));

    ArrayList<Assignment> p2Statements7 = new ArrayList<>();
    p2Statements7.add(new Assignment(mpx, new NodeValue(mpxDataType.valueOf("0", 2))));
    p2Statements7.add(new Assignment(rdy, new NodeValue(rdyDataType.valueOf("0", 2))));
    p2Statements7.add(new Assignment(s2, new NodeValue(s2DataType.valueOf("3", 16))));

    Assignment p2Statement8 = new Assignment(s2, new NodeValue(s2DataType.valueOf("0", 16)));

    // Guarded Actions

    Action p2Action1 = new Action(p2Statements1);
    GuardedAction p2Reset = new GuardedAction(guardTrue, p2Action1);
    p2Reset.setClock(resetEvent);

    Action p2Action2 = new Action(p2Statements4);
    GuardedAction p2ga2 = new GuardedAction(new Guard(p2GuardPredicate2), p2Action2);
    p2ga2.setClock(clockEvent);

    GuardedAction p2ga3 =
        new GuardedAction(new Guard(Node.NOT(p2GuardPredicate2)),
            new Action(p2Statement5));
    p2ga3.setClock(clockEvent);

    Action p2Action4 = new Action(p2Statements2);
    GuardedAction p2ga4 = new GuardedAction(guardTrue, p2Action4);
    p2ga4.setClock(clockEvent);

    Action p2Action5 = new Action(p2Statements3);
    GuardedAction p2ga5 = new GuardedAction(new Guard(p2GuardPredicate1), p2Action5);
    p2ga5.setClock(clockEvent);

    ArrayList<Node> p2ga6Nodes = new ArrayList<>();
    p2ga6Nodes.add(Node.NOT(p2GuardPredicate1));
    p2ga6Nodes.add(p2GuardPredicate3);
    Action p2Action6 = new Action(p2Statements6);
    GuardedAction p2ga6 = new GuardedAction(new Guard(p2ga6Nodes), p2Action6);
    p2ga6.setClock(clockEvent);

    ArrayList<Node> p2ga7Nodes = new ArrayList<>();
    p2ga7Nodes.add(Node.NOT(p2GuardPredicate1));
    p2ga7Nodes.add(Node.NOT(p2GuardPredicate3));
    Action p2Action7 = new Action(p2Statements7);
    GuardedAction p2ga7 = new GuardedAction(new Guard(p2ga7Nodes), p2Action7);
    p2ga7.setClock(clockEvent);

    GuardedAction p2ga8 = new GuardedAction(guardTrue, new Action(p2Statement8));
    p2ga8.setClock(clockEvent);

    // Transitions

    process2.createTransition(gp01State, gp01State, p2Reset);
    process2.createTransition(gp10State, gp01State, p2Reset);
    process2.createTransition(gp11State, gp01State, p2Reset);
    process2.createTransition(gp11wState, gp01State, p2Reset);

    process2.createTransition(gp01State, gp10State, p2ga2);
    process2.createTransition(gp01State, gp01State, p2ga3);
    process2.createTransition(gp10State, gp11State, p2ga4);
    process2.createTransition(gp11State, gp11State, p2ga5);
    process2.createTransition(gp11State, gp10State, p2ga6);
    process2.createTransition(gp11State, gp11wState, p2ga7);
    process2.createTransition(gp11wState, gp01State, p2ga8);

    // process3 efsm

    // States

    EfsmState gIdleState =
        process3.createState(new NodeOperation(StandardOperation.EQ, itfcState,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(G_IDLE), 16))));
    EfsmState gLoadState =
        process3.createState(new NodeOperation(StandardOperation.EQ, itfcState,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(G_LOAD), 16))));
    EfsmState gSendState =
        process3.createState(new NodeOperation(StandardOperation.EQ, itfcState,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(G_SEND), 16))));
    EfsmState gWaitEndState =
        process3.createState(new NodeOperation(StandardOperation.EQ, itfcState,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(G_WAIT_END), 16))));

    process3.setInitialState(gIdleState);

    // Guards

    Node p3GuardPredicate1 =
        new NodeOperation(StandardOperation.EQ, shot, new NodeValue(shotDataType.valueOf(
            "1", 2)));
    Node p3GuardPredicate2 =
        new NodeOperation(StandardOperation.EQ, txEnd, new NodeValue(txEndDataType.valueOf(
            "1", 2)));

    // Assignments

    ArrayList<Assignment> p3Statements1 = new ArrayList<>();
    p3Statements1.add(new Assignment(load, new NodeValue(loadDataType.valueOf("0", 2))));
    p3Statements1.add(new Assignment(send, new NodeValue(sendDataType.valueOf("0", 2))));
    p3Statements1.add(new Assignment(confirm, new NodeValue(confirmDataType.valueOf("0", 2))));

    ArrayList<Assignment> p3Statements2 = new ArrayList<>();
    p3Statements2.add(new Assignment(load, new NodeValue(loadDataType.valueOf("0", 2))));
    p3Statements2.add(new Assignment(send, new NodeValue(sendDataType.valueOf("1", 2))));

    ArrayList<Assignment> p3Statements3 = new ArrayList<>();
    p3Statements3.add(new Assignment(send, new NodeValue(sendDataType.valueOf("0", 2))));

    ArrayList<Assignment> p3Statements4 = new ArrayList<>();
    p3Statements4.add(new Assignment(load, new NodeValue(loadDataType.valueOf("1", 2))));
    p3Statements4.add(new Assignment(confirm, new NodeValue(confirmDataType.valueOf("0", 2))));

    ArrayList<Assignment> p3Statements5 = new ArrayList<>();
    p3Statements5.add(new Assignment(confirm, new NodeValue(confirmDataType.valueOf("0", 2))));

    ArrayList<Assignment> p3Statements6 = new ArrayList<>();
    p3Statements6.add(new Assignment(confirm, new NodeValue(confirmDataType.valueOf("1", 2))));

    // Guarded Actions

    Action p3Action1 = new Action(p3Statements1);
    GuardedAction p3Reset = new GuardedAction(guardTrue, p3Action1);
    p3Reset.setClock(resetEvent);

    Action p3Action2 = new Action(p3Statements4);
    GuardedAction p3ga2 = new GuardedAction(new Guard(p3GuardPredicate1), p3Action2);
    p3ga2.setClock(clockEvent);

    Action p3Action3 = new Action(p3Statements5);
    GuardedAction p3ga3 = new GuardedAction(new Guard(Node.NOT(p3GuardPredicate1)), p3Action3);
    p3ga3.setClock(clockEvent);

    Action p3Action4 = new Action(p3Statements2);
    GuardedAction p3ga4 = new GuardedAction(guardTrue, p3Action4);
    p3ga4.setClock(clockEvent);

    Action p3Action5 = new Action(p3Statements3);
    GuardedAction p3ga5 = new GuardedAction(guardTrue, p3Action5);
    p3ga5.setClock(clockEvent);

    Action p3Action6 = new Action(p3Statements6);
    GuardedAction p3ga6 = new GuardedAction(new Guard(p3GuardPredicate2), p3Action6);
    p3ga6.setClock(clockEvent);

    // Transitions

    process3.createTransition(gIdleState, gIdleState, p3Reset);
    process3.createTransition(gLoadState, gIdleState, p3Reset);
    process3.createTransition(gSendState, gIdleState, p3Reset);
    process3.createTransition(gWaitEndState, gIdleState, p3Reset);

    process3.createTransition(gIdleState, gLoadState, p3ga2);
    process3.createTransition(gIdleState, gIdleState, p3ga3);
    process3.createTransition(gLoadState, gSendState, p3ga4);
    process3.createTransition(gSendState, gWaitEndState, p3ga5);
    process3.createTransition(gWaitEndState, gIdleState, p3ga6);


    // process4 efsm

    EfsmState p4State1 =
        process4.createState(new NodeOperation(StandardOperation.EQ, tre, new NodeValue(
            treDataType.valueOf("0", 2))));
    EfsmState p4State2 =
        process4.createState(new NodeOperation(StandardOperation.EQ, tre, new NodeValue(
            treDataType.valueOf("1", 2))));

    process4.setInitialState(p4State1);


    // Guards

    Node p4GuardPredicate1 =
        new NodeOperation(StandardOperation.EQ, txEnd, new NodeValue(txEndDataType.valueOf(
            "1", 2)));
    Node p4GuardPredicate2 =
        new NodeOperation(StandardOperation.EQ, load, new NodeValue(loadDataType.valueOf(
            "1", 2)));
    Node p4GuardPredicate4 =
        new NodeOperation(StandardOperation.EQ, send, new NodeValue(sendDataType.valueOf(
            "1", 2)));
    Node p4GuardPredicate5 =
        new NodeOperation(StandardOperation.EQ, dsr, new NodeValue(dsrDataType.valueOf("0",
            2)));

    // Assignments

    ArrayList<Assignment> p4Statements1 = new ArrayList<>();
    p4Statements1.add(new Assignment(sendEn, new NodeValue(sendEnDataType.valueOf("0", 2))));
    p4Statements1.add(new Assignment(outReg, new NodeValue(outRegDataType.valueOf("0", 16))));
    p4Statements1.add(new Assignment(tre, new NodeValue(treDataType.valueOf("0", 2))));
    p4Statements1.add(new Assignment(error, new NodeValue(errorDataType.valueOf("0", 2))));

    Assignment p4Statement2 = new Assignment(sendEn, new NodeValue(sendEnDataType.valueOf("0", 2)));
    Assignment p4Statement3 = new Assignment(tre, new NodeValue(treDataType.valueOf("1", 2)));
    Assignment p4Statement4 = new Assignment(outReg, dataIn);
    Assignment p4Statement6 = new Assignment(error, new NodeValue(errorDataType.valueOf("0", 2)));
    Assignment p4Statement7 = new Assignment(error, new NodeValue(errorDataType.valueOf("1", 2)));
    Assignment p4Statement8 = new Assignment(sendEn, new NodeValue(sendEnDataType.valueOf("1", 2)));

    // Guarded Actions

    Action p4Action1 = new Action(p4Statements1);
    GuardedAction p4Reset = new GuardedAction(guardTrue, p4Action1);
    p4Reset.setClock(resetEvent);

    Action p4Action3 = new Action(p4Statement7);
    ArrayList<Node> guard3Nodes = new ArrayList<>();
    guard3Nodes.add(Node.NOT(p4GuardPredicate1));
    guard3Nodes.add(Node.NOT(p4GuardPredicate2));
    guard3Nodes.add(p4GuardPredicate4);
    GuardedAction p4ga31 = new GuardedAction(new Guard(guard3Nodes), p4Action3);
    p4ga31.setClock(clockEvent);

    guard3Nodes.add(p4GuardPredicate5);
    GuardedAction p4ga32 = new GuardedAction(new Guard(guard3Nodes), p4Action3);
    p4ga32.setClock(clockEvent);

    ArrayList<Assignment> action4Statements = new ArrayList<>();
    action4Statements.add(p4Statement6);
    action4Statements.add(p4Statement8);
    Action p4Action4 = new Action(action4Statements);
    ArrayList<Node> guard4Nodes = new ArrayList<>();
    guard4Nodes.add(Node.NOT(p4GuardPredicate1));
    guard4Nodes.add(Node.NOT(p4GuardPredicate2));
    guard4Nodes.add(p4GuardPredicate4);
    guard4Nodes.add(Node.NOT(p4GuardPredicate5));
    GuardedAction p4ga4 = new GuardedAction(new Guard(guard4Nodes), p4Action4);
    p4ga4.setClock(clockEvent);

    ArrayList<Assignment> action5Statements = new ArrayList<>();
    action5Statements.add(p4Statement4);
    action5Statements.add(p4Statement3);
    action5Statements.add(p4Statement6);
    Action p4Action5 = new Action(action5Statements);
    ArrayList<Node> guard5Nodes = new ArrayList<>();
    guard5Nodes.add(p4GuardPredicate2);
    guard5Nodes.add(Node.NOT(p4GuardPredicate1));
    guard5Nodes.add(Node.NOT(p4GuardPredicate4));
    GuardedAction p4ga5 = new GuardedAction(new Guard(guard5Nodes), p4Action5);
    p4ga5.setClock(clockEvent);

    Action p4Action6 = new Action(p4Statement7);
    GuardedAction p4ga6 = new GuardedAction(new Guard(guard5Nodes), p4Action6);
    p4ga6.setClock(clockEvent);

    ArrayList<Assignment> action7Statements = new ArrayList<>();
    action7Statements.add(p4Statement4);
    action7Statements.add(p4Statement3);
    action7Statements.add(p4Statement7);
    Action p4Action7 = new Action(action7Statements);
    ArrayList<Node> guard7Nodes = new ArrayList<>();
    guard7Nodes.add(Node.NOT(p4GuardPredicate1));
    guard7Nodes.add(p4GuardPredicate2);
    guard7Nodes.add(p4GuardPredicate4);
    GuardedAction p4ga7 = new GuardedAction(new Guard(guard7Nodes), p4Action7);
    p4ga7.setClock(clockEvent);

    Action p4Action9 = new Action(p4Statement7);
    ArrayList<Node> guard9Nodes = new ArrayList<>();
    guard9Nodes.add(Node.NOT(p4GuardPredicate1));
    guard9Nodes.add(p4GuardPredicate2);
    guard9Nodes.add(p4GuardPredicate4);
    guard9Nodes.add(p4GuardPredicate5);
    GuardedAction p4ga9 = new GuardedAction(new Guard(guard9Nodes), p4Action9);
    p4ga9.setClock(clockEvent);

    ArrayList<Assignment> action10Statements = new ArrayList<>();
    action10Statements.add(p4Statement6);
    action10Statements.add(p4Statement8);
    Action p4Action10 = new Action(action10Statements);
    ArrayList<Node> guard10Nodes = new ArrayList<>();
    guard10Nodes.add(Node.NOT(p4GuardPredicate1));
    guard10Nodes.add(p4GuardPredicate2);
    guard10Nodes.add(p4GuardPredicate4);
    guard10Nodes.add(Node.NOT(p4GuardPredicate5));
    GuardedAction p4ga10 = new GuardedAction(new Guard(guard10Nodes), p4Action10);
    p4ga10.setClock(clockEvent);

    ArrayList<Assignment> action11Statements = new ArrayList<>();
    action11Statements.add(p4Statement2);
    action11Statements.add(p4Statement3);
    Action p4Action11 = new Action(action11Statements);
    ArrayList<Node> guard11Nodes = new ArrayList<>();
    guard11Nodes.add(p4GuardPredicate1);
    guard11Nodes.add(Node.NOT(p4GuardPredicate2));
    guard11Nodes.add(Node.NOT(p4GuardPredicate4));
    GuardedAction p4ga11 = new GuardedAction(new Guard(guard11Nodes), p4Action11);
    p4ga11.setClock(clockEvent);

    ArrayList<Assignment> action12Statements = new ArrayList<>();
    action12Statements.add(p4Statement2);
    action12Statements.add(p4Statement3);
    action12Statements.add(p4Statement7);
    Action p4Action12 = new Action(action12Statements);
    ArrayList<Node> guard12Nodes = new ArrayList<>();
    guard12Nodes.add(p4GuardPredicate1);
    guard12Nodes.add(Node.NOT(p4GuardPredicate2));
    guard12Nodes.add(p4GuardPredicate4);
    GuardedAction p4ga121 = new GuardedAction(new Guard(guard12Nodes), p4Action12);
    p4ga121.setClock(clockEvent);

    guard12Nodes.add(p4GuardPredicate5);
    GuardedAction p4ga122 = new GuardedAction(new Guard(guard12Nodes), p4Action12);
    p4ga122.setClock(clockEvent);

    ArrayList<Assignment> action13Statements = new ArrayList<>();
    action13Statements.add(p4Statement8);
    action13Statements.add(p4Statement3);
    action13Statements.add(p4Statement6);
    Action p4Action13 = new Action(action13Statements);
    ArrayList<Node> guard13Nodes = new ArrayList<>();
    guard13Nodes.add(p4GuardPredicate1);
    guard13Nodes.add(Node.NOT(p4GuardPredicate2));
    guard13Nodes.add(p4GuardPredicate4);
    guard13Nodes.add(Node.NOT(p4GuardPredicate5));
    GuardedAction p4ga13 = new GuardedAction(new Guard(guard13Nodes), p4Action13);
    p4ga13.setClock(clockEvent);

    Action p4Action14 = new Action(action12Statements);
    ArrayList<Node> guard14Nodes = new ArrayList<>();
    guard14Nodes.add(p4GuardPredicate1);
    guard14Nodes.add(p4GuardPredicate2);
    guard14Nodes.add(Node.NOT(p4GuardPredicate4));
    GuardedAction p4ga14 = new GuardedAction(new Guard(guard14Nodes), p4Action14);
    p4ga14.setClock(clockEvent);

    Action p4Action15 = new Action(action12Statements);
    ArrayList<Node> guard15Nodes = new ArrayList<>();
    guard15Nodes.add(p4GuardPredicate1);
    guard15Nodes.add(p4GuardPredicate2);
    guard15Nodes.add(p4GuardPredicate4);
    guard15Nodes.add(p4GuardPredicate5);
    GuardedAction p4ga15 = new GuardedAction(new Guard(guard15Nodes), p4Action15);
    p4ga15.setClock(clockEvent);

    Action p4Action16 = new Action(action13Statements);
    ArrayList<Node> guard16Nodes = new ArrayList<>();
    guard16Nodes.add(p4GuardPredicate1);
    guard16Nodes.add(p4GuardPredicate2);
    guard16Nodes.add(p4GuardPredicate4);
    guard16Nodes.add(Node.NOT(p4GuardPredicate5));
    GuardedAction p4ga16 = new GuardedAction(new Guard(guard16Nodes), p4Action16);
    p4ga16.setClock(clockEvent);


    ArrayList<Assignment> action17Statements = new ArrayList<>();
    action17Statements.add(p4Statement2);
    action17Statements.add(p4Statement3);
    action17Statements.add(p4Statement6);
    action17Statements.add(p4Statement4);
    Action p4Action17 = new Action(action17Statements);
    ArrayList<Node> guard17Nodes = new ArrayList<>();
    guard17Nodes.add(p4GuardPredicate1);
    guard17Nodes.add(p4GuardPredicate2);
    guard17Nodes.add(Node.NOT(p4GuardPredicate4));
    GuardedAction p4ga17 = new GuardedAction(new Guard(guard17Nodes), p4Action17);
    p4ga17.setClock(clockEvent);

    ArrayList<Assignment> action18Statements = new ArrayList<>();
    action18Statements.add(p4Statement2);
    action18Statements.add(p4Statement3);
    action18Statements.add(p4Statement7);
    action18Statements.add(p4Statement4);
    Action p4Action18 = new Action(action18Statements);
    ArrayList<Node> guard18Nodes = new ArrayList<>();
    guard18Nodes.add(p4GuardPredicate1);
    guard18Nodes.add(p4GuardPredicate2);
    guard18Nodes.add(p4GuardPredicate4);
    guard18Nodes.add(p4GuardPredicate5);
    GuardedAction p4ga18 = new GuardedAction(new Guard(guard18Nodes), p4Action18);
    p4ga18.setClock(clockEvent);

    process4.createTransition(p4State1, p4State1, p4Reset);
    process4.createTransition(p4State2, p4State1, p4Reset);

    process4.createTransition(p4State1, p4State1, p4ga31);
    process4.createTransition(p4State2, p4State2, p4ga32);

    process4.createTransition(p4State2, p4State2, p4ga4);

    process4.createTransition(p4State1, p4State2, p4ga5);

    process4.createTransition(p4State2, p4State2, p4ga6);

    process4.createTransition(p4State1, p4State2, p4ga7);

    process4.createTransition(p4State2, p4State2, p4ga9);

    process4.createTransition(p4State2, p4State2, p4ga10);

    process4.createTransition(p4State1, p4State2, p4ga11);
    process4.createTransition(p4State2, p4State2, p4ga11);

    process4.createTransition(p4State1, p4State2, p4ga121);
    process4.createTransition(p4State2, p4State2, p4ga122);

    process4.createTransition(p4State2, p4State2, p4ga13);

    process4.createTransition(p4State2, p4State2, p4ga14);

    process4.createTransition(p4State2, p4State2, p4ga15);

    process4.createTransition(p4State2, p4State2, p4ga16);

    process4.createTransition(p4State1, p4State2, p4ga17);

    process4.createTransition(p4State1, p4State2, p4ga18);


    // process5 efsm

    EfsmState startBitState =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(START_BIT), 16))));
    EfsmState stopBitState =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(STOP_BIT), 16))));
    EfsmState bit0State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT0), 16))));
    EfsmState bit1State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT1), 16))));
    EfsmState bit2State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT2), 16))));
    EfsmState bit3State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT3), 16))));
    EfsmState bit4State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT4), 16))));
    EfsmState bit5State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT5), 16))));
    EfsmState bit6State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT6), 16))));
    EfsmState bit7State =
        process5.createState(new NodeOperation(StandardOperation.EQ, nextBit,
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(BIT7), 16))));

    process5.setInitialState(startBitState);

    // Guards

    Node p5GuardPredicate1 =
        new NodeOperation(StandardOperation.EQ, sendEn, new NodeValue(
            sendEnDataType.valueOf("1", 2)));
    Node p5GuardPredicate2 =
        new NodeOperation(StandardOperation.GREATER, txConta, new NodeValue(
            txContaDataType.valueOf("104", 10)));


    // Assignments

    ArrayList<Assignment> p5Statements1 = new ArrayList<>();
    p5Statements1.add(new Assignment(txEnd, new NodeValue(txEndDataType.valueOf("0", 2))));
    p5Statements1.add(new Assignment(dataOut, new NodeValue(dataOutDataType.valueOf("0", 2))));
    p5Statements1.add(new Assignment(txConta, new NodeValue(txContaDataType.valueOf("0", 10))));

    ArrayList<Assignment> p5Statements2 = new ArrayList<>();
    p5Statements2.add(new Assignment(txEnd, new NodeValue(txEndDataType.valueOf("0", 2))));
    p5Statements2.add(new Assignment(dataOut, new NodeValue(dataOutDataType.valueOf("1", 2))));

    Assignment p5Statement21 = new Assignment(txEnd, new NodeValue(txEndDataType.valueOf("0", 2)));

    ArrayList<Assignment> p5Statements3 = new ArrayList<>();
    p5Statements3.add(new Assignment(txConta, new NodeOperation(
        StandardOperation.ADD, txConta, new NodeValue(txContaDataType.valueOf("1", 10)))));

    ArrayList<Assignment> p5Statements4 = new ArrayList<>();
    p5Statements4.add(new Assignment(dataOut, new NodeValue(dataOutDataType.valueOf("0", 2))));

    ArrayList<Assignment> p5Statements5 = new ArrayList<>();
    p5Statements5.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "7", 16))))));

    ArrayList<Assignment> p5Statements6 = new ArrayList<>();
    p5Statements6.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "6", 16))))));

    ArrayList<Assignment> p5Statements7 = new ArrayList<>();
    p5Statements7.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "5", 16))))));

    ArrayList<Assignment> p5Statements8 = new ArrayList<>();
    p5Statements8.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "4", 16))))));

    ArrayList<Assignment> p5Statements9 = new ArrayList<>();
    p5Statements9.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "3", 16))))));

    ArrayList<Assignment> p5Statements10 = new ArrayList<>();
    p5Statements10.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "2", 16))))));

    ArrayList<Assignment> p5Statements11 = new ArrayList<>();
    p5Statements11.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "1", 16))))));

    ArrayList<Assignment> p5Statements12 = new ArrayList<>();
    p5Statements12.add(new Assignment(dataOut, new NodeOperation(
        StandardOperation.BVAND, outReg, new NodeOperation(StandardOperation.BVLSHL,
            new NodeValue(outRegDataType.valueOf("1", 16)), new NodeValue(DataType.INTEGER.valueOf(
                "0", 16))))));

    ArrayList<Assignment> p5Statements13 = new ArrayList<>();
    p5Statements13.add(new Assignment(dataOut, new NodeValue(dataOutDataType.valueOf("1", 2))));
    p5Statements13.add(new Assignment(txEnd, new NodeValue(txEndDataType.valueOf("1", 2))));

    ArrayList<Assignment> p5Statements14 = new ArrayList<>();
    p5Statements14.add(new Assignment(txConta, new NodeValue(txContaDataType.valueOf("0", 10))));

    // Guarded Actions

    Action p5Action1 = new Action(p5Statements1);
    GuardedAction p5Reset = new GuardedAction(guardTrue, p5Action1);
    p5Reset.setClock(resetEvent);

    ArrayList<Assignment> p5Action2Statements = new ArrayList<>();
    p5Action2Statements.addAll(p5Statements2);
    p5Action2Statements.addAll(p5Statements3);
    Action p5Action2 = new Action(p5Action2Statements);
    ArrayList<Node> p5ga2Nodes = new ArrayList<>();
    p5ga2Nodes.add(p5GuardPredicate1);
    p5ga2Nodes.add(Node.NOT(p5GuardPredicate2));
    GuardedAction p5ga2 = new GuardedAction(new Guard(p5ga2Nodes), p5Action2);
    p5ga2.setClock(clockEvent);

    Action p5Action3 = new Action(p5Statements2);
    GuardedAction p5ga3 = new GuardedAction(new Guard(Node.NOT(p5GuardPredicate1)), p5Action3);
    p5ga3.setClock(clockEvent);

    ArrayList<Node> p5CommonGuardNodes = new ArrayList<>();
    p5CommonGuardNodes.add(p5GuardPredicate1);
    p5CommonGuardNodes.add(p5GuardPredicate2);

    ArrayList<Assignment> p5Action4Statements = new ArrayList<>();
    p5Action4Statements.add(p5Statement21);
    p5Action4Statements.addAll(p5Statements4);
    p5Action4Statements.addAll(p5Statements14);
    Action p5Action4 = new Action(p5Action4Statements);
    GuardedAction p5ga4 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action4);
    p5ga4.setClock(clockEvent);

    ArrayList<Assignment> p5Action5Statements = new ArrayList<>();
    p5Action5Statements.add(p5Statement21);
    p5Action5Statements.addAll(p5Statements5);
    p5Action5Statements.addAll(p5Statements14);
    Action p5Action5 = new Action(p5Action5Statements);
    GuardedAction p5ga5 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action5);
    p5ga5.setClock(clockEvent);

    ArrayList<Assignment> p5Action6Statements = new ArrayList<>();
    p5Action6Statements.add(p5Statement21);
    p5Action6Statements.addAll(p5Statements6);
    p5Action6Statements.addAll(p5Statements14);
    Action p5Action6 = new Action(p5Action6Statements);
    GuardedAction p5ga6 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action6);
    p5ga6.setClock(clockEvent);

    ArrayList<Assignment> p5Action7Statements = new ArrayList<>();
    p5Action7Statements.add(p5Statement21);
    p5Action7Statements.addAll(p5Statements7);
    p5Action7Statements.addAll(p5Statements14);
    Action p5Action7 = new Action(p5Action7Statements);
    GuardedAction p5ga7 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action7);
    p5ga7.setClock(clockEvent);

    ArrayList<Assignment> p5Action8Statements = new ArrayList<>();
    p5Action8Statements.add(p5Statement21);
    p5Action8Statements.addAll(p5Statements8);
    p5Action8Statements.addAll(p5Statements14);
    Action p5Action8 = new Action(p5Action8Statements);
    GuardedAction p5ga8 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action8);
    p5ga8.setClock(clockEvent);

    ArrayList<Assignment> p5Action9Statements = new ArrayList<>();
    p5Action9Statements.add(p5Statement21);
    p5Action9Statements.addAll(p5Statements9);
    p5Action9Statements.addAll(p5Statements14);
    Action p5Action9 = new Action(p5Action9Statements);
    GuardedAction p5ga9 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action9);
    p5ga9.setClock(clockEvent);

    ArrayList<Assignment> p5Action10Statements = new ArrayList<>();
    p5Action10Statements.add(p5Statement21);
    p5Action10Statements.addAll(p5Statements10);
    p5Action10Statements.addAll(p5Statements14);
    Action p5Action10 = new Action(p5Action10Statements);
    GuardedAction p5ga10 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action10);
    p5ga10.setClock(clockEvent);

    ArrayList<Assignment> p5Action11Statements = new ArrayList<>();
    p5Action11Statements.add(p5Statement21);
    p5Action11Statements.addAll(p5Statements11);
    p5Action11Statements.addAll(p5Statements14);
    Action p5Action11 = new Action(p5Action11Statements);
    GuardedAction p5ga11 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action11);
    p5ga11.setClock(clockEvent);

    ArrayList<Assignment> p5Action12Statements = new ArrayList<>();
    p5Action12Statements.add(p5Statement21);
    p5Action12Statements.addAll(p5Statements12);
    p5Action12Statements.addAll(p5Statements14);
    Action p5Action12 = new Action(p5Action12Statements);
    GuardedAction p5ga12 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action12);
    p5ga12.setClock(clockEvent);

    ArrayList<Assignment> p5Action13Statements = new ArrayList<>();
    p5Action13Statements.addAll(p5Statements13);
    p5Action13Statements.addAll(p5Statements14);
    Action p5Action13 = new Action(p5Action13Statements);
    GuardedAction p5ga13 = new GuardedAction(new Guard(p5CommonGuardNodes), p5Action13);
    p5ga13.setClock(clockEvent);


    process5.createTransition(startBitState, startBitState, p5Reset);
    process5.createTransition(bit0State, startBitState, p5Reset);
    process5.createTransition(bit1State, startBitState, p5Reset);
    process5.createTransition(bit2State, startBitState, p5Reset);
    process5.createTransition(bit3State, startBitState, p5Reset);
    process5.createTransition(bit4State, startBitState, p5Reset);
    process5.createTransition(bit5State, startBitState, p5Reset);
    process5.createTransition(bit6State, startBitState, p5Reset);
    process5.createTransition(bit7State, startBitState, p5Reset);
    process5.createTransition(stopBitState, startBitState, p5Reset);

    process5.createTransition(startBitState, startBitState, p5ga2);
    process5.createTransition(bit0State, bit0State, p5ga2);
    process5.createTransition(bit1State, bit1State, p5ga2);
    process5.createTransition(bit2State, bit2State, p5ga2);
    process5.createTransition(bit3State, bit3State, p5ga2);
    process5.createTransition(bit4State, bit4State, p5ga2);
    process5.createTransition(bit5State, bit5State, p5ga2);
    process5.createTransition(bit6State, bit6State, p5ga2);
    process5.createTransition(bit7State, bit7State, p5ga2);
    process5.createTransition(stopBitState, stopBitState, p5ga2);

    process5.createTransition(startBitState, startBitState, p5ga3);
    process5.createTransition(bit0State, bit0State, p5ga3);
    process5.createTransition(bit1State, bit1State, p5ga3);
    process5.createTransition(bit2State, bit2State, p5ga3);
    process5.createTransition(bit3State, bit3State, p5ga3);
    process5.createTransition(bit4State, bit4State, p5ga3);
    process5.createTransition(bit5State, bit5State, p5ga3);
    process5.createTransition(bit6State, bit6State, p5ga3);
    process5.createTransition(bit7State, bit7State, p5ga3);
    process5.createTransition(stopBitState, stopBitState, p5ga3);

    process5.createTransition(startBitState, bit0State, p5ga4);
    process5.createTransition(bit0State, bit1State, p5ga5);
    process5.createTransition(bit1State, bit2State, p5ga6);
    process5.createTransition(bit2State, bit3State, p5ga7);
    process5.createTransition(bit3State, bit4State, p5ga8);
    process5.createTransition(bit4State, bit5State, p5ga9);
    process5.createTransition(bit5State, bit6State, p5ga10);
    process5.createTransition(bit6State, bit7State, p5ga11);
    process5.createTransition(bit7State, stopBitState, p5ga12);
    process5.createTransition(stopBitState, startBitState, p5ga13);


  }


}
