/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.retrascope.parser.test;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.engine.efsm.generator.test.fate.EfsmFateTestGenerator;
import ru.ispras.retrascope.engine.test.printer.xml.TestXmlPrinter;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.examples.SimpleEfsm;
import ru.ispras.retrascope.parser.test.TestFileNames;
import ru.ispras.retrascope.parser.test.XmlTestParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * The test for {@link XmlTestParser}. Test description:
   * <ol>
   *    <li>test generation for {@link SimpleEfsm} by {@link EfsmFateTestGenerator};
   *    <li>marshalling the test to an XML file;
   *    <li>unmarshalling the test to an XML file;
   *    <li>comparison between the generated and unmarshalled tests. Test is considered as
   *    successfully passed if and only if the comparison is successful;
   *    <li>deleting the XML file.
   * </ol>
 * 
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class XmlTestParserTestCase {
  private final Logger logger = Log.getLogger(this.getClass());
  private final String logHeader = "TestXmlPrinterTestCase: ";

  private final List<ru.ispras.retrascope.result.testbench.Test> tests = new ArrayList<>();
  private final Set<String> fileNames = new HashSet<>();

  /**
   * Generates a test for {@link SimpleEfsm} by {@link EfsmFateTestGenerator} and stores it.
   *
   * @throws IOException if some I/O errors occurs while processing generated files
   */
  @Before
  public void generateTests() {
    Retrascope.main(null);
    EfsmModel efsmModel = new EfsmModel(SimpleEfsm.EFSM);
    EfsmFateTestGenerator generator = new EfsmFateTestGenerator();
    HashMap<EntityType, Entity> inputs = new HashMap<>();
    inputs.put(efsmModel.getEntityType(), efsmModel);
    this.tests.add(generator.start(inputs));
  }

  /**
   * Marshals the stored tests to the XML files, unmarshals them and compares the original and
   * parsed ones.
   * 
   * @throws IOException if some I/O errors occurs while processing generated files
   */
  @Test
  public void marshalUnmarshal() throws IOException {
    TestXmlPrinter printer = new TestXmlPrinter();
    XmlTestParser parser = new XmlTestParser();

    for (ru.ispras.retrascope.result.testbench.Test test: this.tests) {
      HashMap<EntityType, Entity> inputsForMarshalling = new HashMap<>();
      inputsForMarshalling.put(test.getEntityType(), test);
      TestFileNames marshallingResult = printer.start(inputsForMarshalling);
      this.fileNames.add(marshallingResult.getList().get(0));

      HashMap<EntityType, Entity> inputsForUnmarshalling = new HashMap<>();
      inputsForUnmarshalling.put(marshallingResult.getEntityType(), marshallingResult);
      Assert.assertEquals(test, parser.start(inputsForUnmarshalling));
    }
  }

  /**
   * Deletes the generated XML files.
   * 
   * @throws IOException if some I/O errors occurs while processing generated files
   */
  @After
  public void deleteFiles() throws IOException {
    for (String fileName: this.fileNames) {
      this.logger.log(LogLevel.INFO, this.logHeader + "deleting the " + fileName + " file");

      Files.deleteIfExists(Paths.get(fileName));
    }
  }
}
