/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.apache.log4j.Level;
import org.junit.Test;
import org.zamia.BuildPath;
import org.zamia.ERManager;
import org.zamia.SourceFile;
import org.zamia.ZamiaException;
import org.zamia.ZamiaLogger;
import org.zamia.ZamiaProject;
import org.zamia.ZamiaProjectBuilder;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGModule;
import org.zamia.zdb.ZDBException;

import ru.ispras.retrascope.VhdlFiles;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Map;

/**
 * This test runs ZamiaCAD tool on every HDL file from project test suite.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class IgElaborationTestCase {

  @Test
  public void runTest() throws IOException { elaborateAll(); }

  private static final String TEST_DESCRIPTION = "elaboration";

  private void elaborateAll() {
    Map<String, Map<String, String>> vhdlProjectMap = VhdlFiles.getFiles();
    for (String projectName : vhdlProjectMap.keySet()) {

      Map<String, String> projectFileMap = vhdlProjectMap.get(projectName);
      for (String fileName : projectFileMap.keySet()) {

        String topLevelModuleName = projectFileMap.get(fileName);

        Log.getLogger().log(LogLevel.INFO, topLevelModuleName + " " + TEST_DESCRIPTION);
        ZamiaProject zamiaProject = getProject(fileName, topLevelModuleName);

        ArrayList<IGModule> igModules = getIGModules(zamiaProject);

        if (igModules.size() == 0) {
          Log.getLogger().log(LogLevel.INFO, "No modules found");
        }
        zamiaProject.shutdown();
      }
    }
  }

  private ZamiaProject getProject(String relativeFilePath, String toplevelName) {

    ZamiaLogger.setup(Level.OFF);
    ZamiaProject zamiaProject = null;

    /*
    * Create BuildPath.txt
    */

    // get full path to VHDL file directory
    String sourceDirPath;
    int separatorIndex = relativeFilePath.lastIndexOf(File.separatorChar);

    if (separatorIndex >= 0){
      sourceDirPath = relativeFilePath.substring(0, separatorIndex);
    } else {

      // source file is in current dir
      sourceDirPath = System.getProperty("user.dir");
    }

    File buildFile = new File(sourceDirPath + File.separator + "BuildPath.txt");

    try {
      if (!buildFile.exists() && !buildFile.createNewFile()) {
        throw new IOException("Can't create " + buildFile.toPath().toString());
      }

      FileWriter writer = new FileWriter(buildFile);

      // clean file
      writer.flush();

      // write "local "WORK path/to/source/file""
      writer.write("local WORK " + "\"" + sourceDirPath + "\"\n");

      // write "toplevel UPPER_CASE_MODULE_NAME"
      writer.write("toplevel " + toplevelName.toUpperCase() + "\n");
      writer.close();

      /*
       * Create Zamia IG
       */
      zamiaProject = new ZamiaProject(
        "", sourceDirPath, new SourceFile(buildFile));
    } catch (IOException | ZamiaException | ZDBException e) {
      e.printStackTrace();
    }

    return zamiaProject;
  }

  private ArrayList<IGModule> getIGModules(ZamiaProject fZPrj) {
    try {
      fZPrj.clean();

      ZamiaProjectBuilder builder = fZPrj.getBuilder();
      ZamiaLogger.getInstance().getLogger().setLevel(Level.OFF);

      builder.build(true, true, null);
    } catch (IOException | ZamiaException e) {
      e.printStackTrace();
      assert false;
    }
    IGManager igm = fZPrj.getIGM();

    BuildPath bp = fZPrj.getBuildPath();

    ERManager erm = fZPrj.getERM();

    int errNum = erm.getNumErrors();

    if (errNum > 0) {
      for (int i = 0; i < errNum; i++) {
        try {
          throw erm.getError(i);
        } catch (ZamiaException e) {
          e.printStackTrace();
          assert false;
        }
      }
    }

    int topLevelNum = bp.getNumToplevels();

    ArrayList<IGModule> result = new ArrayList<>();

    if (topLevelNum > 0) {
      for (int i = 0; i < topLevelNum; i++) {
        IGModule module = igm.findModule(bp.getToplevel(i));
        result.add(module);
      }
    }
    return result;
  }
}
