/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.util;

import java.io.File;
import java.util.Map;

import org.apache.commons.lang.ArrayUtils;

import ru.ispras.retrascope.Retrascope;

/**
 * Utilities using VHDL files for tests.
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class VhdlUtilTest {
  // The "target" special for the command line
  private static final String VHDL_TOOL_TARGET = "--target";
  // The "top level" special for the command line
  private static final String VHDL_TOOL_TOP_LEVEL = "--toplevel";

  /**
   * Run Retrascope by the VHDL use on input
   * 
   * @param map VHDL files for processing
   * @param target for command line
   */
  public static void runVhdl(Map<String, String> map, String target) {
    runVhdl(map, target, "");
  }

  /**
   * Run Retrascope by the VHDL use on input
   * 
   * @param map VHDL files for processing
   * @param target for command line
   * @param additionalOption an option to run the test
   */
  public static void runVhdl(Map<String, String> map, String target, String additionalOption) {
    String[] fileNames = getFileNames(map);

    String topLevelNames = getTopLevel(map);

    runRetrascope(fileNames, createCommandLine(target, topLevelNames, additionalOption));
  }

  /**
   * Runs Retrascope by the VHDL use on input.
   * 
   * @param map VHDL files for processing
   * @param target value of command-line "target" option
   * @param additionalOptions command-line options
   */
  public static void runVhdl(Map<String, String> map, String target, String[] additionalOptions) {
    String[] fileNames = getFileNames(map);

    String topLevelNames = getTopLevel(map);

    runRetrascope(fileNames, createCommandLine(target, topLevelNames, additionalOptions));
  }

  /**
   * Create "TopLevel" parameter for command line
   * 
   * @param map VHDL files for processing
   */
  private static String getTopLevel(Map<String, String> map) {
    String topLevelNames = "";

    int i = 0;
    for (String fileName : map.keySet()) {
      topLevelNames += map.get(fileName);

      if (i < map.size() - 1) {
        topLevelNames += File.pathSeparator;
      }
      i++;
    }

    return topLevelNames;
  }

  /**
   * Create a list of files for the command line
   * 
   * @param map VHDL files for processing
   */
  private static String[] getFileNames(Map<String, String> map) {
    String[] fileNames = new String[map.size()];

    int i = 0;
    for (String fileName : map.keySet()) {
      fileNames[i] = (new File(fileName)).getAbsolutePath();
      i++;
    }

    return fileNames;
  }

  /**
   * Create command line
   * 
   * @param target for Retrascope
   * @param topLevelNames special parameter for command line
   * @param additionalOption an option to run the test
   */
  private static String[] createCommandLine(String target, String topLevelNames,
      String additionalOption) {
    return new String[] {VHDL_TOOL_TARGET, target, VHDL_TOOL_TOP_LEVEL, topLevelNames,
        additionalOption};
  }

  /**
   * Creates command line.
   * 
   * @param target value of command line "target" option
   * @param topLevelNames value of command line "toplevel" option
   * @param additionalOptions additional options
   * @return command line arguments
   */
  private static String[] createCommandLine(String target, String topLevelNames,
      String[] additionalOptions) {
    return (String[]) ArrayUtils.addAll(new String[] {VHDL_TOOL_TARGET, target,
        VHDL_TOOL_TOP_LEVEL, topLevelNames}, additionalOptions);
  }

  /**
   * Run Retrascope
   * 
   * @param fileNames name of VHDL files for processing
   * @param options command line
   */
  private static void runRetrascope(String[] fileNames, String[] options) {
    Retrascope.main((String[]) ArrayUtils.addAll(fileNames, options));
  }
}
